/* file: "os.c" */

/* Copyright (C) 1994-2003 by Marc Feeley, All Rights Reserved. */

/*
 * This module implements the operating system specific routines
 * including:
 *
 *  - OS specific initialization/finalization
 *  - program startup
 *  - process termination
 *  - error handling
 *  - conversion of error codes to error messages
 *  - low-level memory allocation
 *  - time management
 *  - process times (user time, system time and real time).
 *  - heartbeat interrupt handling
 *  - user interrupt handling
 *  - access to OS environment variables
 *  - shell command
 *  - dynamic loading
 *  - dynamic C compilation
 *  - floating point environment setup
 *  - virtual memory statistics
 *  - filesystem path expansion
 *  - formatting of source code position
 *  - operations on I/O devices
 */

#define ___INCLUDED_FROM_OS
#define ___VERSION 40052
#include "gambit.h"

#include "os.h"
#include "setup.h"
#include "mem.h"
#include "c_intf.h"

#define USE_LINEEDIT
#define LINEEDIT_WITH_NONBLOCKING_IO
#define LINEEDIT_SUPPORT_F5_TO_F12
#define LINEEDIT_WITH_LOCAL_CLIPBOARD
#define LINEEDIT_REPORT_CHAR_ENCODING_ERRORS

#define ___CONSOLE_TERMINAL_TYPE_ENV_VAR "TERM"
#define ___EMACS_ENV_VAR "EMACS"

/*---------------------------------------------------------------------------*/

/* Conversion of OS error codes to Scheme error codes. */

#define ___ERR_CODE_EAGAIN ___FIX(___ERRNO_ERR(EAGAIN))
#define ___ERR_CODE_ENOENT ___FIX(___ERRNO_ERR(ENOENT))
#define ___ERR_CODE_ERROR_FILE_NOT_FOUND ___FIX(___WIN32_ERR(ERROR_FILE_NOT_FOUND))

#define SHOW_OS_ERRORS_no

#ifdef USE_errno

#ifdef SHOW_OS_ERRORS
___HIDDEN ___SCMOBJ err_code_from_errno_lineno
   ___P((int lineno),
        (lineno)
int lineno;)
#else
___HIDDEN ___SCMOBJ err_code_from_errno ___PVOID
#endif
{
  int e = errno;

#ifdef SHOW_OS_ERRORS
  printf ("\"%s\"@%d.1 -- %s\n", __FILE__, lineno, strerror (e));
  fflush (stdout);
#endif

  if (e == 0)
    return ___FIX(___UNKNOWN_ERR);

  return ___FIX(___ERRNO_ERR(e));
}

#define fnf_or_err_code_from_errno() err_code_from_errno()

#ifdef SHOW_OS_ERRORS
#define err_code_from_errno() err_code_from_errno_lineno(__LINE__)
#endif

#endif


#ifdef USE_h_errno

___HIDDEN ___SCMOBJ err_code_from_h_errno ___PVOID
{
  int e = h_errno;

  if (e == NETDB_INTERNAL)
    return err_code_from_errno ();

#ifdef NETDB_WORKS_PROPERLY

  if (e == NETDB_SUCCESS)
    return ___FIX(___UNKNOWN_ERR);

#else

  /* 
   * Linux sometimes returns NETDB_SUCCESS when it should return
   * NETDB_INTERNAL.
   */

  if (e == NETDB_SUCCESS)
    return err_code_from_errno ();

#endif

  return ___FIX(___H_ERRNO_ERR(e));
}

#endif


#ifdef USE_GetLastError

void error (char* where)//////////////////////////
{
  int code = GetLastError ();
  char buf[1000];
  DWORD len = FormatMessageA ////////////// add wide character support!
                (FORMAT_MESSAGE_FROM_SYSTEM|FORMAT_MESSAGE_MAX_WIDTH_MASK,
                 NULL,
                 code,
                 MAKELANGID(LANG_NEUTRAL,SUBLANG_DEFAULT),
                 buf,
                 1000,
                 NULL);
  printf ("error in %s, GetLastError=%d: %s\n", where, code, buf); fflush(stdout);
}

___HIDDEN ___SCMOBJ err_code_from_GetLastError ___PVOID
{
  DWORD e = GetLastError ();

  error ("err_code_from_GetLastError");///////////////////

  if (e == NO_ERROR)
    return ___FIX(___UNKNOWN_ERR);

  return ___FIX(___WIN32_ERR(e));
}

#define fnf_or_err_code_from_GetLastError() err_code_from_GetLastError()

#endif


#define USE_WSAGetLastErrorzzzzz/////////////////////////
#ifdef USE_WSAGetLastError

___HIDDEN ___SCMOBJ err_code_from_WSAGetLastError ___PVOID
{
  DWORD e = WSAGetLastError ();

  error ("err_code_from_WSAGetLastError");///////////////////

  if (e == NO_ERROR)
    return ___FIX(___UNKNOWN_ERR);

  return ___FIX(___WIN32_ERR(e));
}

#endif


#ifdef USE_OSErr

___HIDDEN ___ERR_CODE err_code_from_OSErr
   ___P((OSErr e),
        (e)
OSErr e;)
{
  if (e == noErr)
    return ___UNKNOWN_ERR;

  return ___OSERR_ERR(e);
}

#endif


/*---------------------------------------------------------------------------*/

/* OS specific initialization/finalization. */

#ifndef USE_MACOS_CLASSIC

___HIDDEN void setup_os_specific ___PVOID
{
}

___HIDDEN void cleanup_os_specific ___PVOID
{
}

#endif

#ifdef USE_MACOS_CLASSIC

/* Which features are available. */

___HIDDEN Boolean has_GetUTCDateTime = 0;
___HIDDEN Boolean has_GetDateTime = 0;
___HIDDEN Boolean has_ReadLocation = 0;
___HIDDEN Boolean has_Delay = 0;
___HIDDEN Boolean has_IdleUpdate = 0;
___HIDDEN Boolean has_WaitNextEvent = 0;
___HIDDEN Boolean has_OSDispatch = 0;
___HIDDEN Boolean has_FindFolder = 0;
___HIDDEN Boolean has_AliasMgr = 0;
___HIDDEN Boolean has_AppleEvents = 0;


#define test_bit(n,i) ((n)&(1<<(i)))


___HIDDEN TrapType get_trap_type
   ___P((short trap_num),
        (trap_num)
short trap_num;)
{
  /* OS traps start with A0, Tool traps with A8 or AA. */

  if (trap_num & 0x0800)
    return ToolTrap;
  else
    return OSTrap;
}


___HIDDEN short nb_toolbox_traps ___PVOID
{
  /* InitGraf (trap $A86E) is always implemented. */

  if (NGetTrapAddress (0xA86E, ToolTrap) == NGetTrapAddress (0xAA6E, ToolTrap))
    return (0x200);
  else
    return (0x400);
}


___HIDDEN Boolean trap_exists
  ___P((short trap_num),
       (trap_num)
short trap_num;)
{
  TrapType typ = get_trap_type (trap_num);
  if ((typ == ToolTrap) && ((trap_num &= 0x07FF) >= nb_toolbox_traps ()))
    return false;
  return (NGetTrapAddress (_Unimplemented, ToolTrap) !=
          NGetTrapAddress (trap_num, typ));
}


/* String conversion utilities. */

___HIDDEN Boolean c2pascal
   ___P((char *cstr,
         StringPtr pstr,
         int max_length),
        (cstr,
         pstr,
         max_length)
char *cstr;
StringPtr pstr;
int max_length;)
{
  StringPtr p1 = pstr+1;
  char *p2 = cstr;
  while (max_length > 0 && *p2 != '\0')
    {
       *p1++ = *p2++;
       max_length--;
    }
  if (*p2 != '\0')
    return 0;
  else
    {
      pstr[0] = p2-cstr;
      return 1;
    }
}


___HIDDEN Boolean pascal2c
   ___P((StringPtr pstr,
         char *cstr,
         int max_length),
        (pstr,
         cstr,
         max_length)
StringPtr pstr;
char *cstr;
int max_length;)
{
  char *p1 = cstr;
  StringPtr p2 = pstr+1;
  int len = pstr[0];
  if (len > max_length)
    return 0;
  else
    {
      while (len-- > 0)
        *p1++ = *p2++;
      *p1++ = '\0';
      return 1;
    }
}


#define DIR_SEPARATOR1 ':'
#define PARENT_HOP ":"

#define DIR_SEPARATOR(c)((c) == DIR_SEPARATOR1)
#define SEPARATOR(c)DIR_SEPARATOR(c)

___HIDDEN OSErr make_ResolvedFSSpec
   ___P((short vol,
         long dir,
         ConstStr255Param path,
         FSSpec *spec),
        (vol,
         dir,
         path,
         spec)
short vol;
long dir;
ConstStr255Param path;
FSSpec *spec;)
{
  OSErr err;
  Str255 name;
  StringPtr start = ___CAST(StringPtr,path+1);
  StringPtr end = start + path[0];
  StringPtr p1 = start;
  StringPtr p2 = name+1;
  CInfoPBRec pb;
  Boolean is_folder;
  Boolean is_aliased;

  if (!has_AliasMgr)
    return unimpErr;

  spec->vRefNum = vol;
  spec->parID = dir;

  do
    {
      p2 = name+1;
      while (p1 < end && DIR_SEPARATOR(*p1))  /* copy leading ':'s */
        *p2++ = *p1++;
      while (p1 < end && !DIR_SEPARATOR(*p1)) /* copy name that follows */
        *p2++ = *p1++;
      if (p1 < end && DIR_SEPARATOR(*p1))     /* end with a ':' if folder */
        *p2++ = DIR_SEPARATOR1;
      name[0] = p2 - (name+1);

      err = FSMakeFSSpec (spec->vRefNum, spec->parID, name, spec);
      if (err == fnfErr && p1 == end)
        return noErr;
      if (err != noErr)
        return err;

      if ((err = ResolveAliasFile (spec, 1, &is_folder, &is_aliased)) != noErr)
        return err;
      if (is_folder)
        {
          pb.dirInfo.ioNamePtr = spec->name;
          pb.dirInfo.ioVRefNum = spec->vRefNum;
          pb.dirInfo.ioDrDirID = spec->parID;
          pb.dirInfo.ioFDirIndex = 0;
          if ((err = PBGetCatInfoSync (&pb)) != noErr)
            return err;
          spec->parID = pb.hFileInfo.ioDirID;
          spec->name[0] = 0;
        }
      else if (p1 < end)
        return dirNFErr;
    } while (p1 < end);

  return noErr;
}


___HIDDEN OSErr ResolvedFSSpec_to_fullpath
   ___P((FSSpec *spec,
         StringPtr fullpath),
        (spec,
         fullpath)
FSSpec *spec;
StringPtr fullpath;)
{
  OSErr err;
  int i;
  Str255 result;
  StringPtr p = result + sizeof(result);
  CInfoPBRec pb;
  Str31 name;

  for (i = spec->name[0]; i > 0; i--)
    *--p = spec->name[i];

  pb.dirInfo.ioNamePtr = name;
  pb.dirInfo.ioVRefNum = spec->vRefNum;
  pb.dirInfo.ioDrParID = spec->parID;
  pb.dirInfo.ioFDirIndex = -1;

  do
    {
      pb.dirInfo.ioDrDirID = pb.dirInfo.ioDrParID;
      if ((err = PBGetCatInfoSync (&pb)) != noErr)
        return err;
      if (p-name[0]-1 < result)
        return bdNamErr; /* file name is too long */
      *--p = DIR_SEPARATOR1;
      for (i = name[0]; i > 0; i--)
        *--p = name[i];
    } while (pb.dirInfo.ioDrDirID != fsRtDirID);

  i = result + sizeof(result) - p;
  *fullpath++ = i;
  while (i > 0)
    {
      *fullpath++ = *p++;
      i--;
    }

  return noErr;
}


___HIDDEN ___SCMOBJ path_expand_to_absolute
   ___P((char *path,
         char *directory,///////////////////currently ignored
         char *new_path,
         long max_length),
        (path,
         directory,
         new_path,
         max_length)
char *path;
char *directory;
char *new_path;
long max_length)
{
  ___BOOL result = 0;
  FSSpec spec;
  short vol;
  long dir;
  char tmp[___PATH_MAX_LENGTH+1];
  Str255 ppath;

  if (path[0] == '~')
    {
      if (path[1] == '~')
        {
          /* "~~" or "~~:xxx..." */

          int i = 0;
          int j = 0;
          int sep = 0;
          char *tilde_dir;

          if (!has_FindFolder)
            goto ret;

          if (path[2]!='\0' && !DIR_SEPARATOR(path[2]))
            goto ret;

          tilde_dir = ___setup_params.gambcdir;
          if (tilde_dir == 0)
#ifdef ___GAMBCDIR
            tilde_dir = ___GAMBCDIR;
#else
            tilde_dir = ":Gambit-C";
#endif

          i += 2;

          while (*tilde_dir != '\0')
            if (j < ___PATH_MAX_LENGTH)
              {
                tmp[j] = *tilde_dir++;
                j++;
              }
            else
              goto ret;

          while (path[i] != '\0')
            if (j < ___PATH_MAX_LENGTH)
              {
                if (DIR_SEPARATOR(path[i]))
                  sep = 1;
                tmp[j++] = path[i++];
              }
            else
              goto ret;

          if (!sep)
            if (j < ___PATH_MAX_LENGTH)
              tmp[j++] = DIR_SEPARATOR1;
            else
              goto ret;

          tmp[j] = '\0';
          path = tmp;

          if (FindFolder (kOnSystemDisk,
                          kPreferencesFolderType,
                          0,
                          &vol,
                          &dir)
              != noErr)
            goto ret;
        }
      else if (path[1]!='\0' && !DIR_SEPARATOR(path[1]))
        {
          /* "~user" or "~user:xxx..." */

          goto ret; /* no equivalent on Macintosh */
        }
      else
        {
          /* "~" or "~:xxx..." */

          path++;
          vol = 0; /* use default volume and directory
                      (folder containing application) */
          dir = 0;
        }
    }
  else
    {
      vol = 0; /* use default volume and directory
                  (folder containing application) */
      dir = 0;
    }

  if (!c2pascal (path, ppath, 255) ||
      make_ResolvedFSSpec (vol, dir, ppath, &spec) != noErr ||
      ResolvedFSSpec_to_fullpath (&spec, ppath) != noErr ||
      !pascal2c (ppath, new_path, max_length))
    goto ret;

  result = 1;

 ret:

  return result;
}


___HIDDEN OSErr copy_file_sectors
   ___P((short src_refnum,
         short dst_refnum),
        (src_refnum,
         dst_refnum)
short src_refnum;
short dst_refnum;)
{
  OSErr err1, err2;
  char buf[2048];
  long count1, count2;

  do
    {
      count1 = sizeof (buf);
      err1 = FSRead (src_refnum, &count1, buf);
      if (err1 != noErr && err1 != eofErr)
        return err1;
      count2 = count1;
      err2 = FSWrite (dst_refnum, &count2, buf);
      if (err2 != noErr || count1 != count2)
        return err2;
    } while (err1 != eofErr);

  return noErr;
}


___HIDDEN OSErr copy_file
   ___P((FSSpec src_spec,
         FSSpec dst_spec),
        (src_spec,
         dst_spec)
FSSpec src_spec;
FSSpec dst_spec;)
{
  OSErr err, err2;
  short src_refnum, dst_refnum;
  FInfo src_info;

  if (((err = FSpDelete (&dst_spec)) == noErr || err == fnfErr) &&
      (err = FSpGetFInfo (&src_spec, &src_info)) == noErr &&
      (err = FSpCreate (&dst_spec, 0x3f3f3f3f, 0x3f3f3f3f, 0)) == noErr)
    {
      src_info.fdFlags = src_info.fdFlags & ~kHasBeenInited;
      if ((err = FSpSetFInfo (&dst_spec, &src_info) == noErr) &&
          (err = FSpOpenRF (&src_spec, fsRdPerm, &src_refnum) == noErr))
        {
          if ((err = FSpOpenRF (&dst_spec, fsWrPerm, &dst_refnum)) == noErr)
            {
              err = copy_file_sectors (src_refnum, dst_refnum);
              err2 = FSClose (dst_refnum);
              if (err == noErr)
                err = err2;
            }
          err2 = FSClose (src_refnum);
          if (err == noErr)
            err = err2;
          if (err == noErr &&
              (err = FSpOpenDF (&src_spec, fsRdPerm, &src_refnum) == noErr))
            {
              if ((err = FSpOpenDF (&dst_spec, fsWrPerm, &dst_refnum)) == noErr)
                {
                  err = copy_file_sectors (src_refnum, dst_refnum);
                  err2 = FSClose (dst_refnum);
                  if (err == noErr)
                    err = err2;
                }
              err2 = FSClose (src_refnum);
              if (err == noErr)
                err = err2;
            }
        }
      if (err != noErr)
        FSpDelete (&dst_spec);
    }

  return err;
}


___HIDDEN void setup_os_specific ___PVOID
{
  has_GetUTCDateTime = trap_exists (_UTCDateTime);
  has_GetDateTime = trap_exists (_GetDateTime);
  has_ReadLocation = trap_exists (_ReadLocation);
  has_Delay = trap_exists (_Delay);

  has_IdleUpdate = trap_exists (_IdleUpdate);
  has_WaitNextEvent = trap_exists (_WaitNextEvent);
  has_OSDispatch = trap_exists (_OSDispatch);

  {
    long response;

    has_FindFolder = (Gestalt (gestaltFindFolderAttr, &response) == noErr &&
                      test_bit (response, gestaltFindFolderPresent));
    has_AliasMgr = (Gestalt (gestaltAliasMgrAttr, &response) == noErr &&
                    test_bit (response, gestaltAliasMgrPresent));
    has_AppleEvents = (Gestalt (gestaltAppleEventsAttr, &response) == noErr &&
                       test_bit (response, gestaltAppleEventsPresent));
  }
}


___HIDDEN void cleanup_os_specific ___PVOID
{
}

#endif


/*---------------------------------------------------------------------------*/

/* Program startup */

program_startup_info ___program_startup_info =
{
  0,
  0
  PROGRAM_STARTUP_INFO_INIT_OS_SPECIFIC
};


___EXP_FUNC(int,___main_char)
  ___P((int argc,
        char *argv[],
        ___mod_or_lnk (*linker)(___global_state_struct*),
        char *script_line),
       (argc,
        argv,
        linker,
        script_line)
int argc;
char *argv[];
___mod_or_lnk (*linker)();
char *script_line;)
{
  int result;

  if (___NONNULLCHARSTRINGLIST_to_NONNULLUCS2STRINGLIST
        (argv,
         &___program_startup_info.argv)
      != ___FIX(___NO_ERR))
    result = ___EXIT_CODE_SOFTWARE;
  else
    {
      if (___CHARSTRING_to_UCS2STRING
            (script_line,
             &___program_startup_info.script_line)
          != ___FIX(___NO_ERR))
        result = ___EXIT_CODE_SOFTWARE;
      else
        {
          result = ___main (linker);

          ___free_UCS2STRING (___program_startup_info.script_line);
        }

      ___free_NONNULLUCS2STRINGLIST (___program_startup_info.argv);
    }

  return result;
}


___EXP_FUNC(int,___main_ucs2)
  ___P((int argc,
        ___UCS2STRING argv[],
        ___mod_or_lnk (*linker)(___global_state_struct*),
        char *script_line),
       (argc,
        argv,
        linker,
        script_line)
int argc;
___UCS2STRING argv[];
___mod_or_lnk (*linker)();
char *script_line;)
{
  int result;

  ___program_startup_info.argv = argv;

  if (___CHARSTRING_to_UCS2STRING
        (script_line,
         &___program_startup_info.script_line)
      != ___FIX(___NO_ERR))
    result = ___EXIT_CODE_SOFTWARE;
  else
    {
      result = ___main (linker);

      ___free_UCS2STRING (___program_startup_info.script_line);
    }

  return result;
}


#ifdef USE_WIN32

#ifdef UNICODE
#define ___CMDLINE_CE_SELECT(latin1,utf8,ucs2,ucs4,native) ucs2
#else
#define ___CMDLINE_CE_SELECT(latin1,utf8,ucs2,ucs4,native) latin1
#endif

___HIDDEN ___SCMOBJ parse_windows_command_line
   ___P((___CHAR_TYPE(___CMDLINE_CE_SELECT) *cmdline,
         ___UCS2STRING **argv_return),
        (cmdline,
         argv_return)
___CHAR_TYPE(___CMDLINE_CE_SELECT) *cmdline;
___UCS2STRING **argv_return;)
{
  int argc;
  ___UCS2STRING *argv;
  ___UCS2STRING args;
  int total_arg_len;
  int pass;

  for (pass=0; pass<2; pass++)
    {
      int in_double_quotes;
      int nb_backslashes;
      ___CHAR_TYPE(___CMDLINE_CE_SELECT) *p;
      ___CHAR_TYPE(___CMDLINE_CE_SELECT) c;

      if (pass != 0)
        {
          if ((argv = ___CAST(___UCS2STRING*,
                              ___alloc_mem ((argc + 1)
                                            * sizeof (___UCS2STRING)))) == 0)
            return ___FIX(___HEAP_OVERFLOW_ERR);

          if (total_arg_len > 0)
            {
              if ((args = ___CAST(___UCS2STRING,
                                  ___alloc_mem (total_arg_len
                                                * sizeof (___UCS2)))) == 0)
                {
                  ___free_mem (argv);
                  return ___FIX(___HEAP_OVERFLOW_ERR);
                }
            }
        }

      total_arg_len = 0;
      argc = 0;
      p = cmdline;

      for (;;)
        {
          while ((c = *p) != '\0' && c <= ' ')
            p++;

          if (c == '\0')
            break;

          in_double_quotes = 0;
          nb_backslashes = 0;

          if (pass != 0)
            argv[argc] = args;
            
          while ((c = *p) != '\0' && (in_double_quotes || c > ' '))
            {
              if (c == '\\')
                nb_backslashes++;
              else
                {
                  if (c != '"')
                    nb_backslashes = (nb_backslashes<<1) + 1;
                  else
                    {
                      if ((nb_backslashes & 1) == 0)
                        in_double_quotes ^= 1;
#ifndef PROCESS_PROGRAM_LIKE_OTHER_ARGS
                      if (argc == 0)
                        nb_backslashes = (nb_backslashes<<1) + 1;
#endif
                    }

                  total_arg_len += ((nb_backslashes+1)>>1);

                  if (pass != 0)
                    {
                      while (nb_backslashes > 1)
                        {
                          *args++ = '\\';
                          nb_backslashes -= 2;
                        }
                      if (nb_backslashes != 0)
                        *args++ = c;
                    }

                  nb_backslashes = 0;
                }
              p++;
            }

          total_arg_len += nb_backslashes+1;

          if (pass != 0)
            {
              while (nb_backslashes-- > 0)
                *args++ = '\\';
              *args++ = '\0';
            }

          argc++;
        }
    }

  argv[argc] = 0;

  *argv_return = argv;

  return ___FIX(___NO_ERR);
}


___HIDDEN void free_windows_command_line
   ___P((___UCS2STRING *argv),
        (argv)
___UCS2STRING *argv;)
{
  if (argv[0] != 0)
    ___free_mem (argv[0]);

  ___free_mem (argv);
}


___EXP_FUNC(int,___winmain_char)
  ___P((HINSTANCE hInstance,
        HINSTANCE hPrevInstance,
        LPSTR lpCmdLine,
        int nCmdShow,
        ___mod_or_lnk (*linker)(___global_state_struct*),
        char *script_line),
       (hInstance,
        hPrevInstance,
        lpCmdLine,
        nCmdShow,
        linker,
        script_line)
HINSTANCE hInstance;
HINSTANCE hPrevInstance;
LPSTR lpCmdLine;
int nCmdShow;
___mod_or_lnk (*linker)(___global_state_struct*);
char *script_line;)
{
  int result;

  if (parse_windows_command_line
        (GetCommandLine (),
         &___program_startup_info.argv)
      != ___FIX(___NO_ERR))
    result = ___EXIT_CODE_SOFTWARE;
  else
    {
      if (___CHARSTRING_to_UCS2STRING
            (script_line,
             &___program_startup_info.script_line)
          != ___FIX(___NO_ERR))
        result = ___EXIT_CODE_SOFTWARE;
      else
        {
          ___program_startup_info.hInstance     = hInstance;
          ___program_startup_info.hPrevInstance = hPrevInstance;
          ___program_startup_info.lpCmdLine     = lpCmdLine;
          ___program_startup_info.nCmdShow      = nCmdShow;

          result = ___main (linker);

          ___free_UCS2STRING (___program_startup_info.script_line);
        }

      free_windows_command_line (___program_startup_info.argv);
    }

  return result;
}

#endif


/*---------------------------------------------------------------------------*/

/* Process termination. */

void ___exit_process
   ___P((int status),
        (status)
int status;)
{
  exit (status);
}


/*---------------------------------------------------------------------------*/

/* Error handling. */

void ___display_error
   ___P((char **msgs),
        (msgs)
char **msgs;)
{
  if (___setup_params.display_error != 0)
    ___setup_params.display_error (msgs);
  else if (___DEBUG_SETTINGS_LEVEL(___setup_params.debug_settings) > 0)
    {
      while (*msgs != 0)
        fprintf (stderr, "%s", *msgs++);
    }
}


void ___fatal_error
   ___P((char **msgs),
        (msgs)
char **msgs;)
{
  if (___setup_params.fatal_error != 0)
    ___setup_params.fatal_error (msgs);
  else
    {
      char *new_msgs[100];
      int i;
      new_msgs[0] = "*** FATAL ERROR -- ";
      for (i=0; i<100-2; i++)
        {
          if (msgs[i] == 0)
            break;
          new_msgs[i+1] = msgs[i];
        }
      new_msgs[i+1] = "\n";
      new_msgs[i+2] = 0;
      ___display_error (new_msgs);
    }

  ___exit_process (___EXIT_CODE_SOFTWARE);
}


/*---------------------------------------------------------------------------*/

/* Conversion of error codes to error messages. */

___HIDDEN char *c_type_name_table[] =
{
  "int8",
  "unsigned-int8",
  "int16",
  "unsigned-int16",
  "int32",
  "unsigned-int32",
  "int64",
  "unsigned-int64",
  "float32",
  "float64",
  "char",
  "signed-char",
  "unsigned-char",
  "latin1",
  "ucs2",
  "ucs4",
  "short",
  "unsigned-short",
  "int",
  "unsigned-int",
  "long",
  "unsigned-long",
  "long-long",
  "unsigned-long-long",
  "float",
  "double",
  "struct",
  "union",
  "type",
  "pointer",
  "nonnull-pointer",
  "function",
  "nonnull-function",
  "bool",
  "char-string",
  "nonnull-char-string",
  "nonnull-char-string-list",
  "latin1-string",
  "nonnull-latin1-string",
  "nonnull-latin1-string-list",
  "utf8-string",
  "nonnull-utf8-string",
  "nonnull-utf8-string-list",
  "ucs2-string",
  "nonnull-ucs2-string",
  "nonnull-ucs2-string-list",
  "ucs4-string",
  "nonnull-ucs4-string",
  "nonnull-ucs4-string-list",
  "VARIANT",
  "(heap overflow)"
};


#ifdef USE_POSIX
#define ___ERR_CODE_CE_SELECT(latin1,utf8,ucs2,ucs4,native) native
#endif

#ifdef USE_WIN32
#ifdef UNICODE
#define ___ERR_CODE_CE_SELECT(latin1,utf8,ucs2,ucs4,native) ucs2
#else
#define ___ERR_CODE_CE_SELECT(latin1,utf8,ucs2,ucs4,native) latin1
#endif
#endif


___HIDDEN void append_charstring
   ___P((___STRING_TYPE(___ERR_CODE_CE_SELECT) buf,
         int *pos,
         char *str),
        (buf,
         pos,
         str)
___STRING_TYPE(___ERR_CODE_CE_SELECT) buf;
int *pos;
char *str;)
{
  int i = 0;
  int p = *pos;

  while (str[i] != '\0')
    {
      if (p >= ___ERR_MAX_LENGTH)
        break;
      buf[p++] = ___CAST(___CHAR_TYPE(___ERR_CODE_CE_SELECT),
                         ___CAST(unsigned char,str[i++]));
    }

  buf[p] = 0;

  *pos = p;
}


___SCMOBJ ___os_err_code_to_string
   ___P((___SCMOBJ err),
        (err)
___SCMOBJ err;)
{
  ___SCMOBJ e;
  ___SCMOBJ result;
  ___ERR_CODE err_code = ___INT(err);
  int facility = ___ERR_CODE_FACILITY(err_code);
  ___CHAR_TYPE(___ERR_CODE_CE_SELECT) buf[___ERR_MAX_LENGTH+1];
  int pos = 0;

  buf[0] = 0;

  if (facility >= ___ERR_CODE_FACILITY_SYSTEM)
    {
      /* System specific error code */

      if (err_code == ___UNWIND_C_STACK)
        append_charstring (buf, &pos, "C stack can't be unwound further");
      else if (err_code == ___SFUN_HEAP_OVERFLOW_ERR)
        append_charstring (buf, &pos, "Heap overflow while allocating stack marker");
      else if (err_code == ___IMPL_LIMIT_ERR)
        append_charstring (buf, &pos, "Implementation limit encountered");
      else if (err_code == ___HEAP_OVERFLOW_ERR)
        append_charstring (buf, &pos, "Heap overflow");
      else if (err_code == ___CLOSED_DEVICE_ERR)
        append_charstring (buf, &pos, "Device is closed");
      else if (err_code == ___INVALID_OP_ERR)
        append_charstring (buf, &pos, "Invalid operation");
      else if (err_code == ___MODULE_VERSION_TOO_OLD_ERR)
        append_charstring (buf, &pos, "Module was compiled with an older version of the compiler");
      else if (err_code == ___MODULE_VERSION_TOO_NEW_ERR)
        append_charstring (buf, &pos, "Module was compiled with a newer version of the compiler");
      else if (err_code == ___MODULE_ALREADY_LOADED_ERR)
        append_charstring (buf, &pos, "Can't load a given object file more than once");
      else if (err_code == ___DYNAMIC_LOADING_NOT_AVAILABLE_ERR)
        append_charstring (buf, &pos, "Dynamic loading is not available on this platform");
      else if (err_code == ___DYNAMIC_LOADING_LOOKUP_ERR)
        append_charstring (buf, &pos, "The object file did not contain the required function");
      else if ((err_code >= ___STOC_BASE && err_code <= ___STOC_MAX) ||
               (err_code >= ___CTOS_BASE && err_code <= ___CTOS_MAX))
        {
          int arg_num, c_type_index;
          char *dir;
          if (err_code <= ___STOC_MAX)
            {
              arg_num = (err_code-___STOC_BASE) & ((1<<7)-1);
              c_type_index = (err_code-___STOC_BASE) >> 7;
              dir = "to C";
            }
          else
            {
              arg_num = (err_code-___CTOS_BASE) & ((1<<7)-1);
              c_type_index = (err_code-___CTOS_BASE) >> 7;
              dir = "from C";
            }
          if (arg_num == ___RETURN_POS)
            append_charstring (buf, &pos, "Can't convert result ");
          else if (arg_num == 0)
            append_charstring (buf, &pos, "Can't convert ");
          else
            {
              char digit[2];
              int d = 1;

              while (d < arg_num/10)
                d *= 10;

              append_charstring (buf, &pos, "(Argument ");

              digit[1] = '\0';
              while (d > 0)
                {
                  digit[0] = (arg_num / d % 10) + '0';
                  append_charstring (buf, &pos, digit);
                  d /= 10;
                }

              append_charstring (buf, &pos, ") Can't convert ");
            }
          append_charstring (buf, &pos, dir);
          append_charstring (buf, &pos, " ");
          append_charstring (buf, &pos, c_type_name_table[c_type_index]);
        }
      else
        append_charstring (buf, &pos, "Unknown error");
    }
  else if (facility >= ___ERR_CODE_FACILITY_MACOS)
    {
      /* MACOS error code */

      append_charstring (buf, &pos, "Unknown MACOS error");
    }
  else if (facility >= ___ERR_CODE_FACILITY_ERRNO)
    {
      /* ANSI-C errno error code */

#ifdef USE_errno

      char *msg = strerror (___ERRNO_FROM_ERR_CODE(err_code));

      if (msg == NULL)
        msg = "Unknown error";

      append_charstring (buf, &pos, msg);

#endif
    }
  else if (facility >= ___ERR_CODE_FACILITY_H_ERRNO)
    {
      /* netdb h_errno error code */

#ifdef USE_h_errno

      char *msg;
      int i = ___H_ERRNO_FROM_ERR_CODE(err_code);

#ifdef USE_hstrerror

      msg = ___CAST(char*,hstrerror (i));

#else

      static char *h_errno_messages[] =
      {
        "Resolver Error 0 (no error)",
        "Unknown host",
        "Host name lookup failure",
        "Unknown server error",
        "No address associated with name"
      };

      if (i < 0 || i > 4)
        msg = "Unknown resolver error";
      else
        msg = h_errno_messages[i];

#endif

      append_charstring (buf, &pos, msg);

#endif
    }
  else
    {
      /* Windows HRESULT error code */

#ifdef USE_FormatMessage

      DWORD len =
        FormatMessage (FORMAT_MESSAGE_FROM_SYSTEM |
                       FORMAT_MESSAGE_MAX_WIDTH_MASK,
                       NULL,
                       ___WIN32_FROM_ERR_CODE(err_code),
                       MAKELANGID(LANG_NEUTRAL,SUBLANG_DEFAULT),
                       buf,
                       ___ERR_MAX_LENGTH,
                       NULL);

      if (len == 0)
        buf[0] = 0;

#endif
    }

  if ((e = ___NONNULLSTRING_to_SCMOBJ
             (buf,
              &result,
              ___RETURN_POS,
              ___CE(___ERR_CODE_CE_SELECT)))
      != ___FIX(___NO_ERR))
    result = e;

  return result;
}


/*---------------------------------------------------------------------------*/

/* Low-level memory allocation. */

#ifdef ___DEBUG_ALLOC_MEM

___HIDDEN long alloc_mem_calls = 0;
___HIDDEN long free_mem_calls = 0;

#endif


___HIDDEN void setup_alloc_mem ___PVOID
{
#ifdef ___DEBUG_ALLOC_MEM

  alloc_mem_calls = 0;
  free_mem_calls = 0;

#endif
}


___HIDDEN void cleanup_alloc_mem ___PVOID
{
#ifdef ___DEBUG_ALLOC_MEM

  if (alloc_mem_calls != free_mem_calls)
    fprintf (stderr,
             "*** MEMORY LEAK: alloc_mem_calls = %ld  free_mem_calls = %ld\n",
             alloc_mem_calls,
             free_mem_calls);

#endif
}


void *___alloc_mem
   ___P((unsigned long bytes),
        (bytes)
unsigned long bytes;)
{
  void *ptr;

#ifdef ___DEBUG_ALLOC_MEM
#ifdef USE_WIN32

  InterlockedIncrement (&alloc_mem_calls);

#else

  alloc_mem_calls++;

#endif
#endif

#ifdef USE_TempNewHandle

  if (has_OSDispatch)
    {
      OSErr e;
      Ptr p;
      Handle h = TempNewHandle (sizeof (Handle) + bytes, &err);
      if (e != noErr || h == 0)
        return 0;
      HLock (h);
      p = *h;
      *___CAST(Handle*,p) = h;
      ptr = p + sizeof (Handle);
    }
  else
    ptr = malloc (bytes);

#else

  ptr = malloc (bytes);

#endif

#ifdef ___DEBUG_ALLOC_MEM_TRACE
  printf ("%p allocated (%lu bytes)\n", ptr, bytes);
#endif

  return ptr;
}


void ___free_mem
   ___P((void *ptr),
        (ptr)
void *ptr;)
{
#ifdef ___DEBUG_ALLOC_MEM_TRACE
  printf ("%p freed\n", ptr);
#endif

#ifdef ___DEBUG_ALLOC_MEM
#ifdef USE_WIN32

  InterlockedIncrement (&free_mem_calls);

#else

  free_mem_calls++;

#endif
#endif

#ifdef USE_TempNewHandle

  if (has_OSDispatch)
    {
      OSErr e;
      Handle h = *___CAST(Handle*,___CAST(Ptr,ptr) - sizeof (Handle));
      HUnlock (h);
      TempDisposeHandle (h, &e);
    }

#endif

  free (ptr);
}


/*---------------------------------------------------------------------------*/

/* Time management. */

#define TIME_REFERENCE_POINT 1970

#define DAYS_UNTIL_JAN_1_2000(year) \
(730119 - ((year)-1)*365 - ((year)-1)/4 + ((year)-1)/100 - ((year)-1)/400)

#define DAYS_UNTIL_TIME_REFERENCE_POINT(year) \
(DAYS_UNTIL_JAN_1_2000(year) - DAYS_UNTIL_JAN_1_2000(TIME_REFERENCE_POINT))

#define JAN_1(year)(DAYS_UNTIL_TIME_REFERENCE_POINT(year) * 24 * 60 * 60)

#ifdef DIVISION_BY_ZERO_WORKS_PROPERLY
#define POS_INFINITY (1.0/0.0)  /* positive infinity */
#define NEG_INFINITY (-1.0/0.0) /* negative infinity */
#else
#define POS_INFINITY (1.7976931348623157e308)  /* positive infinity */
#define NEG_INFINITY (-1.7976931348623157e308) /* negative infinity */
#endif

#ifdef ___FLOAT_TIME_REPRESENTATION

#define TIME_POS_INFINITY POS_INFINITY
#define TIME_NEG_INFINITY NEG_INFINITY
#define TIME_ZERO 0.0

#endif

#ifdef ___INT_TIME_REPRESENTATION

#define TIME_POS_INFINITY { 2147483647, 999999999 }
#define TIME_NEG_INFINITY { -2147483648, 0 }
#define TIME_ZERO { 0, 0 }

#endif


___time ___time_pos_infinity = TIME_POS_INFINITY;
___time ___time_neg_infinity = TIME_NEG_INFINITY;
___time ___time_zero = TIME_ZERO;


void ___time_from_nsecs
   ___P((___time *tim,
         ___SM32 secs,
         ___SM32 nsecs),
        (tim,
         secs,
         nsecs)
___time *tim;
___SM32 secs;
___SM32 nsecs;)
{
#ifdef ___FLOAT_TIME_REPRESENTATION
  *tim = secs + nsecs / 1000000000.0;
#endif

#ifdef ___INT_TIME_REPRESENTATION
  tim->secs = secs + nsecs / 1000000000;
  tim->nsecs = nsecs % 1000000000;
#endif
}


void ___time_add
   ___P((___time *tim1,
         ___time tim2),
        (tim1,
         tim2)
___time *tim1;
___time tim2;)
{
#ifdef ___FLOAT_TIME_REPRESENTATION
  *tim1 += tim2;
#endif

#ifdef ___INT_TIME_REPRESENTATION
  ___SM32 total_secs = tim1->secs + tim2.secs;
  ___SM32 total_nsecs = tim1->nsecs + tim2.nsecs;
  tim1->secs += tim2.secs + total_nsecs / 1000000000;
  tim1->nsecs = total_nsecs % 1000000000;
#endif
}


void ___time_subtract
   ___P((___time *tim1,
         ___time tim2),
        (tim1,
         tim2)
___time *tim1;
___time tim2;)
{
#ifdef ___FLOAT_TIME_REPRESENTATION
  *tim1 -= tim2;
#endif

#ifdef ___INT_TIME_REPRESENTATION
  ___SM32 total_nsecs;

  if (tim2.nsecs > tim1->nsecs)
    {
      tim1->secs -= tim2.secs + 1;
      tim1->nsecs += 1000000000 - tim2.nsecs;
    }
  else
    {
      tim1->secs -= tim2.secs;
      tim1->nsecs -= tim2.nsecs;
    }
#endif
}


___BOOL ___time_equal
   ___P((___time tim1,
         ___time tim2),
        (tim1,
         tim2)
___time tim1;
___time tim2;)
{
#ifdef ___FLOAT_TIME_REPRESENTATION
  return (tim1 == tim2);
#endif

#ifdef ___INT_TIME_REPRESENTATION
  return (tim1.secs == tim2.secs && tim1.nsecs == tim2.nsecs);
#endif
}


___BOOL ___time_less
   ___P((___time tim1,
         ___time tim2),
        (tim1,
         tim2)
___time tim1;
___time tim2;)
{
#ifdef ___FLOAT_TIME_REPRESENTATION
  return (tim1 < tim2);
#endif

#ifdef ___INT_TIME_REPRESENTATION
  return (tim1.secs < tim2.secs ||
          (tim1.secs == tim2.secs && tim1.nsecs < tim2.nsecs));
#endif
}


___BOOL ___time_positive
   ___P((___time tim),
        (tim)
___time tim;)
{
#ifdef ___FLOAT_TIME_REPRESENTATION
  return (tim > 0.0);
#endif

#ifdef ___INT_TIME_REPRESENTATION
  return (tim.secs > 0 ||
          (tim.secs == 0 && tim.nsecs > 0));
#endif
}


void ___time_get_current_time
   ___P((___time *tim),
        (tim)
___time *tim;)
{
#ifndef USE_clock_gettime
#ifndef USE_getclock
#ifndef USE_GetSystemTime
#ifndef USE_gettimeofday
#ifndef USE_ftime
#ifndef USE_time
#ifndef USE_MACOS_CLASSIC

  *tim = ___time_neg_infinity;

#endif
#endif
#endif
#endif
#endif
#endif
#endif

#ifdef USE_clock_gettime

  /* The clock_gettime function is in POSIX.1b (realtime programming). */

  struct timespec ts;
  if (clock_gettime (CLOCK_REALTIME, &ts) == 0)
    ___time_from_nsecs (tim, ts.tv_sec-JAN_1(1970), ts.tv_nsec);
  else
    *tim = ___time_neg_infinity;

#endif

#ifdef USE_getclock

  /* The getclock function is in the Tru64 UNIX standard C library. */

  struct timespec ts;
  if (getclock (TIMEOFDAY, &ts) == 0)
    ___time_from_nsecs (tim, ts.tv_sec-JAN_1(1970), ts.tv_nsec);
  else
    *tim = ___time_neg_infinity;

#endif

#ifdef USE_GetSystemTime

  SYSTEMTIME st;
  FILETIME ft;
  GetSystemTime (&st);
  if (SystemTimeToFileTime (&st, &ft))
    {
      LONGLONG x = *___CAST(LONGLONG*,&ft);
      ___SM32 secs = x / 10000000 - JAN_1(1601);
      ___SM32 nsecs = x % 10000000 * 100;
      ___time_from_nsecs (tim, secs, nsecs);
    }
  else
    *tim = ___time_neg_infinity;

#endif

#ifdef USE_gettimeofday

  struct timeval tv;
  if (gettimeofday (&tv, NULL) == 0)
    ___time_from_nsecs (tim, tv.tv_sec-JAN_1(1970), tv.tv_usec*1000);
  else
    *tim = ___time_neg_infinity;

#endif

#ifdef USE_ftime

  struct timeb tb;
  ftime (&tb);
  ___time_from_nsecs (tim, tb.time-JAN_1(1970), tb.millitm*1000000);

#endif

#ifdef USE_time

  time_t t;
  time (&t);
  ___time_from_nsecs (tim, t-JAN_1(1970), 0);

#endif

#ifdef USE_MACOS_CLASSIC

  if (has_GetUTCDateTime)
    {
      UTCDateTime dt;

      if (GetUTCDateTime (&dt, kUTCDefaultOptions) == noErr)
        ___time_from_nsecs (tim,
                            dt.lowSeconds-JAN_1(1904),
                            ___CAST(___SM32,dt.fraction)*15258 +
                            ((___CAST(___SM32,dt.fraction)*101)>>7));
      else
        *tim = ___time_neg_infinity;
    }
  else if (has_GetDateTime)
    {
      unsigned long t;

      GetDateTime (&t);  /* Get seconds since 01-01-1904 in local time. */

      if (has_ReadLocation)
        {
          /* Adjust to UTC based on machine location. */

          MachineLocation here;
          long offset;

          ReadLocation (&here);
          offset = here.u.gmtDelta & 0x00ffffff;
          if (offset & 0x00800000)
            offset |= 0xff000000;
          t -= offset;
        }

      ___time_from_nsecs (tim, t-JAN_1(1904), 0);
    }
  else
    *tim = ___time_neg_infinity;

#endif
}


void ___time_to_seconds
   ___P((___time tim,
         ___F64 *seconds),
        (tim,
         seconds)
___time tim;
___F64 *seconds;)
{
#ifdef ___FLOAT_TIME_REPRESENTATION
  *seconds = tim;
#endif

#ifdef ___INT_TIME_REPRESENTATION
  *seconds = tim.secs + tim.nsecs / 1000000000.0;
#endif
}


void ___time_from_seconds
   ___P((___time *tim,
         ___F64 seconds),
        (tim,
         seconds)
___time *tim;
___F64 seconds;)
{
#ifdef ___FLOAT_TIME_REPRESENTATION
  *tim = seconds;
#endif

#ifdef ___INT_TIME_REPRESENTATION
  if (seconds >= 2147483648.0) /* more than 68 years is pos infinity! */
    *tim = ___time_pos_infinity;
  else if (seconds < -2147483648.0) /* less than -68 years is neg infinity! */
    *tim = ___time_neg_infinity;
  else
    {
      ___SM32 secs = seconds;
      ___SM32 nsecs = (seconds - secs) * 1000000000.0;
      if (nsecs < 0)
        {
          secs--;
          nsecs += 1000000000;
        }
      tim->secs = secs;
      tim->nsecs = nsecs;
    }
#endif
}


___HIDDEN void absolute_time_to_relative_time
   ___P((___time tim,
         ___time *rtim),
        (tim,
         rtim)
___time tim;
___time *rtim;)
{
  if (___time_less (tim, ___time_pos_infinity))
    {
      ___time now;
      *rtim = tim;
      ___time_get_current_time (&now);
      ___time_subtract (rtim, now);
    }
  else
    *rtim = ___time_pos_infinity;
}


#ifndef LINEEDIT_WITH_NONBLOCKING_IO

___HIDDEN void absolute_time_sleep
   ___P((___time tim),
        (tim)
___time tim;)
{
  absolute_time_to_relative_time (tim, &tim);

  if (___time_positive (tim))
    {
      ___SM32 nsecs;

#ifdef ___FLOAT_TIME_REPRESENTATION
      if (tim > 2.0)
        nsecs = 2000000000;
      else
        nsecs = tim * 1000000000.0;
#endif

#ifdef ___INT_TIME_REPRESENTATION
      if (tim.secs > 2)
        nsecs = 2000000000;
      else
        nsecs = tim.secs * 1000000000 + tim.nsecs;
#endif

#ifndef USE_nanosleep
#ifndef USE_Sleep
#ifndef USE_sleep
#ifndef USE_MACOS_CLASSIC
      /* Can't sleep... so stay awake! */
#endif
#endif
#endif
#endif

#ifdef USE_nanosleep
      {
        /* The nanosleep function is in POSIX.1b (realtime programming). */
        struct timespec ts;
        ts.tv_sec = nsecs / 1000000000;
        ts.tv_nsec = nsecs;
        nanosleep (&ts, NULL);
      }
#endif

#ifdef USE_Sleep
      Sleep (nsecs / 1000000);
#endif

#ifdef USE_sleep
      sleep (nsecs / 1000000000);
#endif

#ifdef USE_MACOS_CLASSIC
      {
        ___U32 ticks = nsecs / 16666666;
        if (has_IdleUpdate)
          IdleUpdate ();
        if (has_WaitNextEvent)
          {
            EventRecord er;
            WaitNextEvent (0, &er, ticks, NULL);
          }
        else if (has_Delay)
          Delay (ticks, &ticks);
      }
#endif
    }
}

#endif


#ifdef USE_select

___HIDDEN void absolute_time_to_nonnegative_timeval
   ___P((___time tim,
         struct timeval **tv),
        (tim,
         tv)
___time tim;
struct timeval **tv;)
{
  if (___time_less (tim, ___time_pos_infinity))
    {
      struct timeval *t = *tv;

      if (___time_positive (tim))
        {
#ifdef ___FLOAT_TIME_REPRESENTATION
          if (tim >= 2147483648.0) /* upper bound is 68 years! */
            {
              t->tv_sec = 2147483647;
              t->tv_usec = 999999;
            }
          else
            {
              t->tv_sec = ___CAST(int,tim);
              t->tv_usec = ___CAST(int,(tim - t->tv_sec) * 1000000.0);
            }
#endif

#ifdef ___INT_TIME_REPRESENTATION
          t->tv_sec = tim.secs;
          t->tv_usec = tim.nsecs / 1000;
#endif
        }
      else
        {
          t->tv_sec = 0;
          t->tv_usec = 0;
        }
    }
  else
    *tv = NULL;
}

#endif


/*---------------------------------------------------------------------------*/

/* Process times (user time, system time and real time). */

___HIDDEN ___F64 process_start_seconds; /* process start in real time */

void ___process_times
   ___P((___F64 *user,
         ___F64 *sys,
         ___F64 *real),
        (user,
         sys,
         real)
___F64 *user;
___F64 *sys;
___F64 *real;)
{
#ifndef USE_GetProcessTimes
#ifndef USE_getrusage
#ifndef USE_times
#ifndef USE_clock
#ifndef USE_DosQuerySysInfo
#ifndef USE_MACOS_CLASSIC

  *user = 0.0; /* can't get time... result is 0 */
  *sys = 0.0;

#endif
#endif
#endif
#endif
#endif
#endif

#ifdef USE_GetProcessTimes

  HANDLE p;
  FILETIME creation_time, exit_time, sys_time, user_time;

  p = GetCurrentProcess ();

  if (GetProcessTimes (p, &creation_time, &exit_time, &sys_time, &user_time))
    {
      *user = (user_time.dwHighDateTime * 4294967296.0 +
               user_time.dwLowDateTime) / 1.0e7;
      *sys = (sys_time.dwHighDateTime * 4294967296.0 +
              sys_time.dwLowDateTime) / 1.0e7;
    }
  else
    {
      *user = 0.0; /* can't get time... result is 0 */
      *sys = 0.0;
    }

#endif

#ifdef USE_getrusage

  struct rusage ru;

  if (getrusage (RUSAGE_SELF, &ru) == 0)
    {
      *user = ru.ru_utime.tv_sec + ___CAST(___F64,ru.ru_utime.tv_usec) / 1.0e6;
      *sys = ru.ru_stime.tv_sec + ___CAST(___F64,ru.ru_stime.tv_usec) / 1.0e6;
    }
  else
    {
      *user = 0.0; /* can't get time... result is 0 */
      *sys = 0.0;
    }

#endif

#ifdef USE_times

#ifdef CLK_TCK
#define CLOCK_TICKS_PER_SEC CLK_TCK
#else
#ifdef USE_sysconf
#ifdef _SC_CLK_TCK
#define CLOCK_TICKS_PER_SEC sysconf (_SC_CLK_TCK)
#else
#ifdef _SC_CLOCKS_PER_SEC
#define CLOCK_TICKS_PER_SEC sysconf (_SC_CLOCKS_PER_SEC)
#endif
#endif
#endif
#endif

#ifndef CLOCK_TICKS_PER_SEC
@error Cannot find a definition for CLOCK_TICKS_PER_SEC
#endif

  struct tms t;

  static long clock_ticks_per_sec = 0;

  if (clock_ticks_per_sec == 0)
    clock_ticks_per_sec = CLOCK_TICKS_PER_SEC;

  times (&t);

  *user = ___CAST(___F64,t.tms_utime) / clock_ticks_per_sec;
  *sys = ___CAST(___F64,t.tms_stime) / clock_ticks_per_sec;

#endif

#ifdef USE_clock

  *user = ___CAST(___F64,clock ()) / CLOCKS_PER_SEC;
  *sys = 0.0; /* fake system time */

#endif

#ifdef USE_DosQuerySysInfo

  static ___U32 origin = 0;
  ___U32 now;

  while (origin == 0)
    DosQuerySysInfo (QSV_MS_COUNT, QSV_MS_COUNT, &origin, sizeof (origin));

  DosQuerySysInfo (QSV_MS_COUNT, QSV_MS_COUNT, &now, sizeof (now));

  *user = ___CAST(___F64,(now-origin)) / 1.0e3; /* are the units right? */
  *sys = 0.0; /* fake system time */

#endif

#ifdef USE_MACOS_CLASSIC

  static ___U32 origin = 0;
  ___U32 now;

  while (origin == 0)
    origin = TickCount ();

  now = TickCount ();

  *user = ___CAST(___F64,(now-origin)) / 60;
  *sys = 0.0; /* fake system time */

#endif

  {
    ___time now;
    ___F64 seconds;

    ___time_get_current_time (&now);
    ___time_to_seconds (now, &seconds);

    *real = seconds - process_start_seconds;
  }
}


___HIDDEN void setup_process_times ___PVOID
{
  ___F64 user, sys;

  process_start_seconds = 0.0;
  ___process_times (&user, &sys, &process_start_seconds);

}


___HIDDEN void cleanup_process_times ___PVOID
{
}


/*---------------------------------------------------------------------------*/

/* Heartbeat interrupt handling. */

/*
 * Use the process-time timer unless only the real-time timer is
 * available (e.g. DJGPP).  Note that on some systems (e.g. MkLinux)
 * ITIMER_VIRTUAL is an enum type, not a macro.
 */

#ifdef ITIMER_VIRTUAL
#define HEARTBEAT_ITIMER ITIMER_VIRTUAL
#define HEARTBEAT_SIG SIGVTALRM
#else
#ifdef ITIMER_REAL
#define HEARTBEAT_ITIMER ITIMER_REAL
#define HEARTBEAT_SIG SIGALRM
#else
#define HEARTBEAT_ITIMER ITIMER_VIRTUAL
#define HEARTBEAT_SIG SIGVTALRM
#endif
#endif


#ifndef USE_setitimer
#ifndef USE_dos_setvect_1Ch
#ifndef USE_DosStartTimer
#ifndef USE_VInstall
#ifndef USE_CreateThread

___F64 ___set_heartbeat_interval
   ___P((___F64 seconds),
        (seconds)
___F64 seconds;)
{
  return 0.0;
}


___HIDDEN void disable_heartbeat_interrupt ___PVOID
{
}


___HIDDEN void enable_heartbeat_interrupt ___PVOID
{
}


___HIDDEN void setup_heartbeat_interrupt_handling ___PVOID
{
}


___HIDDEN void cleanup_heartbeat_interrupt_handling ___PVOID
{
}


#endif
#endif
#endif
#endif
#endif


/*   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   */

#ifdef USE_setitimer

___HIDDEN void heartbeat_interrupt_handler
   ___P((int sig),
        (sig)
int sig;)
{
#ifdef USE_signal
  signal (HEARTBEAT_SIG, heartbeat_interrupt_handler);
#endif
  ___raise_interrupt (___INTR_HEARTBEAT);
}


___F64 ___set_heartbeat_interval
   ___P((___F64 seconds),
        (seconds)
___F64 seconds;)
{
  struct itimerval tv;

  if (seconds < 0.0) /* turn heartbeat off */
    {
      tv.it_interval.tv_sec  = 0;
      tv.it_interval.tv_usec = 0;
      tv.it_value.tv_sec     = 0;
      tv.it_value.tv_usec    = 0;
      setitimer (HEARTBEAT_ITIMER, &tv, 0);
      return 0.0;
    }
  else
    {
      int secs;
      int usecs;

      if (seconds >= 2147483648.0) /* upper bound is 68 years! */
        {
          secs = 2147483647;
          usecs = 999999;
        }
      else
        {
          secs = ___CAST(int,seconds);
          usecs = ___CAST(int,(seconds - secs) * 1000000.0);
          if (secs <= 0 && usecs == 0) /* use smallest interval */
            {
              secs = 0;
              usecs = 1;
            }
        }

      tv.it_interval.tv_sec  = secs;
      tv.it_interval.tv_usec = usecs;
      tv.it_value.tv_sec     = secs;
      tv.it_value.tv_usec    = usecs;
      setitimer (HEARTBEAT_ITIMER, &tv, 0);
      getitimer (HEARTBEAT_ITIMER, &tv);
      return tv.it_interval.tv_sec + tv.it_interval.tv_usec / 1000000.0;
    }
}


___HIDDEN void disable_heartbeat_interrupt ___PVOID
{
#ifdef USE_sigaction
  struct sigaction act;
  act.sa_handler = SIG_IGN;
  act.sa_flags = 0;
  sigemptyset (&act.sa_mask);
  sigaction (HEARTBEAT_SIG, &act, 0);
#endif

#ifdef USE_signal
  signal (HEARTBEAT_SIG, SIG_IGN);
#endif
}


___HIDDEN void enable_heartbeat_interrupt ___PVOID
{
#ifdef USE_sigaction
  struct sigaction act;
  act.sa_handler = heartbeat_interrupt_handler;
  act.sa_flags = 0;
#ifdef SA_INTERRUPT
  act.sa_flags |= SA_INTERRUPT;
#endif
  sigemptyset (&act.sa_mask);
  sigaction (HEARTBEAT_SIG, &act, 0);
#endif

#ifdef USE_signal
  signal (HEARTBEAT_SIG, heartbeat_interrupt_handler);
#endif
}


___HIDDEN void setup_heartbeat_interrupt_handling ___PVOID
{
  ___set_heartbeat_interval (-1.0);
  enable_heartbeat_interrupt ();
}


___HIDDEN void cleanup_heartbeat_interrupt_handling ___PVOID
{
  ___set_heartbeat_interval (-1.0);
  disable_heartbeat_interrupt ();
}


#endif


/*   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   */

#ifdef USE_dos_setvect_1Ch

/* 1Ch = timer tick vector */
___HIDDEN ___BOOL heartbeat_enabled;
___HIDDEN void (__interrupt __far *prev_vector_1Ch) ___PVOID;
___HIDDEN long heartbeat_interval;
___HIDDEN long heartbeat_countdown;


___HIDDEN void __interrupt __far heartbeat_interrupt_handler ___PVOID
{
  if (heartbeat_countdown > 1)
    heartbeat_countdown--;
  else if (heartbeat_countdown == 1 && heartbeat_enabled)
    {
      heartbeat_countdown = heartbeat_interval;
      ___raise_interrupt (___INTR_HEARTBEAT);
    }
  _chain_intr (prev_vector_1Ch);
}


___F64 ___set_heartbeat_interval
   ___P((___F64 seconds),
        (seconds)
___F64 seconds;)
{
  if (seconds < 0.0) /* turn heartbeat off */
    {
      _dos_setvect (0x1C, prev_vector_1Ch);
      return 0.0;
    }
  else
    {
      ___F64 ticks = seconds * 18.2; /* 18.2 ticks/sec */
      int t;

      if (ticks >= 2147483648.0) /* upper bound is 3.7 years! */
        t = 2147483647;
      else
        {
          t = ticks;
          if (t <= 0) /* use smallest interval */
            t = 1;
        }

      heartbeat_interval = t;
      heartbeat_countdown = heartbeat_interval;
      _dos_setvect (0x1C, heartbeat_interrupt_handler);
      return t / 18.2;
    }
}


___HIDDEN void disable_heartbeat_interrupt ___PVOID
{
  heartbeat_enabled = 0;
}


___HIDDEN void enable_heartbeat_interrupt ___PVOID
{
  heartbeat_enabled = 1;
}


___HIDDEN void setup_heartbeat_interrupt_handling ___PVOID
{
  prev_vector_1Ch = _dos_getvect (0x1C);
  heartbeat_enabled = 1;
  ___set_heartbeat_interval (-1.0);
}


___HIDDEN void cleanup_heartbeat_interrupt_handling ___PVOID
{
  ___set_heartbeat_interval (-1.0);
  heartbeat_enabled = 0;
}


#endif


/*   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   */

#ifdef USE_DosStartTimer

___HIDDEN ___BOOL heartbeat_enabled;
___HIDDEN ___BOOL heartbeat_hev;
___HIDDEN ___BOOL heartbeat_tid;
___HIDDEN ___BOOL heartbeat_htimer;
___HIDDEN HEV    hev;
___HIDDEN TID    tid;
___HIDDEN HTIMER htimer;


___HIDDEN void APIENTRY heartbeat_thread_code
   ___P((ULONG param),
        (param)
ULONG param;)
{
  ULONG pc = 0;
  for (;;)
    {
      DosWaitEventSem (hev, SEM_INDEFINITE_WAIT);
      DosResetEventSem (hev, &pc);
      if (heartbeat_enabled)
        {
          DosEnterCritSec ();
          ___raise_interrupt (___INTR_HEARTBEAT);
          DosExitCritSec ();
        }
    }
}


___F64 ___set_heartbeat_interval
   ___P((___F64 seconds),
        (seconds)
___F64 seconds;)
{
  if (!heartbeat_hev || !heartbeat_tid)
    return 0.0;
  if (heartbeat_htimer)
    {
      DosStopTimer (htimer);
      heartbeat_htimer = 0;
    }
  if (seconds < 0.0) /* turn heartbeat off */
    return 0.0;
  else
    {
      ___F64 msecs = seconds * 1000.0;
      ULONG m;

      if (msecs >= 4294967295.0) /* upper bound is 49.7 days! */
        m = 4294967295;
      else
        {
          m = msecs;
          if (m == 0) /* use smallest interval */
            m = 1;
        }

      if (DosStartTimer (m, ___CAST(HSEM,hev), &htimer) == NO_ERROR)
        {
          heartbeat_htimer = 1;
          return m / 1000.0;
        }
      else
        return 0.0;
    }
}


___HIDDEN void disable_heartbeat_interrupt ___PVOID
{
  heartbeat_enabled = 0;
}


___HIDDEN void enable_heartbeat_interrupt ___PVOID
{
  heartbeat_enabled = 1;
}


___HIDDEN void setup_heartbeat_interrupt_handling ___PVOID
{
  heartbeat_enabled = 0;
  heartbeat_hev     = 0;
  heartbeat_tid     = 0;
  heartbeat_htimer  = 0;
  if (DosCreateEventSem (0, &hev, DC_SEM_SHARED, FALSE) != NO_ERROR)
    return;
  heartbeat_hev = 1;
  if (DosCreateThread (&tid,
                       heartbeat_thread_code,
                       0,
                       CREATE_READY|STACK_COMMITTED,
                       4096)
      != NO_ERROR)
    return;
  heartbeat_tid = 1;
  DosSetPriority (PRTYS_THREAD, PRTYC_NOCHANGE, PRTYD_MAXIMUM, tid);
  heartbeat_enabled = 1;
}


___HIDDEN void cleanup_heartbeat_interrupt_handling ___PVOID
{
  heartbeat_enabled = 0;
  if (heartbeat_htimer)
    {
      DosStopTimer (htimer);
      heartbeat_htimer = 0;
    }
  if (heartbeat_tid)
    {
      DosKillThread (tid);
      heartbeat_tid = 0;
    }
  if (heartbeat_hev)
    {
      DosCloseEventSem (hev);
      heartbeat_hev = 0;
    }
}


#endif


/*   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   */

#ifdef USE_VInstall

___HIDDEN ___BOOL heartbeat_task_installed;
___HIDDEN short heartbeat_task_ticks;
___HIDDEN VBLTask heartbeat_task;
___HIDDEN ___BOOL heartbeat_enabled;


___HIDDEN void heartbeat_task_code ___PVOID
{
  if (heartbeat_enabled)
    ___raise_interrupt (___INTR_HEARTBEAT);
  heartbeat_task.vblCount = heartbeat_task_ticks;
}


#ifdef __POWERPC__

#define heartbeat_task_proc heartbeat_task_code

#else

___HIDDEN asm void heartbeat_task_proc ___PVOID
{
    move.l a5,-(sp)
    move.l 0x904,a5
    jsr    heartbeat_task_code
    move.l (sp)+,a5
    rts
}

#endif


___F64 ___set_heartbeat_interval
   ___P((___F64 seconds),
        (seconds)
___F64 seconds;)
{
  if (heartbeat_task_installed)
    if (VRemove (___CAST(QElemPtr,&heartbeat_task)) != noErr)
      {
        heartbeat_task_installed = 0;
        return 0.0;
      }

  if (seconds < 0.0) /* turn heartbeat off */
    return 0.0;
  else
    {
      ___F64 ticks = seconds * 60.0; /* 60 ticks/sec */
      short t;

      if (ticks >= 32767.0) /* upper bound is 9 minutes! */
        t = 32767;
      else
        {
          t = ticks;
          if (t <= 0) /* use smallest interval */
            t = 1;
        }

      heartbeat_task_ticks    = t;
      heartbeat_task.qType    = vType;
      heartbeat_task.vblAddr  = NewVBLProc (heartbeat_task_proc);
      heartbeat_task.vblCount = t;
      heartbeat_task.vblPhase = 0;
      if (VInstall (___CAST(QElemPtr,&heartbeat_task)) == noErr)
        {
          heartbeat_task_installed = 1;
          return t / 60.0;
        }
      else
        return 0.0;
    }
}


___HIDDEN void disable_heartbeat_interrupt ___PVOID
{
  heartbeat_enabled = 0;
}


___HIDDEN void enable_heartbeat_interrupt ___PVOID
{
  heartbeat_enabled = 1;
}


___HIDDEN void setup_heartbeat_interrupt_handling ___PVOID
{
  heartbeat_enabled = 1;
  heartbeat_task_installed = 0;
}


___HIDDEN void cleanup_heartbeat_interrupt_handling ___PVOID
{
  ___set_heartbeat_interval (-1.0);
  heartbeat_enabled = 0;
}


#endif


/*   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   */

#ifdef USE_CreateThread

/*
 * The HEARTBEAT_PRIORITY should be high so that heartbeats will occur
 * as close as possible to the correct time.
 */

#define HEARTBEAT_PRIORITY THREAD_PRIORITY_HIGHEST

___HIDDEN HANDLE heartbeat_thread;
___HIDDEN HANDLE heartbeat_update; /* heartbeat needs updating when signaled */
___HIDDEN DWORD heartbeat_interval;
___HIDDEN ___BOOL heartbeat_enabled;


___HIDDEN DWORD WINAPI heartbeat_thread_proc
   ___P((LPVOID param),
        (param)
LPVOID param;)
{
  DWORD interval = heartbeat_interval;

  for (;;)
    {
      DWORD n = WaitForSingleObject (heartbeat_update, interval);

      interval = heartbeat_interval;

      if (interval == 0)
        return 0;

      switch (n)
        {
        case WAIT_ABANDONED:
        case WAIT_OBJECT_0:
          break;

        case WAIT_TIMEOUT:
          if (heartbeat_enabled)
            ___raise_interrupt (___INTR_HEARTBEAT);
          break;

        default: /* WAIT_FAILED */
          return 0;
        }
    }

  return 0;
}


___F64 ___set_heartbeat_interval
   ___P((___F64 seconds),
        (seconds)
___F64 seconds;)
{
  if (seconds < 0.0) /* turn heartbeat off */
    {
      heartbeat_interval = INFINITE;
      SetEvent (heartbeat_update); /* ignore error */
      return 0.0;
    }
  else
    {
      ___F64 msecs = seconds * 1000.0;
      DWORD m;

      if (msecs >= 4294967294.0) /* upper bound is 49.6 days! */
        m = 4294967294UL; /* note that INFINITE = 4294967295 */
      else
        {
          m = msecs;
          if (m == 0) /* use smallest interval */
            m = 1;
        }

      heartbeat_interval = m;
      SetEvent (heartbeat_update); /* ignore error */
      return m / 1000.0;
    }
}


___HIDDEN void disable_heartbeat_interrupt ___PVOID
{
  heartbeat_enabled = 0;
}


___HIDDEN void enable_heartbeat_interrupt ___PVOID
{
  heartbeat_enabled = 1;
}


___HIDDEN void setup_heartbeat_interrupt_handling ___PVOID
{
  DWORD thread_id;

  heartbeat_enabled = 1;
  heartbeat_interval = INFINITE;

  heartbeat_update =
    CreateEvent (NULL,  /* can't inherit */
                 FALSE, /* auto reset */
                 FALSE, /* not signaled */
                 NULL); /* no name */

  if (heartbeat_update == NULL)
    {
      heartbeat_thread = NULL;
      return;
    }

  heartbeat_thread =
    CreateThread (NULL,  /* no security attributes        */
                  65536, /* committed stack size          */
                  heartbeat_thread_proc,
                  0,     /* argument to thread function   */
                  0,     /* use default creation flags    */
                  &thread_id);

  if (heartbeat_thread == NULL)
    {
      CloseHandle (heartbeat_update); /* ignore error */
      heartbeat_update = NULL;
      heartbeat_thread = NULL;
      return;
    }

  SetThreadPriority (heartbeat_thread, HEARTBEAT_PRIORITY); /* ignore error */

  return;
}


___HIDDEN void cleanup_heartbeat_interrupt_handling ___PVOID
{
  if (heartbeat_thread != NULL)
    {
      heartbeat_enabled = 0;
      heartbeat_interval = 0;
      SetEvent (heartbeat_update); /* ignore error */
      WaitForSingleObject (heartbeat_thread, INFINITE); /* ignore error */
      CloseHandle (heartbeat_update); /* ignore error */
      heartbeat_update = NULL;
      heartbeat_thread = NULL;
    }
}


#endif


/*   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   */

/* User interrupt handling. */

/***************** osoleted by console device? */

#ifndef USE_sigaction_SIGINT
#ifndef USE_signal_SIGINT
#ifndef USE_dos_setvect_23h

___HIDDEN void setup_user_interrupt_handling ___PVOID
{
}

___HIDDEN void cleanup_user_interrupt_handling ___PVOID
{
}

___HIDDEN void disable_user_interrupt ___PVOID
{
}

___HIDDEN void enable_user_interrupt ___PVOID
{
}

#endif
#endif
#endif

/*   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   */

#ifdef USE_sigaction_SIGINT

___HIDDEN void user_interrupt_handler
   ___P((int sig),
        (sig)
int sig;)
{
  ___raise_interrupt (___INTR_USER);
}

___HIDDEN void setup_user_interrupt_handling ___PVOID
{
  struct sigaction act;

  act.sa_handler = user_interrupt_handler;
  act.sa_flags = 0;
#ifdef SA_INTERRUPT
  act.sa_flags |= SA_INTERRUPT;
#endif
  sigemptyset (&act.sa_mask);
  sigaction (SIGINT, &act, 0);
}

___HIDDEN void cleanup_user_interrupt_handling ___PVOID
{
  struct sigaction act;

  act.sa_handler = SIG_DFL;
  act.sa_flags = 0;
  sigemptyset (&act.sa_mask);
  sigaction (SIGINT, &act, 0);
}

___HIDDEN void disable_user_interrupt ___PVOID
{
  sigset_t mask;

  sigemptyset (&mask);
  sigaddset (&mask, SIGINT);
  sigprocmask (SIG_BLOCK, &mask, 0);
}

___HIDDEN void enable_user_interrupt ___PVOID
{
  sigset_t mask;

  sigemptyset (&mask);
  sigaddset (&mask, SIGINT);
  sigprocmask (SIG_UNBLOCK, &mask, 0);
}

#endif

/*   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   */

#ifdef USE_signal_SIGINT

___HIDDEN void user_interrupt_handler
   ___P((int sig),
        (sig)
int sig;)
{
  signal (SIGINT, user_interrupt_handler);
  ___raise_interrupt (___INTR_USER);
}

___HIDDEN void setup_user_interrupt_handling ___PVOID
{
  signal (SIGINT, user_interrupt_handler);
}

___HIDDEN void cleanup_user_interrupt_handling ___PVOID
{
  signal (SIGINT, SIG_DFL);
}

___HIDDEN void disable_user_interrupt ___PVOID
{
  signal (SIGINT, SIG_IGN);
}

___HIDDEN void enable_user_interrupt ___PVOID
{
  signal (SIGINT, user_interrupt_handler);
}

#endif

/*   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   */

#ifdef USE_dos_setvect_23h

/* 23h = ctrl-c vector */
___HIDDEN void (__interrupt __far *prev_vector_23h) ___PVOID;

/* 1Bh = ctrl-break vector */
___HIDDEN void (__interrupt __far *prev_vector_1Bh) ___PVOID;

___HIDDEN void __interrupt __far user_interrupt_handler ___PVOID
{
  ___raise_interrupt (___INTR_USER);
}

___HIDDEN void setup_user_interrupt_handling ___PVOID
{
  prev_vector_23h = _dos_getvect (0x23);
  prev_vector_1Bh = _dos_getvect (0x1B);
  _dos_setvect (0x23, user_interrupt_handler);
  _dos_setvect (0x1B, user_interrupt_handler);
}

___HIDDEN void cleanup_user_interrupt_handling ___PVOID
{
  _dos_setvect (0x23, prev_vector_23h);
  _dos_setvect (0x1B, prev_vector_1Bh);
}

___HIDDEN void disable_user_interrupt ___PVOID
{
}

___HIDDEN void enable_user_interrupt ___PVOID
{
}

#endif


/*   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   */

___HIDDEN void setup_interrupt_handling ___PVOID
{
  setup_user_interrupt_handling ();
  setup_heartbeat_interrupt_handling ();
}

___HIDDEN void cleanup_interrupt_handling ___PVOID
{
  cleanup_heartbeat_interrupt_handling ();
  cleanup_user_interrupt_handling ();
}

___HIDDEN void disable_interrupts ___PVOID
{
  disable_heartbeat_interrupt ();
  disable_user_interrupt ();
}

___HIDDEN void enable_interrupts ___PVOID
{
  enable_user_interrupt ();
  enable_heartbeat_interrupt ();
}


/*---------------------------------------------------------------------------*/

/* Access to OS environment variables. */

char *___getenv
   ___P((char *name),
        (name)
char *name;)
{
#ifndef USE_POSIX
#ifndef USE_WIN32

  return 0;

#endif
#endif

#ifdef USE_POSIX

  return getenv (name);

#endif

#ifdef USE_WIN32

  return getenv (name);  ////////////////////////

#endif
}


#define ENV_CHAR_BYTES 1/////////////


___LOCAL int environ_unused_at_end = 0;
___LOCAL ___BOOL environ_was_extended = 0;


___SCMOBJ ___getenv_ucs2
   ___P((___UCS2STRING name,
         ___UCS2STRING *value),
        (name,
         value)
___UCS2STRING name;
___UCS2STRING *value;)
{
  ___UCS2STRING v;
  ___UCS2STRING p1;

  /* reject strings that contain "=" except as the first character */

  p1 = name;

  if (*p1 == '=')
    p1++;

  while (*p1 != '\0')
    {
#if ENV_CHAR_BYTES == 1
      if (*p1 > 255)
        return ___FIX(___IMPL_LIMIT_ERR);
#endif
      if (*p1++ == '=')
        return ___FIX(___IMPL_LIMIT_ERR);
    }

  /* find in the environment a string of the form name=value */

#ifdef USE_POSIX

  {
    extern char **environ;

    char **probe;
    char *p2;

    probe = environ;

    while ((p2 = *probe++) != 0)
      {
        p1 = name;

        while (*p1 != '\0' &&
               *p1 == ___CAST(___UCS2,___CAST(unsigned char,*p2)))
          {
            p1++;
            p2++;
          }

        if (*p1 == '\0' && *p2 == '=')
          {
            int len = 0;

            p2++;

            while (p2[len] != '\0')
              len++;

            v = ___CAST(___UCS2STRING,
                        ___alloc_mem (sizeof (___UCS2) * (len+1)));

            if (v == 0)
              return ___FIX(___HEAP_OVERFLOW_ERR);

            v[len] = '\0';

            while (len > 0)
              {
                len--;
                v[len] = ___CAST(___UCS2,___CAST(unsigned char,p2[len]));
              }

            *value = v;

            return ___FIX(___NO_ERR);
          }
      }
  }

#endif

  *value = 0;

  return ___FIX(___NO_ERR);
}


___SCMOBJ ___setenv_ucs2
   ___P((___UCS2STRING name,
         ___UCS2STRING value),
        (name,
         value)
___UCS2STRING name;
___UCS2STRING value;)
{
  ___UCS2STRING p1;
  int name_len = 0;
  int value_len = 0;

  /* reject strings that contain "=" except as the first character */

  p1 = name;

  if (*p1 == '=')
    {
      p1++;
      name_len++;
    }

  while (*p1 != '\0')
    {
#if ENV_CHAR_BYTES == 1
      if (*p1 > 255)
        return ___FIX(___IMPL_LIMIT_ERR);
#endif
      if (*p1++ == '=')
        return ___FIX(___IMPL_LIMIT_ERR);
      name_len++;
    }

  p1 = value;

  while (*p1 != '\0')
    {
#if ENV_CHAR_BYTES == 1
      if (*p1 > 255)
        return ___FIX(___IMPL_LIMIT_ERR);
#endif
      p1++;
      value_len++;
    }

  /* find in the environment a string of the form name=value */

#ifdef USE_POSIX

  {
    extern char **environ;

    char **old_environ = environ;
    char **probe;
    char *p2;

    char *name_value = ___CAST(char*,
                               ___alloc_mem (name_len + value_len + 2));

    if (name_value == 0)
      return ___FIX(___HEAP_OVERFLOW_ERR);

    p2 = name_value;

    p1 = name;

    while (name_len > 0)
      {
        *p2++ = ___CAST(char,*p1++);
        name_len--;
      }

    *p2++ = '=';

    p1 = value;

    while (value_len > 0)
      {
        *p2++ = ___CAST(char,*p1++);
        value_len--;
      }

    *p2++ = '\0';

    probe = old_environ;

    while ((p2 = *probe++) != 0)
      {
        p1 = name;

        while (*p1 != '\0' &&
               *p1 == ___CAST(___UCS2,___CAST(unsigned char,*p2)))
          {
            p1++;
            p2++;
          }

        if (*p1 == '\0' && *p2 == '=')
          {
            probe[-1] = name_value;
            return ___FIX(___NO_ERR);
          }
      }

    if (environ_unused_at_end > 0)
      {
        probe[-1] = name_value;
        probe[0] = 0;
        environ_unused_at_end--;
        return ___FIX(___NO_ERR);
      }
    else
      {
        char **new_environ;
        int n = probe - old_environ; /* length including null pointer at end */

        environ_unused_at_end = n/2 + 1;

        new_environ =
          ___CAST(char**,
                  ___alloc_mem ((n + environ_unused_at_end) * sizeof (char*)));

        if (new_environ == 0)
          {
            ___free_mem (name_value);
            return ___FIX(___HEAP_OVERFLOW_ERR);
          }

        environ = new_environ;
        probe = old_environ;

        while (--n > 0)
          *new_environ++ = *probe++;

        *new_environ++ = name_value;
        *new_environ++ = 0;
          
        environ_unused_at_end--;

        if (environ_was_extended)
          ___free_mem (old_environ);

        environ_was_extended = 1;
      }
  }

#endif

  return ___FIX(___NO_ERR);
}


___SCMOBJ ___unsetenv_ucs2
   ___P((___UCS2STRING name),
        (name)
___UCS2STRING name;)
{
  ___UCS2STRING p1;

  /* reject strings that contain "=" except as the first character */

  p1 = name;

  if (*p1 == '=')
    p1++;

  while (*p1 != '\0')
    {
#if ENV_CHAR_BYTES == 1
      if (*p1 > 255)
        return ___FIX(___IMPL_LIMIT_ERR);
#endif
      if (*p1++ == '=')
        return ___FIX(___IMPL_LIMIT_ERR);
    }

  /* find in the environment a string of the form name=value */

#ifdef USE_POSIX

  {
    extern char **environ;

    char **probe;
    char *p2;

    probe = environ;

    while ((p2 = *probe++) != 0)
      {
        p1 = name;

        while (*p1 != '\0' &&
               *p1 == ___CAST(___UCS2,___CAST(unsigned char,*p2)))
          {
            p1++;
            p2++;
          }

        if (*p1 == '\0' && *p2 == '=')
          {
            environ_unused_at_end++;

            while ((probe[-1] = probe[0]) != 0)
              probe++;

            return ___FIX(___NO_ERR);
          }
      }
  }

#endif

  return ___FIX(___NO_ERR);
}


___SCMOBJ ___os_getenv
   ___P((___SCMOBJ name),
        (name)
___SCMOBJ name;)
{
  ___SCMOBJ e;
  ___SCMOBJ result;

#ifndef USE_POSIX
#ifndef USE_WIN32

  result = ___FIX(___UNKNOWN_ERR);

#endif
#endif

#ifdef USE_POSIX

  ___UCS2STRING cname;

  if ((e = ___SCMOBJ_to_NONNULLUCS2STRING
             (name,
              &cname,
              1))
      != ___FIX(___NO_ERR))
    result = e;
  else
    {
      ___UCS2STRING cvalue;

      if ((e = ___getenv_ucs2 (cname, &cvalue)) != ___FIX(___NO_ERR))
        result = e;
      else
        {
          if ((e = ___UCS2STRING_to_SCMOBJ
                     (cvalue,
                      &result,
                      ___RETURN_POS))
              != ___FIX(___NO_ERR))
            result = e;
          else
            ___release_scmobj (result);
          if (cvalue != 0)
            ___free_mem (cvalue);
        }
      ___release_string (cname);
    }

#endif

#ifdef USE_WIN32

#ifdef UNICODE
#define ___GETENV_CE_SELECT(latin1,utf8,ucs2,ucs4,native) ucs2
#else
#define ___GETENV_CE_SELECT(latin1,utf8,ucs2,ucs4,native) latin1
#endif

#define GETENV_VALUE_STATIC_SIZE 1024

  void *cname;
  ___CHAR_TYPE(___GETENV_CE_SELECT) cvalue[GETENV_VALUE_STATIC_SIZE];
  ___CHAR_TYPE(___GETENV_CE_SELECT) *cvalue_ptr = cvalue;
  int n;

  if ((e = ___SCMOBJ_to_NONNULLSTRING
             (name,
              &cname,
              1,
              ___CE(___GETENV_CE_SELECT),
              0))
      != ___FIX(___NO_ERR))
    result = e;
  else
    {
      n = GetEnvironmentVariable
            (___CAST(___STRING_TYPE(___GETENV_CE_SELECT),cname),
             cvalue_ptr,
             GETENV_VALUE_STATIC_SIZE);

      if (n >= GETENV_VALUE_STATIC_SIZE)
        {
          cvalue_ptr = ___CAST(___CHAR_TYPE(___GETENV_CE_SELECT)*,
                               ___alloc_mem (sizeof (*cvalue_ptr) * n));

          if (cvalue_ptr != 0)
            n = GetEnvironmentVariable
                  (___CAST(___STRING_TYPE(___GETENV_CE_SELECT),cname),
                   cvalue_ptr,
                   n);
        }

      if (cvalue_ptr == 0)
        result = ___FIX(___HEAP_OVERFLOW_ERR);
      else
        {
          if (n <= 0)
            result = ___FAL;
          else if ((e = ___NONNULLSTRING_to_SCMOBJ
                          (cvalue_ptr,
                           &result,
                           ___RETURN_POS,
                           ___CE(___GETENV_CE_SELECT)))
                   != ___FIX(___NO_ERR))
            result = e;
          else
            ___release_scmobj (result);

          if (cvalue_ptr != cvalue)
            ___free_mem (cvalue_ptr);
        }

      ___release_string (cname);
    }

#endif

  return result;
}


___SCMOBJ ___os_setenv
   ___P((___SCMOBJ name,
         ___SCMOBJ value),
        (name,
         value)
___SCMOBJ name;
___SCMOBJ value;)
{
  ___SCMOBJ e;

#ifndef USE_POSIX
#ifndef USE_WIN32

  e = ___FIX(___UNKNOWN_ERR);

#endif
#endif

#ifdef USE_POSIX

  ___UCS2STRING cname;
  ___UCS2STRING cvalue;

  if ((e = ___SCMOBJ_to_NONNULLUCS2STRING
             (name,
              &cname,
              1))
      == ___FIX(___NO_ERR))
    {
      if (value == ___FAL)
        e = ___unsetenv_ucs2 (cname);
      else if ((e = ___SCMOBJ_to_NONNULLUCS2STRING
                      (value,
                       &cvalue,
                       2))
               == ___FIX(___NO_ERR))
        {
          e = ___setenv_ucs2 (cname, cvalue);
          ___release_string (cvalue);
        }

      ___release_string (cname);
    }

#endif

#ifdef USE_WIN32

#ifdef UNICODE
#define ___SETENV_CE_SELECT(latin1,utf8,ucs2,ucs4,native) ucs2
#else
#define ___SETENV_CE_SELECT(latin1,utf8,ucs2,ucs4,native) latin1
#endif

  void *cname;
  void *cvalue;

  if ((e = ___SCMOBJ_to_NONNULLSTRING
             (name,
              &cname,
              1,
              ___CE(___SETENV_CE_SELECT),
              0))
      == ___FIX(___NO_ERR))
    {
      if (value == ___FAL)
        {
          if (!SetEnvironmentVariable
                 (___CAST(___STRING_TYPE(___SETENV_CE_SELECT),cname),
                  0))
            {
              e = err_code_from_GetLastError ();

              /*
               * Apparently an error is signaled if the environment
               * variable being removed does not exist (the Microsoft
               * documentation does not mention this).
               */

              if (e == ___FIX(___WIN32_ERR(ERROR_ENVVAR_NOT_FOUND)))
                e = ___FIX(___NO_ERR);
            }
        }
      else if ((e = ___SCMOBJ_to_NONNULLSTRING
                      (value,
                       &cvalue,
                       2,
                       ___CE(___SETENV_CE_SELECT),
                       0))
               == ___FIX(___NO_ERR))
        {
          if (!SetEnvironmentVariable
                 (___CAST(___STRING_TYPE(___SETENV_CE_SELECT),cname),
                  ___CAST(___STRING_TYPE(___SETENV_CE_SELECT),cvalue)))
            e = err_code_from_GetLastError ();
          ___release_string (cvalue);
        }

      ___release_string (cname);
    }

#endif

  return e;
}


___SCMOBJ ___os_environ ___PVOID
{
  ___SCMOBJ e;
  ___SCMOBJ result;

#ifndef USE_POSIX
#ifndef USE_WIN32

  e = ___NUL;

#endif
#endif

#ifdef USE_POSIX

  extern char **environ;

  if ((e = ___NONNULLCHARSTRINGLIST_to_SCMOBJ
             (environ,
              &result,
              ___RETURN_POS))
      != ___FIX(___NO_ERR))
    result = e;
  else
    ___release_scmobj (result);

#endif

#ifdef USE_WIN32

#ifdef UNICODE
#define ___ENVIRON_CE_SELECT(latin1,utf8,ucs2,ucs4,native) ucs2
#else
#define ___ENVIRON_CE_SELECT(latin1,utf8,ucs2,ucs4,native) latin1
#endif

  ___STRING_TYPE(___ENVIRON_CE_SELECT) env;
  ___STRING_TYPE(___ENVIRON_CE_SELECT) ptr;
  ___SCMOBJ pair;
  ___SCMOBJ str;

  result = ___NUL;

  env = GetEnvironmentStrings ();

  if (env != 0 && *env != 0)
    {
      ptr = env;

      /* find end of environment strings. */

      do
        {
          do { ptr++; } while (*ptr != 0);
          ptr++; /* skip null char at end of string */
        } while (*ptr != 0);

      while (ptr > env)
        {
          ptr--; /* move ptr to terminating null char of previous string */

          while (ptr > env && ptr[-1] != 0)
            ptr--;

          if ((e = ___NONNULLSTRING_to_SCMOBJ
                     (ptr,
                      &str,
                      ___RETURN_POS,
                      ___CE(___ENVIRON_CE_SELECT)))
              != ___FIX(___NO_ERR))
            break;

          pair = ___make_pair (str, result, ___STILL);

          ___release_scmobj (str);
          ___release_scmobj (result);

          if (___FIXNUMP(pair))
            {
              e = ___FIX(___CTOS_HEAP_OVERFLOW_ERR+___RETURN_POS);
              break;
            }

          result = pair;
        }

      ___release_scmobj (result);
    }

  if (env != 0)
    if (!FreeEnvironmentStrings (env))
      result = err_code_from_GetLastError ();

#endif

  return result;
}


/*---------------------------------------------------------------------------*/

/* Shell command. */

___SCMOBJ ___os_shell_command
   ___P((___SCMOBJ cmd),
        (cmd)
___SCMOBJ cmd;)
{
  ___SCMOBJ e;

#ifndef USE_POSIX
#ifndef USE_WIN32

  e = ___FIX(___UNKNOWN_ERR);

#endif
#endif

#ifdef USE_POSIX

  char *ccmd;

  if ((e = ___SCMOBJ_to_NONNULLCHARSTRING
             (cmd,
              &ccmd,
              1))
      == ___FIX(___NO_ERR))
    {
      int code;

      disable_interrupts ();

      code = system (ccmd);

      if (code == -1)
        e = err_code_from_errno ();
      else
        e = ___FIX(code & ___MAX_FIX);

      enable_interrupts ();

      ___release_string (ccmd);
    }

#endif

#ifdef USE_WIN32

#ifdef UNICODE
#define ___SHELL_COMMAND_CE_SELECT(latin1,utf8,ucs2,ucs4,native) ucs2
#else
#define ___SHELL_COMMAND_CE_SELECT(latin1,utf8,ucs2,ucs4,native) latin1
#endif

  void *ccmd;

  if ((e = ___SCMOBJ_to_NONNULLSTRING
             (cmd,
              &ccmd,
              1,
              ___CE(___SHELL_COMMAND_CE_SELECT),
              0))
      == ___FIX(___NO_ERR))
    {
#ifdef DO_NOT_USE_system

      /*
       * This code does not really cause the shell to run the command.
       * This means that the shell builtin commands (such as "DIR"
       * cannot be executed.  It is better to use "system" and
       * "_wsystem".
       */

      DWORD code;
      STARTUPINFO si;
      PROCESS_INFORMATION pi;

      ZeroMemory (&si, sizeof (si));
      si.cb = sizeof (si);
      ZeroMemory (&pi, sizeof (pi));

      if (!CreateProcess (NULL,  /* module name                              */
                          ___CAST(___STRING_TYPE(___SETENV_CE_SELECT),ccmd),
                          NULL,  /* process handle not inheritable           */
                          NULL,  /* thread handle not inheritable            */
                          FALSE, /* set handle inheritance to FALSE          */
                          0,     /* no creation flags                        */
                          NULL,  /* use parent's environment block           */
                          NULL,  /* use parent's starting directory          */
                          &si,   /* pointer to STARTUPINFO structure         */
                          &pi))  /* pointer to PROCESS_INFORMATION structure */
        e = err_code_from_GetLastError ();
      else
        {
          if (WaitForSingleObject (pi.hProcess, INFINITE) == WAIT_FAILED ||
              !GetExitCodeProcess (pi.hProcess, &code))
            e = err_code_from_GetLastError ();
          else
            e = ___FIX(code & ___MAX_FIX);

          CloseHandle (pi.hProcess); /* ignore error */
          CloseHandle (pi.hThread); /* ignore error */
        }

#else

      int code;

#ifdef UNICODE
      code = _wsystem (___CAST(___STRING_TYPE(___SETENV_CE_SELECT),ccmd));
#else
      code = system (___CAST(___STRING_TYPE(___SETENV_CE_SELECT),ccmd));
#endif

      if (code == -1)
        e = err_code_from_errno ();
      else
        e = ___FIX(code & ___MAX_FIX);

#endif

      ___release_string (ccmd);
    }

#endif

  return e;
}


/*---------------------------------------------------------------------------*/

/* Dynamic loading. */

#ifdef USE_dlopen
#define DL_DESCR void *
#define ___DL_PATH_CE_SELECT(latin1,utf8,ucs2,ucs4,native) native
#define ___DL_MODNAME_CE_SELECT(latin1,utf8,ucs2,ucs4,native) native
#endif

#ifdef USE_shl_load
#define DL_DESCR shl_t
#define ___DL_PATH_CE_SELECT(latin1,utf8,ucs2,ucs4,native) native
#define ___DL_MODNAME_CE_SELECT(latin1,utf8,ucs2,ucs4,native) native
#endif

#ifdef USE_LoadLibrary
#define DL_DESCR HANDLE
#ifdef UNICODE
#define ___DL_PATH_CE_SELECT(latin1,utf8,ucs2,ucs4,native) ucs2
#else
#define ___DL_PATH_CE_SELECT(latin1,utf8,ucs2,ucs4,native) latin1
#endif
#define ___DL_MODNAME_CE_SELECT(latin1,utf8,ucs2,ucs4,native) native
#endif

#ifdef USE_DosLoadModule
#define DL_DESCR HMODULE
#define ___DL_PATH_CE_SELECT(latin1,utf8,ucs2,ucs4,native) native
#define ___DL_MODNAME_CE_SELECT(latin1,utf8,ucs2,ucs4,native) native
#endif

#ifdef USE_dxe_load
#define DL_DESCR void *
#define ___DL_PATH_CE_SELECT(latin1,utf8,ucs2,ucs4,native) native
#define ___DL_MODNAME_CE_SELECT(latin1,utf8,ucs2,ucs4,native) native
#endif

#ifdef USE_GetDiskFragment
#define DL_DESCR CFragConnectionID
#define ___DL_PATH_CE_SELECT(latin1,utf8,ucs2,ucs4,native) native
#define ___DL_MODNAME_CE_SELECT(latin1,utf8,ucs2,ucs4,native) native
#endif

#ifdef USE_NSLinkModule
#define DL_DESCR NSModule
#define ___DL_PATH_CE_SELECT(latin1,utf8,ucs2,ucs4,native) native
#define ___DL_MODNAME_CE_SELECT(latin1,utf8,ucs2,ucs4,native) native
#endif


#ifdef DL_DESCR

typedef struct dl_entry
  {
    struct dl_entry *next;
    DL_DESCR descr;
  } dl_entry;

___HIDDEN dl_entry *dl_list;

#endif


___HIDDEN void setup_dynamic_load ___PVOID
{
#ifdef DL_DESCR

  dl_list = 0;

#endif
}


___HIDDEN ___SCMOBJ dynamic_load_module
   ___P((___STRING_TYPE(___DL_PATH_CE_SELECT) cpath,
         ___STRING_TYPE(___DL_MODNAME_CE_SELECT) cmodname,
         void **linker),
        (cpath,
         cmodname,
         linker)
___STRING_TYPE(___DL_PATH_CE_SELECT) cpath;
___STRING_TYPE(___DL_MODNAME_CE_SELECT) cmodname;
void **linker;)
{
  ___SCMOBJ e;
  ___SCMOBJ result = ___FIX(___UNKNOWN_ERR);

#ifndef DL_DESCR

  result = ___FIX(___DYNAMIC_LOADING_NOT_AVAILABLE_ERR);

#else

  dl_entry *p;

  p = ___CAST(dl_entry*,
              ___alloc_mem (sizeof (dl_entry)));

  if (p == 0)
    return ___FIX(___HEAP_OVERFLOW_ERR);

#ifdef USE_dlopen

#ifdef RTLD_NOW
  p->descr = dlopen (cpath, RTLD_NOW);
#else
  p->descr = dlopen (cpath, 1);
#endif

  if (p->descr != 0 &&
      (*linker = dlsym (p->descr, cmodname)) != 0)
    result = ___FIX(___NO_ERR);
  else
    {
      if ((e = ___NONNULLCHARSTRING_to_SCMOBJ
                 (___CAST(char*,dlerror ()),
                  &result,
                  ___RETURN_POS))
          != ___FIX(___NO_ERR))
        result = e;

      if (p->descr != 0)
        dlclose (p->descr);
    }

#endif

#ifdef USE_shl_load

  p->descr = shl_load (cpath, BIND_IMMEDIATE, 0);

  if (p->descr == 0)
    result = err_code_from_errno ();
  else if (!shl_findsym (&p->descr, cmodname, TYPE_PROCEDURE, linker) &&
           *linker != 0)
    result = ___FIX(___NO_ERR);
  else
    {
      result = ___FIX(___DYNAMIC_LOADING_LOOKUP_ERR);
      shl_unload (p->descr);
    }

#endif

#ifdef USE_LoadLibrary

  p->descr = LoadLibrary (cpath);

  if (p->descr != 0 &&
      (*linker = GetProcAddress (p->descr, cmodname)) != 0)
    result = ___FIX(___NO_ERR);
  else
    {
      result = err_code_from_GetLastError ();

      if (p->descr != 0)
        FreeLibrary (p->descr);
    }

#endif

#ifdef USE_DosLoadModule

  {
    HMODULE hmodule;
    APIRET rc;
    char real_path[280];
    int i;

    e = ___FIX(___NO_ERR);

    i = 0;
    while (cpath[i] != '\0')
      i++;

    if (i <= 4 ||
        cpath[i-1] != 'l' && cpath[i-1] != 'L' ||
        cpath[i-2] != 'l' && cpath[i-2] != 'L' ||
        cpath[i-3] != 'd' && cpath[i-3] != 'D' ||
        cpath[i-4] != '.')
      {
        /* if path doesn't end in ".dll" we must find the real ".dll" path */

        FILE *f = fopen (cpath, "r");

        if (f == 0 || !fgets (real_path, 280, f))
          e = err_code_from_errno ();
        else
          cpath = real_path;

        if (f != 0)
          fclose (f);
      }

    if (e != ___FIX(___NO_ERR))
      result = e;
    else
      {
        char *errmsg = 0;
        char errbuf[256];

        if (DosLoadModule (errbuf, 256, cpath, &hmodule) != NO_ERROR)
          errmsg = errbuf;
        else
          {
            p->descr = hmodule;

            rc = DosQueryProcAddr (hmodule, 0L, cmodname, linker);

            if (rc != NO_ERROR || *linker == 0)
              {
                switch (rc)
                  {
                  case ERROR_INVALID_HANDLE:
                    errmsg = "Invalid handle";
                    break;
                  case ERROR_INVALID_NAME:
                    errmsg = "Invalid name";
                    break;
                  case ERROR_INVALID_ORDINAL:
                    errmsg = "Invalid ordinal";
                    break;
                  case ERROR_ENTRY_IS_CALLGATE:
                    errmsg = "Entry is callgate";
                    break;
                  default:
                    errmsg = "Unknown error";
                    break;
                  }

                DosFreeModule (hmodule);
              }
          }

        if (errmsg != 0)
          {
            if ((e = ___NONNULLCHARSTRING_to_SCMOBJ
                       (errmsg,
                        &result,
                        ___RETURN_POS))
                != ___FIX(___NO_ERR))
              result = e;
          }
      }
  }

#endif

#ifdef USE_dxe_load

  p->descr = _dxe_load (cpath);

  if ((*linker = p->descr) != 0)
    result = ___FIX(___NO_ERR);
  else
    result = ___FIX(___DYNAMIC_LOADING_LOOKUP_ERR);

#endif

#ifdef USE_GetDiskFragment

  {
    OSErr err;
    Ptr mainadr, procadr;
    Str63 ppath;
    Str255 pmodname;
    Str255 pmsg;
    char msg[256];
    FSSpec spec;

    if (!c2pascal (cpath, ppath, sizeof(ppath)-1))
      *errmsg = "Path is too long";
    else if (!c2pascal (cmodname, pmodname, sizeof(pmodname)-1))
      *errmsg = "Module name is too long";
    else if (make_ResolvedFSSpec (0, 0, ppath, &spec) != noErr)
      *errmsg = "Invalid path";
    else
      {
        err = GetDiskFragment (&spec, 0, kCFragGoesToEOF, ppath,
                               kPrivateCFragCopy, &p->descr, &mainadr, pmsg);
        if (err != noErr)
          {
            pascal2c (pmsg, msg, 255);
            sprintf (dl_error_buffer,
                     "GetDiskFragment failed with error code %d on \"%s\"",
                     err,
                     msg);
            *errmsg = dl_error_buffer;
          }
        else
          {
            if (FindSymbol (p->descr, pmodname, &procadr, kCodeCFragSymbol)
                != noErr)
              {
                *errmsg = "FindSymbol failed";
                CloseConnection (&p->descr);
              }
            else
              *linker = ___CAST(void*,procadr);
          }
      }
  }

#endif

#ifdef USE_NSLinkModule

  {
    NSObjectFileImage img;
    NSSymbol sym;

    if (NSCreateObjectFileImageFromFile (cpath, &img)
        != NSObjectFileImageSuccess)
      result = ___FIX(___UNKNOWN_ERR);
    else
      {
        p->descr = NSLinkModule (img, cpath, NSLINKMODULE_OPTION_BINDNOW);

        if (p->descr != 0 &&
            (sym = NSLookupSymbolInModule (p->descr, cmodname)) != 0 &&
            (*linker = NSAddressOfSymbol (sym)) != 0)
          result = ___FIX(___NO_ERR);
        else
          {
            result = ___FIX(___DYNAMIC_LOADING_LOOKUP_ERR);

            if (p->descr != 0)
              NSUnLinkModule (p->descr, NSUNLINKMODULE_OPTION_NONE);
          }

        NSDestroyObjectFileImage (img);
      }
  }

#endif

  if (result != ___FIX(___NO_ERR))
    ___free_mem (p);
  else
    {
      p->next = dl_list;
      dl_list = p;
    }

#endif

  return result;
}


___HIDDEN void cleanup_dynamic_load ___PVOID
{
#ifdef DL_DESCR

  dl_entry *p = dl_list;
  while (p != 0)
    {
      dl_entry *next = p->next;

#ifdef USE_dlopen
      dlclose (p->descr);
#endif

#ifdef USE_shl_load
      shl_unload (p->descr);
#endif

#ifdef USE_LoadLibrary
      FreeLibrary (p->descr);
#endif

#ifdef USE_DosLoadModule
      DosFreeModule (p->descr);
#endif

#ifdef USE_dxe_load
#endif

#ifdef USE_GetDiskFragment
      CloseConnection (&p->descr);
#endif

#ifdef USE_NSLinkModule
      NSUnLinkModule (p->descr, NSUNLINKMODULE_OPTION_NONE);
#endif

      ___free_mem (p);
      p = next;
    }
  dl_list = 0;

#endif  
}


___HIDDEN char c_id_prefix[] =
#ifdef ___IMPORTED_ID_PREFIX
___IMPORTED_ID_PREFIX
#endif
___C_ID_PREFIX;

#define c_id_prefix_length (sizeof (c_id_prefix) - 1)

___HIDDEN char c_id_suffix[] =
#ifdef ___IMPORTED_ID_SUFFIX
___IMPORTED_ID_SUFFIX
#endif
"";

#define c_id_suffix_length (sizeof (c_id_suffix) - 1)

___HIDDEN char hex_digits[] = "0123456789abcdef";

#define c_id_subsequent(c) \
(((c)>='A'&&(c)<='Z') || \
 ((c)>='a'&&(c)<='z') || \
 ((c)>='0'&&(c)<='9') || \
 ((c)=='_'))


___HIDDEN ___SCMOBJ ___SCMOBJ_to_MODNAMESTRING
   ___P((___SCMOBJ obj,
         void **x,
         int arg_num),
        (obj,
         x,
         arg_num)
___SCMOBJ obj;
void **x;
int arg_num;)
{
  ___STRING_TYPE(___DL_MODNAME_CE_SELECT) r;
  int len;
  int i;
  int j;
  ___SCMOBJ ___temp; /* used by ___STRINGP */

  if (!___STRINGP(obj))
    return ___FIX(___DL_MODNAME_CE_SELECT(___STOC_NONNULLLATIN1STRING_ERR,
                                          ___STOC_NONNULLUTF8STRING_ERR,
                                          ___STOC_NONNULLUCS2STRING_ERR,
                                          ___STOC_NONNULLUCS4STRING_ERR,
                                          ___STOC_NONNULLCHARSTRING_ERR) +
                  arg_num);

  len = ___INT(___STRINGLENGTH(obj));
  i = len - 1;
  j = c_id_prefix_length + c_id_suffix_length;

  for (i=len-1; i>=0; i--)
    {
      ___UCS4 c = ___INT(___STRINGREF(obj,___FIX(i)));
      if (c == '_')
        j += 2;
      else if (c_id_subsequent(c))
        j++;
      else
        {
          j += 3;
          while (c > 15)
            {
              c >>= 4;
              j++;
            }
        }
    }

  r = ___CAST(___STRING_TYPE(___DL_MODNAME_CE_SELECT),
              ___alloc_mem ((j+1) *
                            sizeof (___CHAR_TYPE(___DL_MODNAME_CE_SELECT))));

  if (r == 0)
    return ___FIX(___STOC_HEAP_OVERFLOW_ERR+arg_num);

  r[j--] = '\0';

  for (i=c_id_suffix_length-1; i>=0; i--)
    r[j--] = c_id_suffix[i];

  for (i=len-1; i>=0; i--)
    {
      ___UCS4 c = ___INT(___STRINGREF(obj,___FIX(i)));
      if (c == '_')
        {
          r[j--] = '_';
          r[j--] = '_';
        }
      else if (c_id_subsequent(c))
        r[j--] = c;
      else
        {
          r[j--] = '_';
          do
            {
              r[j--] = hex_digits[c & 15];
              c >>= 4;
            } while (c != 0);
          r[j--] = '_';
        }
    }

  for (i=c_id_prefix_length-1; i>=0; i--)
    r[j--] = c_id_prefix[i];

  *x = r;

  return ___FIX(___NO_ERR);
}


___SCMOBJ ___dynamic_load
   ___P((___SCMOBJ path,
         ___SCMOBJ modname,
         void **linker),
        (path,
         modname,
         linker)
___SCMOBJ path;
___SCMOBJ modname;
void **linker;)
{
  ___SCMOBJ e;
  ___SCMOBJ result;
  void *cpath;
  void *cmodname;

  if ((e = ___SCMOBJ_to_NONNULLSTRING
             (path,
              &cpath,
              1,
              ___CE(___DL_PATH_CE_SELECT),
              0))
      != ___FIX(___NO_ERR))
    result = e;
  else
    {
      if ((e = ___SCMOBJ_to_MODNAMESTRING
                 (modname,
                  &cmodname,
                  2))
          != ___FIX(___NO_ERR))
        result = e;
      else
        {
          result = dynamic_load_module
                     (___CAST(___STRING_TYPE(___DL_PATH_CE_SELECT),cpath),
                      ___CAST(___STRING_TYPE(___DL_MODNAME_CE_SELECT),cmodname),
                      linker);

          ___free_mem (cmodname);
        }
      ___release_string (cpath);
    }

  return result;
}


/*---------------------------------------------------------------------------*/

/* Dynamic C compilation. */

#ifdef USE_CWIDE

/* 
 * On the Macintosh with MACOS Classic, a C module can be dynamically
 * compiled by sending AppleEvents to the CodeWarrior IDE.  This code
 * has been tested with CWPro 1 on a PowerPC based Macintosh.
 */

#define CWIDE_EVENT_CLASS 'MMPR'
#define CWIDE_SIGNATURE 'CWIE'

#define CWIDE_no_result     0
#define CWIDE_list_of_short 1
#define CWIDE_list_of_ErrM  2

enum
  {
    errShell_ActionFailed  = 1,
    errShell_FileNotFound  = 2,
    errShell_DuplicateFile = 3,
    errShell_CompileError  = 4,
    errShell_MakeFailed    = 5,
    errShell_NoOpenProject = 6,
    errShell_WindowNotOpen = 7
  };


___HIDDEN AEAddressDesc CWIDE_address;


___HIDDEN OSErr setup_CWIDE ___PVOID
{
  long signature = CWIDE_SIGNATURE;

  CWIDE_address.dataHandle = nil;

  return AECreateDesc (typeApplSignature,
                       ___CAST(Ptr,&signature),
                       sizeof (long),
                       &CWIDE_address);
}


___HIDDEN OSErr cleanup_CWIDE ___PVOID
{
  if (CWIDE_address.dataHandle != nil)
    return AEDisposeDesc (&CWIDE_address);
  else
    return noErr;
}


___HIDDEN OSErr CWIDE_send_ae
   ___P((AppleEvent *ae,
         int res_kind,
         long *CWIDE_err),
        (ae,
         res_kind,
         CWIDE_err)
AppleEvent ae;
int res_kind;
long *CWIDE_err;)
{
  OSErr err;
  AppleEvent reply_ae;

  reply_ae.dataHandle = nil;

  err = AESend (ae,
                &reply_ae,
                kAEWaitReply,
                kAENormalPriority,
                kNoTimeOut,
                nil,
                nil);

  if (err == noErr && res_kind != CWIDE_no_result)
    switch (res_kind)
      {
        case CWIDE_list_of_short:
        case CWIDE_list_of_ErrM:
          {
            AEDescList list;
            list.dataHandle = nil;
            err = AEGetParamDesc (&reply_ae, keyAEResult, typeAEList, &list);
            if (err == noErr)
              if (res_kind == CWIDE_list_of_short)
                {
                  short error_code;
                  AEKeyword key;
                  DescType typ;
                  Size siz;
                  err = AEGetNthPtr (&list,
                                     1,
                                     typeShortInteger,
                                     &key,
                                     &typ,
                                     ___CAST(Ptr,&error_code),
                                     sizeof (short),
                                     &siz);
                  if (err == noErr)
                    *CWIDE_err = error_code;
                }
              else
                {
                  long nb_errors;
                  err = AECountItems (&list, &nb_errors);
                  if (err == noErr)
                    *CWIDE_err = nb_errors;
                }
            if (list.dataHandle != nil)
              AEDisposeDesc (&list);
            break;
          }
      }

  if (reply_ae.dataHandle != nil)
    AEDisposeDesc (&reply_ae);

  return err;
}


___HIDDEN OSErr CWIDE_make_ae
   ___P((AEEventClass clas,
         AEEventID id,
         AppleEvent *ae),
        (clas,
         id,
         ae)
AEEventClass clas;
AEEventID id;
AppleEvent *ae;)
{
  return AECreateAppleEvent (clas,
                             id,
                             &CWIDE_address,
                             kAutoGenerateReturnID,
                             kAnyTransactionID,
                             ae);
}


___HIDDEN OSErr CWIDE_file_param
   ___P((FSSpec *file,
         AEKeyword key,
         AppleEvent *ae),
        (file,
         key,
         ae)
FSSpec *file;
AEKeyword key;
AppleEvent *ae;)
{
  OSErr err;
  AEDescList list;

  list.dataHandle = nil;

  err = AECreateList (nil, 0, false, &list);
  if (err == noErr)
    {
      err = AEPutPtr (&list,
                      1,
                      typeFSS,
                      ___CAST(Ptr,file),
                      sizeof (FSSpec));
      if (err == noErr)
        err = AEPutParamDesc (ae, key, &list);
    }

  if (list.dataHandle != nil)
    AEDisposeDesc (&list);

  return err;
}


___HIDDEN OSErr CWIDE_no_param
   ___P((AEEventID id,
         int res_kind,
         long *CWIDE_err),
        (id,
         res_kind,
         CWIDE_err)
AEEventID id;
int res_kind;
long *CWIDE_err;)
{
  OSErr err;
  AppleEvent ae;

  ae.dataHandle = nil;

  err = CWIDE_make_ae (CWIDE_EVENT_CLASS, id, &ae);
  if (err == noErr)
    err = CWIDE_send_ae (&ae, res_kind, CWIDE_err);

  if (ae.dataHandle != nil)
    AEDisposeDesc (&ae);

  return err;
}


___HIDDEN OSErr CWIDE_open_project
   ___P((FSSpec *file),
        (file)
FSSpec *file;)
{
  long CWIDE_err;
  OSErr err;
  AppleEvent ae;

  ae.dataHandle = nil;

  err = CWIDE_make_ae ('aevt', 'odoc', &ae);
  if (err == noErr)
    {
      err = CWIDE_file_param (file, keyDirectObject, &ae);
      if (err == noErr)
        err = CWIDE_send_ae (&ae, CWIDE_no_result, &CWIDE_err);
    }

  if (ae.dataHandle != nil)
    AEDisposeDesc (&ae);

  return err;
}


___HIDDEN OSErr CWIDE_close_project ___PVOID
{
  long CWIDE_err;
  return CWIDE_no_param ('ClsP', CWIDE_no_result, &CWIDE_err);
}


___HIDDEN OSErr CWIDE_make_project
   ___P((long *CWIDE_err),
        (CWIDE_err)
long *CWIDE_err;)
{
  OSErr err;
  AppleEvent ae;

  ae.dataHandle = nil;

  err = CWIDE_make_ae (CWIDE_EVENT_CLASS, 'Make', &ae);
  if (err == noErr)
    {
      Boolean extern_editor = true;
      err = AEPutParamPtr (&ae,
                           'Errs',
                           typeBoolean,
                           ___CAST(Ptr,&extern_editor),
                           sizeof (Boolean));
      if (err == noErr)
        err = CWIDE_send_ae (&ae, CWIDE_list_of_ErrM, CWIDE_err);
    }

  if (ae.dataHandle != nil)
    AEDisposeDesc (&ae);

  return err;
}


___HIDDEN OSErr CWIDE_add_file
   ___P((FSSpec *file,
         short segment_num,
         long *CWIDE_err),
        (file,
         segment_num,
         CWIDE_err)
FSSpec *file;
short segment_num;
long *CWIDE_err;)
{
  OSErr err;
  AppleEvent ae;

  ae.dataHandle = nil;

  err = CWIDE_make_ae (CWIDE_EVENT_CLASS, 'AddF', &ae);
  if (err == noErr)
    {
      err = CWIDE_file_param (file, keyDirectObject, &ae);
      if (err == noErr)
        {
          err = AEPutParamPtr (&ae,
                               'Segm',
                               typeShortInteger,
                               ___CAST(Ptr,&segment_num),
                               sizeof (short));
          if (err == noErr)
            err = CWIDE_send_ae (&ae, CWIDE_list_of_short, CWIDE_err);
        }
    }

  if (ae.dataHandle != nil)
    AEDisposeDesc (&ae);

  return err;
}


___HIDDEN OSErr CWIDE_set_prefs
   ___P((Str255 name,
         AERecord *rec),
        (name,
         rec)
Str255 name;
AERecord *rec;)
{
  long CWIDE_err;
  OSErr err;
  AppleEvent ae;

  ae.dataHandle = nil;

  err = CWIDE_make_ae (CWIDE_EVENT_CLASS, 'Pref', &ae);
  if (err == noErr)
    {
      err = AEPutParamPtr (&ae,
                           'PNam',
                           typeChar,
                           ___CAST(Ptr,(name+1)),
                           name[0]);
      if (err == noErr)
        {
          err = AEPutParamDesc (&ae, 'PRec', rec);
          if (err == noErr)
            err = CWIDE_send_ae (&ae, CWIDE_no_result, &CWIDE_err);
        }
    }

  if (ae.dataHandle != nil)
    AEDisposeDesc (&ae);

  return err;
}


___HIDDEN OSErr CWIDE_set_output
   ___P((Str255 path),
        (path)
Str255 path;)
{
  OSErr err;
  AERecord rec;
  err = AECreateList (nil, 0, true, &rec);
  if (err == noErr)
    {
      err = AEPutKeyPtr (&rec,
                         'PR02',
                         typeChar,
                         ___CAST(Ptr,(path+1)),
                         path[0]);
      if (err == noErr)
        err = CWIDE_set_prefs ("\pPPC Project", &rec);
    }

  if (rec.dataHandle != nil)
    AEDisposeDesc (&rec);

  return err;
}


___HIDDEN int CWIDE_is_a_regular_file
   ___P((char *path),
        (path)
char *path;)
{
  struct stat sbuf;
  return (stat (path, &sbuf) == 0 && S_ISREG(sbuf.st_mode));
}


___BOOL ___dynamic_cc
   ___P((char *cc_cmd,
         char *ld_cmd,
         char *root_path,
         char *output_path,
         char *cc_options,
         char *ld_options,
         ___BOOL ld_flip,
         char **errmsg),
        (cc_cmd,
         ld_cmd,
         root_path,
         output_path,
         cc_options,
         ld_options,
         ld_flip,
         errmsg)
char *cc_cmd;
char *ld_cmd;
char *root_path;
char *output_path;
char *cc_options;
char *ld_options;
___BOOL ld_flip;
char **errmsg;)
{
  long CWIDE_err;
  OSErr err;
  FSSpec source_spec, project_spec, model_spec;
  char source_path[___PATH_MAX_LENGTH+1];
  char project_path[___PATH_MAX_LENGTH+1];
  char model_path[___PATH_MAX_LENGTH+1];
  Str255 source_ppath;
  Str255 project_ppath;
  Str255 output_ppath;
  Str255 model_ppath;
  char *p;

  /******* we should check that we will not overflow the strings ******/

  sprintf (source_path, "%s.c", root_path);
  sprintf (project_path, "%s.prj", output_path);

  p = output_path; /* strip directory */
  while (*p != '\0')
    {
      if (DIR_SEPARATOR(*p))
        output_path = p+1;
      p++;
    }

  if (path_expand_to_absolute (cc_cmd, 0, model_path, ___PATH_MAX_LENGTH)
      != ___FIX(___NO_ERR) ||
      !CWIDE_is_a_regular_file (model_path))
    {
      *errmsg = "The model project file is missing";
      return 1;
    }

  if (!c2pascal (source_path, source_ppath, sizeof (source_ppath)-1) ||
      !c2pascal (project_path, project_ppath, sizeof (project_ppath)-1) ||
      !c2pascal (output_path, output_ppath, sizeof (output_ppath)-1) ||
      !c2pascal (model_path, model_ppath, sizeof (model_ppath)-1) ||
      (err = make_ResolvedFSSpec (0, 0, source_ppath, &source_spec)) != noErr ||
      (err = make_ResolvedFSSpec (0, 0, project_ppath, &project_spec)) != noErr ||
      (err = make_ResolvedFSSpec (0, 0, model_ppath, &model_spec)) != noErr ||
      (err = copy_file (model_spec, project_spec)) != noErr ||
      (err = setup_CWIDE ()) != noErr)
    {
      *errmsg = "The setting up of the CodeWarrior IDE failed";
      return 1;
    }

  if ((err = CWIDE_close_project ()) != noErr ||
      (err = CWIDE_open_project (&project_spec)) != noErr)
    {
      if (err == connectionInvalid)
        *errmsg = "The CodeWarrior IDE is not running";
      else
        *errmsg = "The CodeWarrior IDE failed to open the project file";
    }
  else if ((err = CWIDE_add_file (&source_spec, 0, &CWIDE_err)) != noErr)
    *errmsg = "The CodeWarrior IDE failed to add the source file to the project";
  else if (CWIDE_err == errShell_FileNotFound)
    {
      err = fnfErr;
      *errmsg = "The source file could not be found";
    }
  else if ((err = CWIDE_set_output (output_ppath)) != noErr)
    *errmsg = "The CodeWarrior IDE failed to set the output file";
  else if ((err = CWIDE_make_project (&CWIDE_err)) != noErr)
    *errmsg = "The CodeWarrior IDE failed to make the project";
  else if (CWIDE_err > 0)
    {
      err = fnfErr; /* anything can do */
      *errmsg = "Compilation failed at C level";
    }
  else if ((err = CWIDE_close_project ()) != noErr)
    *errmsg = "The CodeWarrior IDE could not close the project file";
  else
    remove (project_path);

  cleanup_CWIDE ();

  return err != noErr;
}

#endif

#ifndef USE_CWIDE

___BOOL ___dynamic_cc
   ___P((char *cc_cmd,
         char *ld_cmd,
         char *root_path,
         char *output_path,
         char *cc_options,
         char *ld_options,
         ___BOOL ld_flip,
         char **errmsg),
        (cc_cmd,
         ld_cmd,
         root_path,
         output_path,
         cc_options,
         ld_options,
         ld_flip,
         errmsg)
char *cc_cmd;
char *ld_cmd;
char *root_path;
char *output_path;
char *cc_options;
char *ld_options;
___BOOL ld_flip;
char **errmsg;)
{
  char cmd[___CMD_MAX_LENGTH+1];

  /******* we should check that we will not overflow the string ******/

  sprintf (cmd, cc_cmd, cc_options, root_path, root_path);

  if (system (cmd) != 0)
    {
      *errmsg = "Compilation failed at C level";
      return 1;
    }

  if (ld_flip)
    sprintf (cmd, ld_cmd, ld_options, output_path, root_path);
  else
    sprintf (cmd, ld_cmd, ld_options, root_path, output_path);

  if (system (cmd) != 0)
    {
      *errmsg = "Linking failed at C level";
      return 1;
    }

  return 0;
}

#endif


/*---------------------------------------------------------------------------*/

/* Allocation of C closures. */

#define USE_dynamic_code_gen///////////////

#ifdef USE_dynamic_code_gen

#ifdef ___CPU_x86
#define C_CLOSURE_CODE_SIZE 16
#endif

#ifdef ___CPU_ppc
#define C_CLOSURE_CODE_SIZE (9*4)
#ifndef __GNUC__
#undef C_CLOSURE_CODE_SIZE
#endif
#endif

#ifdef ___CPU_sparc
#if ___WORD_SIZE == 32
#define C_CLOSURE_CODE_SIZE (7*4)
#else
#define C_CLOSURE_CODE_SIZE (19*4)
#endif
#ifndef __GNUC__
#ifndef USE_MACOS_CLASSIC /*******?????????????******/
#undef C_CLOSURE_CODE_SIZE
#endif
#endif
#endif

#endif


#ifdef C_CLOSURE_CODE_SIZE

___LOCAL void sync_icache_and_dcache (void *start, int length)
{
#ifdef ___CPU_x86

  /* The x86 processor always keeps the icache and dcache in sync. */

#endif

#ifdef ___CPU_ppc

#ifdef __GNUC__

#define CACHE_BLOCK_SIZE 4 /* we are conservative! */

  ___U8 *s = ___CAST(___U8*,___CAST(long,start) & -CACHE_BLOCK_SIZE);

  do
    {
      __asm__ __volatile__ ("dcbf 0,%0" : : "g" (s) : "memory");
      __asm__ __volatile__ ("sync" : : : "memory");
      __asm__ __volatile__ ("icbi 0,%0" : : "g" (s) : "memory");
      s += CACHE_BLOCK_SIZE;
      length -= CACHE_BLOCK_SIZE;
    } while (length > 0);

  __asm__ __volatile__ ("sync" : : : "memory");
  __asm__ __volatile__ ("isync" : : : "memory");

#else

#ifdef USE_MACOS_CLASSIC

  MakeDataExecutable (start, length);

#endif

#endif

#ifdef ___CPU_sparc

#ifdef __GNUC__

#define MACHINE_WORD_SIZE 4

  ___U8 *s = ___CAST(___U8*,___CAST(long,start) & -MACHINE_WORD_SIZE);

  do
    {
      __asm__ __volatile__ ("flush %0" : : "g" (s) : "memory");
      s += MACHINE_WORD_SIZE;
      length -= MACHINE_WORD_SIZE;
    } while (length > 0);

#endif

#endif

#endif
}

#endif


___LOCAL void *c_closure_self; /* set by the C closure trampoline code */


void *___make_c_closure
   ___P((___SCMOBJ proc,
         void *converter),
        (proc,
         converter)
___SCMOBJ proc;
void *converter;)
{
  void *c_closure;

#ifndef C_CLOSURE_CODE_SIZE

  c_closure = 0;

#else

  if ((c_closure = ___alloc_rc (C_CLOSURE_CODE_SIZE)) == 0)
    return 0;

  ___set_data_rc (c_closure, proc);

  /* Generate the trampoline code of the C closure. */

  /*
   * A "C closure" is a short piece of machine code that does the following:
   *
   *   1) stores its own address in the global variable "c_closure_self"
   *   2) jumps to the "converter" function (which is a C function generated
   *      by the Gambit-C compiler's C-interface)
   *
   * The code must not change any processor register or stack location
   * that is used in the C calling convention.  In this way, the
   * converter function will think that it received the parameters
   * directly from the caller.  The only side-effect of the C closure
   * code is that the global variable "c_closure_self" will point to
   * the C closure, allowing the converter function to access other
   * data stored with the C closure (namely a reference to the Scheme
   * procedure that the converter function needs to call).
   *
   * After generating the trampoline code of the C closure, it is
   * important to synchronize the instruction and data caches so that
   * the processor does not execute stale instructions in the instruction
   * cache.
   */

#ifdef ___CPU_x86

  {
    ___U8 *p = ___CAST(___U8*,c_closure);

                                     /* x86 machine code                     */

    p[ 0] = 0x68;                    /* pushl $p                             */
    p[ 1] = ___CAST_U8(___CAST_U32(p));
    p[ 2] = ___CAST_U8(___CAST_U32(p)>>8);
    p[ 3] = ___CAST_U8(___CAST_U32(p)>>16);
    p[ 4] = ___CAST_U8(___CAST_U32(p)>>24);
    p[ 5] = 0x8f;                    /* popl  c_closure_self                 */
    p[ 6] = 0x05;
    p[ 7] = ___CAST_U8(___CAST_U32(&c_closure_self));
    p[ 8] = ___CAST_U8(___CAST_U32(&c_closure_self)>>8);
    p[ 9] = ___CAST_U8(___CAST_U32(&c_closure_self)>>16);
    p[10] = ___CAST_U8(___CAST_U32(&c_closure_self)>>24);
    p[11] = 0xe9;                    /* jmp   converter                      */
    p[12] = ___CAST_U8(___CAST_U32(___CAST(___U8*,converter)-(p+16)));
    p[13] = ___CAST_U8(___CAST_U32(___CAST(___U8*,converter)-(p+16))>>8);
    p[14] = ___CAST_U8(___CAST_U32(___CAST(___U8*,converter)-(p+16))>>16);
    p[15] = ___CAST_U8(___CAST_U32(___CAST(___U8*,converter)-(p+16))>>24);
  }

#endif

#ifdef ___CPU_ppc

  {
    ___U32 *p = ___CAST(___U32*,c_closure);

                                     /* PowerPC machine code                 */

    p[0] = 0x39600000                /* li    r11,LO16(converter)            */
           + ___CAST_U32(___CAST_U16(___CAST_U32(converter)));
    p[1] = 0x3d6b0000                /* addis r11,r11,HI16(converter)        */
           + (___CAST_U32(converter) >> 16)
           + ((___CAST_U32(converter) & 0x8000) ? 1 : 0);
    p[2] = 0x7d6903a6;               /* mtspr ctr,r11                        */
    p[3] = 0x39600000                /* li    r11,LO16(p)                    */
           + ___CAST_U32(___CAST_U16(___CAST_U32(p)));
    p[4] = 0x3d6b0000                /* addis r11,r11,HI16(p)                */
           + (___CAST_U32(p) >> 16)
           + ((___CAST_U32(p) & 0x8000) ? 1 : 0);
    p[5] = 0x380b0000;               /* addi  r0,r11,0                       */
    p[6] = 0x3d600000                /* lis   r11,HI16(c_closure_self)       */
           + (___CAST_U32(&c_closure_self) >> 16)
           + ((___CAST_U32(&c_closure_self) & 0x8000) ? 1 : 0);
    p[7] = 0x900b0000                /* stw   r0,LO16(c_closure_self)(r11)   */
           + ___CAST_U32(___CAST_U16(___CAST_U32(&c_closure_self)));
    p[8] = 0x4e800420;               /* bctr                                 */
  }

#endif

#ifdef ___CPU_sparc

  {
    ___U32 *p = ___CAST(___U32*,c_closure);

                                    /* SPARC machine code                   */

#if ___WORD_WIDTH == 32

    p[0] = 0x07000000                /* sethi HI22(p),%g3                    */
           + (___CAST_U32(p) >> 10);
    p[1] = 0x8610E000                /* or    %g3,LO10(p),%g3                */
           + (___CAST_U32(p) & 0x3ff);
    p[2] = 0x05000000                /* sethi HI22(c_closure_self),%g2       */
           + (___CAST_U32(&c_closure_self) >> 10);
    p[3] = 0xC620A000                /* st    %g3,[%g2+LO10(c_closure_self)] */
           + (___CAST_U32(&c_closure_self) & 0x3ff);
    p[4] = 0x05000000                /* sethi HI22(converter),%g2            */
           + (___CAST_U32(converter) >> 10);
    p[5] = 0x81C0A000                /* jmp   %g2+LO10(converter)            */
           + (___CAST_U32(converter) & 0x3ff);
    p[6] = 0x01000000;               /* nop                                  */

#else

    p[ 0] = 0x07000000               /* sethi HHI22(p),%g3                   */
           + (___CAST_U64(p) >> (10+32));
    p[ 1] = 0x8610E000               /* or    %g3,HLO10(p),%g3               */
           + ((___CAST_U64(p) >> 32) & 0x3ff);
    p[ 2] = 0x8728F020;              /* sllx  %g3,32,%g3                     */
    p[ 3] = 0x03000000               /* sethi LHI22(p),%g1                   */
           + ((___CAST_U64(p) >> 10) & 0x3fffff);
    p[ 4] = 0x82106000               /* or    %g1,LLO10(p),%g1               */
           + (___CAST_U64(p) & 0x3ff);
    p[ 5] = 0x8600C001;              /* add   %g3,%g1,%g3                    */
    p[ 6] = 0x05000000               /* sethi HHI22(c_closure_self),%g2      */
           + (___CAST_U64(&c_closure_self) >> (10+32));
    p[ 7] = 0x8410A000               /* or    %g2,HLO10(c_closure_self),%g2  */
           + ((___CAST_U64(&c_closure_self) >> 32) & 0x3ff);
    p[ 8] = 0x8528B020;              /* sllx  %g2,32,%g2                     */
    p[ 9] = 0x03000000               /* sethi LHI22(c_closure_self),%g1      */
           + ((___CAST_U64(&c_closure_self) >> 10) & 0x3fffff);
    p[10] = 0x84008001;              /* add   %g2,%g1,%g2                    */
    p[11] = 0xC670A000               /* stx   %g3,[%g2+LLO10(c_closure_self)]*/
           + (___CAST_U64(&c_closure_self) & 0x3ff);
    p[12] = 0x05000000               /* sethi HHI22(converter),%g2           */
           + (___CAST_U64(converter) >> (10+32));
    p[13] = 0x8410A000               /* or    %g2,HLO10(converter),%g2       */
           + ((___CAST_U64(converter) >> 32) & 0x3ff);
    p[14] = 0x8528B020;              /* sllx  %g2,32,%g2                     */
    p[15] = 0x03000000               /* sethi LHI22(converter),%g1           */
           + ((___CAST_U64(converter) >> 10) & 0x3fffff);
    p[16] = 0x84008001;              /* add   %g2,%g1,%g2                    */
    p[17] = 0x81C0A000               /* jmp   %g2+LLO10(converter)           */
           + (___CAST_U64(converter) & 0x3ff);
    p[18] = 0x01000000;              /* nop                                  */

#endif
  }

#endif

    sync_icache_and_dcache (c_closure, C_CLOSURE_CODE_SIZE);

#endif

  return c_closure;
}


___BOOL ___is_a_c_closure
   ___P((void *fn),
        (fn)
void *fn;)
{
#ifndef C_CLOSURE_CODE_SIZE

  return 0;

#else

  /*
   * We check that the function's code contains the same sequence that
   * is generated by ___make_c_closure.  If there is a match, we are
   * sure that the function was generated by ___make_c_closure because
   * it is impossible for the C compiler to have generated this code
   * for a valid C function.
   */

#ifdef ___CPU_x86

  {
    ___U8 *p = ___CAST(___U8*,fn);

    return p != 0 &&
           p[ 0] == 0x68 &&
           p[ 1] == ___CAST_U8(___CAST_U32(p)) &&
           p[ 2] == ___CAST_U8(___CAST_U32(p)>>8) &&
           p[ 3] == ___CAST_U8(___CAST_U32(p)>>16) &&
           p[ 4] == ___CAST_U8(___CAST_U32(p)>>24) &&
           p[ 5] == 0x8f &&
           p[ 6] == 0x05 &&
           p[ 7] == ___CAST_U8(___CAST_U32(&c_closure_self)) &&
           p[ 8] == ___CAST_U8(___CAST_U32(&c_closure_self)>>8) &&
           p[ 9] == ___CAST_U8(___CAST_U32(&c_closure_self)>>16) &&
           p[10] == ___CAST_U8(___CAST_U32(&c_closure_self)>>24) &&
           p[12] == 0xe9;
  }

#endif

#ifdef ___CPU_ppc

  {
    ___U32 *p = ___CAST(___U32*,fn);

    return p != 0 &&
           (p[0] >> 16) == 0x3d6b &&
           p[1] == 0x7d6903a6 &&
           (p[2] >> 16) == 0x3960 &&
           p[3] == 0x39600000
                   + ___CAST_U32(___CAST_U16(___CAST_U32(p))) &&
           p[4] == 0x3d6b0000
                   + (___CAST_U32(p) >> 16)
                   + ((___CAST_U32(p) & 0x8000) ? 1 : 0) &&
           p[5] == 0x380b0000 &&
           p[6] == 0x3d600000
                   + (___CAST_U32(&c_closure_self) >> 16)
                   + ((___CAST_U32(&c_closure_self) & 0x8000) ? 1 : 0) &&
           p[7] == 0x900b0000
                   + ___CAST_U32(___CAST_U16(___CAST_U32(&c_closure_self))) &&
           p[8] == 0x4e800420;
  }

#endif

#ifdef ___CPU_sparc

  {
    ___U32 *p = ___CAST(___U32*,fn);

#if ___WORD_WIDTH == 32

    return p != 0 &&
           p[0] == 0x07000000
                   + (___CAST_U32(p) >> 10) &&
           p[1] == 0x8610E000
                   + (___CAST_U32(p) & 0x3ff) &&
           p[2] == 0x05000000
                   + (___CAST_U32(&c_closure_self) >> 10) &&
           p[3] == 0xC620A000
                   + (___CAST_U32(&c_closure_self) & 0x3ff) &&
           (p[4] >> 22) == (0x05000000 >> 22) &&
           (p[5] >> 10) == (0x81C0A000 >> 10) &&
           p[6] == 0x01000000;

#else

    return p != 0 &&
           p[ 0] == 0x07000000
                   + (___CAST_U64(p) >> (10+32)) &&
           p[ 1] == 0x8610E000
                   + ((___CAST_U64(p) >> 32) & 0x3ff) &&
           p[ 2] == 0x8728F020 &&
           p[ 3] == 0x03000000
                   + ((___CAST_U64(p) >> 10) & 0x3fffff) &&
           p[ 4] == 0x82106000
                   + (___CAST_U64(p) & 0x3ff) &&
           p[ 5] == 0x8600C001 &&
           p[ 6] == 0x05000000
                   + (___CAST_U64(&c_closure_self) >> (10+32)) &&
           p[ 7] == 0x8410A000
                   + ((___CAST_U64(&c_closure_self) >> 32) & 0x3ff) &&
           p[ 8] == 0x8528B020 &&
           p[ 9] == 0x03000000
                   + ((___CAST_U64(&c_closure_self) >> 10) & 0x3fffff) &&
           p[10] == 0x84008001 &&
           p[11] == 0xC670A000
                   + (___CAST_U64(&c_closure_self) & 0x3ff) &&
           (p[12] >> 22) == (0x05000000 >> 22) &&
           (p[13] >> 10) == (0x8410A000 >> 10) &&
           p[14] == 0x8528B020 &&
           (p[15] >> 22) == (0x03000000 >> 22) &&
           p[16] == 0x84008001 &&
           (p[17] >> 10) == (0x81C0A000 >> 10) &&
           p[18] == 0x01000000;

#endif
  }

#endif

#endif
}


void ___c_closure_release
   ___P((void *c_closure),
        (c_closure)
void *c_closure;)
{
  ___release_rc (c_closure);
}


void *___c_closure_self ___PVOID
{
  return c_closure_self;
}


/*---------------------------------------------------------------------------*/

/* Floating point environment setup. */

___HIDDEN void setup_fp ___PVOID
{
#ifdef HAVE_GET_FPC_CSR

  /* Enable denormalized numbers. */

  union fpc_csr csr;

  csr.fc_word = get_fpc_csr ();
  csr.fc_struct.flush = 0;
  set_fpc_csr (csr.fc_word);

#endif

  
#ifdef USE_control87

#ifdef __LCC__

#define FP_EXC_MASK _MCW_EM
#define FP_EXC_CW \
(_EM_INVALID+_EM_ZERODIVIDE+_EM_OVERFLOW+_EM_UNDERFLOW+_EM_INEXACT+_EM_DENORMAL)

  _control87 (FP_EXC_CW, FP_EXC_MASK);

#else

#define FP_EXC_MASK MCW_EM
#define FP_EXC_CW \
(EM_INVALID+EM_ZERODIVIDE+EM_OVERFLOW+EM_UNDERFLOW+EM_INEXACT+EM_DENORMAL)

  _control87 (FP_EXC_CW, FP_EXC_MASK);

#endif

#endif

#ifdef USE__FPU_SETCW

#define FPU_CW \
(_FPU_MASK_IM+_FPU_MASK_ZM+_FPU_MASK_OM+_FPU_MASK_UM+_FPU_MASK_PM+ \
_FPU_MASK_DM+_FPU_DOUBLE+_FPU_RC_NEAREST)

  fpu_control_t cw = FPU_CW;
  _FPU_SETCW (cw);

#endif
}


___HIDDEN void cleanup_fp ___PVOID
{
}


/*---------------------------------------------------------------------------*/

/* Virtual memory statistics. */

void ___vm_stats
   ___P((long *minflt,
         long *majflt),
        (minflt,
         majflt)
long *minflt;
long *majflt;)
{
#ifndef USE_getrusage

  *minflt = 0; /* can't get statistics... result is 0 */
  *majflt = 0;

#endif

#ifdef USE_getrusage

  struct rusage ru;

  if (getrusage (RUSAGE_SELF, &ru) == 0)
    {
      *minflt = ru.ru_minflt;
      *majflt = ru.ru_majflt;
    }
  else
    {
      *minflt = 0; /* can't get statistics... result is 0 */
      *majflt = 0;
    }

#endif
}


/*---------------------------------------------------------------------------*/

/* Filesystem path expansion. */

#ifdef USE_POSIX
#define ___PATH_CE_SELECT(latin1,utf8,ucs2,ucs4,native) native
#endif

#ifdef USE_WIN32
#ifdef UNICODE
#define ___PATH_CE_SELECT(latin1,utf8,ucs2,ucs4,native) ucs2
#else
#define ___PATH_CE_SELECT(latin1,utf8,ucs2,ucs4,native) latin1
#endif
#endif


___SCMOBJ ___os_path_gambcdir ___PVOID
{
  ___SCMOBJ e;
  ___SCMOBJ result;

  if (___setup_params.gambcdir != 0)
    {
      if ((e = ___NONNULLUCS2STRING_to_SCMOBJ
                 (___setup_params.gambcdir,
                  &result,
                  ___RETURN_POS))
          != ___FIX(___NO_ERR))
        result = e;
      else
        ___release_scmobj (result);
    }
  else
    {

#ifndef ___GAMBCDIR

#ifdef USE_POSIX
#define ___GAMBCDIR "/usr/local/Gambit-C"
#endif

#ifdef USE_WIN32
#define ___GAMBCDIR "c:\\Gambit-C"
#endif

#ifdef USE_MACOS_CLASSIC
#define ___GAMBCDIR ":Gambit-C"
#endif

#endif

      if ((e = ___NONNULLCHARSTRING_to_SCMOBJ
                 (___GAMBCDIR,
                  &result,
                  ___RETURN_POS))
          != ___FIX(___NO_ERR))
        result = e;
      else
        ___release_scmobj (result);
    }

  return result;
}


___SCMOBJ ___os_path_normalize_directory
   ___P((___SCMOBJ path),
        (path)
___SCMOBJ path;)
{
  ___SCMOBJ e;
  ___SCMOBJ result;
  void *cpath;

  if ((e = ___SCMOBJ_to_STRING
             (path,
              &cpath,
              1,
              ___CE(___PATH_CE_SELECT),
              0))
      != ___FIX(___NO_ERR))
    result = e;
  else
    {
      ___STRING_TYPE(___PATH_CE_SELECT) p =
        ___CAST(___STRING_TYPE(___PATH_CE_SELECT),cpath);
      ___STRING_TYPE(___PATH_CE_SELECT) dir;

#ifdef USE_POSIX

      ___CHAR_TYPE(___PATH_CE_SELECT) old_dir[___PATH_MAX_LENGTH+1+1];
      ___CHAR_TYPE(___PATH_CE_SELECT) normalized_dir[___PATH_MAX_LENGTH+1+1];

      dir = normalized_dir;

      if (getcwd (old_dir, ___PATH_MAX_LENGTH) == 0)
        result = err_code_from_errno ();
      else
        {
          if (p == 0)
            dir = old_dir;
          else
            {
              if (chdir (p) < 0)
                e = err_code_from_errno ();
              else
                {
                  if (getcwd (normalized_dir, ___PATH_MAX_LENGTH) == 0)
                    e = err_code_from_errno ();
                  else
                    e = ___FIX(___NO_ERR);
                }
              chdir (old_dir); /* ignore error */
            }

          if (e != ___FIX(___NO_ERR))
            result = e;
          else
            {
              p = dir;

              while (*p != '\0')
                p++;

              if (p == dir || p[-1] != '/')
                {
                  *p++ = '/';
                  *p++ = '\0';
                }

              if ((e = ___NONNULLSTRING_to_SCMOBJ
                         (dir,
                          &result,
                          ___RETURN_POS,
                          ___CE(___PATH_CE_SELECT)))
                  != ___FIX(___NO_ERR))
                result = e;
              else
                ___release_scmobj (result);
            }
        }

#endif
      ___release_string (cpath);
    }

  return result;
}


/*---------------------------------------------------------------------------*/

/* Formatting of source code position. */

char *___format_filepos
   ___P((char *path,
         long filepos,
         ___BOOL pinpoint),
        (path,
         filepos,
         pinpoint)
char *path;
long filepos;
___BOOL pinpoint;)
{
#ifdef USE_MACOS

#ifdef USE_mac_gui

  if (pinpoint)
    mac_gui_highlight (path, filepos);

#endif

#endif

  return 0; /* Use default format for displaying location */
}


/*---------------------------------------------------------------------------*/

/* Operations on I/O devices. */

/* User module setup. */

typedef struct user_module_struct
  {
    int refcount; /* number of threads using this structure */

#ifndef USE_WIN32

#define USER_MODULE_INIT_OS_SPECIFIC

#endif

#ifdef USE_WIN32

    HANDLE always_signaled;  /* this event is always signaled */
    HANDLE abort_select;     /* ___device_select exits when this is signaled */

#define USER_MODULE_INIT_OS_SPECIFIC , NULL, NULL

#endif
  } user_module;

___HIDDEN user_module user_mod =
{
  0
  USER_MODULE_INIT_OS_SPECIFIC
};


#ifdef USE_WIN32

___HIDDEN BOOL WINAPI console_event_handler
   ___P((DWORD dwCtrlType),
        (dwCtrlType)
DWORD dwCtrlType;)
{
#if 0
  /*****************************/
  switch (dwCtrlType)
    {
    case CTRL_C_EVENT:
      user_mod.got_event = 1;
      break;
    case CTRL_BREAK_EVENT:
      user_mod.got_event = 2;
      break;
    case CTRL_CLOSE_EVENT:
      user_mod.got_event = 3;
      break;
    case CTRL_LOGOFF_EVENT:
      user_mod.got_event = 4;
      break;
    case CTRL_SHUTDOWN_EVENT:
      user_mod.got_event = 5;
      break;
    default:
      user_mod.got_event = 999;
    }

  SetEvent (user_mod.abort_select); /* ignore error */

  return TRUE;
#else
  return FALSE;
#endif
}

#endif


___HIDDEN ___SCMOBJ user_module_setup ___PVOID
{
  if (user_mod.refcount == 0)
    {
#ifdef USE_WIN32

      ___SCMOBJ e = ___FIX(___NO_ERR);
      DWORD thread_id;

      user_mod.always_signaled = NULL;
      user_mod.abort_select = NULL;

      user_mod.always_signaled =
        CreateEvent (NULL,  /* can't inherit */
                     TRUE,  /* manual reset */
                     TRUE,  /* signaled */
                     NULL); /* no name */

      if (user_mod.always_signaled == NULL)
        e = err_code_from_GetLastError ();
      else
        {
          user_mod.abort_select =
            CreateEvent (NULL,  /* can't inherit */
                         TRUE,  /* manual reset */
                         FALSE, /* not signaled */
                         NULL); /* no name */

          if (user_mod.abort_select == NULL)
            e = err_code_from_GetLastError ();
          else
            {
              if (!SetConsoleCtrlHandler (console_event_handler, TRUE))/****************zzzzzzzzzzzzzzz*/
                e = err_code_from_GetLastError ();
            }
        }

      if (e != ___FIX(___NO_ERR))
        {
          if (user_mod.abort_select != NULL)
            CloseHandle (user_mod.abort_select); /* ignore error */

          if (user_mod.always_signaled != NULL)
            CloseHandle (user_mod.always_signaled); /* ignore error */

          return e;
        }

#endif
    }

  user_mod.refcount++;

  return ___FIX(___NO_ERR);
}


___HIDDEN void user_module_cleanup ___PVOID
{
  if (--user_mod.refcount == 0)
    {
#ifdef USE_WIN32

      SetConsoleCtrlHandler (console_event_handler, FALSE); /* ignore error *//**************zzzzzzzzzzzzz*/

      CloseHandle (user_mod.abort_select); /* ignore error */

      CloseHandle (user_mod.always_signaled); /* ignore error */

#endif
    }
}


/*---------------------------------------------------------------------------*/

#ifdef USE_PUMPS

/* Nonblocking pipes */

/*
 * A pipe is a unidirectional FIFO data structure that buffers data
 * between a "writer" that writes data to the pipe and a "reader" that
 * reads this data.  Independently from the FIFO buffer, a pipe has a
 * "reader error indicator" and a "writer error indicator".  When
 * created the error indicators are set to "no error" and the FIFO
 * contains no data.

 * Typically the writer writes bytes and out-of-band messages at one
 * end with a "write" operation and the reader reads these bytes and
 * out-of-band messages in the same order at the other end with a
 * "read" operation.  Out-of-band messages are used to request some
 * special action from the reader and are expected to be infrequent.
 * When a read operation extracts an out-of-band message from the pipe
 * it reports that zero bytes can be read from the pipe and sets the
 * out-of-band structure (one of the parameters of the read
 * operation).  The writer can set the "writer error indicator" when
 * it needs to transmit an error code to the reader.  After all
 * previously written bytes and out-of-band messages are read by the
 * reader, the next read operation will return this error code and the
 * error indicator is reset to "no error".  Similarly, the reader can
 * set the "reader error indicator" when it needs to transmit an error
 * code to the writer.  The next write operation will return this
 * error code and the error indicator is reset to "no error".

 * The nonblocking pipes implemented here have read and write
 * operations that don't block the calling thread.  When the data
 * transfer requested can't be performed immediately these operations
 * will return with a special error code (EAGAIN).  This allows the
 * calling thread to do other things before attempting the operation
 * again.  We say a pipe is "ready for an operation" if calling that
 * operation will not return EAGAIN (i.e. the operations will succeed
 * with no error or will fail with some error different from EAGAIN).

 * Specifically,
 *
 *   a pipe is ready for reading when
 *     - the "writer error indicator" is not "no error"
 *     - or
 *       - the "reader error indicator" is "no error"
 *       - and
 *         - an out-of-band message is in the pipe
 *         - or at least one byte is in the FIFO buffer
 *
 *   a pipe is ready for writing when
 *     - the "reader error indicator" is not "no error"
 *     - or
 *       - the "writer error indicator" is "no error"
 *       - and no out-of-band message is in the pipe
 *       - and at least one byte is free in the FIFO buffer
 *       - and the previous read operation did not return an error or
 *         an out-of-band message
 *         (at creation a pipe is not ready for writing and the first
 *         "read" operation will make the pipe ready for writing, in other
 *         words, after it reports an error or out-of-band message, the
 *         writer blocks until the reader shows some interest in more data)

 * This table summarizes the states of a pipe (an R means ready for
 * reading and a W means ready for writing and a w means ready for
 * writing iff the most recent read operation did not return an
 * error or out-of-band message):

 *        no writer :  writer     reader      bytes in FIFO buffer
 *          error   :  error      error?      (N is size of buffer)
 *        +----+----+----+----+
 *        |  w | R  | R  | R  |     no           0 bytes
 *        +----+----+----+----+
 *        | Rw | R  | R  | R  |     no           > 0 bytes but < N
 *        +----+----+----+----+
 *        | R  | R  | R  | R  |     no           = N
 *        +----+----+----+----+
 *        |  W |  W | RW | RW |     yes          0 bytes
 *        +----+----+----+----+
 *        |  W |  W | RW | RW |     yes          > 0 bytes but < N
 *        +----+----+----+----+
 *        |  W |  W | RW | RW |     yes          = N
 *        +----+----+----+----+
 *        :!oob:oob :!oob:oob :

 */

typedef struct nonblocking_pipe_oob_msg_struct
  {
    int op;         /* operation */
    DWORD param[2]; /* parameters */
  } nonblocking_pipe_oob_msg;

typedef struct nonblocking_pipe_struct
  {
    HANDLE mutex;     /* mutex to serialize access to the pipe */
    HANDLE revent;    /* event, signaled iff ready for reading */
    HANDLE wevent;    /* event, signaled iff ready for writing */
    ___SCMOBJ rerr;   /* the reader error indicator */
    ___SCMOBJ werr;   /* the writer error indicator */
    ___BOOL oob;      /* the out-of-band indicator */
    nonblocking_pipe_oob_msg oob_msg; /* out-of-band message */
    int rd;           /* the read pointer of the circular buffer */
    int wr;           /* the write pointer of the circular buffer */
    int size;         /* size of the circular buffer */
    ___U8 *buffer;    /* the circular buffer */
  } nonblocking_pipe;

___HIDDEN ___SCMOBJ nonblocking_pipe_setup
   ___P((nonblocking_pipe *pipe,
         int size),
        (pipe,
         size)
nonblocking_pipe *pipe;
int size;)
{
  ___U8 *buffer;
  HANDLE mutex;
  HANDLE revent;
  HANDLE wevent;

  buffer = ___CAST(___U8*,
                   ___alloc_mem (size));

  if (buffer == NULL)
    return ___FIX(___HEAP_OVERFLOW_ERR);

  mutex = CreateMutex (NULL,  /* can't inherit */
                       FALSE, /* unlocked */
                       NULL); /* no name */

  if (mutex == NULL)
    {
      ___SCMOBJ e = err_code_from_GetLastError ();
      ___free_mem (buffer);
      return e;
    }

  revent = CreateEvent (NULL,  /* can't inherit */
                        TRUE,  /* manual reset */
                        FALSE, /* not signaled */
                        NULL); /* no name */

  if (revent == NULL)
    {
      ___SCMOBJ e = err_code_from_GetLastError ();
      CloseHandle (mutex); /* ignore error */
      ___free_mem (buffer);
      return e;
    }

  wevent = CreateEvent (NULL,  /* can't inherit */
                        TRUE,  /* manual reset */
                        FALSE, /* not signaled */
                        NULL); /* no name */

  if (wevent == NULL)
    {
      ___SCMOBJ e = err_code_from_GetLastError ();
      CloseHandle (revent); /* ignore error */
      CloseHandle (mutex); /* ignore error */
      ___free_mem (buffer);
      return e;
    }

  pipe->mutex = mutex;
  pipe->revent = revent;
  pipe->wevent = wevent;
  pipe->rerr = ___FIX(___NO_ERR);
  pipe->werr = ___FIX(___NO_ERR);
  pipe->oob = 0;
  pipe->rd = 0;
  pipe->wr = 0;
  pipe->size = size;
  pipe->buffer = buffer;

  return ___FIX(___NO_ERR);
}

___HIDDEN ___SCMOBJ nonblocking_pipe_cleanup
   ___P((nonblocking_pipe *pipe),
        (pipe)
nonblocking_pipe *pipe;)
{
  CloseHandle (pipe->wevent); /* ignore error */
  CloseHandle (pipe->revent); /* ignore error */
  CloseHandle (pipe->mutex); /* ignore error */
  ___free_mem (pipe->buffer);

  return ___FIX(___NO_ERR);
}

___HIDDEN ___SCMOBJ nonblocking_pipe_set_reader_err
   ___P((nonblocking_pipe *pipe,
         ___SCMOBJ err),
        (pipe,
         err)
nonblocking_pipe *pipe;
___SCMOBJ err;)
{
  if (WaitForSingleObject (pipe->mutex, INFINITE) == WAIT_FAILED)
    return err_code_from_GetLastError ();

  /* note: the reader error indicator may get overwritten */

  pipe->rerr = err;

  SetEvent (pipe->wevent); /* ignore error */

  if (pipe->werr == ___FIX(___NO_ERR))
    ResetEvent (pipe->revent); /* ignore error */

  ReleaseMutex (pipe->mutex); /* ignore error */

  return ___FIX(___NO_ERR);
}

___HIDDEN ___SCMOBJ nonblocking_pipe_set_writer_err
   ___P((nonblocking_pipe *pipe,
         ___SCMOBJ err),
        (pipe,
         err)
nonblocking_pipe *pipe;
___SCMOBJ err;)
{
  if (WaitForSingleObject (pipe->mutex, INFINITE) == WAIT_FAILED)
    return err_code_from_GetLastError ();

  /* note: the writer error indicator may get overwritten */

  pipe->werr = err;

  SetEvent (pipe->revent); /* ignore error */

  if (pipe->rerr == ___FIX(___NO_ERR))
    ResetEvent (pipe->wevent); /* ignore error */

  ReleaseMutex (pipe->mutex); /* ignore error */

  return ___FIX(___NO_ERR);
}

___HIDDEN ___SCMOBJ nonblocking_pipe_write_oob
   ___P((nonblocking_pipe *pipe,
         nonblocking_pipe_oob_msg *oob_msg),
        (pipe,
         oob_msg)
nonblocking_pipe *pipe;
nonblocking_pipe_oob_msg *oob_msg;)
{
  if (WaitForSingleObject (pipe->mutex, INFINITE) == WAIT_FAILED)
    return err_code_from_GetLastError ();

  if (pipe->werr != ___FIX(___NO_ERR) || pipe->oob)
    {
      ReleaseMutex (pipe->mutex); /* ignore error */
      return ___ERR_CODE_EAGAIN;
    }

  pipe->oob = 1;
  pipe->oob_msg = *oob_msg;

  if (pipe->rerr == ___FIX(___NO_ERR))
    {
      SetEvent (pipe->revent); /* ignore error */
      ResetEvent (pipe->wevent); /* ignore error */
    }

  ReleaseMutex (pipe->mutex); /* ignore error */

  return ___FIX(___NO_ERR);
}

___HIDDEN ___SCMOBJ nonblocking_pipe_read_ready_wait
   ___P((nonblocking_pipe *pipe),
        (pipe)
nonblocking_pipe *pipe;)
{
  ___SCMOBJ werr;

  if (WaitForSingleObject (pipe->revent, INFINITE) == WAIT_FAILED ||
      WaitForSingleObject (pipe->mutex, INFINITE) == WAIT_FAILED)
    return err_code_from_GetLastError ();

  werr = pipe->werr;

  if (werr != ___FIX(___NO_ERR))
    {
      pipe->werr = ___FIX(___NO_ERR);

      if (pipe->rerr != ___FIX(___NO_ERR) ||
          (pipe->rd == pipe->wr && !pipe->oob))
        ResetEvent (pipe->revent); /* ignore error */
    }

  ReleaseMutex (pipe->mutex); /* ignore error */

  return werr;
}

___HIDDEN ___SCMOBJ nonblocking_pipe_write_ready_wait
   ___P((nonblocking_pipe *pipe),
        (pipe)
nonblocking_pipe *pipe;)
{
  ___SCMOBJ rerr;

  if (WaitForSingleObject (pipe->wevent, INFINITE) == WAIT_FAILED ||
      WaitForSingleObject (pipe->mutex, INFINITE) == WAIT_FAILED)
    return err_code_from_GetLastError ();

  rerr = pipe->rerr;

  if (rerr != ___FIX(___NO_ERR))
    {
      pipe->rerr = ___FIX(___NO_ERR);

      ResetEvent (pipe->wevent); /* ignore error */
    }

  ReleaseMutex (pipe->mutex); /* ignore error */

  return rerr;
}

___HIDDEN ___SCMOBJ nonblocking_pipe_read
   ___P((nonblocking_pipe *pipe,
         ___U8 *buf,
         ___stream_index len,
         ___stream_index *len_done,
         nonblocking_pipe_oob_msg *oob_msg),
        (pipe,
         buf,
         len,
         len_done,
         oob_msg)
nonblocking_pipe *pipe;
___U8 *buf;
___stream_index len;
___stream_index *len_done;
nonblocking_pipe_oob_msg *oob_msg;)
{
  ___SCMOBJ rerr;
  ___SCMOBJ werr;
  DWORD rd;
  DWORD wr;
  ___U8 *p;
  DWORD end;
  DWORD n;
  DWORD i;

  if (len <= 0)
    return ___FIX(___UNKNOWN_ERR);

  if (WaitForSingleObject (pipe->mutex, INFINITE) == WAIT_FAILED)
    return err_code_from_GetLastError ();

  rerr = pipe->rerr;
  werr = pipe->werr;
  rd = pipe->rd;
  wr = pipe->wr;

  if (rerr != ___FIX(___NO_ERR))
    {
      /* there is a reader error */

      if (werr == ___FIX(___NO_ERR))
        werr = ___ERR_CODE_EAGAIN;
      else
        {
          pipe->werr = ___FIX(___NO_ERR);
          ResetEvent (pipe->revent); /* ignore error */
        }
      ReleaseMutex (pipe->mutex); /* ignore error */
      return werr;
    }

  /* there is no reader error */

  if (rd == wr)
    {
      /* no bytes in FIFO buffer */

      if (pipe->oob)
        {
          /* out-of-band present */

          *oob_msg = pipe->oob_msg;
          pipe->oob = 0;
          if (werr == ___FIX(___NO_ERR))
            {
              ResetEvent (pipe->revent); /* ignore error */
#if 0
              /******************zzzzzzzzzzzzzz****/
              SetEvent (pipe->wevent); /* ignore error */
#endif
            }
          ReleaseMutex (pipe->mutex); /* ignore error */
          *len_done = 0;
          return ___FIX(___NO_ERR);
        }

      /* out-of-band not present */

      if (werr != ___FIX(___NO_ERR))
        {
          /* there is a writer error */

          pipe->werr = ___FIX(___NO_ERR);

          ResetEvent (pipe->revent); /* ignore error */
          ReleaseMutex (pipe->mutex); /* ignore error */
          return werr;
        }

      /* there is no writer error */

      SetEvent (pipe->wevent); /* ignore error */
      ReleaseMutex (pipe->mutex); /* ignore error */
      return ___ERR_CODE_EAGAIN;
    }

  /* at least one byte in FIFO buffer */

  end = pipe->size - rd; /* number of bytes from rd to end */

  n = wr + end; /* number of bytes available */
  if (n >= pipe->size)
    n -= pipe->size;

  if (n > len) /* don't transfer more than len */
    n = len;

  *len_done = n;

  p = pipe->buffer + rd; /* prepare transfer source */

  rd = rd + n;
  if (rd >= pipe->size)
    rd -= pipe->size;

  pipe->rd = rd;

  if (werr == ___FIX(___NO_ERR) && !pipe->oob)
    {
      /* there is no writer error and out-of-band not present */

      if (rd == wr) /* the FIFO will be empty? */
        ResetEvent (pipe->revent); /* ignore error */

      /* the FIFO will not be full */

      SetEvent (pipe->wevent); /* ignore error */
    }

  if (n <= end)
    {
      /* only need to transfer n bytes starting from original rd */

      for (i=n; i>0; i--)
        *buf++ = *p++;
    }
  else
    {
      /* need to transfer end bytes starting from original rd */

      for (i=end; i>0; i--)
        *buf++ = *p++;

      /* and to transfer n-end bytes starting from 0 */

      p = pipe->buffer;

      for (i=n-end; i>0; i--)
        *buf++ = *p++;
    }

  ReleaseMutex (pipe->mutex); /* ignore error */

  return ___FIX(___NO_ERR);
}

___HIDDEN ___SCMOBJ nonblocking_pipe_write
   ___P((nonblocking_pipe *pipe,
         ___U8 *buf,
         ___stream_index len,
         ___stream_index *len_done),
        (pipe,
         buf,
         len,
         len_done)
nonblocking_pipe *pipe;
___U8 *buf;
___stream_index len;
___stream_index *len_done;)
{
  ___SCMOBJ rerr;
  ___SCMOBJ werr;
  DWORD rd;
  DWORD wr;
  ___U8 *p;
  DWORD end;
  DWORD n;
  DWORD i;

  if (len <= 0)
    return ___FIX(___UNKNOWN_ERR);

  if (WaitForSingleObject (pipe->mutex, INFINITE) == WAIT_FAILED)
    return err_code_from_GetLastError ();

  rerr = pipe->rerr;
  werr = pipe->werr;
  rd = pipe->rd;
  wr = pipe->wr;

  if (werr != ___FIX(___NO_ERR))
    {
      /* there is a writer error */

      if (rerr == ___FIX(___NO_ERR))
        rerr = ___ERR_CODE_EAGAIN;
      else
        {
          pipe->rerr = ___FIX(___NO_ERR);
          ResetEvent (pipe->wevent); /* ignore error */
        }
      ReleaseMutex (pipe->mutex); /* ignore error */
      return rerr;
    }

  /* there is no writer error */

  if (rerr != ___FIX(___NO_ERR))
    {
      /* there is a reader error */

      pipe->rerr = ___FIX(___NO_ERR);

      if (rd != wr || pipe->oob) /* FIFO is not empty */
        SetEvent (pipe->revent); /* ignore error */

      ResetEvent (pipe->wevent); /* ignore error */
      ReleaseMutex (pipe->mutex); /* ignore error */
      return rerr;
    }

  /* there is no reader error */

  if (wr + 1 - rd == 0 || wr + 1 - rd == pipe->size || pipe->oob)
    {
      /* FIFO buffer is full or out-of-band present */

      ReleaseMutex (pipe->mutex); /* ignore error */
      return ___ERR_CODE_EAGAIN;
    }

  /* FIFO buffer is not full and out-of-band not present */

  end = pipe->size - wr; /* number of bytes from wr to end */

  n = rd + end - 1; /* number of bytes available */
  if (n >= pipe->size)
    n -= pipe->size;

  if (n > len) /* don't transfer more than len */
    n = len;

  *len_done = n;

  p = pipe->buffer + wr; /* prepare transfer source */

  wr = wr + n;
  if (wr >= pipe->size)
    wr -= pipe->size;

  pipe->wr = wr;

  if (wr + 1 - rd == 0 || wr + 1 - rd == pipe->size) /* FIFO will be full? */
    ResetEvent (pipe->wevent); /* ignore error */

  /* FIFO will not be empty */

  SetEvent (pipe->revent); /* ignore error */

  if (n <= end)
    {
      /* only need to transfer n bytes starting from original wr */

      for (i=n; i>0; i--)
        *p++ = *buf++;
    }
  else
    {
      /* need to transfer end bytes starting from original wr */

      for (i=end; i>0; i--)
        *p++ = *buf++;

      /* and to transfer n-end bytes starting from 0 */

      p = pipe->buffer;

      for (i=n-end; i>0; i--)
        *p++ = *buf++;
    }

  ReleaseMutex (pipe->mutex); /* ignore error */

  return ___FIX(___NO_ERR);
}

#endif


/*---------------------------------------------------------------------------*/

/* Generic device operations. */

___SCMOBJ ___device_select
   ___P((___device **devs,
         int nb_read_devs,
         int nb_write_devs,
         ___time timeout),
        (devs,
         nb_read_devs,
         nb_write_devs,
         timeout)
___device **devs;
int nb_read_devs;
int nb_write_devs;
___time timeout;)
{
#define SELECT_DEVS_MAX 8192

  int nb_devs;
  ___device_select_state state;
  int pass;
  ___BOOL all_select_setups_done;
  ___device *select_devs[SELECT_DEVS_MAX];
  int i;

  nb_devs = nb_read_devs + nb_write_devs;
  if (nb_devs > SELECT_DEVS_MAX) /* ignore last devices if too many */
    nb_devs = SELECT_DEVS_MAX;
  
  state.devs = devs;

  state.timeout = timeout;
  state.relative_timeout = ___time_pos_infinity;

#ifdef USE_select

  state.highest_fd_plus_1 = 0;

  FD_ZERO(&state.readfds);
  FD_ZERO(&state.writefds);
  FD_ZERO(&state.exceptfds);

#endif

#ifdef USE_MsgWaitForMultipleObjects

#if 0
  state.timeout = 5000;/*******************/
#else
  state.timeout = INFINITE;
#endif

  state.timeout_dev_pos = -1;

  state.message_queue_mask = 0;
  state.message_queue_dev_pos = -1;

  state.wait_objs_buffer[0] = user_mod.abort_select;
  state.wait_objs_buffer[1] = user_mod.heartbeat_thread;

  state.nb_wait_objs = 2;

#endif

  for (i=nb_devs-1; i>=0; i--)
    {
      ___device *d = devs[i];
      select_devs[i] = d;
      if (d != NULL)
        d->select_step = 0;
    }

  pass = ___SELECT_PASS_1;
  all_select_setups_done = 0;

  while (!all_select_setups_done)
    {
      all_select_setups_done = 1;

      for (i=nb_devs-1; i>=0; i--)
        {
          ___SCMOBJ e;
          ___device *d = select_devs[i];
          if (d != NULL)
            {
              if ((e = ___device_select_virt
                         (d,
                          i>=nb_read_devs,
                          i,
                          pass,
                          &state))
                  != ___FIX(___NO_ERR))
                {
                  if (e != ___FIX(___SELECT_SETUP_DONE))
                    return e;
                  select_devs[i] = NULL;
                }
              else
                all_select_setups_done = 0;
            }
        }

      pass++;
    }

#ifdef USE_select

  {
    ___SCMOBJ e;/***************unused*/
    ___time delta;
    struct timeval delta_tv_struct;
    struct timeval *delta_tv = &delta_tv_struct;
    int select_result;

    absolute_time_to_relative_time (state.timeout, &delta);

    if (___time_less (state.relative_timeout, delta))
      delta = state.relative_timeout;

    absolute_time_to_nonnegative_timeval (delta, &delta_tv);

#if 0
    /****************************/
    printf ("timeout = %20.10g\n",state.timeout);
    printf ("delta = %20.10g\n",delta);
    printf ("delta_tv = %p secs=%d  usecs=%d\n",delta_tv,delta_tv_struct.tv_sec,delta_tv_struct.tv_usec);
#if 0
    fcntl (0, F_SETFL, 0); /* set nonblocking mode */
#endif
#endif

#if 0
    /****************************/
    {
      static FILE* logfile = NULL;
      if (logfile == NULL) logfile = fopen ("logfile", "w");
      if (logfile != NULL)
        fprintf(logfile,"select  secs=%d  usecs=%d  highest_fd_plus_1 = %d\n",delta_tv->tv_sec,delta_tv->tv_usec,state.highest_fd_plus_1);
      else
        fprintf(logfile,"select  NULL  highest_fd_plus_1 = %d\n",state.highest_fd_plus_1);
#endif

    select_result =
      select (state.highest_fd_plus_1,
              &state.readfds,
              &state.writefds,
              &state.exceptfds,
              delta_tv);

    state.timeout_reached = (select_result == 0);

#if 0
    if (select_result<0)
        {printf("<<select_result=%d>>\n",select_result);fflush(stdout);}

    /****************************/

    if (FD_ISSET(0,&state.readfds)) fprintf(logfile,"0 ready for reading\n");
    if (FD_ISSET(1,&state.readfds)) fprintf(logfile,"1 ready for reading\n");
    if (FD_ISSET(3,&state.readfds)) fprintf(logfile,"3 ready for reading\n");
    if (FD_ISSET(4,&state.readfds)) fprintf(logfile,"4 ready for reading\n");
    if (FD_ISSET(0,&state.writefds)) fprintf(logfile,"0 ready for writing\n");
    if (FD_ISSET(1,&state.writefds)) fprintf(logfile,"1 ready for writing\n");
    if (FD_ISSET(3,&state.writefds)) fprintf(logfile,"3 ready for writing\n");
    if (FD_ISSET(4,&state.writefds)) fprintf(logfile,"4 ready for writing\n");
    if (FD_ISSET(0,&state.exceptfds)) fprintf(logfile,"0 ready for exceptions\n");
    if (FD_ISSET(1,&state.exceptfds)) fprintf(logfile,"1 ready for exceptions\n");
    if (FD_ISSET(3,&state.exceptfds)) fprintf(logfile,"3 ready for exceptions\n");
    if (FD_ISSET(4,&state.exceptfds)) fprintf(logfile,"4 ready for exceptions\n");

#if 0
    fcntl (0, F_SETFL, O_NONBLOCK); /* set nonblocking mode */
#endif

    fprintf(logfile,"select result = %d        errno = %d\n",select_result, errno);/*****************/

    if (select_result<0)
      fprintf(logfile,"secs=%d  usecs=%d  state.highest_fd_plus_1 = %d    select result = %d\n",delta_tv_struct.tv_sec,delta_tv_struct.tv_usec,state.highest_fd_plus_1,  errno);/*****************/
    fflush(logfile);
    }
#endif

    if (select_result < 0)
      return err_code_from_errno ();
  }

#endif

#ifdef USE_MsgWaitForMultipleObjects

  if (state.timeout == 0)
    goto timed_out; /* don't even call MsgWaitForMultipleObjects */

  while (state.nb_wait_objs > 0 || state.message_queue_mask != 0)
    {
      DWORD n;

#if 0
      n = MsgWaitForMultipleObjects
            (state.nb_wait_objs,
             state.wait_objs_buffer,
             FALSE,
             state.timeout,
             state.message_queue_mask);
#else
      n = 0;/////////////////////////
#endif

      if (n == WAIT_FAILED)
        return err_code_from_GetLastError ();

      if (n >= WAIT_OBJECT_0 &&
          n <= WAIT_OBJECT_0+state.nb_wait_objs)
        n -= WAIT_OBJECT_0;
      else if (n >= WAIT_ABANDONED_0 &&
               n <= WAIT_ABANDONED_0+state.nb_wait_objs-1)
        n -= WAIT_ABANDONED_0;
      else
        {
          /* n == WAIT_TIMEOUT */

          /*
           * The call to MsgWaitForMultipleObjects timed out.  Mark
           * the appropriate device "ready".
           */

          if (state.timeout != 0)
            {
              /* first call to MsgWaitForMultipleObjects */

              timed_out:

              i = state.timeout_dev_pos;
              if (i >= 0)
                devs[i] = NULL;
            }

          break;
        }

      if (n == state.nb_wait_objs)
        {
          /*
           * The message queue contains a message that is of interest.
           * Mark the appropriate device "ready".
           */

          i = state.message_queue_dev_pos;
          if (i >= 0)
            devs[i] = NULL;

          /*
           * Don't check if other devices are ready because this might
           * cause an infinite loop.
           */

          break;
        }
      else if (n == 0)
        {
          /*
           * The call to device_select must be aborted because the
           * abort_select event is set.  This occurs when an interrupt
           * (such as a CTRL-C user interrupt) needs to be serviced
           * promptly by the main program.
           */

          /**********************/
          ResetEvent (user_mod.abort_select); /* ignore error */

          break;
        }
      else if (n == 1)
        {
          /*
           * The heartbeat thread has died.  This is normally due to
           * the program being terminated abruptly by the user (for
           * example by using the thread manager or the "shutdown"
           * item in the start menu).  When this happens we must
           * initiate a clean termination of the program.
           */

          fprintf (stderr, "heartbeat thread died\n");
          fflush (stderr);/***********/

          return ___FIX(___UNKNOWN_ERR);
        }
      else
        {
          /* Mark the appropriate device "ready". */

          i = state.wait_obj_to_dev_pos[n];
          if (i >= 0)
            devs[i] = NULL;

          /* Prepare to check remaining devices. */

          state.wait_obj_to_dev_pos[n] =
            state.wait_obj_to_dev_pos[--state.nb_wait_objs];
        }

      state.timeout = 0; /* next MsgWaitForMultipleObjects will only poll */
    }

#endif

  for (i=nb_devs-1; i>=0; i--)
    {
      ___SCMOBJ e;
      ___device *d = devs[i];
      if (d != NULL)
        if ((e = ___device_select_virt
                   (d,
                    i>=nb_read_devs,
                    i,
                    ___SELECT_PASS_CHECK,
                    &state))
            != ___FIX(___NO_ERR))
{
printf("<<e=%d>>\n",e);fflush(stdout);
          return e;
}
    }

  return ___FIX(___NO_ERR);
}

#ifdef USE_select

___HIDDEN void device_select_add_fd
   ___P((___device_select_state *state,
         int fd,
         ___BOOL for_writing),
        (state,
         fd,
         for_writing)
___device_select_state *state;
int fd;
___BOOL for_writing;)
{
  if (for_writing)
    FD_SET(fd, &state->writefds);
  else
    FD_SET(fd, &state->readfds);

  if (fd >= state->highest_fd_plus_1)
    state->highest_fd_plus_1 = fd+1;
}

___HIDDEN void device_select_add_relative_timeout
   ___P((___device_select_state *state,
         ___F64 seconds),
        (state,
         seconds)
___device_select_state *state;
___F64 seconds;)
{
  ___time relative_timeout;

  ___time_from_seconds (&relative_timeout, seconds);

  if (___time_less (relative_timeout, state->relative_timeout))
    state->relative_timeout = relative_timeout;
}

#endif

#ifdef USE_MsgWaitForMultipleObjects

___HIDDEN void device_select_add_wait_obj
   ___P((___device_select_state *state,
         int i,
         HANDLE wait_obj),
        (state,
         i,
         wait_obj)
___device_select_state *state;
int i;
HANDLE wait_obj;)
{
  int j = state->nb_wait_objs;

  if (j < MAXIMUM_WAIT_OBJECTS)
    {
      state->wait_objs_buffer[j] = wait_obj;
      state->wait_obj_to_dev_pos[j] = i;
      state->nb_wait_objs = j+1;
    }
}

#endif

___SCMOBJ ___device_flush_write
   ___P((___device *self),
        (self)
___device *self;)
{
  return ___device_flush_write_virt (self);
}

___SCMOBJ ___device_close
   ___P((___device *self,
         int direction),
        (self,
         direction)
___device *self;
int direction;)
{
  return ___device_close_virt (self, direction);
}

___HIDDEN void device_add_to_group
   ___P((___device_group *dgroup,
         ___device *dev),
        (dgroup,
         dev)
___device_group *dgroup;
___device *dev;)
{
  ___device *head = dgroup->list;

  dev->group = dgroup;

  if (head == NULL)
    {
      dev->prev = dev;
      dev->next = dev;
      dgroup->list = dev;
    }
  else
    {
      ___device *tail = head->prev;
      dev->prev = tail;
      dev->next = head;
      head->prev = dev;
      tail->next = dev;
    }
}

___HIDDEN void device_remove_from_group
   ___P((___device *dev),
        (dev)
___device *dev;)
{
  ___device_group *dgroup = dev->group;
  ___device *prev = dev->prev;
  ___device *next = dev->next;

  if (prev == dev)
    dgroup->list = NULL;
  else
    {
      next->prev = prev;
      prev->next = next;
      dev->prev = dev;
      dev->next = dev;
      if (dgroup->list == dev)
        dgroup->list = next;
    }

  dev->group = NULL;
}

___HIDDEN void device_add_ref
   ___P((___device *self),
        (self)
___device *self;)
{
#ifdef USE_PUMPS
  InterlockedIncrement (&self->refcount);
#else
  self->refcount++;
#endif
}

___SCMOBJ ___device_release
   ___P((___device *self),
        (self)
___device *self;)
{
  ___SCMOBJ e = ___FIX(___NO_ERR);

  if (
#ifdef USE_PUMPS
      !InterlockedDecrement (&self->refcount)
#else
      --self->refcount == 0
#endif
      )
    {
      e = ___device_release_virt (self);
      ___free_mem (self);
    }

  return e;
}

___SCMOBJ ___device_cleanup
   ___P((___device *self),
        (self)
___device *self;)
{
  ___SCMOBJ e;
  ___device *devs[1];

  if (self->group == NULL)
    return ___FIX(___UNKNOWN_ERR);

  device_remove_from_group (self);

  for (;;)
    {
      e = ___device_close (self, ___DIRECTION_RD);
      if (e == ___FIX(___NO_ERR))
        break;
      if (e != ___ERR_CODE_EAGAIN)
        return e;

      devs[0] = self;
      e = ___device_select (devs, 1, 0, ___time_pos_infinity);
      if (e != ___FIX(___NO_ERR))
        return e;
    }

  for (;;)
    {
      e = ___device_close (self, ___DIRECTION_WR);
      if (e == ___FIX(___NO_ERR))
        break;
      if (e != ___ERR_CODE_EAGAIN)
        return e;

      devs[0] = self;
      e = ___device_select (devs, 0, 1, ___time_pos_infinity);
      if (e != ___FIX(___NO_ERR))
        return e;
    }

  return ___device_release (self);
}

___SCMOBJ ___device_group_setup
   ___P((___device_group **dgroup),
        (dgroup)
___device_group **dgroup;)
{
  ___SCMOBJ e;
  ___device_group *g;

  g = ___CAST(___device_group*,
              ___alloc_mem (sizeof (___device_group)));

  if (g == NULL)
    return ___FIX(___HEAP_OVERFLOW_ERR);

  e = user_module_setup ();

  if (e != ___FIX(___NO_ERR))
    {
      ___free_mem (g);
      return e;
    }

  g->list = NULL;

  *dgroup = g;

  return ___FIX(___NO_ERR);
}

void ___device_group_cleanup
   ___P((___device_group *dgroup),
        (dgroup)
___device_group *dgroup;)
{
  while (dgroup->list != NULL)
    if (___device_cleanup (dgroup->list) != ___FIX(___NO_ERR))
      break;

  user_module_cleanup ();

  ___free_mem (dgroup);
}

/*   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   */

/* Timer device. */

/*
 * Timer devices are not particularly useful, given that the scheduler
 * implements timeouts.  Use this as an example.
 */

___HIDDEN int device_timer_kind
   ___P((___device *self),
        (self)
___device *self;)
{
  return ___TIMER_KIND;
}

___HIDDEN ___SCMOBJ device_timer_select_virt
   ___P((___device *self,
         ___BOOL for_writing,
         int i,
         int pass,
         ___device_select_state *state),
        (self,
         for_writing,
         i,
         pass,
         state)
___device *self;
___BOOL for_writing;
int i;
int pass;
___device_select_state *state;)
{
  ___device_timer *d = ___CAST(___device_timer*,self);

  if (pass == ___SELECT_PASS_1)
    {
      if (___time_less (d->expiry, state->timeout))
        state->timeout = d->expiry;
      return ___FIX(___SELECT_SETUP_DONE);
    }

  /* pass == ___SELECT_PASS_CHECK */

  if (state->timeout_reached)
    {
      if (___time_equal (d->expiry, state->timeout))
        state->devs[i] = NULL;
    }

  return ___FIX(___NO_ERR);
}

___HIDDEN ___SCMOBJ device_timer_release_virt
   ___P((___device *self),
        (self)
___device *self;)
{
  return ___FIX(___NO_ERR);
}

___HIDDEN ___SCMOBJ device_timer_flush_write_virt
   ___P((___device *self),
        (self)
___device *self;)
{
  return ___FIX(___NO_ERR);
}

___HIDDEN ___SCMOBJ device_timer_close_virt
   ___P((___device *self,
         int direction),
        (self,
         direction)
___device *self;
int direction;)
{
  return ___FIX(___NO_ERR);
}

___HIDDEN ___device_timer_vtbl ___device_timer_table =
{
  {
    device_timer_kind,
    device_timer_select_virt,
    device_timer_release_virt,
    device_timer_flush_write_virt,
    device_timer_close_virt
  }
};

___SCMOBJ ___device_timer_setup
   ___P((___device_timer **dev,
         ___device_group *dgroup),
        (dev,
         dgroup)
___device_timer **dev;
___device_group *dgroup;)
{
  ___device_timer *d;

  d = ___CAST(___device_timer*,
              ___alloc_mem (sizeof (___device_timer)));

  if (d == NULL)
    return ___FIX(___HEAP_OVERFLOW_ERR);

  d->base.vtbl = &___device_timer_table;
  d->base.refcount = 1;
  d->base.direction = ___DIRECTION_RD | ___DIRECTION_WR;
  d->base.read_stage = ___STAGE_OPEN;
  d->base.write_stage = ___STAGE_OPEN;

  d->expiry = ___time_pos_infinity;

  *dev = d;

  device_add_to_group (dgroup, &d->base);

  return ___FIX(___NO_ERR);
}

void ___device_timer_set_expiry
   ___P((___device_timer *dev,
         ___time expiry),
        (dev,
         expiry)
___device_timer *dev;
___time expiry;)
{
  dev->expiry = expiry;
}

/*   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   */

/* Byte stream devices. */

#ifdef USE_PUMPS

#define OOB_EOS          0
#define OOB_FLUSH_WRITE  1
#define OOB_SEEK_ABS     2
#define OOB_SEEK_REL     3
#define OOB_SEEK_REL_END 4

typedef struct ___device_stream_pump_struct
  {
    HANDLE thread;
    nonblocking_pipe pipe;
  } ___device_stream_pump;

#define PIPE_BUFFER_SIZE 16384

/*
 * It is important that the PUMP_PRIORITY be above that of the main
 * program (which is THREAD_PRIORITY_NORMAL) so that I/O operations,
 * such as flushing output buffers, will be done as soon as the pump
 * reads it from the nonblocking pipe.  Thus the operation will appear
 * to be synchronous to the main program.
 */

#define PUMP_PRIORITY THREAD_PRIORITY_ABOVE_NORMAL

___HIDDEN ___SCMOBJ ___device_stream_pump_setup
   ___P((___device_stream_pump **pump,
         DWORD committed_stack_size,
         LPTHREAD_START_ROUTINE proc,
         LPVOID arg),
        (pump,
         committed_stack_size,
         proc,
         arg)
___device_stream_pump **pump;
DWORD committed_stack_size;
LPTHREAD_START_ROUTINE proc;
LPVOID arg;)
{
  ___SCMOBJ e;
  ___device_stream_pump *p;
  HANDLE thread_handle;
  DWORD thread_id;

  p = ___CAST(___device_stream_pump*,
              ___alloc_mem (sizeof (___device_stream_pump)));

  if (p == NULL)
    return ___FIX(___HEAP_OVERFLOW_ERR);

  e = nonblocking_pipe_setup (&p->pipe, PIPE_BUFFER_SIZE+1);

  if (e != ___FIX(___NO_ERR))
    {
      ___free_mem (p);
      return e;
    }

  *pump = p; /* set before thread created to avoid race condition */

  thread_handle =
    CreateThread (NULL,                 /* no security attributes       */
                  committed_stack_size, /* committed stack size         */
                  proc,                 /* thread procedure             */
                  arg,                  /* argument to thread procedure */
                  0,                    /* use default creation flags   */
                  &thread_id);

  if (thread_handle == NULL ||
      !SetThreadPriority (thread_handle, PUMP_PRIORITY))
    {
      e = err_code_from_GetLastError ();
      nonblocking_pipe_cleanup (&p->pipe);
      ___free_mem (p);
      *pump = NULL; /* make sure caller does not think a pump was created */
      return e;
    }

  p->thread = thread_handle;

  return ___FIX(___NO_ERR);
}

___HIDDEN ___SCMOBJ ___device_stream_pump_reader_kill
   ___P((___device_stream_pump *pump),
        (pump)
___device_stream_pump *pump;)
{
  return nonblocking_pipe_set_reader_err
           (&pump->pipe,
            ___FIX(___KILL_PUMP));
}

___HIDDEN ___SCMOBJ ___device_stream_pump_writer_kill
   ___P((___device_stream_pump *pump),
        (pump)
___device_stream_pump *pump;)
{
  return nonblocking_pipe_set_writer_err
           (&pump->pipe,
            ___FIX(___KILL_PUMP));
}

___HIDDEN ___SCMOBJ ___device_stream_pump_wait
   ___P((___device_stream_pump *pump),
        (pump)
___device_stream_pump *pump;)
{
  DWORD code;

  code = WaitForSingleObject (pump->thread, 0);

  if (code == WAIT_FAILED)
    return err_code_from_GetLastError ();

  if (code == WAIT_TIMEOUT)
    return ___ERR_CODE_EAGAIN;

  return ___FIX(___NO_ERR);
}

___HIDDEN ___SCMOBJ ___device_stream_pump_cleanup
   ___P((___device_stream_pump *pump),
        (pump)
___device_stream_pump *pump;)
{
  CloseHandle (pump->thread); /* ignore error */
  nonblocking_pipe_cleanup (&pump->pipe); /* ignore error */
  ___free_mem (pump);

  return ___FIX(___NO_ERR);
}

#endif

___HIDDEN ___SCMOBJ ___device_stream_select_virt
   ___P((___device *self,
         ___BOOL for_writing,
         int i,
         int pass,
         ___device_select_state *state),
        (self,
         for_writing,
         i,
         pass,
         state)
___device *self;
___BOOL for_writing;
int i;
int pass;
___device_select_state *state;)
{
  ___device_stream *d = ___CAST(___device_stream*,self);

#ifdef USE_PUMPS

  if (for_writing)
    {
      ___device_stream_pump *p = d->write_pump;
      if (p != NULL)
        {
          HANDLE wait_obj;

          if (d->base.write_stage == ___STAGE_OPEN)
            wait_obj = p->pipe.wevent;
          else
            wait_obj = p->thread;

          device_select_add_wait_obj (state, i, wait_obj);

          return ___FIX(___SELECT_SETUP_DONE);
        }
    }
  else
    {
      ___device_stream_pump *p = d->read_pump;
      if (p != NULL)
        {
          HANDLE wait_obj;

          if (d->base.read_stage == ___STAGE_OPEN)
            wait_obj = p->pipe.revent;
          else
            wait_obj = p->thread;

          device_select_add_wait_obj (state, i, wait_obj);

          return ___FIX(___SELECT_SETUP_DONE);
        }
    }

#endif

  return ___device_stream_select_raw_virt
           (d,
            for_writing,
            i,
            pass,
            state);
}

___HIDDEN ___SCMOBJ ___device_stream_release_virt
   ___P((___device *self),
        (self)
___device *self;)
{
  ___SCMOBJ e;
  ___device_stream *d = ___CAST(___device_stream*,self);

  e = ___device_stream_release_raw_virt (d);

#ifdef USE_PUMPS

  {
    ___device_stream_pump *p;

    p = d->read_pump;
    if (p != NULL)
      ___device_stream_pump_cleanup (p); /* ignore error */

    p = d->write_pump;
    if (p != NULL)
      ___device_stream_pump_cleanup (p); /* ignore error */
  }

#endif

  return e;
}

___HIDDEN ___SCMOBJ ___device_stream_flush_write_virt
   ___P((___device *self),
        (self)
___device *self;)
{
  ___device_stream *d = ___CAST(___device_stream*,self);

#ifdef USE_PUMPS

  {
    ___device_stream_pump *p = d->write_pump;

    if (p != NULL)
      {
        nonblocking_pipe_oob_msg oob_msg;

        oob_msg.op = OOB_FLUSH_WRITE;

        return nonblocking_pipe_write_oob (&p->pipe, &oob_msg);
      }
  }

#endif

  return ___device_stream_flush_write_raw_virt (d);
}

___SCMOBJ ___device_stream_seek
   ___P((___device_stream *self,
         ___stream_index pos,
         int whence),
        (self,
         pos,
         whence)
___device_stream *self;
___stream_index pos;
int whence;)
{
#ifdef USE_PUMPS

  {
    ___device_stream_pump *p = self->write_pump;

    if (p != NULL)
      {
        nonblocking_pipe_oob_msg oob_msg;

        oob_msg.op = OOB_SEEK_ABS + whence;
        oob_msg.param[0] = pos;

        return nonblocking_pipe_write_oob (&p->pipe, &oob_msg);
      }
  }

#endif

  return ___device_stream_seek_raw_virt (self, pos, whence);
}

___SCMOBJ ___device_stream_read
   ___P((___device_stream *self,
         ___U8 *buf,
         ___stream_index len,
         ___stream_index *len_done),
        (self,
         buf,
         len,
         len_done)
___device_stream *self;
___U8 *buf;
___stream_index len;
___stream_index *len_done;)
{
#ifdef USE_PUMPS

  {
    ___device_stream_pump *p = self->read_pump;

    if (p != NULL)
      {
        nonblocking_pipe_oob_msg oob_msg;
        return nonblocking_pipe_read (&p->pipe, buf, len, len_done, &oob_msg);
      }
  }

#endif

  return ___device_stream_read_raw_virt (self, buf, len, len_done);
}

___SCMOBJ ___device_stream_write
   ___P((___device_stream *self,
         ___U8 *buf,
         ___stream_index len,
         ___stream_index *len_done),
        (self,
         buf,
         len,
         len_done)
___device_stream *self;
___U8 *buf;
___stream_index len;
___stream_index *len_done;)
{
#ifdef USE_PUMPS

  {
    ___device_stream_pump *p = self->write_pump;

    if (p != NULL)
      return nonblocking_pipe_write (&p->pipe, buf, len, len_done);
  }

#endif

  return ___device_stream_write_raw_virt (self, buf, len, len_done);
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

#ifdef USE_PUMPS

___HIDDEN DWORD WINAPI ___device_stream_read_pump_proc
   ___P((LPVOID param),
        (param)
LPVOID param;)
{
  ___device_stream *dev = (___device_stream*)param;
  nonblocking_pipe *p = &dev->read_pump->pipe;
  ___SCMOBJ e;
  DWORD len;
  DWORD n;
  DWORD i;
  ___U8 buf[PIPE_BUFFER_SIZE];
  nonblocking_pipe_oob_msg oob_msg;

  for (;;)
    {
      /* wait until the reader needs some data */

      e = nonblocking_pipe_write_ready_wait (p);

      /* read some characters from device */

      if (e == ___FIX(___NO_ERR))
        e = ___device_stream_read_raw_virt (dev, buf, PIPE_BUFFER_SIZE, &len);

      if (e == ___FIX(___NO_ERR))
        {
          if (len == 0)
            {
              /* we reached the end-of-stream */

              oob_msg.op = OOB_EOS;

              while ((e = nonblocking_pipe_write_oob (p, &oob_msg))
                     == ___ERR_CODE_EAGAIN)
                {
                  /* suspend thread until operation can be performed */

                  e = nonblocking_pipe_write_ready_wait (p);
                  if (e != ___FIX(___NO_ERR))
                    break;
                }
              if (e != ___FIX(___NO_ERR))
                break;
            }
          else
            {
              /* write to the pipe the bytes that were read */

              i = 0;

              while (i < len)
                {
                  while ((e = nonblocking_pipe_write (p, buf+i, len-i, &n))
                         == ___ERR_CODE_EAGAIN)
                    {
                      /* suspend thread until operation can be performed */

                      e = nonblocking_pipe_write_ready_wait (p);
                      if (e != ___FIX(___NO_ERR))
                        break;
                    }
                  if (e != ___FIX(___NO_ERR))
                    break;
                  i += n;
                }
            }
        }

      if (e != ___FIX(___NO_ERR))
        {
          if (e == ___FIX(___KILL_PUMP)) /* terminate? */
            break;

          if (e == ___ERR_CODE_EAGAIN)
            continue;

          /* report the failure back through the pipe */

          e = nonblocking_pipe_set_writer_err (p, e);

          if (e != ___FIX(___NO_ERR))
            {
              /*
               * The failure could not be reported. To avoid an
               * infinite loop the thread is terminated.
               */

              ExitThread (0);
            }
        }
    }

  device_release (&dev->base); /* ignore error */

  return 0;
}

___HIDDEN DWORD WINAPI ___device_stream_write_pump_proc
   ___P((LPVOID param),
        (param)
LPVOID param;)
{
  ___device_stream *dev = (___device_stream*)param;
  nonblocking_pipe *p = &dev->write_pump->pipe;
  ___SCMOBJ e;
  DWORD len;
  DWORD n;
  DWORD i;
  ___U8 buf[PIPE_BUFFER_SIZE];
  nonblocking_pipe_oob_msg oob_msg;

  for (;;)
    {
      /* get from the pipe some bytes to write to the device */

      while ((e = nonblocking_pipe_read (p, buf, PIPE_BUFFER_SIZE, &len, &oob_msg))
             == ___ERR_CODE_EAGAIN)
        {
          /* suspend thread until operation can be performed */

          e = nonblocking_pipe_read_ready_wait (p);
          if (e != ___FIX(___NO_ERR))
            break;
        }

      if (e == ___FIX(___NO_ERR))
        {
          if (len > 0)
            {
              /* write to the device the bytes that were read from the pipe */

              i = 0;

              while (i < len)
                {
                  e = ___device_stream_write_raw_virt (dev, buf+i, len-i, &n);
                  if (e != ___FIX(___NO_ERR))
                    break;
                  i += n;
                }
            }
          else
            {
              switch (oob_msg.op)
                {
                case OOB_FLUSH_WRITE:
                  e = ___device_stream_flush_write_raw_virt (dev);
                  break;
                case OOB_SEEK_ABS:
                case OOB_SEEK_REL:
                case OOB_SEEK_REL_END:
                  fprintf (stderr,"***** got OOB_SEEK %d\n", oob_msg.param[0]);fflush (stderr);
                  e = ___device_stream_seek_raw_virt (dev, oob_msg.param[0]);
                  break;
                case OOB_EOS:
                  fprintf (stderr,"***** got OOB_EOS\n");fflush (stderr);
                  break;
                }
            }
        }

      if (e != ___FIX(___NO_ERR))
        {
          if (err == ___FIX(___KILL_PUMP)) /* terminate? */
            break;

          /* report the failure back through the pipe */

          e = nonblocking_pipe_set_reader_err (p, e);

          if (e != ___FIX(___NO_ERR))
            {
              /*
               * The failure could not be reported. To avoid an
               * infinite loop the thread is terminated.
               */

              ExitThread (0);
            }
        }
    }

  device_release (&dev->base); /* ignore error */

  return 0;
}

#endif

___SCMOBJ ___device_stream_setup
   ___P((___device_stream *dev,
         ___device_group *dgroup,
         int direction,
         int pumps_on),
        (dev,
         dgroup,
         direction,
         pumps_on)/*********************/
___device_stream *dev;
___device_group *dgroup;
int direction;
int pumps_on;)
{
  dev->base.refcount = 1;
  dev->base.direction = direction;
  dev->base.read_stage = ___STAGE_CLOSED;
  dev->base.write_stage = ___STAGE_CLOSED;

#ifdef USE_PUMPS
  dev->read_pump = NULL;
  dev->write_pump = NULL;
#endif

  device_add_to_group (dgroup, &dev->base);

  if (direction & ___DIRECTION_RD)
    {
      dev->base.read_stage = ___STAGE_OPEN;

#ifdef USE_PUMPS

      if (pumps_on & ___DIRECTION_RD)
        {
          ___SCMOBJ e;

          device_add_ref (&dev->base);

          if ((e = ___device_stream_pump_setup
                     (&dev->read_pump,
                      65536,/********************/
                      ___device_stream_read_pump_proc,
                      dev))
              != ___FIX(___NO_ERR))
            {
              device_release (&dev->base); /* ignore error */
              device_cleanup (&dev->base); /* ignore error */
              return e;
            }
        }

#endif
    }

  if (direction & ___DIRECTION_WR)
    {
      dev->base.write_stage = ___STAGE_OPEN;

#ifdef USE_PUMPS

      if (pumps_on & ___DIRECTION_WR)
        {
          ___SCMOBJ e;

          device_add_ref (&dev->base);

          if ((e = ___device_stream_pump_setup
                     (&dev->write_pump,
                      65536,/********************/
                      ___device_stream_write_pump_proc,
                      dev))
              != ___FIX(___NO_ERR))
            {
              device_release (&dev->base); /* ignore error */
              device_cleanup (&dev->base); /* ignore error */
              return e;
            }
        }

#endif
    }

  return ___FIX(___NO_ERR);
}

/*---------------------------------------------------------------------------*/

#ifdef USE_CURSES

#ifdef USE_TERMCAP
#define USE_TERMCAP_OR_TERMINFO
#endif

#ifdef USE_TERMINFO
#define USE_TERMCAP_OR_TERMINFO
#endif

#endif

#ifdef USE_CURSES

/*================================================================*/
#define UNICODE_NUL         0
#define UNICODE_BELL        7
#define UNICODE_BACKSPACE   8
#define UNICODE_TAB         9
#define UNICODE_LINEFEED    10
#define UNICODE_RETURN      13
#define UNICODE_ESCAPE      27
#define UNICODE_SPACE       32
#define UNICODE_DOUBLEQUOTE 34
#define UNICODE_SHARP       35
#define UNICODE_QUOTE       39
#define UNICODE_COMMA       44
#define UNICODE_1           49
#define UNICODE_2           50
#define UNICODE_3           51
#define UNICODE_4           52
#define UNICODE_6           54
#define UNICODE_8           56
#define UNICODE_SEMICOLON   59
#define UNICODE_AT          64
#define UNICODE_UPPER_F     70
#define UNICODE_LOWER_F     102
#define UNICODE_UPPER_S     83
#define UNICODE_LOWER_S     115
#define UNICODE_UPPER_U     85
#define UNICODE_LOWER_U     117
#define UNICODE_LPAREN      40
#define UNICODE_RPAREN      41
#define UNICODE_LBRACKET    91
#define UNICODE_BACKSLASH   92
#define UNICODE_RBRACKET    93
#define UNICODE_LBRACE      123
#define UNICODE_VBAR        124
#define UNICODE_RBRACE      125
#define UNICODE_RUBOUT      127
#define UNICODE_REPLACEMENT 0xfffd

/*---------------------------------------------------------------------------*/

/* Data types.  */

/* Type of characters that can be read from and written to a tty */

typedef ___CS_SELECT(___LATIN1,___UCS2,___UCS4) tty_char;

#define TTY_CHAR_MAX ___MAX_CHR

/* Tty management data structures */

#define TTY_TAB_STOP 8

#define TTY_RBUF_SIZE 100
#define TTY_WBUF_SIZE 100

#define DEFAULT_EMACS_BINDINGS 1
#define DEFAULT_PAREN_BALANCE_DURATION_NSECS 500000000 /* 1/2 second */

typedef int tty_input_mode;

#define RAW_INPUT_MODE      1
#define LINEEDIT_INPUT_MODE 256

typedef int tty_term_mode;

#define INITIAL_TERM_MODE 0
#define CBREAK_TERM_MODE  1
#define RAW_TERM_MODE     2

typedef int tty_encoding;

#define ASCII_ENCODING  0
#define LATIN1_ENCODING 1
#define UTF8_ENCODING   2
#define UCS2LE_ENCODING 3
#define UCS2BE_ENCODING 4
#define UCS4LE_ENCODING 5
#define UCS4BE_ENCODING 6
#define NATIVE_ENCODING 7

typedef int tty_text_attrs;

#define TEXT_STYLE_NORMAL  0
#define TEXT_STYLE_BOLD    1
#define TEXT_STYLE_REVERSE 2

#define TEXT_COLOR_BLACK   0
#define TEXT_COLOR_RED     1
#define TEXT_COLOR_GREEN   2
#define TEXT_COLOR_YELLOW  3
#define TEXT_COLOR_BLUE    4
#define TEXT_COLOR_MAGENTA 5
#define TEXT_COLOR_CYAN    6
#define TEXT_COLOR_WHITE   7
#define DEFAULT_TEXT_COLOR 8

#define GET_STYLE(x)(((x)>>8)&3)
#define GET_FOREGROUND_COLOR(x)((x)&15)
#define GET_BACKGROUND_COLOR(x)(((x)>>4)&15)

#define MAKE_TEXT_ATTRS(s,f,b)(((s)<<8)+(f)+((b)<<4))

#if 0
/*********************************/
#define DEFAULT_TEXT_ATTRS \
MAKE_TEXT_ATTRS(TEXT_STYLE_NORMAL,DEFAULT_TEXT_COLOR,DEFAULT_TEXT_COLOR)
#else
#define INITIAL_TEXT_ATTRS \
MAKE_TEXT_ATTRS(TEXT_STYLE_NORMAL,DEFAULT_TEXT_COLOR,DEFAULT_TEXT_COLOR)
#define DEFAULT_INPUT_TEXT_ATTRS \
MAKE_TEXT_ATTRS(TEXT_STYLE_BOLD,DEFAULT_TEXT_COLOR,DEFAULT_TEXT_COLOR)
#define DEFAULT_OUTPUT_TEXT_ATTRS INITIAL_TEXT_ATTRS
#endif

typedef int tty_signal;

#define USER_INTERRUPT_SIGNAL 0

#ifndef USE_POSIX
#ifndef USE_WIN32
#ifndef USE_MACOS

#define NB_TTY_SIGNALS 1

#endif
#endif
#endif

#ifdef USE_POSIX

#define WINDOW_SIZE_SIGNAL 1
#define CONTINUE_SIGNAL    2
#define NB_TTY_SIGNALS     3

#endif

#ifdef USE_WIN32

#define NB_TTY_SIGNALS 1

#endif

#ifdef USE_MACOS

/********************/

#endif

typedef struct tty_state_struct
  {
    struct tty_state_struct *prev; /* circular list links */
    struct tty_state_struct *next;

    int refcount; /* nb of "tty" structures that point to this "tty_state" */

    tty_input_mode input_mode; /* input mode (RAW or not, LINEEDIT or not) */
    tty_term_mode term_mode;   /* terminal mode (INITIAL, CBREAK, RAW) */
    tty_encoding encoding;     /* external char encoding (ASCII, UTF8, ...) */

    /* read buffer, needed for decoding characters */
    int rlo;
    int rhi;
    ___U8 rbuf[TTY_RBUF_SIZE];

    /* write buffer, needed for encoding characters */
    int wlo;
    int whi;
    ___U8 wbuf[TTY_WBUF_SIZE];
    ___SCMOBJ werr; /* write error flag for latest write operation */

    ___BOOL received_signal[NB_TTY_SIGNALS]; /* signals that were received */

#ifdef USE_POSIX
    int fd;                         /* file descriptor of the terminal */
    struct termios initial_termios; /* initial terminal driver settings */
#endif

#ifdef USE_WIN32
    HANDLE input;
    HANDLE output;
#define INPUT_RBUF_SIZE 10
    INPUT_RECORD input_rbuf[INPUT_RBUF_SIZE];
    int input_rlo;
    int input_rhi;
#endif

#ifdef USE_MACOS
    /*****************/
#endif

#ifdef USE_LINEEDIT
    struct lineedit_state_struct *les; /* line edit state */
#endif
  } tty_state;

typedef struct tty_struct
  {
    tty_state *state; /* NULL means not yet open console */
  } tty;

typedef struct tty_module_struct
  {
    tty_state *list; /* circular list of tty_state's */
    tty console;     /* the console is a special tty */
    void (*user_interrupt_handler) ___PVOID; /* called on a user interrupt */
  } tty_module;

___HIDDEN tty_module tty_mod; /* tty module's global state */

/*---------------------------------------------------------------------------*/

/*********************/

___HIDDEN ___SCMOBJ tty_set_term_mode
   ___P((tty_state *state,
         tty_term_mode mode),
        ());

___HIDDEN ___SCMOBJ tty_set_input_mode
   ___P((tty *t,
         tty_input_mode input_mode),
        ());

___HIDDEN ___SCMOBJ tty_set_terminal_type
   ___P((tty *t,
         char *terminal_type,
         ___BOOL emacs_bindings),
        ());

___HIDDEN ___SCMOBJ tty_set_encoding
   ___P((tty *t,
         tty_encoding encoding),
        ());

___HIDDEN ___SCMOBJ tty_cleanup
   ___P((tty *t),
        ());

/*---------------------------------------------------------------------------*/

/*qqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqq*/

#ifdef USE_LINEEDIT

/* Extensible string data structure */

typedef tty_char extensible_string_char;

typedef struct extensible_string_struct
  {
    extensible_string_char *buffer;
    int length;
    int max_length;
  } extensible_string;

/* Line editing data structures */

#define LINEEDIT_EV_NONE           0
#define LINEEDIT_EV_KEY            1
#define LINEEDIT_EV_RETURN         2
#define LINEEDIT_EV_BACK           3
#define LINEEDIT_EV_BACK_WORD      4
#define LINEEDIT_EV_TAB            5
#define LINEEDIT_EV_MARK           6
#define LINEEDIT_EV_PASTE          7
#define LINEEDIT_EV_CUT            8
#define LINEEDIT_EV_CUT_RIGHT      9
#define LINEEDIT_EV_CUT_LEFT       10
#define LINEEDIT_EV_REFRESH        11
#define LINEEDIT_EV_TRANSPOSE      12
#define LINEEDIT_EV_TRANSPOSE_WORD 13
#define LINEEDIT_EV_UP             14
#define LINEEDIT_EV_DOWN           15
#define LINEEDIT_EV_RIGHT          16
#define LINEEDIT_EV_RIGHT_WORD     17
#define LINEEDIT_EV_LEFT           18
#define LINEEDIT_EV_LEFT_WORD      19
#define LINEEDIT_EV_HOME           20
#define LINEEDIT_EV_HOME_DOC       21
#define LINEEDIT_EV_INSERT         22
#define LINEEDIT_EV_DELETE         23
#define LINEEDIT_EV_DELETE_WORD    24
#define LINEEDIT_EV_END            25
#define LINEEDIT_EV_END_DOC        26
#define LINEEDIT_EV_F1             27
#define LINEEDIT_EV_META_F1        28
#define LINEEDIT_EV_F2             29
#define LINEEDIT_EV_META_F2        30
#define LINEEDIT_EV_F3             31
#define LINEEDIT_EV_META_F3        32
#define LINEEDIT_EV_F4             33
#define LINEEDIT_EV_META_F4        34

#ifdef LINEEDIT_SUPPORT_F5_TO_F12
#define LINEEDIT_EV_F5             35
#define LINEEDIT_EV_META_F5        36
#define LINEEDIT_EV_F6             37
#define LINEEDIT_EV_META_F6        38
#define LINEEDIT_EV_F7             39
#define LINEEDIT_EV_META_F7        40
#define LINEEDIT_EV_F8             41
#define LINEEDIT_EV_META_F8        42
#define LINEEDIT_EV_F9             43
#define LINEEDIT_EV_META_F9        44
#define LINEEDIT_EV_F10            45
#define LINEEDIT_EV_META_F10       46
#define LINEEDIT_EV_F11            47
#define LINEEDIT_EV_META_F11       48
#define LINEEDIT_EV_F12            49
#define LINEEDIT_EV_META_F12       50
#define LINEEDIT_EV_LAST           LINEEDIT_EV_META_F12
#else
#define LINEEDIT_EV_LAST           LINEEDIT_EV_META_F4
#endif

#define LINEEDIT_CAP_CLEAR 0
#define LINEEDIT_CAP_ED    1
#define LINEEDIT_CAP_CUU1  2
#define LINEEDIT_CAP_CUD1  3
#define LINEEDIT_CAP_CUU   4
#define LINEEDIT_CAP_CUD   5
#define LINEEDIT_CAP_CUF   6
#define LINEEDIT_CAP_CUB   7
#define LINEEDIT_CAP_SGR0  8
#define LINEEDIT_CAP_BOLD  9
#define LINEEDIT_CAP_REV   10
#define LINEEDIT_CAP_SETAF 11
#define LINEEDIT_CAP_SETAB 12
#define LINEEDIT_CAP_LAST  12

typedef struct lineedit_event_struct
  {
    int event_kind;
    tty_char event_char;
  } lineedit_event;

typedef struct lineedit_history_struct
  {
    struct lineedit_history_struct *prev; /* circular list links */
    struct lineedit_history_struct *next;
    extensible_string actual;
    extensible_string edited; /* buffer != NULL when being edited */
  } lineedit_history;

#define WITH_ESC_PREFIX ___CAST_U8(1<<7)

/*
 * LARGEST_DECODER is the size of the largest input decoder over all
 * terminals defined in the termcap database.  The largest input
 * decoder is for the "ti926-8" terminal.  Note that this value must
 * be updated when the keyboard bindings are changed.
 */

#ifdef LINEEDIT_SUPPORT_F5_TO_F12
#define LARGEST_DECODER 177
#else
#define LARGEST_DECODER 128
#endif

/*
 * The type lineedit_input_decoder_state must at least contain the
 * range 0 .. LINEEDIT_EV_LAST + LARGEST_DECODER + 1.
 */

#if LINEEDIT_EV_LAST + LARGEST_DECODER + 1 <= 255

typedef ___U8 lineedit_input_decoder_state;
#define LINEEDIT_INPUT_DECODER_STATE_MAX 255

#else

typedef ___U16 lineedit_input_decoder_state;
#define LINEEDIT_INPUT_DECODER_STATE_MAX 65535

#endif

#define LINEEDIT_INPUT_DECODER_MAX_LENGTH \
(LINEEDIT_INPUT_DECODER_STATE_MAX-LINEEDIT_EV_LAST-1)

typedef struct lineedit_input_test_struct
  {
    ___U8 trigger; /* when this byte is read, action is performed */
    lineedit_input_decoder_state action; /* event type or new state */
    lineedit_input_decoder_state next;   /* next test when byte != trigger */
  } lineedit_input_test;

typedef struct lineedit_input_decoder_struct
  {
    lineedit_input_test *buffer;
    int length;
    int max_length;
  } lineedit_input_decoder;

#define LINEEDIT_MODE_SCHEME 1

struct lineedit_state_undo_struct
  {
    lineedit_history *hist;
    int edit_point;     /* position in buffer where edit ops take place */
    int mark_point;     /* position in buffer of mark */
    int screen_cursor;  /* position in screen of cursor */
    int cursor_to_point;/* offset to add to cursor to get buffer position */
    ___BOOL delayed_wrap;
    ___BOOL paren_balance_trigger;
    ___BOOL paren_balance_in_progress;
    ___time paren_balance_end;
    tty_text_attrs attrs;
  };

typedef struct lineedit_state_struct
  {
    struct tty_state_struct *ts;

    int edit_mode;

    ___BOOL paste_cancel;
    int paste_index;
    tty_char *paste_text;

    int input_rlo;           /* index of next character to read */
    extensible_string input; /* last input line that was entered */

    int output_wlo;
    extensible_string output;

    int max_history_length;
    int history_length; /* number of history lines, not counting last */
    lineedit_history *hist_last;

    struct lineedit_state_undo_struct current;

    int paren_balance_duration_nsecs;

    tty_text_attrs input_attrs;
    tty_text_attrs output_attrs;

    int screen_rows;
    int screen_cols;
    int screen_size;
    ___BOOL has_auto_left_margin;
    ___BOOL has_auto_right_margin;
    ___BOOL has_eat_newline_glitch;
    ___BOOL linefeed_moves_to_left_margin;
    int tab_stop;

    lineedit_input_decoder input_decoder;
    lineedit_input_decoder_state input_decoder_state;

#ifdef LINEEDIT_WITH_LOCAL_CLIPBOARD
    extensible_string clipboard;
#endif

#define USE_CURSES

#ifdef USE_CURSES
    char *capability[LINEEDIT_CAP_LAST+1];
#endif
  } lineedit_state;

#define LINEEDIT_MAX_TAB_STOP 8

typedef struct lineedit_module_struct
  {
    tty_char tab_spaces[LINEEDIT_MAX_TAB_STOP];
#ifdef USE_TERMCAP_OR_TERMINFO
    lineedit_state *current_state; /* parameter of lineedit_output_putc */
    ___SCMOBJ last_err;            /* parameter of lineedit_output_putc */
#endif
  } lineedit_module;

___HIDDEN lineedit_module lineedit_mod;

/* begin CHECK CHECK CHECK CHECK CHECK CHECK CHECK CHECK */

/* Forward declarations */

___HIDDEN ___SCMOBJ tty_read_noncanonical
   ___P((tty_state *state,
         tty_char *buf,
         ___stream_index len,
         ___stream_index *len_done),
        ());

___HIDDEN void lineedit_track_cursor_movement
   ___P((lineedit_state *state,
         tty_char *buf,
         ___stream_index len),
        ());

___HIDDEN ___SCMOBJ tty_write_to_wbuf
   ___P((tty_state *state,
         tty_char *buf,
         ___stream_index len,
         ___stream_index *len_done),
        ());

___HIDDEN void tty_wbuf_drain
   ___P((tty_state *state),
        ());

/* end CHECK CHECK CHECK CHECK CHECK CHECK CHECK CHECK */

/*---------------------------------------------------------------------------*/

/* Extensible string routines */

___HIDDEN ___SCMOBJ extensible_string_setup
   ___P((extensible_string *str,
         int n),
        (str,
         n)
extensible_string *str;
int n;)
{
#define EXTENSIBLE_STRING_INITIAL_BUFFER_SIZE 16

  if (n < EXTENSIBLE_STRING_INITIAL_BUFFER_SIZE)
    n = EXTENSIBLE_STRING_INITIAL_BUFFER_SIZE;

  str->buffer = ___CAST(extensible_string_char*,
                        ___alloc_mem (n * sizeof (extensible_string_char)));

  if (str->buffer == NULL)
    return ___FIX(___HEAP_OVERFLOW_ERR);

  str->length = 0;
  str->max_length = n;

  return ___FIX(___NO_ERR);
}

___HIDDEN void extensible_string_cleanup
   ___P((extensible_string *str),
        (str)
extensible_string *str;)
{
  ___free_mem (str->buffer);
}

___HIDDEN ___SCMOBJ extensible_string_copy
   ___P((extensible_string_char *src,
         int len,
         extensible_string *dst,
         int fudge),
        (src,
         len,
         dst,
         fudge)
extensible_string_char *src;
int len;
extensible_string *dst;
int fudge;)
{
  extensible_string_char *buf;

  buf = ___CAST(extensible_string_char*,
                ___alloc_mem ((len+fudge) * sizeof (extensible_string_char)));

  if (buf == NULL)
    return ___FIX(___HEAP_OVERFLOW_ERR);

  dst->buffer = buf;
  dst->length = len;
  dst->max_length = len+fudge;

  while (len-- > 0)
    buf[len] = src[len];

  return ___FIX(___NO_ERR);
}

___HIDDEN ___SCMOBJ extensible_string_set_length
   ___P((extensible_string *str,
         int len,
         int fudge),
        (str,
         len,
         fudge)
extensible_string *str;
int len;
int fudge;)
{
  if (len > str->max_length || 2*len+1 < str->max_length)
    {
      int i;
      int new_max_length = (fudge<0) ? 3*len/2+1 : len+fudge;
      extensible_string_char *old_buffer = str->buffer;
      extensible_string_char *new_buffer =
        ___CAST(extensible_string_char*,
                ___alloc_mem (new_max_length *
                              sizeof (extensible_string_char)));

      if (new_buffer == NULL)
        return ___FIX(___HEAP_OVERFLOW_ERR);

      i = str->length;
      if (i > len)
        i = len;

      while (i-- > 0)
        new_buffer[i] = old_buffer[i];

      ___free_mem (old_buffer);
      str->buffer = new_buffer;
      str->max_length = new_max_length;
    }

  str->length = len;

  return ___FIX(___NO_ERR);
}

___HIDDEN void extensible_string_delete
   ___P((extensible_string *str,
         int pos,
         int len),
        (str,
         pos,
         len)
extensible_string *str;
int pos;
int len;)
{
  int i;

  if (pos < 0)
    pos = 0;
  else if (pos > str->length)
    pos = str->length;

  i = str->length - pos;
  if (len < 0)
    len = 0;
  else if (len > i)
    len = i;

  for (i=pos; i<str->length-len; i++)
    str->buffer[i] = str->buffer[i+len];

  str->length -= len;
}

___HIDDEN ___SCMOBJ extensible_string_insert
   ___P((extensible_string *str,
         int pos,
         int len,
         extensible_string_char *chars),
        (str,
         pos,
         len,
         chars)
extensible_string *str;
int pos;
int len;
extensible_string_char *chars;)
{
  ___SCMOBJ e;
  int i;

  if (pos < 0)
    pos = 0;
  else if (pos > str->length)
    pos = str->length;

  if (len < 0)
    len = 0;

  if ((e = extensible_string_set_length (str, str->length+len, -1))
      == ___FIX(___NO_ERR))
    {
      for (i=str->length-len-1; i>=pos; i--)
        str->buffer[i+len] = str->buffer[i];

      for (i=len-1; i>=0; i--)
        str->buffer[i+pos] = chars[i];
    }

  return e;
}

/*---------------------------------------------------------------------------*/

/* Line editing input decoder routines */

___HIDDEN ___SCMOBJ lineedit_input_decoder_setup
   ___P((lineedit_input_decoder *decoder),
        (decoder)
lineedit_input_decoder *decoder;)
{
#define LINEEDIT_INPUT_DECODER_INITIAL_BUFFER_SIZE 150

  int n = LINEEDIT_INPUT_DECODER_INITIAL_BUFFER_SIZE;

  decoder->buffer = ___CAST(lineedit_input_test*,
                            ___alloc_mem (n * sizeof (lineedit_input_test)));

  if (decoder->buffer == NULL)
    return ___FIX(___HEAP_OVERFLOW_ERR);

  decoder->length = 0;
  decoder->max_length = n;

  return ___FIX(___NO_ERR);
}

___HIDDEN void lineedit_input_decoder_cleanup
   ___P((lineedit_input_decoder *decoder),
        (decoder)
lineedit_input_decoder *decoder;)
{
  ___free_mem (decoder->buffer);
}

___HIDDEN ___SCMOBJ lineedit_input_decoder_set_length
   ___P((lineedit_input_decoder *decoder,
         int len,
         int fudge),
        (decoder,
         len,
         fudge)
lineedit_input_decoder *decoder;
int len;
int fudge;)
{
  if (len > LINEEDIT_INPUT_DECODER_MAX_LENGTH)
    return ___FIX(___UNKNOWN_ERR);

  if (len > decoder->max_length)
    {
      int i;
      int new_max_length = (fudge<0) ? 3*len/2+1 : len+fudge;
      lineedit_input_test *old_buffer = decoder->buffer;
      lineedit_input_test *new_buffer =
        ___CAST(lineedit_input_test*,
                ___alloc_mem (new_max_length *
                              sizeof (lineedit_input_test)));

      if (new_buffer == NULL)
        return ___FIX(___HEAP_OVERFLOW_ERR);

      i = decoder->length;
      if (i > len)
        i = len;

      while (i-- > 0)
        new_buffer[i] = old_buffer[i];

      ___free_mem (old_buffer);
      decoder->buffer = new_buffer;
      decoder->max_length = new_max_length;
    }

  decoder->length = len;

  return ___FIX(___NO_ERR);
}

___HIDDEN ___SCMOBJ lineedit_input_decoder_add
   ___P((lineedit_input_decoder *decoder,
         char *seq,
         ___U8 event),
        (decoder,
         seq,
         event)
lineedit_input_decoder *decoder;
char *seq;
___U8 event;)
{
  ___SCMOBJ e;
  ___U8 b;
  int i = 0;
  char *p = seq;

  if (event & WITH_ESC_PREFIX)
    b = UNICODE_ESCAPE;
  else
    b = ___CAST_U8(*p++);

  if (decoder->length > 0)
    {
      while (b != ___CAST_U8('\0') || p == seq+1)
        {
          if (decoder->buffer[i].trigger == b)
            {
              int a = decoder->buffer[i].action;
              if (a < decoder->length)
                {
                  i = a;
                  b = ___CAST_U8(*p++);
                }
              else
                return ___FIX(___NO_ERR); /* prefix of sequence exists */
            }
          else
            {
              int n = decoder->buffer[i].next;
              if (n < decoder->length)
                i = n;
              else
                {
                  decoder->buffer[i].next = decoder->length;
                  break;
                }
            }
        }
    }

  if (b != ___CAST_U8('\0') || p == seq+1)
    {
      /* sequence is not a prefix of a preexisting sequence */

      while (b != ___CAST_U8('\0') || p == seq+1)
        {
          i = decoder->length;

          if ((e = lineedit_input_decoder_set_length (decoder, i+1, -1))
              != ___FIX(___NO_ERR))
            return e;

          decoder->buffer[i].trigger = b;
          decoder->buffer[i].action = i+1;
          decoder->buffer[i].next = LINEEDIT_INPUT_DECODER_STATE_MAX;

          b = ___CAST_U8(*p++);
        }
      decoder->buffer[i].action =
        LINEEDIT_INPUT_DECODER_STATE_MAX-(event&~WITH_ESC_PREFIX);
    }

  return ___FIX(___NO_ERR);
}

typedef struct lineedit_defseq_struct
  {
    char *seq;
    ___U8 event;
  } lineedit_defseq;

___HIDDEN lineedit_defseq lineedit_defseq_common[] =
{
  /* sequences that all terminals support */

   { "\012",     LINEEDIT_EV_RETURN                           }
  ,{ "\015",     LINEEDIT_EV_RETURN                           }
  ,{ "\010",     LINEEDIT_EV_BACK                             }
  ,{ "\010",     LINEEDIT_EV_BACK_WORD      | WITH_ESC_PREFIX }
  ,{ "\011",     LINEEDIT_EV_TAB                              }
};

___HIDDEN lineedit_defseq lineedit_defseq_map_rubout_to_backspace[] =
{
  /* sequences that map the rubout key to backspace */

   { "\177",     LINEEDIT_EV_BACK                             }
  ,{ "\177",     LINEEDIT_EV_BACK_WORD      | WITH_ESC_PREFIX }
};

___HIDDEN lineedit_defseq lineedit_defseq_emacs[] =
{
  /* sequences specific to emacs mode */

   { "\0",       LINEEDIT_EV_MARK                             }
  ,{ "\031",     LINEEDIT_EV_PASTE                            }
  ,{ "\027",     LINEEDIT_EV_CUT                              }
  ,{ "\013",     LINEEDIT_EV_CUT_RIGHT                        }
  ,{ "\025",     LINEEDIT_EV_CUT_LEFT                         }
  ,{ "\014",     LINEEDIT_EV_REFRESH                          }
  ,{ "\024",     LINEEDIT_EV_TRANSPOSE                        }
  ,{ "\024",     LINEEDIT_EV_TRANSPOSE_WORD | WITH_ESC_PREFIX }
  ,{ "t",        LINEEDIT_EV_TRANSPOSE_WORD | WITH_ESC_PREFIX }
  ,{ "T",        LINEEDIT_EV_TRANSPOSE_WORD | WITH_ESC_PREFIX }

  ,{ "\020",     LINEEDIT_EV_UP                               }
  ,{ "\016",     LINEEDIT_EV_DOWN                             }
  ,{ "\006",     LINEEDIT_EV_RIGHT                            }
  ,{ "\002",     LINEEDIT_EV_LEFT                             }
  ,{ "\006",     LINEEDIT_EV_RIGHT_WORD     | WITH_ESC_PREFIX }
  ,{ "f",        LINEEDIT_EV_RIGHT_WORD     | WITH_ESC_PREFIX }
  ,{ "F",        LINEEDIT_EV_RIGHT_WORD     | WITH_ESC_PREFIX }
  ,{ "\002",     LINEEDIT_EV_LEFT_WORD      | WITH_ESC_PREFIX }
  ,{ "b",        LINEEDIT_EV_LEFT_WORD      | WITH_ESC_PREFIX }
  ,{ "B",        LINEEDIT_EV_LEFT_WORD      | WITH_ESC_PREFIX }

  ,{ "\001",     LINEEDIT_EV_HOME                             }
  ,{ "\004",     LINEEDIT_EV_DELETE                           }
  ,{ "\005",     LINEEDIT_EV_END                              }
  ,{ "\004",     LINEEDIT_EV_DELETE_WORD    | WITH_ESC_PREFIX }
};

___HIDDEN lineedit_defseq lineedit_defseq_widespread[] =
{
  /* sequences that many types of terminals support */

   { "[A",       LINEEDIT_EV_UP             | WITH_ESC_PREFIX }
  ,{ "[B",       LINEEDIT_EV_DOWN           | WITH_ESC_PREFIX }
  ,{ "[C",       LINEEDIT_EV_RIGHT          | WITH_ESC_PREFIX }
  ,{ "[D",       LINEEDIT_EV_LEFT           | WITH_ESC_PREFIX }

  ,{ "\033[A",   LINEEDIT_EV_NONE           | WITH_ESC_PREFIX }
  ,{ "\033[B",   LINEEDIT_EV_NONE           | WITH_ESC_PREFIX }
  ,{ "\033[C",   LINEEDIT_EV_RIGHT_WORD     | WITH_ESC_PREFIX }
  ,{ "\033[D",   LINEEDIT_EV_LEFT_WORD      | WITH_ESC_PREFIX }

  ,{ "OA",       LINEEDIT_EV_UP             | WITH_ESC_PREFIX }
  ,{ "OB",       LINEEDIT_EV_DOWN           | WITH_ESC_PREFIX }
  ,{ "OC",       LINEEDIT_EV_RIGHT          | WITH_ESC_PREFIX }
  ,{ "OD",       LINEEDIT_EV_LEFT           | WITH_ESC_PREFIX }

  ,{ "\033OA",   LINEEDIT_EV_NONE           | WITH_ESC_PREFIX }
  ,{ "\033OB",   LINEEDIT_EV_NONE           | WITH_ESC_PREFIX }
  ,{ "\033OC",   LINEEDIT_EV_RIGHT_WORD     | WITH_ESC_PREFIX }
  ,{ "\033OD",   LINEEDIT_EV_LEFT_WORD      | WITH_ESC_PREFIX }

  ,{ "A",        LINEEDIT_EV_UP             | WITH_ESC_PREFIX }
  ,{ "B",        LINEEDIT_EV_DOWN           | WITH_ESC_PREFIX }
  ,{ "C",        LINEEDIT_EV_RIGHT          | WITH_ESC_PREFIX }
  ,{ "D",        LINEEDIT_EV_LEFT           | WITH_ESC_PREFIX }

  ,{ "\033A",    LINEEDIT_EV_NONE           | WITH_ESC_PREFIX }
  ,{ "\033B",    LINEEDIT_EV_NONE           | WITH_ESC_PREFIX }
  ,{ "\033C",    LINEEDIT_EV_RIGHT_WORD     | WITH_ESC_PREFIX }
  ,{ "\033D",    LINEEDIT_EV_LEFT_WORD      | WITH_ESC_PREFIX }

#ifdef USE_XTERM_CTRL_COMBINATIONS
  ,{ "[5A",      LINEEDIT_EV_NONE           | WITH_ESC_PREFIX }
  ,{ "[5B",      LINEEDIT_EV_NONE           | WITH_ESC_PREFIX }
  ,{ "[5C",      LINEEDIT_EV_RIGHT_WORD     | WITH_ESC_PREFIX }
  ,{ "[5D",      LINEEDIT_EV_LEFT_WORD      | WITH_ESC_PREFIX }
#endif

  ,{ "[1~",      LINEEDIT_EV_HOME           | WITH_ESC_PREFIX }
  ,{ "[2~",      LINEEDIT_EV_INSERT         | WITH_ESC_PREFIX }
  ,{ "[3~",      LINEEDIT_EV_DELETE         | WITH_ESC_PREFIX }
  ,{ "[4~",      LINEEDIT_EV_END            | WITH_ESC_PREFIX }
  ,{ "[5~",      LINEEDIT_EV_HOME_DOC       | WITH_ESC_PREFIX }
  ,{ "[6~",      LINEEDIT_EV_END_DOC        | WITH_ESC_PREFIX }

  ,{ "\033[1~",  LINEEDIT_EV_HOME_DOC       | WITH_ESC_PREFIX }
  ,{ "\033[2~",  LINEEDIT_EV_NONE           | WITH_ESC_PREFIX }
  ,{ "\033[3~",  LINEEDIT_EV_DELETE_WORD    | WITH_ESC_PREFIX }
  ,{ "\033[4~",  LINEEDIT_EV_END_DOC        | WITH_ESC_PREFIX }
  ,{ "\033[5~",  LINEEDIT_EV_NONE           | WITH_ESC_PREFIX }
  ,{ "\033[6~",  LINEEDIT_EV_NONE           | WITH_ESC_PREFIX }

  ,{ "[H",       LINEEDIT_EV_HOME           | WITH_ESC_PREFIX }
  ,{ "[L",       LINEEDIT_EV_INSERT         | WITH_ESC_PREFIX }
  ,{ "[F",       LINEEDIT_EV_END            | WITH_ESC_PREFIX }
  ,{ "\177",     LINEEDIT_EV_DELETE                           }
  ,{ "\033[H",   LINEEDIT_EV_HOME_DOC       | WITH_ESC_PREFIX }
  ,{ "\177",     LINEEDIT_EV_DELETE_WORD    | WITH_ESC_PREFIX }
  ,{ "\033[F",   LINEEDIT_EV_END_DOC        | WITH_ESC_PREFIX }

#ifdef USE_XTERM_CTRL_COMBINATIONS
  ,{ "[1;5~",    LINEEDIT_EV_HOME_DOC       | WITH_ESC_PREFIX }
  ,{ "[2;5~",    LINEEDIT_EV_NONE           | WITH_ESC_PREFIX }
  ,{ "[3;5~",    LINEEDIT_EV_DELETE_WORD    | WITH_ESC_PREFIX }
  ,{ "[4;5~",    LINEEDIT_EV_END_DOC        | WITH_ESC_PREFIX }
  ,{ "[5;5~",    LINEEDIT_EV_NONE           | WITH_ESC_PREFIX }
  ,{ "[6;5~",    LINEEDIT_EV_NONE           | WITH_ESC_PREFIX }

  ,{ "[5H",      LINEEDIT_EV_HOME_DOC       | WITH_ESC_PREFIX }
  ,{ "[5L",      LINEEDIT_EV_NONE           | WITH_ESC_PREFIX }
  ,{ "[5F",      LINEEDIT_EV_END_DOC        | WITH_ESC_PREFIX }
#endif

  ,{ "OP",       LINEEDIT_EV_F1             | WITH_ESC_PREFIX }
  ,{ "OQ",       LINEEDIT_EV_F2             | WITH_ESC_PREFIX }
  ,{ "OR",       LINEEDIT_EV_F3             | WITH_ESC_PREFIX }
  ,{ "OS",       LINEEDIT_EV_F4             | WITH_ESC_PREFIX }

  ,{ "\033OP",   LINEEDIT_EV_META_F1        | WITH_ESC_PREFIX }
  ,{ "\033OQ",   LINEEDIT_EV_META_F2        | WITH_ESC_PREFIX }
  ,{ "\033OR",   LINEEDIT_EV_META_F3        | WITH_ESC_PREFIX }
  ,{ "\033OS",   LINEEDIT_EV_META_F4        | WITH_ESC_PREFIX }

#ifdef USE_XTERM_CTRL_COMBINATIONS
  ,{ "O5P",      LINEEDIT_EV_META_F1        | WITH_ESC_PREFIX }
  ,{ "O5Q",      LINEEDIT_EV_META_F2        | WITH_ESC_PREFIX }
  ,{ "O5R",      LINEEDIT_EV_META_F3        | WITH_ESC_PREFIX }
  ,{ "O5S",      LINEEDIT_EV_META_F4        | WITH_ESC_PREFIX }
#endif

  ,{ "[11~",     LINEEDIT_EV_F1             | WITH_ESC_PREFIX }
  ,{ "[12~",     LINEEDIT_EV_F2             | WITH_ESC_PREFIX }
  ,{ "[13~",     LINEEDIT_EV_F3             | WITH_ESC_PREFIX }
  ,{ "[14~",     LINEEDIT_EV_F4             | WITH_ESC_PREFIX }

  ,{ "\033[11~", LINEEDIT_EV_META_F1        | WITH_ESC_PREFIX }
  ,{ "\033[12~", LINEEDIT_EV_META_F2        | WITH_ESC_PREFIX }
  ,{ "\033[13~", LINEEDIT_EV_META_F3        | WITH_ESC_PREFIX }
  ,{ "\033[14~", LINEEDIT_EV_META_F4        | WITH_ESC_PREFIX }

#ifdef USE_XTERM_CTRL_COMBINATIONS
  ,{ "[11;5~",   LINEEDIT_EV_META_F1        | WITH_ESC_PREFIX }
  ,{ "[12;5~",   LINEEDIT_EV_META_F2        | WITH_ESC_PREFIX }
  ,{ "[13;5~",   LINEEDIT_EV_META_F3        | WITH_ESC_PREFIX }
  ,{ "[14;5~",   LINEEDIT_EV_META_F4        | WITH_ESC_PREFIX }
#endif

#ifdef LINEEDIT_SUPPORT_F5_TO_F12

  ,{ "[15~",     LINEEDIT_EV_F5             | WITH_ESC_PREFIX }
  ,{ "[17~",     LINEEDIT_EV_F6             | WITH_ESC_PREFIX }
  ,{ "[18~",     LINEEDIT_EV_F7             | WITH_ESC_PREFIX }
  ,{ "[19~",     LINEEDIT_EV_F8             | WITH_ESC_PREFIX }
  ,{ "[20~",     LINEEDIT_EV_F9             | WITH_ESC_PREFIX }
  ,{ "[21~",     LINEEDIT_EV_F10            | WITH_ESC_PREFIX }
  ,{ "[23~",     LINEEDIT_EV_F11            | WITH_ESC_PREFIX }
  ,{ "[24~",     LINEEDIT_EV_F12            | WITH_ESC_PREFIX }

  ,{ "\033[15~", LINEEDIT_EV_META_F5        | WITH_ESC_PREFIX }
  ,{ "\033[17~", LINEEDIT_EV_META_F6        | WITH_ESC_PREFIX }
  ,{ "\033[18~", LINEEDIT_EV_META_F7        | WITH_ESC_PREFIX }
  ,{ "\033[19~", LINEEDIT_EV_META_F8        | WITH_ESC_PREFIX }
  ,{ "\033[20~", LINEEDIT_EV_META_F9        | WITH_ESC_PREFIX }
  ,{ "\033[21~", LINEEDIT_EV_META_F10       | WITH_ESC_PREFIX }
  ,{ "\033[23~", LINEEDIT_EV_META_F11       | WITH_ESC_PREFIX }
  ,{ "\033[24~", LINEEDIT_EV_META_F12       | WITH_ESC_PREFIX }

#ifdef USE_XTERM_CTRL_COMBINATIONS
  ,{ "[15;5~",   LINEEDIT_EV_META_F5        | WITH_ESC_PREFIX }
  ,{ "[17;5~",   LINEEDIT_EV_META_F6        | WITH_ESC_PREFIX }
  ,{ "[18;5~",   LINEEDIT_EV_META_F7        | WITH_ESC_PREFIX }
  ,{ "[19;5~",   LINEEDIT_EV_META_F8        | WITH_ESC_PREFIX }
  ,{ "[20;5~",   LINEEDIT_EV_META_F9        | WITH_ESC_PREFIX }
  ,{ "[21;5~",   LINEEDIT_EV_META_F10       | WITH_ESC_PREFIX }
  ,{ "[23;5~",   LINEEDIT_EV_META_F11       | WITH_ESC_PREFIX }
  ,{ "[24;5~",   LINEEDIT_EV_META_F12       | WITH_ESC_PREFIX }
#endif

#endif
};

___HIDDEN ___SCMOBJ lineedit_defseq_install_table
   ___P((lineedit_input_decoder *decoder,
         lineedit_defseq *table,
         int len),
        (decoder,
         table,
         len)
lineedit_input_decoder *decoder;
lineedit_defseq *table;
int len;)
{
  ___SCMOBJ e;
  int i;

  for (i=0; i<len; i++)
    if ((e = lineedit_input_decoder_add
               (decoder,
                table[i].seq,
                table[i].event))
        != ___FIX(___NO_ERR))
      return e;

  return ___FIX(___NO_ERR);
}

#ifdef USE_CURSES

typedef struct lineedit_defkey_struct
  {
    char *cap;
    ___U8 event_no_escape;
    ___U8 event_with_escape;
  } lineedit_defkey;

#ifdef USE_TERMCAP
#define DEFKEY(tcap_name,tinfo_name,xterm_def,event_no_esc,event_with_esc) \
{ tcap_name, event_no_esc, event_with_esc | WITH_ESC_PREFIX }
#else
#ifdef USE_TERMINFO
#define DEFKEY(tcap_name,tinfo_name,xterm_def,event_no_esc,event_with_esc) \
{ tinfo_name, event_no_esc, event_with_esc | WITH_ESC_PREFIX }
#else
#define DEFKEY(tcap_name,tinfo_name,xterm_def,event_no_esc,event_with_esc) \
{ xterm_def, event_no_esc, event_with_esc | WITH_ESC_PREFIX }
#endif
#endif

typedef struct lineedit_defcap_struct
  {
    char *cap;
  } lineedit_defcap;

#ifdef USE_TERMCAP
#define DEFCAP(tcap_name,tinfo_name,xterm_def) { tcap_name }
#else
#ifdef USE_TERMINFO
#define DEFCAP(tcap_name,tinfo_name,xterm_def) { tinfo_name }
#else
#define DEFCAP(tcap_name,tinfo_name,xterm_def) { xterm_def }
#endif
#endif

___HIDDEN lineedit_defcap lineedit_defcap_table[] =
{
  /* order must match LINEEDIT_CAP_XXX */

   DEFCAP( "cl", "clear", "\033[H\033[2J" ) /* erase screen and home cursor */
  ,DEFCAP( "cd", "ed",    "\033[J"        ) /* erase to end of screen       */
  ,DEFCAP( "up", "cuu1",  "\033[A"        ) /* move up one line             */
  ,DEFCAP( "do", "cud1",  "\012"          ) /* move down one line           */
  ,DEFCAP( "UP", "cuu",   "\033[%p1%dA"   ) /* move up N lines              */
  ,DEFCAP( "DO", "cud",   "\033[%p1%dB"   ) /* move down N lines            */
  ,DEFCAP( "RI", "cuf",   "\033[%p1%dC"   ) /* move right N columns         */
  ,DEFCAP( "LE", "cub",   "\033[%p1%dD"   ) /* move left N columns          */
  ,DEFCAP( "me", "sgr0",  "\033[m\017"    ) /* turn off all attributes      */
  ,DEFCAP( "md", "bold",  "\033[1m"       ) /* turn on bold attribute       */
  ,DEFCAP( "mr", "rev",   "\033[7m"       ) /* turn on reverse attribute    */
  ,DEFCAP( "AF", "setaf", "\033[3%p1%dm"  ) /* ANSI set foreground color    */
  ,DEFCAP( "AB", "setab", "\033[4%p1%dm"  ) /* ANSI set background color    */
};

___HIDDEN lineedit_defkey lineedit_defkey_table[] =
{
   DEFKEY("ku","kcuu1","\033OA",  LINEEDIT_EV_UP,      LINEEDIT_EV_NONE       )
  ,DEFKEY("kd","kcud1","\033OB",  LINEEDIT_EV_DOWN,    LINEEDIT_EV_NONE       )
  ,DEFKEY("kr","kcuf1","\033OC",  LINEEDIT_EV_RIGHT,   LINEEDIT_EV_RIGHT_WORD )
  ,DEFKEY("kl","kcub1","\033OD",  LINEEDIT_EV_LEFT,    LINEEDIT_EV_LEFT_WORD  )
  ,DEFKEY("kh","khome","\033[1~", LINEEDIT_EV_HOME,    LINEEDIT_EV_HOME_DOC   )
  ,DEFKEY("kI","kich1","\033[2~", LINEEDIT_EV_INSERT,  LINEEDIT_EV_NONE       )
  ,DEFKEY("kD","kdch1","\033[3~", LINEEDIT_EV_DELETE,  LINEEDIT_EV_DELETE_WORD)
  ,DEFKEY("@7","kend", "\033[4~", LINEEDIT_EV_END,     LINEEDIT_EV_END_DOC    )
  ,DEFKEY("kP","kpp",  "\033[5~", LINEEDIT_EV_HOME_DOC,LINEEDIT_EV_NONE       )
  ,DEFKEY("kN","knp",  "\033[6~", LINEEDIT_EV_END_DOC, LINEEDIT_EV_NONE       )
  ,DEFKEY("k1","kf1",  "\033OP",  LINEEDIT_EV_F1,      LINEEDIT_EV_META_F1    )
  ,DEFKEY("k2","kf2",  "\033OQ",  LINEEDIT_EV_F2,      LINEEDIT_EV_META_F2    )
  ,DEFKEY("k3","kf3",  "\033OR",  LINEEDIT_EV_F3,      LINEEDIT_EV_META_F3    )
  ,DEFKEY("k4","kf4",  "\033OS",  LINEEDIT_EV_F4,      LINEEDIT_EV_META_F4    )
#ifdef LINEEDIT_SUPPORT_F5_TO_F12
  ,DEFKEY("k5","kf5",  "\033[15~",LINEEDIT_EV_F5,      LINEEDIT_EV_META_F5    )
  ,DEFKEY("k6","kf6",  "\033[17~",LINEEDIT_EV_F6,      LINEEDIT_EV_META_F6    )
  ,DEFKEY("k7","kf7",  "\033[18~",LINEEDIT_EV_F7,      LINEEDIT_EV_META_F7    )
  ,DEFKEY("k8","kf8",  "\033[19~",LINEEDIT_EV_F8,      LINEEDIT_EV_META_F8    )
  ,DEFKEY("k9","kf9",  "\033[20~",LINEEDIT_EV_F9,      LINEEDIT_EV_META_F9    )
  ,DEFKEY("k;","kf10", "\033[21~",LINEEDIT_EV_F10,     LINEEDIT_EV_META_F10   )
  ,DEFKEY("F1","kf11", "\033[23~",LINEEDIT_EV_F11,     LINEEDIT_EV_META_F11   )
  ,DEFKEY("F2","kf12", "\033[24~",LINEEDIT_EV_F12,     LINEEDIT_EV_META_F12   )
#endif
};

___HIDDEN ___SCMOBJ lineedit_defkey_install
   ___P((lineedit_input_decoder *decoder,
         char *cap,
         ___U8 event_no_escape,
         ___U8 event_with_escape),
        (decoder,
         cap,
         event_no_escape,
         event_with_escape)
lineedit_input_decoder *decoder;
char *cap;
___U8 event_no_escape;
___U8 event_with_escape;)
{
  ___SCMOBJ e = ___FIX(___NO_ERR);
  char *seq;

#ifdef USE_TERMCAP
  seq = tgetstr (cap, NULL);
#else
#ifdef USE_TERMINFO
  seq = tigetstr (cap);
#else
  seq = cap;
#endif
#endif

  if (seq != ___CAST(char*,-1) && seq != NULL)
    if ((e = lineedit_input_decoder_add (decoder, seq, event_no_escape))
        != ___FIX(___NO_ERR) &&
        event_with_escape != (LINEEDIT_EV_NONE|WITH_ESC_PREFIX))
      e = lineedit_input_decoder_add (decoder, seq, event_with_escape);

  return e;
}

___HIDDEN ___SCMOBJ lineedit_defkey_install_table
   ___P((lineedit_input_decoder *decoder,
         lineedit_defkey *table,
         int len),
        (decoder,
         table,
         len)
lineedit_input_decoder *decoder;
lineedit_defkey *table;
int len;)
{
  ___SCMOBJ e;
  int i;

  for (i=0; i<len; i++)
    if ((e = lineedit_defkey_install
               (decoder,
                table[i].cap,
                table[i].event_no_escape,
                table[i].event_with_escape))
        != ___FIX(___NO_ERR))
      return e;

  return ___FIX(___NO_ERR);
}

#endif

#define NBELEMS(table)(sizeof (table) / sizeof (table[0]))

___HIDDEN ___SCMOBJ lineedit_input_decoder_init
   ___P((lineedit_input_decoder *decoder,
         ___BOOL map_rubout_to_backspace,
         ___BOOL emacs_bindings,
         ___BOOL curses_keys),
        (decoder,
         map_rubout_to_backspace,
         emacs_bindings,
         curses_keys)
lineedit_input_decoder *decoder;
___BOOL map_rubout_to_backspace;
___BOOL emacs_bindings;
___BOOL curses_keys;)
{
  ___SCMOBJ e;

  decoder->length = 0;

  if ((e = lineedit_defseq_install_table
             (decoder,
              lineedit_defseq_common,
              NBELEMS(lineedit_defseq_common)))
      != ___FIX(___NO_ERR))
    return e;

  if (map_rubout_to_backspace)
    if ((e = lineedit_defseq_install_table
               (decoder,
                lineedit_defseq_map_rubout_to_backspace,
                NBELEMS(lineedit_defseq_map_rubout_to_backspace)))
        != ___FIX(___NO_ERR))
      return e;

  if (emacs_bindings)
    if ((e = lineedit_defseq_install_table
               (decoder,
                lineedit_defseq_emacs,
                NBELEMS(lineedit_defseq_emacs)))
        != ___FIX(___NO_ERR))
      return e;

#ifdef USE_CURSES

  if (curses_keys)
    if ((e = lineedit_defkey_install_table
               (decoder,
                lineedit_defkey_table,
                NBELEMS(lineedit_defkey_table)))
        != ___FIX(___NO_ERR))
      return e;

#endif

  if ((e = lineedit_defseq_install_table
             (decoder,
              lineedit_defseq_widespread,
              NBELEMS(lineedit_defseq_widespread)))
      != ___FIX(___NO_ERR))
    return e;

  return ___FIX(___NO_ERR);
}

/*---------------------------------------------------------------------------*/

/* Line editing history routines */

___HIDDEN ___SCMOBJ lineedit_history_begin_edit
   ___P((lineedit_state *state,
         lineedit_history *h),
        (state,
         h)
lineedit_state *state;
lineedit_history *h;)
{
#define LINEEDIT_FUDGE 80

  ___SCMOBJ e;

  if (h->edited.buffer == NULL)
    {
      if ((e = extensible_string_copy
                 (h->actual.buffer,
                  h->actual.length,
                  &h->edited,
                  LINEEDIT_FUDGE))
          != ___FIX(___NO_ERR))
        return e;
    }

  return ___FIX(___NO_ERR);
}

___HIDDEN void lineedit_history_end_edit
   ___P((lineedit_state *state,
         lineedit_history *h),
        (state,
         h)
lineedit_state *state;
lineedit_history *h;)
{
  if (h->edited.buffer != NULL)
    {
      extensible_string_cleanup (&h->edited);
      h->edited.buffer = NULL;
    }
}

___HIDDEN ___SCMOBJ lineedit_history_setup
   ___P((lineedit_state *state,
         lineedit_history **hist),
        (state,
         hist)
lineedit_state *state;
lineedit_history **hist;)
{
  ___SCMOBJ e;
  lineedit_history *h;

  h = ___CAST(lineedit_history*,
              ___alloc_mem (sizeof (lineedit_history)));

  if (h == NULL)
    return ___FIX(___HEAP_OVERFLOW_ERR);

  if ((e = extensible_string_setup (&h->actual, 0)) != ___FIX(___NO_ERR))
    {
      ___free_mem (h);
      return e;
    }

  h->edited.buffer = NULL;

  h->prev = h; /* create a circular list */
  h->next = h;

  *hist = h;

  return ___FIX(___NO_ERR);
}

___HIDDEN ___SCMOBJ lineedit_history_cleanup
   ___P((lineedit_state *state,
         lineedit_history *h),
        (state,
         h)
lineedit_state *state;
lineedit_history *h;)
{
  lineedit_history_end_edit (state, h);

  extensible_string_cleanup (&h->actual);

  ___free_mem (h);

  return ___FIX(___NO_ERR);
}

___HIDDEN void lineedit_history_remove
   ___P((lineedit_state *state,
         lineedit_history *item),
        (state,
         item)
lineedit_state *state;
lineedit_history *item;)
{
  lineedit_history *prev = item->prev;
  lineedit_history *next = item->next;

  if (prev == item)
    state->hist_last = NULL;
  else
    {
      next->prev = prev;
      prev->next = next;
      item->prev = item;
      item->next = item;
      if (state->hist_last == item)
        state->hist_last = prev;
    }
}

___HIDDEN void lineedit_history_trim
   ___P((lineedit_state *state),
        (state)
lineedit_state *state;)
{
  while (state->history_length > state->max_history_length)
    {
      lineedit_history *first = state->hist_last->next;
      lineedit_history_remove (state, first);
      lineedit_history_cleanup (state, first); /* ignore error */
      state->history_length--;
    }
}

___HIDDEN void lineedit_set_max_history_length
   ___P((lineedit_state *state,
         int max_history_length),
        (state,
         max_history_length)
lineedit_state *state;
int max_history_length;)
{
  if (max_history_length >= 0)
    {
      state->max_history_length = max_history_length;
      lineedit_history_trim (state);
    }
}

___HIDDEN void lineedit_history_add
   ___P((lineedit_state *state,
         lineedit_history *item),
        (state,
         item)
lineedit_state *state;
lineedit_history *item;)
{
  lineedit_history *dest = state->hist_last;

  if (dest == NULL)
    {
      item->prev = item;
      item->next = item;
    }
  else
    {
      lineedit_history *after_dest = dest->next;
      item->next = after_dest;
      item->prev = dest;
      dest->next = item;
      after_dest->prev = item;
    }

  state->hist_last = item;
  state->history_length++;
}


/*---------------------------------------------------------------------------*/

/* Line editing routines */

___HIDDEN ___SCMOBJ lineedit_output_drain
   ___P((lineedit_state *state),
        (state)
lineedit_state *state;)
{
  /*
   * Drain the output generated by the previous line editing
   * operation.
   */

  ___SCMOBJ e;
  ___stream_index len_done;
  int wlo;
  int length;
  tty_state *ts = state->ts;

  if ((e = ts->werr) != ___FIX(___NO_ERR))
    {
      ts->werr = ___FIX(___NO_ERR);
      return e;
    }

  tty_wbuf_drain (ts); /* error code returned in ts->werr */

  wlo = state->output_wlo;
  length = state->output.length;

  while (wlo < length)
    {
      if ((e = ts->werr) != ___FIX(___NO_ERR))
        {
          ts->werr = ___FIX(___NO_ERR);
          state->output_wlo = wlo;
          return e;
        }

      e = tty_write_to_wbuf (ts, &state->output.buffer[wlo],
                             length - wlo,
                             &len_done);

      if (e != ___FIX(___NO_ERR))
        {
#ifdef LINEEDIT_REPORT_CHAR_ENCODING_ERRORS
          if (e == ___FIX(___STOC_LATIN1STRING_ERR))
            wlo++;
#endif
          state->output_wlo = wlo;
          return e;
        }

      if (len_done == 0)
        {
          state->output_wlo = wlo;
          return e;
        }

      wlo += len_done;

      tty_wbuf_drain (ts); /* error code returned in ts->werr */
    }

  if (extensible_string_set_length (&state->output, 0, 1) == ___FIX(___NO_ERR))
    wlo = 0;

  state->output_wlo = wlo;

  return ___FIX(___NO_ERR);
}

___HIDDEN void lineedit_track_cursor_movement
   ___P((lineedit_state *state,
         tty_char *buf,
         ___stream_index len),
        (state,
         buf,
         len)
lineedit_state *state;
tty_char *buf;
___stream_index len;)
{
  /*
   * Even with the termcap/terminfo database it is impossible to tell
   * exactly how the cursor will move when an arbitrary sequence of
   * characters is sent to the terminal.  So we make general
   * assumptions about the terminal behavior with a bias towards
   * "xterm", which is probably the terminal that will most often be
   * used under UNIX.
   */

  /*
   * The fields state->current.screen_cursor,
   * state->current.cursor_to_point, and state->current.delayed_wrap
   * are updated accordingly.
   */

  int screen_cols = state->screen_cols;
  int screen_cursor = state->current.screen_cursor;
  int cursor_to_point = state->current.cursor_to_point;
  ___BOOL delayed_wrap = state->current.delayed_wrap;
  tty_char c;
  int col;

  while (len-- > 0)
    {
      c = *buf++;

      switch (c)
        {
        case UNICODE_BELL:
          break;

        case UNICODE_BACKSPACE:
          if (screen_cursor % screen_cols > 0 ||
              (screen_cursor > 0 && state->has_auto_left_margin))
            screen_cursor--;
          delayed_wrap = 0;
          break;

        case UNICODE_TAB:
          col = ((screen_cursor % screen_cols) / state->tab_stop + 1)
                * state->tab_stop;
          if (col >= screen_cols)
            {
              if (state->has_auto_right_margin &&
                  !state->has_eat_newline_glitch)
                {
                  screen_cursor = (screen_cursor / screen_cols + 1)
                                  * screen_cols;
                  if (screen_cursor >= state->screen_size)
                    {
                      screen_cursor -= screen_cols;
                      cursor_to_point += screen_cols;
                    }
                }
              else
                {
                  col = screen_cols-1; /* don't wrap */
                  screen_cursor = (screen_cursor / screen_cols) * screen_cols
                                  + col;
                }
            }
          else
            screen_cursor = (screen_cursor / screen_cols) * screen_cols + col;
          break;

        case UNICODE_LINEFEED:
          if (state->linefeed_moves_to_left_margin)
            screen_cursor = (screen_cursor / screen_cols + 1) * screen_cols;
          else
            screen_cursor = screen_cursor + screen_cols;
          if (screen_cursor >= state->screen_size)
            {
              screen_cursor -= screen_cols;
              cursor_to_point += screen_cols;
            }
          delayed_wrap = 0;
          break;

        case UNICODE_RETURN:
          screen_cursor = (screen_cursor / screen_cols) * screen_cols;
          delayed_wrap = 0;
          break;

        default:
          if (delayed_wrap)
            {
              screen_cursor++;
              if (screen_cursor >= state->screen_size)
                {
                  screen_cursor -= screen_cols;
                  cursor_to_point += screen_cols;
                }
            }

          screen_cursor++;

          if (screen_cursor % screen_cols != 0)
            delayed_wrap = 0;
          else if (state->has_auto_right_margin)
            {
              if (state->has_eat_newline_glitch)
                {
                  screen_cursor--;
                  delayed_wrap = 1; /* wrap later */
                }
              else
                {
                  if (screen_cursor >= state->screen_size)
                    {
                      screen_cursor -= screen_cols;
                      cursor_to_point += screen_cols;
                    }
                  delayed_wrap = 0;
                }
            }
          else
            {
              screen_cursor--;
              delayed_wrap = 0;
            }
        }
    }

  state->current.screen_cursor = screen_cursor;
  state->current.cursor_to_point = cursor_to_point;
  state->current.delayed_wrap = delayed_wrap;
}

___HIDDEN ___SCMOBJ lineedit_output_string
   ___P((lineedit_state *state,
         char *str,
         int rep),
        (state,
         str,
         rep)
lineedit_state *state;
char *str;
int rep;)
{
  /*
   * Output the C string "str" multiple times ("rep" is the repeat
   * factor).
   */

  while (rep-- > 0)
    {
      char *p = str;
      while (*p != '\0')
        {
          ___SCMOBJ e;
          tty_char c = *p++;
          if ((e = extensible_string_insert
                     (&state->output,
                      state->output.length,
                      1,
                      &c))
              != ___FIX(___NO_ERR))
            return e;
#if 0
printf ("%c", c); fflush(stdout);/******************/
#endif
        }
    }

  return ___FIX(___NO_ERR);
}

#ifdef USE_CURSES

#ifdef USE_TERMCAP_OR_TERMINFO

___HIDDEN int lineedit_output_putc
   ___P((int c),
        (c)
int c;)
{
  /*
   * This routine outputs a C character.  It is needed for the call to
   * the termcap/terminfo "tputs" routine.  The parameters it uses are
   * stored in the lineedit module structure "lineedit_mod".
   */

  if (lineedit_mod.last_err == ___FIX(___NO_ERR))
    {
      tty_char ch = c;
      lineedit_mod.last_err = extensible_string_insert
                                (&lineedit_mod.current_state->output,
                                 lineedit_mod.current_state->output.length,
                                 1,
                                 &ch);
    }

  return c;
}

#endif

___HIDDEN ___SCMOBJ lineedit_output_cap
   ___P((lineedit_state *state,
         int cap,
         int arg,
         int rep),
        (state,
         cap,
         arg,
         rep)
lineedit_state *state;
int cap;
int arg;
int rep;)
{
  /*
   * Output the sequence for the terminal capability "cap" with
   * integer parameter "arg" (-1 means no parameter).  "rep" is the
   * repeat factor.
   */

  while (rep > 0)
    {
      char *str = state->capability[cap];

#ifdef USE_TERMCAP_OR_TERMINFO

      if (arg >= 0)
        {
#ifdef USE_TERMCAP
          str = tgoto (str, arg, arg);
#else
          str = tparm (str, arg);
#endif
        }

      if (cap < LINEEDIT_CAP_SGR0)
        state->current.delayed_wrap = 0; /* cursor cmds cancel delayed wrap */

      lineedit_mod.current_state = state;
      lineedit_mod.last_err = ___FIX(___NO_ERR);

      if (tputs (str, 1, lineedit_output_putc) == ERR &&
          lineedit_mod.last_err == ___FIX(___NO_ERR))
        lineedit_mod.last_err = ___FIX(___UNKNOWN_ERR);

      if (lineedit_mod.last_err != ___FIX(___NO_ERR))
        return lineedit_mod.last_err;

#else

      while (*str != '\0')
        {
          ___SCMOBJ e;
          tty_char ch = *str++;

          if (ch=='%') /* the parameter "arg" has to be inserted here */
            {
              int n = 1;

              while (n*10 <= arg)
                n *= 10;

              while (n > 0)
                {
                  ch = '0' + (arg / n) % 10;

                  if ((e = extensible_string_insert
                             (&state->output,
                              state->output.length,
                              1,
                              &ch))
                      != ___FIX(___NO_ERR))
                    return e;

                  n /= 10;
                }

              str += 4; /* skip "%p1%d" */
            }
          else
            {
              if ((e = extensible_string_insert
                         (&state->output,
                          state->output.length,
                          1,
                          &ch))
                  != ___FIX(___NO_ERR))
                return e;
            }
        }

#endif

      rep--;
    }

  return ___FIX(___NO_ERR);
}

#endif


___HIDDEN ___SCMOBJ lineedit_output_set_text_attrs
   ___P((lineedit_state *state,
         tty_text_attrs attrs),
        (state,
         attrs)
lineedit_state *state;
tty_text_attrs attrs;)
{
  /*
   * Output the control sequence to set the text attributes of the
   * next characters sent to the terminal.
   */

#ifdef USE_CURSES

  ___SCMOBJ e;
  int turn_on;
  tty_text_attrs current_attrs = state->current.attrs;

  if (current_attrs == attrs)
    return ___FIX(___NO_ERR);

  turn_on = GET_STYLE(attrs);

  if ((GET_STYLE(current_attrs) & ~turn_on) != 0 ||
      (GET_FOREGROUND_COLOR(attrs) >= DEFAULT_TEXT_COLOR &&
       GET_FOREGROUND_COLOR(current_attrs) < DEFAULT_TEXT_COLOR) ||
      (GET_BACKGROUND_COLOR(attrs) >= DEFAULT_TEXT_COLOR &&
       GET_BACKGROUND_COLOR(current_attrs) < DEFAULT_TEXT_COLOR))
    {
      if (state->capability[LINEEDIT_CAP_SGR0] != NULL)
        if ((e = lineedit_output_cap (state, LINEEDIT_CAP_SGR0, -1, 1))
            != ___FIX(___NO_ERR))
          return e;
      current_attrs = MAKE_TEXT_ATTRS(0,DEFAULT_TEXT_COLOR,DEFAULT_TEXT_COLOR);
    }
  else
    turn_on = (~GET_STYLE(current_attrs) & turn_on);

  if ((turn_on & TEXT_STYLE_BOLD) &&
      state->capability[LINEEDIT_CAP_BOLD] != NULL)
    if ((e = lineedit_output_cap (state, LINEEDIT_CAP_BOLD, -1, 1))
        != ___FIX(___NO_ERR))
      return e;

  if ((turn_on & TEXT_STYLE_REVERSE) &&
      state->capability[LINEEDIT_CAP_REV] != NULL)
    if ((e = lineedit_output_cap (state, LINEEDIT_CAP_REV, -1, 1))
        != ___FIX(___NO_ERR))
      return e;

  if (GET_FOREGROUND_COLOR(attrs) < DEFAULT_TEXT_COLOR &&
      GET_FOREGROUND_COLOR(attrs) != GET_FOREGROUND_COLOR(current_attrs) &&
      state->capability[LINEEDIT_CAP_SETAF] != NULL)
    if ((e = lineedit_output_cap
               (state,
                LINEEDIT_CAP_SETAF,
                GET_FOREGROUND_COLOR(attrs),
                1))
        != ___FIX(___NO_ERR))
      return e;

  if (GET_BACKGROUND_COLOR(attrs) < DEFAULT_TEXT_COLOR &&
      GET_BACKGROUND_COLOR(attrs) != GET_BACKGROUND_COLOR(current_attrs) &&
      state->capability[LINEEDIT_CAP_SETAB] != NULL)
    if ((e = lineedit_output_cap
               (state,
                LINEEDIT_CAP_SETAB,
                GET_BACKGROUND_COLOR(attrs),
                1))
        != ___FIX(___NO_ERR))
      return e;

  state->current.attrs = attrs;

#endif

#ifdef USE_WIN32

  {
    /*********************/

    WORD attributes = 0;

    if (GET_STYLE(attrs) & TEXT_STYLE_BOLD)
      attributes |= FOREGROUND_INTENSITY;

    if (GET_FOREGROUND_COLOR(attrs) == DEFAULT_TEXT_COLOR)
      attributes |= (FOREGROUND_RED|FOREGROUND_GREEN|FOREGROUND_BLUE);
    else
      {
        if (GET_FOREGROUND_COLOR(attrs) & 1)
          attributes |= FOREGROUND_RED;
        if (GET_FOREGROUND_COLOR(attrs) & 2)
          attributes |= FOREGROUND_GREEN;
        if (GET_FOREGROUND_COLOR(attrs) & 4)
          attributes |= FOREGROUND_BLUE;
      }

    if (GET_BACKGROUND_COLOR(attrs) == DEFAULT_TEXT_COLOR)
      attributes |= 0;
    else
      {
        if (GET_BACKGROUND_COLOR(attrs) & 1)
          attributes |= BACKGROUND_RED;
        if (GET_BACKGROUND_COLOR(attrs) & 2)
          attributes |= BACKGROUND_GREEN;
        if (GET_BACKGROUND_COLOR(attrs) & 4)
          attributes |= BACKGROUND_BLUE;
      }

    lineedit_output_drain (state); /* ignore error */

    if (!SetConsoleTextAttribute
           (state->ts->output,
            attributes))
      return err_code_from_GetLastError ();
  }
#endif

  return ___FIX(___NO_ERR);
}


___HIDDEN ___SCMOBJ lineedit_output_chars
   ___P((lineedit_state *state,
         tty_char *buf,
         ___stream_index len,
         tty_text_attrs attrs),
        (state,
         buf,
         len,
         attrs)
lineedit_state *state;
tty_char *buf;
___stream_index len;
tty_text_attrs attrs;)
{
  /*
   * Output "len" characters from the buffer "buf" and keep track of
   * cursor motion.
   */

  ___SCMOBJ e;

  if ((e = lineedit_output_set_text_attrs (state, attrs))
      != ___FIX(___NO_ERR))
    return e;

  e = extensible_string_insert
        (&state->output,
         state->output.length,
         len,
         buf);
  if (0) {int i; for (i=0;i<len;i++) printf ("%c", buf[i]); fflush(stdout);/******************/}

  lineedit_track_cursor_movement (state, buf, len);

  return e;
}


___HIDDEN ___SCMOBJ lineedit_output_char_repetition
   ___P((lineedit_state *state,
         tty_char c,
         int rep,
         tty_text_attrs attrs),
        (state,
         c,
         rep,
         attrs)
lineedit_state *state;
tty_char c;
int rep;
tty_text_attrs attrs;)
{
  /*
   * Output a character "rep" times and keep track of cursor motion.
   */

#define CHAR_BUFFER_SIZE (80*50)

  ___SCMOBJ e;
  tty_char char_buffer[CHAR_BUFFER_SIZE];
  int n;

  n = rep;
  if (n > CHAR_BUFFER_SIZE)
    n = CHAR_BUFFER_SIZE;

  while (n > 0)
    char_buffer[--n] = c;

  while (rep > 0)
    {
      n = rep;
      if (n > CHAR_BUFFER_SIZE)
        n = CHAR_BUFFER_SIZE;
      if ((e = lineedit_output_chars (state, char_buffer, n, attrs))
          != ___FIX(___NO_ERR))
        return e;
      rep -= n;
    }

  return ___FIX(___NO_ERR);
}


___HIDDEN tty_text_attrs lineedit_erase_attrs
   ___P((lineedit_state *state),
        (state)
lineedit_state *state;)
{
  tty_text_attrs output_attrs = state->output_attrs;
  int output_style = GET_STYLE(output_attrs);
  int current_style = GET_STYLE(state->current.attrs);

  return MAKE_TEXT_ATTRS((current_style & TEXT_STYLE_BOLD) |
                         (output_style & TEXT_STYLE_REVERSE),
                         GET_FOREGROUND_COLOR(output_attrs),
                         GET_BACKGROUND_COLOR(output_attrs));
}


___HIDDEN ___SCMOBJ lineedit_output_spaces
   ___P((lineedit_state *state,
         int rep),
        (state,
         rep)
lineedit_state *state;
int rep;)
{
  return lineedit_output_char_repetition
           (state,
            UNICODE_SPACE,
            rep,
            lineedit_erase_attrs (state));
}


___HIDDEN ___SCMOBJ lineedit_erase_to_end_of_screen
   ___P((lineedit_state *state,
         int pos,
         int chars_to_erase),
        (state,
         pos,
         chars_to_erase)
lineedit_state *state;
int pos;
int chars_to_erase;)
{
  /*
   * Erases from the current position up to the end of the screen.
   * "pos" is the start position and "chars_to_erase" is the number of
   * characters that are erased.
   */

/***************************strange interface**/

  ___SCMOBJ e;

#ifdef USE_CURSES

  if (chars_to_erase > 3 &&
      state->capability[LINEEDIT_CAP_ED] != NULL)
    {
      if ((e = lineedit_output_set_text_attrs
                 (state,
                  lineedit_erase_attrs (state)))
          == ___FIX(___NO_ERR))
        e = lineedit_output_cap (state, LINEEDIT_CAP_ED, -1, 1);
    }
  else

#endif

    e = lineedit_output_spaces (state, chars_to_erase);

  return e;
}


___HIDDEN ___SCMOBJ lineedit_output_chars_at
   ___P((lineedit_state *state,
         tty_char *str,
         int chars,
         int chars_to_erase,
         int pos),
        (state,
         str,
         chars,
         chars_to_erase,
         pos)
lineedit_state *state;
tty_char *str;
int chars;
int chars_to_erase;
int pos;)
{
  ___SCMOBJ e;/**************unused*/

  /******************not used!

  if ((e = lineedit_prepare_to_write_at (state, pos))
      != ___FIX(___NO_ERR))
    return e;

  if ((e = lineedit_output_chars (state, str, chars, state->input_attrs))
       != ___FIX(___NO_ERR))
    return e;
  */

  return lineedit_erase_to_end_of_screen (state, pos+chars, chars_to_erase);
}


___HIDDEN ___SCMOBJ lineedit_output_current_hist
   ___P((lineedit_state *state,
         int start,
         int len),
        (state,
         start,
         len)
lineedit_state *state;
int start;
int len;)
{
  ___SCMOBJ e;
  extensible_string *edited = &state->current.hist->edited;
  int to_end;
  int n;

  if (start < 0)
    {
      n = -start;
      if (n > len)
        n = len;
      if ((e = lineedit_output_spaces (state, n)) != ___FIX(___NO_ERR))
        return e;
      start += n;
      len -= n;
    }

  to_end = edited->length - start;

  if (to_end > 0)
    {
      if (to_end > len)
        n = len;
      else
        n = to_end;

      if ((e = lineedit_output_chars
                 (state,
                  &edited->buffer[start],
                  n,
                  state->input_attrs))
          != ___FIX(___NO_ERR))
        return e;

      len -= n;

      if (n == to_end &&
          state->current.delayed_wrap &&
          state->current.screen_cursor + 1 < state->screen_size &&
          len < 1)
        len = 1;
    }

  return lineedit_erase_to_end_of_screen (state, 999999, len);/*******/
}


___HIDDEN ___SCMOBJ lineedit_output_force_delayed_wrap
   ___P((lineedit_state *state),
        (state)
lineedit_state *state;)
{
  /*
   * Forces the cursor to wrap to the next line if the wrap was
   * delayed.
   */

  if (state->current.delayed_wrap)
    return lineedit_output_current_hist
             (state,
              state->current.screen_cursor+1+state->current.cursor_to_point,
              1);

  return ___FIX(___NO_ERR);
}


___HIDDEN ___SCMOBJ lineedit_move_screen_cursor
   ___P((lineedit_state *state,
         int pos),
        (state,
         pos)
lineedit_state *state;
int pos;)
{
  /*
   * This routine sends the appropriate commands to the terminal to
   * move the cursor to position "pos".
   */

  /*
   * The fields state->current.screen_cursor and
   * state->current.delayed_wrap are updated accordingly.
   */

  ___SCMOBJ e;
  int screen_cursor = state->current.screen_cursor;
  int screen_cols;
  int pos_col;
  int col_dist;
  int row_dist;
  int dist;

  if (pos == screen_cursor)
    return ___FIX(___NO_ERR);

  screen_cols = state->screen_cols;
  pos_col = pos % screen_cols;
  col_dist = pos_col - screen_cursor % screen_cols;
  row_dist = pos / screen_cols - screen_cursor / screen_cols;

#ifdef USE_CURSES

  for (;;)
    {
      /*
       * Check if the cursor can be moved backward by writing no more
       * than 4 backspace characters (if on the same row or column) or
       * 8 backspace characters (otherwise).  This is probably no more
       * than is required by cursor movement commands.
       */

      if (pos < screen_cursor &&
          (row_dist == 0 || state->has_auto_left_margin) &&
          screen_cursor - pos <= ((row_dist == 0 || col_dist == 0) ? 4 : 8))
        break;

      /*
       * Check if the cursor can be moved forward by writing no more
       * than 4 plain characters (if on the same row or column) or 8
       * plain characters (otherwise).  This is probably no more than
       * is required by cursor movement commands.
       */

      if (pos > screen_cursor &&
          screen_cursor + state->current.cursor_to_point >= 0 &&
          (row_dist == 0 || state->has_auto_right_margin) &&
          pos - (screen_cursor + state->current.delayed_wrap)
              + ((pos_col == 0 && state->has_eat_newline_glitch) ? 2 : 0) <=
              ((row_dist == 0 || col_dist == 0) ? 4 : 8))
        break;

      /*
       * If the cursor is on the target row, move the cursor to the
       * target column using cursor movement commands, if the terminal
       * supports this.
       */

      if (row_dist == 0)
        {
          if (col_dist == 0 ||
              ((col_dist > 0)
               ? (state->capability[LINEEDIT_CAP_CUF] == NULL)
               : (state->capability[LINEEDIT_CAP_CUB] == NULL)))
            break;

          if ((e = lineedit_output_cap
                     (state,
                      (col_dist > 0) ? LINEEDIT_CAP_CUF : LINEEDIT_CAP_CUB,
                      (col_dist > 0) ? col_dist : -col_dist,
                      1))
              != ___FIX(___NO_ERR))
            return e;

          screen_cursor += col_dist;
          state->current.screen_cursor = screen_cursor;
          break;
        }

      /*
       * Move the cursor to the target row using cursor movement
       * commands, if the terminal supports this.
       */

      if ((row_dist > 0)
          ? (state->capability[LINEEDIT_CAP_CUD] == NULL &&
             state->capability[LINEEDIT_CAP_CUD1] == NULL)
          : (state->capability[LINEEDIT_CAP_CUU] == NULL &&
             state->capability[LINEEDIT_CAP_CUU1] == NULL))
        break;

      if ((row_dist > 0)
          ? (state->capability[LINEEDIT_CAP_CUD] == NULL ||
             (state->capability[LINEEDIT_CAP_CUD1] != NULL && row_dist == 1))
          : (state->capability[LINEEDIT_CAP_CUU] == NULL ||
             (state->capability[LINEEDIT_CAP_CUU1] != NULL && row_dist == -1)))
        e = lineedit_output_cap
              (state,
               (row_dist > 0) ? LINEEDIT_CAP_CUD1 : LINEEDIT_CAP_CUU1,
               -1,
               (row_dist > 0) ? row_dist : -row_dist);
      else
        e = lineedit_output_cap
              (state,
               (row_dist > 0) ? LINEEDIT_CAP_CUD : LINEEDIT_CAP_CUU,
               (row_dist > 0) ? row_dist : -row_dist,
               1);

      if (e != ___FIX(___NO_ERR))
        return e;

      /* Cursor commands to change row were successful. */

      screen_cursor += row_dist * screen_cols;
      state->current.screen_cursor = screen_cursor;
      row_dist = 0;
    }

#endif

  /*
   * If the cursor is still not at the target position, move the
   * cursor by writing plain characters or backspace characters.  We
   * assume that the terminal has auto left/right margins if the
   * cursor needs to change row.
   */

  dist = pos - screen_cursor;

  if (dist != 0)
    {
      if (dist < 0)
        {
          if ((e = lineedit_output_char_repetition
                     (state,
                      UNICODE_BACKSPACE,
                      -dist,
                      state->current.attrs))
              != ___FIX(___NO_ERR))
            return e;
        }
      else
        {
          int start = screen_cursor + state->current.cursor_to_point +
                      state->current.delayed_wrap;
          int len = dist - state->current.delayed_wrap;
          ___BOOL need_backspace =
            (pos_col == 0 && state->has_eat_newline_glitch);

          if ((e = lineedit_output_current_hist
                     (state,
                      start,
                      len+need_backspace))
              != ___FIX(___NO_ERR))
            return e;

          if (need_backspace)
            {
              if ((e = lineedit_output_char_repetition
                         (state,
                          UNICODE_BACKSPACE,
                          1,
                          state->current.attrs))
                  != ___FIX(___NO_ERR))
                return e;
            }
        }

      screen_cursor += dist;
      state->current.screen_cursor = screen_cursor;
    }

  return ___FIX(___NO_ERR);
}


___HIDDEN ___SCMOBJ lineedit_left_margin_of_next_row
   ___P((lineedit_state *state),
        (state)
lineedit_state *state;)
{
  ___SCMOBJ e;

  if (!state->linefeed_moves_to_left_margin)
    if ((e = lineedit_output_char_repetition
               (state,
                UNICODE_RETURN,
                1,
                state->output_attrs))
        != ___FIX(___NO_ERR))
      return e;

  return lineedit_output_char_repetition
           (state,
            UNICODE_LINEFEED,
            1,
            state->output_attrs);
}


___HIDDEN ___SCMOBJ lineedit_prepare_to_write_at
   ___P((lineedit_state *state,
         int pos),
        (state,
         pos)
lineedit_state *state;
int pos;)
{
  /*
   * This routine sends commands to the terminal such that the next
   * character sent to the terminal will show up at position "pos".
   * The position can be equal to state->screen_size, in which case
   * either the screen is scrolled right away or the next character
   * sent to the terminal will cause the screen to scroll.
   */

  ___SCMOBJ e;

  if (state->current.screen_cursor + state->current.delayed_wrap == pos)
    return ___FIX(___NO_ERR);

  if (pos >= state->screen_size)
    {
      if ((e = lineedit_move_screen_cursor (state, state->screen_size-1))
          != ___FIX(___NO_ERR))
        return e;
    }
  else
    {
      if ((e = lineedit_move_screen_cursor (state, pos))
          != ___FIX(___NO_ERR))
        return e;

      if (!state->current.delayed_wrap)
        return ___FIX(___NO_ERR);

      /*
       * Output a backspace and a character.  Note that this assumes
       * the screen is at least 2 columns wide.
       */

      if ((e = lineedit_output_char_repetition
                 (state,
                  UNICODE_BACKSPACE,
                  1,
                  state->current.attrs))
          != ___FIX(___NO_ERR))
        return e;
    }

  return lineedit_output_current_hist
           (state,
            state->current.screen_cursor + state->current.cursor_to_point,
            1);
}


___HIDDEN ___SCMOBJ lineedit_newline
   ___P((lineedit_state *state),
        (state)
lineedit_state *state;)
{
  ___SCMOBJ e;
  extensible_string *edited = &state->current.hist->edited;
  int pos = edited->length - state->current.cursor_to_point;

  if (pos < 0)
    pos = 0;
  else if (pos >= state->screen_size)
    pos = state->screen_size-1;

  if ((e = lineedit_prepare_to_write_at (state, pos))
      != ___FIX(___NO_ERR))
    return e;

  return lineedit_left_margin_of_next_row (state);
}


___HIDDEN ___SCMOBJ lineedit_erase_screen
   ___P((lineedit_state *state),
        (state)
lineedit_state *state;)
{
  ___SCMOBJ e;

#ifdef USE_CURSES

  if (state->capability[LINEEDIT_CAP_CLEAR] != NULL)
    {
      if ((e = lineedit_output_set_text_attrs
                 (state,
                  lineedit_erase_attrs (state)))
          == ___FIX(___NO_ERR))
        e = lineedit_output_cap (state, LINEEDIT_CAP_CLEAR, -1, 1);
    }
  else

#endif

    e = lineedit_newline (state);

  if (e == ___FIX(___NO_ERR))
    {
      state->current.screen_cursor = 0; /****************/
    }

  return e;
}


___HIDDEN ___SCMOBJ lineedit_copy_to_clipboard
   ___P((lineedit_state *state,
         tty_char *buf,
         int len),
        (state,
         buf,
         len)
lineedit_state *state;
tty_char *buf;
int len;)
{
  /*
   * Copies "len" characters from the buffer "buf" to the clipboard.
   */

#ifdef LINEEDIT_WITH_LOCAL_CLIPBOARD

  ___SCMOBJ e;
  extensible_string str;

  if ((e = extensible_string_copy (buf, len, &str, 0)) != ___FIX(___NO_ERR))
    return e;

  extensible_string_cleanup (&state->clipboard);

  state->clipboard = str;

#else

  /***********************/

#endif

  return ___FIX(___NO_ERR);
}


___HIDDEN ___SCMOBJ lineedit_paste_from_clipboard
   ___P((lineedit_state *state),
        (state)
lineedit_state *state;)
{
  /*
   * Paste the content of the clipboard to the input stream (the next
   * characters read from the terminal will in fact come from the
   * pasted text).
   */

#ifdef LINEEDIT_WITH_LOCAL_CLIPBOARD

  int len = state->clipboard.length;
  tty_char *str;

  str = ___CAST(tty_char*,
                ___alloc_mem ((len+1) * sizeof (tty_char)));

  if (str == NULL)
    return ___FIX(___HEAP_OVERFLOW_ERR);

  str[len] = 0;
  while (len-- > 0)
    str[len] = state->clipboard.buffer[len];

  if (state->paste_text != NULL)
    ___free_mem (state->paste_text);

  state->paste_index = 0;
  state->paste_text = str;

#else

  /*************************/

#endif

  return ___FIX(___NO_ERR);
}


___HIDDEN ___BOOL lineedit_read_ready
   ___P((lineedit_state *state),
        (state)
lineedit_state *state;)
{
  tty_state *ts = state->ts;
  return (ts->rlo < ts->rhi || state->paste_text != NULL);
}


___HIDDEN ___SCMOBJ lineedit_get_event
   ___P((lineedit_state *state,
         lineedit_event *ev),
        (state,
         ev)
lineedit_state *state;
lineedit_event *ev;)
{
  /*
   * This routine obtains the next line editing event from the
   * terminal.  An event can be a key press or some editing operation
   * (like "delete next word").
   */

  lineedit_input_decoder_state s = state->input_decoder_state;
  tty_char c;
  ___BOOL first_char = (s == 0);
  ___stream_index len_done;
  ___SCMOBJ e;

  next_char:

  if ((e = tty_read_noncanonical (state->ts, &c, 1, &len_done))
       == ___FIX(___NO_ERR) &&
      len_done == 1)
    {
      while (s < state->input_decoder.length)
        {
          if (state->input_decoder.buffer[s].trigger == c)
            {
              int a = state->input_decoder.buffer[s].action;
              if (a < LINEEDIT_INPUT_DECODER_MAX_LENGTH)
                {
                  s = a;
                  first_char = 0;
                  goto next_char;
                }
              state->input_decoder_state = 0;
              ev->event_kind = LINEEDIT_INPUT_DECODER_STATE_MAX-a;
              return ___FIX(___NO_ERR);
            }
          else
            s = state->input_decoder.buffer[s].next;
        }
      if (first_char)
        {
          ev->event_kind = LINEEDIT_EV_KEY;
          ev->event_char = c;
          return ___FIX(___NO_ERR);
        }
      s = 0; /* ignore the sequence including last character read */
    }

  state->input_decoder_state = s;
  ev->event_kind = LINEEDIT_EV_NONE;

  if (e != ___FIX(___NO_ERR) && e != ___ERR_CODE_EAGAIN)
    return e;

  return ___FIX(___NO_ERR);
}


___HIDDEN ___SCMOBJ lineedit_set_terminal_type
   ___P((lineedit_state *state,
         char *terminal_type,
         ___BOOL emacs_bindings),
        (state,
         terminal_type,
         emacs_bindings)
lineedit_state *state;
char *terminal_type;
___BOOL emacs_bindings;)
{
  ___SCMOBJ e;
  int screen_rows = 24;
  int screen_cols = 80;
  ___BOOL has_auto_left_margin = 1;
  ___BOOL has_auto_right_margin = 1;
  ___BOOL has_eat_newline_glitch = 0;

#ifdef USE_CURSES

  int i;

  if (terminal_type != NULL)
    {
#ifdef USE_TERMCAP

      char termcap_buffer[4080]; /* the value 4080 is used by GNU's readline */

      if (tgetent (termcap_buffer, terminal_type) != 1)
        return ___FIX(___UNKNOWN_ERR);

      screen_rows = tgetnum ("li");
      screen_cols = tgetnum ("co");
      has_auto_left_margin = (tgetflag ("bw") != 0);
      has_auto_right_margin = (tgetflag ("am") != 0);
      has_eat_newline_glitch = (tgetflag ("xn") != 0);

#else
#ifdef USE_TERMINFO

      int errret;

      if (setupterm (terminal_type, 1, &errret) == ERR)
        return ___FIX(___UNKNOWN_ERR);

      screen_rows = tigetnum ("lines");
      screen_cols = tigetnum ("cols");
      has_auto_left_margin = (tigetflag ("bw") != 0);
      has_auto_right_margin = (tigetflag ("am") != 0);
      has_eat_newline_glitch = (tigetflag ("xenl") != 0);

      printf ("%d %d %d %d %d\n",
              screen_rows,
              screen_cols,
              has_auto_left_margin,
              has_auto_right_margin,
              has_eat_newline_glitch
              );

#else

      screen_rows = 24;
      screen_cols = 80;
      has_auto_left_margin = 0;
      has_auto_right_margin = 1;
      has_eat_newline_glitch = 1;

#endif
#endif
    }

  /*
   * Setup the capability table.
   */

  for (i=0; i<LINEEDIT_CAP_LAST+1; i++)
    {
      char *seq = state->capability[i];///////////uninitialized????

      if (seq != NULL)
        {
          ___free_mem (seq);
          state->capability[i] = NULL;
        }
    }

  if (terminal_type != NULL)
    for (i=0; i<LINEEDIT_CAP_LAST+1; i++)
      {
        char *seq;

#ifdef USE_TERMCAP
        seq = tgetstr (lineedit_defcap_table[i].cap, NULL);
#else
#ifdef USE_TERMINFO
        seq = tigetstr (lineedit_defcap_table[i].cap);
#else
        seq = lineedit_defcap_table[i].cap;
#endif
#endif

        if (seq != (char*)-1 && seq != NULL)
          {
            char *p;

            /*
             * Reject any sequence that includes a linefeed if the
             * terminal driver automatically adds a carriage return.
             */

            if (state->linefeed_moves_to_left_margin)
              {
                p = seq;
                while (*p != '\0')
                  if (*p != UNICODE_LINEFEED)
                    p++;
                  else
                    break;
                if (*p != '\0')
                  seq = NULL;
              }

            /*
             * Keep a copy of the sequence.
             */

            if (seq != NULL)
              {
                int len = 0;
                while (seq[len] != '\0')
                  len++;
                p = ___CAST(char*,___alloc_mem (len));
                if (p != NULL)
                  {
                    p[len] = '\0';
                    while (len-- > 0)
                      p[len] = seq[len];
                  }
                seq = p;
              }
          }
        else
          seq = NULL;

        state->capability[i] = seq;
      }

#endif

  /*
   * Initialize input decoder so that it recognizes special keys.  Map
   * rubout key to backspace if terminal driver is setup to erase when
   * a rubout is typed.
   */

  state->input_decoder_state = 0;

  if ((e = lineedit_input_decoder_init
             (&state->input_decoder,
#ifdef USE_POSIX
              (state->ts->initial_termios.c_cc[VERASE] == UNICODE_RUBOUT),
#else
              0,
#endif
              emacs_bindings,
              terminal_type != NULL))
      != ___FIX(___NO_ERR))
    return e;

  if (screen_rows <= 0)
    screen_rows = 24;
  if (screen_cols <= 0)
    screen_cols = 80;

  state->screen_rows = screen_rows;
  state->screen_cols = screen_cols;
  state->screen_size = screen_rows * screen_cols;
  state->has_auto_left_margin = has_auto_left_margin;
  state->has_auto_right_margin = has_auto_right_margin;
  state->has_eat_newline_glitch = has_eat_newline_glitch;

#if 0
  printf ("%s screen_rows=%d screen_cols=%d alm=%d arm=%d eng=%d\n", terminal_type,screen_rows, screen_cols,
          has_auto_left_margin,
          has_auto_right_margin,
          has_eat_newline_glitch);fflush (stdout);
#endif

  return ___FIX(___NO_ERR);
}


___HIDDEN ___SCMOBJ lineedit_set_paren_balance_duration
   ___P((lineedit_state *state,
         int nsecs),
        (state,
         nsecs)
lineedit_state *state;
int nsecs;)
{
  state->paren_balance_duration_nsecs = nsecs;
  return ___FIX(___NO_ERR);
}


___HIDDEN ___SCMOBJ lineedit_setup
   ___P((lineedit_state **les,
         tty_state *ts),
        (les,
         ts)
lineedit_state **les;
tty_state *ts;)
{
  ___SCMOBJ e;
  lineedit_history *h;
  lineedit_state *state;

  *les = NULL;

  /* allocate a new lineedit_state */

  state = ___CAST(lineedit_state*,
                  ___alloc_mem (sizeof (lineedit_state)));

  if (state == NULL)
    return ___FIX(___HEAP_OVERFLOW_ERR);

  *les = state;

  state->ts = ts;

  state->edit_mode = LINEEDIT_MODE_SCHEME;

  state->paste_cancel = 0;
  state->paste_index = 0;
  state->paste_text = NULL;

  state->input_rlo = 0;
  state->input.buffer = NULL;

  state->output_wlo = 0;

  state->max_history_length = 1000000; /* set a very high limit */
  state->history_length = -1;
  state->hist_last = NULL;

  state->current.edit_point = 0;
  state->current.mark_point = 0;
  state->current.screen_cursor = 0;
  state->current.cursor_to_point = 0;
  state->current.delayed_wrap = 0;
  state->current.paren_balance_in_progress = 0;
  state->current.attrs = INITIAL_TEXT_ATTRS;

  state->paren_balance_duration_nsecs = DEFAULT_PAREN_BALANCE_DURATION_NSECS;

  state->input_attrs = DEFAULT_INPUT_TEXT_ATTRS;
  state->output_attrs = DEFAULT_OUTPUT_TEXT_ATTRS;

  state->screen_rows = 24;
  state->screen_cols = 80;
  state->screen_size = 24*80;

  state->has_auto_left_margin = 1;
  state->has_auto_right_margin = 1;
  state->has_eat_newline_glitch = 0;
  state->linefeed_moves_to_left_margin = 1;

  state->tab_stop = TTY_TAB_STOP;

  if ((e = extensible_string_setup (&state->output, 0)) == ___FIX(___NO_ERR))
    {
      if ((e = lineedit_history_setup (state, &h)) == ___FIX(___NO_ERR))
        {
          lineedit_history_add (state, h);
          state->current.hist = h;
          if ((e = lineedit_history_begin_edit (state, h))
              == ___FIX(___NO_ERR))
            {
              if ((e = lineedit_input_decoder_setup (&state->input_decoder))
                  == ___FIX(___NO_ERR))
                {
                  if ((e = lineedit_set_terminal_type
                             (state,
                              NULL,
                              DEFAULT_EMACS_BINDINGS))
                      == ___FIX(___NO_ERR))
                    {
#ifdef LINEEDIT_WITH_LOCAL_CLIPBOARD
                      if ((e = extensible_string_setup (&state->clipboard, 0))
                          == ___FIX(___NO_ERR))
                        {
#endif
#ifdef USE_CURSES
                          int i;
                          for (i=0; i<LINEEDIT_CAP_LAST+1; i++)
                            state->capability[i] = NULL;
#endif
#ifdef USE_WIN32
                          /***********************/
#endif
#ifdef USE_MACOS
                          /***********************/
#endif
#ifdef LINEEDIT_WITH_LOCAL_CLIPBOARD
                          if (e != ___FIX(___NO_ERR))
                            extensible_string_cleanup (&state->clipboard);
                        }
#endif
                    }
                  if (e != ___FIX(___NO_ERR))
                    lineedit_input_decoder_cleanup (&state->input_decoder);
                }
            }
          if (e != ___FIX(___NO_ERR))
            lineedit_history_cleanup (state, h); /* ignore error */
        }
      if (e != ___FIX(___NO_ERR))
        extensible_string_cleanup (&state->output);
    }

  if (e != ___FIX(___NO_ERR))
    ___free_mem (state);

  return e;
}


___HIDDEN void lineedit_cleanup
   ___P((lineedit_state *state),
        (state)
lineedit_state *state;)
{
  lineedit_output_set_text_attrs
    (state,
     INITIAL_TEXT_ATTRS); /* ignore error */

  lineedit_output_drain (state); /* ignore error */

  if (state->paste_text != NULL)
    ___free_mem (state->paste_text); /* discard paste text */

  if (state->input.buffer != NULL)
    extensible_string_cleanup (&state->input);

  extensible_string_cleanup (&state->output);

  state->max_history_length = -1;
  lineedit_history_trim (state);

#ifdef LINEEDIT_WITH_LOCAL_CLIPBOARD
  extensible_string_cleanup (&state->clipboard);
#endif

#ifdef USE_CURSES
  {
    int i;

    lineedit_input_decoder_cleanup (&state->input_decoder);

    for (i=0; i<LINEEDIT_CAP_LAST+1; i++)
      {
        char *seq = state->capability[i];
        if (seq != NULL)
          ___free_mem (seq);
      }
  }
#endif

#ifdef USE_WIN32
  /*********************/
#endif

#ifdef USE_MACOS
  /*********************/
#endif

  ___free_mem (state);
}


___HIDDEN ___SCMOBJ lineedit_move_edit_point
   ___P((lineedit_state *state,
         int pos),
        (state,
         pos)
lineedit_state *state;
int pos;)
{
  /*
   * Moves the edit point of the current edited line.  The screen may
   * get scrolled in order to keep the edit point visible.
   */

  ___SCMOBJ e;
  extensible_string *edited = &state->current.hist->edited;
  int screen_cursor;
  int screen_cols;
  int screen_rows;
  int screen_size;
  int cursor_to_point;
  int cursor_row;
  int first_row;
  int last_row;
  int nb_rows;
  int all_fits;
  int n;
  int start;

  if (pos < 0 ||
      pos > edited->length)
    return ___FIX(___INVALID_OP_ERR);

  screen_cols = state->screen_cols;
  screen_rows = state->screen_rows;
  screen_size = state->screen_size;
  cursor_to_point = state->current.cursor_to_point;
  screen_cursor = pos - cursor_to_point; /* new cursor position in screen */

  /*
   * Check if the text from the cursor to the position just after end
   * of the edited line can all fit on the screen (while maintaining
   * column alignment).
   */

  cursor_row = screen_cursor;
  if (cursor_row < 0)
    cursor_row -= screen_cols - 1;
  cursor_row /= screen_cols;

  first_row = -cursor_to_point;
  if (first_row < 0)
    first_row -= screen_cols - 1;
  first_row /= screen_cols;

  last_row = edited->length - cursor_to_point;
  if (last_row < 0)
    last_row -= screen_cols - 1;
  last_row /= screen_cols;

  all_fits = (last_row - cursor_row < screen_rows);

  /*
   * Determine how many rows to scroll and in which direction (nb_rows
   * is positive when scrolling forward and negative when scrolling
   * backward).
   */

  if (all_fits)
    {
      if (cursor_row >= 0 &&
          last_row < screen_rows)
        /* Don't scroll if text from cursor to end is visible right now. */
        nb_rows = 0;
      else
        /* Scroll so that last row of edited line is at bottom of screen. */
        nb_rows = last_row - (screen_rows - 1);
    }
  else
    {
      if (cursor_row >= 0 &&
          cursor_row < screen_rows)
        /* Don't scroll if cursor is visible right now. */
        nb_rows = 0;
      else
        {
          /*
           * Scroll cursor row to bottom of screen unless this makes
           * the beginning of the edited line visible, in which case
           * align the beginning of the edited line to the top of the
           * screen.
           */

          if (cursor_row < first_row + (screen_rows - 1))
            nb_rows = first_row;
          else
            nb_rows = cursor_row - (screen_rows - 1);
        }
    }

  /*
   * Perform scrolling of screen.
   */

  if (nb_rows != 0)
    {
      screen_cursor -= screen_cols * nb_rows;

      if (nb_rows < 0)
        {
          /* scroll screen backward by whole rows */

          if (pos < screen_cursor)
            screen_cursor = pos; /* align start of buffer with screen home */

          start = pos - screen_cursor;

          n = screen_size; /* write to the whole screen */

          if ((e = lineedit_prepare_to_write_at (state, 0))
              != ___FIX(___NO_ERR))
            return e;
        }
      else
        {
          /* scroll screen forward by whole rows */

          start = screen_size + cursor_to_point;

          n = screen_cols * nb_rows; /* write characters that scroll in */

          if ((e = lineedit_prepare_to_write_at (state, screen_size))
              != ___FIX(___NO_ERR))
            return e;
        }

      if (!state->has_eat_newline_glitch)/*****************/
        n--; /* to avoid wraparound at end of screen */

      if (n > edited->length - start)
        n = edited->length - start;

      if ((e = lineedit_output_current_hist (state, start, n))
          != ___FIX(___NO_ERR))
        return e;

      if (all_fits)
        if ((e = lineedit_output_force_delayed_wrap (state))
            != ___FIX(___NO_ERR))
          return e;
    }

  if ((e = lineedit_move_screen_cursor (state, screen_cursor))
      != ___FIX(___NO_ERR))
    return e;

  state->current.edit_point = pos;
  state->current.cursor_to_point = pos - screen_cursor;

  return ___FIX(___NO_ERR);
}


___HIDDEN ___SCMOBJ lineedit_update_region
   ___P((lineedit_state *state,
         int start,
         int end,
         ___BOOL allow_scrolling),
        (state,
         start,
         end,
         allow_scrolling)
lineedit_state *state;
int start;
int end;
___BOOL allow_scrolling;)
{
  ___SCMOBJ e;
  int screen_size = state->screen_size;
  int cursor_to_point = state->current.cursor_to_point;
  int pos = start - cursor_to_point;

  if (!state->has_eat_newline_glitch)/*****************/
    screen_size--; /* to avoid wraparound at end of screen */

  if (pos >= screen_size ||
      end - cursor_to_point <= 0)
    return ___FIX(___NO_ERR);

  if (pos < 0)
    pos = 0;

  start = pos + cursor_to_point;

  if ((e = lineedit_prepare_to_write_at (state, pos))
      != ___FIX(___NO_ERR))
    return e;

  return lineedit_output_current_hist (state, start, end - start);
}


___HIDDEN ___SCMOBJ lineedit_refresh
   ___P((lineedit_state *state),
        (state)
lineedit_state *state;)
{
  ___SCMOBJ e;
  extensible_string *edited = &state->current.hist->edited;

  if ((e = lineedit_erase_screen (state))
      != ___FIX(___NO_ERR))
    return e;

  state->current.cursor_to_point = -state->current.screen_cursor;

  if ((e = lineedit_update_region (state, 0, edited->length, 1))
      != ___FIX(___NO_ERR))
    return e;

  return lineedit_move_edit_point (state, state->current.edit_point);
}


___HIDDEN ___BOOL lineedit_word_boundary
   ___P((lineedit_state *state,
         int back,
         int pos,
         int *final_pos),
        (state,
         back,
         pos,
         final_pos)
lineedit_state *state;
int back;
int pos;
int *final_pos;)
{
  extensible_string *edited = &state->current.hist->edited;
  tty_char *buf = edited->buffer;
  int len = edited->length;
  int depth = 0;
  tty_char in_string = UNICODE_NUL; /* assume not inside a string */
  tty_char c;

#define OPEN_PAREN(x) \
((x) == UNICODE_LPAREN || (x) == UNICODE_LBRACKET || (x) == UNICODE_LBRACE)

#define CLOSE_PAREN(x) \
((x) == UNICODE_RPAREN || (x) == UNICODE_RBRACKET || (x) == UNICODE_RBRACE)

#define STRING_DELIM(x) \
((x) == UNICODE_DOUBLEQUOTE || (x) == UNICODE_VBAR)

#define INT_VECTOR(x) \
((x) == UNICODE_LOWER_S || (x) == UNICODE_UPPER_S || \
 (x) == UNICODE_LOWER_U || (x) == UNICODE_UPPER_U)

#define FLOAT_VECTOR(x) \
((x) == UNICODE_LOWER_F || (x) == UNICODE_UPPER_F)

  if (back == 0)
    {
      /* skip whitespace */

      while (pos < len)
        {
          c = buf[pos];
          if (c <= UNICODE_SPACE)
            pos++;
          else
            break;
        }

      /* skip standard read-macros */

      while (pos < len)
        {
          c = buf[pos];
          if (c == UNICODE_QUOTE)
            pos++;
          else if (c == UNICODE_COMMA)
            {
              pos++;
              if (pos < len && buf[pos] == UNICODE_AT)
                pos++;
            }
          else
            break;
        }

      if (pos == len)
        {
          depth = -1; /* force error */
          goto done_forward;
        }

      /* handle head specially */

      c = buf[pos++];

      if (pos < len && c == UNICODE_SHARP)
        {
          /* handle #\xxx, #(...), #f32(...), etc */

          c = buf[pos++];
          if (c == UNICODE_BACKSLASH)
            {
              if (pos < len)
                pos++;
            }
          else if (OPEN_PAREN(c))
            depth++;
          else
            {
              ___BOOL f = FLOAT_VECTOR(c);
              if (pos < len && (f || INT_VECTOR(c)))
                {
                  c = buf[pos++];
                  if ((!f && pos < len && c == UNICODE_8) ||
                      (pos+1 < len &&
                       ((!f && c == UNICODE_1 && buf[pos] == UNICODE_6) ||
                        (c == UNICODE_3 && buf[pos] == UNICODE_2) ||
                        (c == UNICODE_6 && buf[pos] == UNICODE_4)) &&
                       (pos++, 1)))
                    {
                      c = buf[pos++];
                      if (OPEN_PAREN(c))
                        depth++;
                      else
                        pos--;
                    }
                }
            }
        }
      else if (OPEN_PAREN(c))
        depth++;
      else if (CLOSE_PAREN(c))
        {
          depth = -1; /* force error */
          goto done_forward;
        }
      else if (STRING_DELIM(c))
        in_string = c;

      while (pos < len)
        if (in_string == UNICODE_NUL)
          {
            c = buf[pos];

            if (c == UNICODE_SHARP)
              {
                if (pos+2 < len && buf[pos+1] == UNICODE_BACKSLASH)
                  pos += 2;
              }
            else if (c <= UNICODE_SPACE ||
                     c == UNICODE_SEMICOLON ||
                     c == UNICODE_QUOTE ||
                     c == UNICODE_COMMA)
              {
                if (depth == 0)
                  break;
              }
            else if (OPEN_PAREN(c))
              {
                if (depth == 0)
                  break;
                depth++;
              }
            else if (CLOSE_PAREN(c))
              {
                if (depth == 0)
                  break;
                depth--;
                if (depth == 0)
                  {
                    pos++;
                    break;
                  }
              }
            else if (STRING_DELIM(c))
              {
                if (depth == 0)
                  break;
                in_string = c;
              }

            pos++;
          }
        else
          {
            c = buf[pos++];

            if (c == UNICODE_BACKSLASH) /* character is escaped */
              {
                if (pos < len)
                  pos++;
              }
            else if (c == in_string) /* end of string reached */
              {
                if (depth == 0)
                  break;
                in_string = UNICODE_NUL;
              }
          }

      done_forward:;
    }
  else
    {
      /* skip whitespace */

      while (pos > 0)
        {
          c = buf[pos-1];
          if (c <= UNICODE_SPACE)
            pos--;
          else
            break;
        }

      if (pos == 0)
        {
          depth = -1; /* force error */
          goto done_backward;
        }

      /* handle tail specially */

      c = buf[--pos];

      if (pos >= 2 &&
          buf[pos-2] == UNICODE_SHARP &&
          buf[pos-1] == UNICODE_BACKSLASH)
        pos -= 2;
      else if (OPEN_PAREN(c))
        {
          depth = -1; /* force error */
          goto done_backward;
        }
      else if (CLOSE_PAREN(c))
        depth++;
      else if (STRING_DELIM(c))
        in_string = c;
      else if (c == UNICODE_QUOTE ||
               c == UNICODE_COMMA ||
               (pos >= 1 && c == UNICODE_AT && buf[pos-1] == UNICODE_COMMA))
        {
          pos++;
          goto skip_read_macros;
        }

      while (pos > 0)
        if (in_string == UNICODE_NUL)
          {
            c = buf[pos-1];

            if (pos >= 3 &&
                buf[pos-3] == UNICODE_SHARP &&
                buf[pos-2] == UNICODE_BACKSLASH)
              pos -= 2;
            else if (c <= UNICODE_SPACE ||
                     c == UNICODE_SEMICOLON ||
                     c == UNICODE_QUOTE ||
                     c == UNICODE_COMMA)
              {
                if (depth == 0)
                  break;
              }
            else if (pos >= 2 &&
                     c == UNICODE_AT &&
                     buf[pos-2] == UNICODE_COMMA)
              {
                if (depth == 0)
                  break;
                pos--;
              }
            else if (OPEN_PAREN(c))
              {
                if (depth == 0)
                  break;
                depth--;
                if (depth == 0)
                  {
                    pos--;
                    break;
                  }
              }
            else if (CLOSE_PAREN(c))
              {
                if (depth == 0)
                  break;
                depth++;
              }
            else if (STRING_DELIM(c))
              {
                if (depth == 0)
                  break;
                in_string = c;
              }

            pos--;
          }
        else
          {
            int i;

            i = --pos;

            while (i > 0 &&
                   buf[i-1] == UNICODE_BACKSLASH) /* find nb of backslashes */
              i--;

            if ((pos-i) & 1) /* character is escaped */
              pos = i;
            else if (buf[pos] == in_string) /* beginning of string reached */
              {
                if (depth == 0)
                  break;
                in_string = UNICODE_NUL;
              }
          }

      done_backward:;

      if (back == 2)
        {
          /* handle #(...), #f32(...), etc */

          if (pos < len)
            {
              c = buf[pos];
              if (OPEN_PAREN(c))
                {
                  if (pos >= 1 && buf[pos-1] == UNICODE_SHARP)
                    pos--;
                  else if (pos >= 3 && buf[pos-3] == UNICODE_SHARP)
                    {
                      c = buf[pos-2];
                      if (INT_VECTOR(c) && buf[pos-1] == UNICODE_8)
                        pos -= 3;
                    }
                  else if (pos >= 4 && buf[pos-4] == UNICODE_SHARP)
                    {
                      ___BOOL f;
                      c = buf[pos-3];
                      f = FLOAT_VECTOR(c);
                      if (f || INT_VECTOR(c))
                        {
                          c = buf[pos-2];
                          if ((!f &&
                               c == UNICODE_1 &&
                               buf[pos-1] == UNICODE_6) ||
                              (c == UNICODE_3 && buf[pos-1] == UNICODE_2) ||
                              (c == UNICODE_6 && buf[pos-1] == UNICODE_4))
                            pos -= 4;
                        }
                    }
                }
            }

          /* skip standard read-macros */

          skip_read_macros:

          while (pos > 0)
            {
              c = buf[pos-1];
              if (c == UNICODE_QUOTE || c == UNICODE_COMMA)
                pos--;
              else if (pos >= 2 &&
                       c == UNICODE_AT &&
                       buf[pos-2] == UNICODE_COMMA)
                pos -= 2;
              else
                break;
            }
        }
    }

  *final_pos = pos;

  return (depth == 0 && in_string == UNICODE_NUL);
}


___HIDDEN ___SCMOBJ lineedit_insert_chars
   ___P((lineedit_state *state,
         tty_char *buf,
         int len),
        (state,
         buf,
         len)
lineedit_state *state;
tty_char *buf;
int len;)
{
  ___SCMOBJ e;
  extensible_string *edited = &state->current.hist->edited;
  int edit_point = state->current.edit_point;
  int open_paren_point;
  tty_char c;

  if (len <= 0)
    return ___FIX(___NO_ERR);

  if ((e = extensible_string_insert (edited, edit_point, len, buf))
      != ___FIX(___NO_ERR))
    return e;

  if (state->current.mark_point > edit_point)
    state->current.mark_point += len;

  if ((e = lineedit_update_region (state, edit_point, edited->length, 1))
      != ___FIX(___NO_ERR))
    return e;

  edit_point += len;

  if (state->paren_balance_duration_nsecs > 0 &&
      state->ts->rlo >= state->ts->rhi &&
      (c = buf[len-1], CLOSE_PAREN(c)) &&
      lineedit_word_boundary (state, 1, edit_point, &open_paren_point) &&
      (c = edited->buffer[open_paren_point], OPEN_PAREN(c)))
    {
      if ((e = lineedit_move_edit_point (state, open_paren_point))
          != ___FIX(___NO_ERR))
        return e;

      if ((e = lineedit_output_drain (state)) != ___FIX(___NO_ERR))
        return e;

      state->current.edit_point = edit_point;
      state->current.paren_balance_trigger = 1;

      return ___FIX(___NO_ERR);
    }

  return lineedit_move_edit_point (state, edit_point);
}


___HIDDEN ___SCMOBJ lineedit_delete_chars
   ___P((lineedit_state *state,
         int pos,
         ___BOOL copy_to_clipboard),
        (state,
         pos,
         copy_to_clipboard)
lineedit_state *state;
int pos;
___BOOL copy_to_clipboard;)
{
  ___SCMOBJ e;
  extensible_string *edited = &state->current.hist->edited;
  int start;
  int end;
  int n;

  if (pos < 0 ||
      pos > edited->length)
    return ___FIX(___INVALID_OP_ERR);

  if (pos < state->current.edit_point)
    {
      start = pos;
      end = state->current.edit_point;
    }
  else
    {
      start = state->current.edit_point;
      end = pos;
    }

  n = end - start;

  if (n == 0)
    return ___FIX(___NO_ERR);

  if (copy_to_clipboard &&
      ((e = lineedit_copy_to_clipboard (state, &edited->buffer[start], n))
       != ___FIX(___NO_ERR)))
    return e;

  extensible_string_delete (edited, start, n);

  if (state->current.mark_point >= start)
    {
      if (state->current.mark_point >= end)
        state->current.mark_point -= n;
      else
        state->current.mark_point = start;
    }

  if ((e = lineedit_update_region (state, start, edited->length + n, 1))
      != ___FIX(___NO_ERR))
    return e;

  return lineedit_move_edit_point (state, start);
}


___HIDDEN void lineedit_rotate_left_chars
   ___P((lineedit_state *state,
         int start,
         int end,
         int n),
        (state,
         start,
         end,
         n)
lineedit_state *state;
int start;
int end;
int n;)
{
  extensible_string *edited = &state->current.hist->edited;
  tty_char *p = &edited->buffer[start];
  tty_char temp;
  int len = end-start;
  int left = len;
  int i;
  int j;
  int k;

  i = 0;
  while (left > 0)
    {
      temp = p[i];
      j = i;
      k = (j+n) % len;
      left--;
      while (k != i)
        {
          p[j] = p[k];
          j = k;
          k = (j+n) % len;
          left--;
        }
      p[j] = temp;
      i++;
    }
}


___HIDDEN ___SCMOBJ lineedit_transpose_chars
   ___P((lineedit_state *state,
         int start1,
         int end1,
         int start2,
         int end2),
        (state,
         start1,
         end1,
         start2,
         end2)
lineedit_state *state;
int start1;
int end1;
int start2;
int end2;)
{
  ___SCMOBJ e;

  if (start1 > end1 ||
      end1 > start2 ||
      start2 > end2)
    return ___FIX(___INVALID_OP_ERR);

  lineedit_rotate_left_chars
    (state,
     start1,
     end2,
     end1-start1);

  lineedit_rotate_left_chars
    (state,
     start1,
     end2-(end1-start1),
     start2-end1);

  if ((e = lineedit_update_region (state, start1, end2, 1))
      != ___FIX(___NO_ERR))
    return e;

  return lineedit_move_edit_point (state, end2);
}


___HIDDEN ___SCMOBJ lineedit_move_history
   ___P((lineedit_state *state,
         lineedit_history *h),
        (state,
         h)
lineedit_state *state;
lineedit_history *h;)
{
  ___SCMOBJ e;
  int old_length;
  int new_length;

  if ((e = lineedit_history_begin_edit (state, h))
      != ___FIX(___NO_ERR))
    return e;

  if ((e = lineedit_move_edit_point (state, 0))
      != ___FIX(___NO_ERR))
    return e;

  old_length = state->current.hist->edited.length;
  new_length = h->edited.length;

  state->current.mark_point = 0;
  state->current.hist = h;

  if ((e = lineedit_update_region
             (state,
              0,
              (new_length > old_length) ? new_length : old_length,
              1))
      != ___FIX(___NO_ERR))
    return e;

  return lineedit_move_edit_point (state, new_length);
}


___HIDDEN ___SCMOBJ lineedit_move_history_relative
   ___P((lineedit_state *state,
         ___BOOL back),
        (state,
         back)
lineedit_state *state;
___BOOL back;)
{
  lineedit_history *h;
  lineedit_history *limit;

  if (back)
    {
      h = state->current.hist->prev;
      limit = state->hist_last;
    }
  else
    {
      h = state->current.hist->next;
      limit = state->hist_last->next;
    }

  if (h == limit)
    return ___FIX(___INVALID_OP_ERR);

  return lineedit_move_history (state, h);
}


___HIDDEN ___SCMOBJ lineedit_line_done
   ___P((lineedit_state *state,
         ___BOOL end_of_file),
        (state,
         end_of_file)
lineedit_state *state;
___BOOL end_of_file;)
{
  ___SCMOBJ e;
  extensible_string *edited = &state->current.hist->edited;
  extensible_string *input = &state->input;

  /*
   * Transfer the current edited line to the input.
   */

  if (input->buffer == NULL)
    {
      state->input_rlo = 0;

      if ((e = extensible_string_copy
                 (edited->buffer,
                  edited->length,
                  input,
                  1))
          != ___FIX(___NO_ERR))
        return e;

      if (!end_of_file)
        {
          int old_length = input->length;
          if ((e = extensible_string_set_length (input, old_length+1, 0))
              != ___FIX(___NO_ERR))
            {
              extensible_string_cleanup (input);
              state->input.buffer = NULL;
              return e;
            }
          input->buffer[old_length] = UNICODE_LINEFEED;
        }
    }

  /*
   * Update history (empty lines are not added).
   */

  if (edited->length == 0 || end_of_file)
    {
      if ((e = lineedit_move_edit_point (state, edited->length))
          == ___FIX(___NO_ERR) &&
          (end_of_file ||
           (e = lineedit_newline (state))
           == ___FIX(___NO_ERR)))
        extensible_string_set_length
          (&state->hist_last->edited,
           0,
           1); /* ignore error */
    }
  else
    {
      /*
       * Make a copy of the current edited line for the history.
       */

      extensible_string actual;

      if ((e = extensible_string_copy
                 (edited->buffer,
                  edited->length,
                  &actual,
                  0))
          == ___FIX(___NO_ERR))
        {
          /*
           * Add the current edited line to the history.
           */

          lineedit_history *next_line;

          if ((e = lineedit_history_setup (state, &next_line))
              == ___FIX(___NO_ERR))
            {
              if ((e = lineedit_history_begin_edit (state, next_line))
                  == ___FIX(___NO_ERR) &&
                  (e = lineedit_move_edit_point (state, edited->length))
                  == ___FIX(___NO_ERR) &&
                  (e = lineedit_newline (state))
                  == ___FIX(___NO_ERR))
                {
                  extensible_string_cleanup (&state->hist_last->actual);
                  state->hist_last->actual = actual;
                  lineedit_history_add (state, next_line);
                  lineedit_history_trim (state);
                }
              else
                lineedit_history_cleanup (state, next_line); /* ignore error */
            }

          if (e != ___FIX(___NO_ERR))
            extensible_string_cleanup (&actual);
        }
    }

  if (e == ___FIX(___NO_ERR))
    {
      /*
       * Revert history lines to their nonedited state.
       */

      lineedit_history *h = state->hist_last;

      for (;;)
        {
          h = h->prev;
          if (h == state->hist_last)
            break;
          lineedit_history_end_edit (state, h);
        }

      /*
       * Prepare editing of next line.
       */

      state->current.edit_point = 0;
      state->current.mark_point = 0;
      state->current.cursor_to_point = -state->current.screen_cursor;
      state->current.hist = state->hist_last;
    }

  return e;
}


___HIDDEN ___SCMOBJ lineedit_end_paren_balance
   ___P((lineedit_state *state),
        (state)
lineedit_state *state;)
{
  if (state->current.paren_balance_in_progress)
    {
      state->current.paren_balance_in_progress = 0;
      return lineedit_move_edit_point
               (state,
                state->current.edit_point); /* ignore error */
    }

  return ___FIX(___NO_ERR);
}


___HIDDEN ___SCMOBJ lineedit_process_single_event
   ___P((lineedit_state *state,
         lineedit_event *ev),
        (state,
         ev)
lineedit_state *state;
lineedit_event *ev;)
{
  extensible_string *edited = &state->current.hist->edited;

  switch (ev->event_kind)
    {
    case LINEEDIT_EV_NONE:
      return ___FIX(___NO_ERR);

    case LINEEDIT_EV_KEY:
      if (ev->event_char < UNICODE_SPACE || /* discard control characters */
          ev->event_char == UNICODE_RUBOUT)
        return ___FIX(___INVALID_OP_ERR);
      return lineedit_insert_chars (state, &ev->event_char, 1);

    case LINEEDIT_EV_RETURN:
      return lineedit_line_done (state, 0);

    case LINEEDIT_EV_BACK:
      return lineedit_delete_chars (state, state->current.edit_point-1, 0);

    case LINEEDIT_EV_BACK_WORD:
      {
        int i;
        if (!lineedit_word_boundary (state, 2, state->current.edit_point, &i))
          return ___FIX(___INVALID_OP_ERR);
        return lineedit_delete_chars (state, i, 0);
      }

    case LINEEDIT_EV_TAB:
      return lineedit_insert_chars
               (state,
                lineedit_mod.tab_spaces,
                state->tab_stop-(state->current.edit_point%state->tab_stop));

    case LINEEDIT_EV_MARK:
      state->current.mark_point = state->current.edit_point;
      return ___FIX(___NO_ERR);

    case LINEEDIT_EV_PASTE:
      return lineedit_paste_from_clipboard (state);

    case LINEEDIT_EV_CUT:
      return lineedit_delete_chars (state, state->current.mark_point, 1);

    case LINEEDIT_EV_CUT_RIGHT:
      return lineedit_delete_chars (state, edited->length, 1);

    case LINEEDIT_EV_CUT_LEFT:
      return lineedit_delete_chars (state, 0, 1);

    case LINEEDIT_EV_REFRESH:
      return lineedit_refresh (state);

    case LINEEDIT_EV_TRANSPOSE:
      {
        int i = state->current.edit_point;
        if (i <= 0 ||
            edited->length < 2)
          return ___FIX(___INVALID_OP_ERR);
        if (i == edited->length)
          return lineedit_transpose_chars (state, i-2, i-1, i-1, i);
        else
          return lineedit_transpose_chars (state, i-1, i, i, i+1);
      }

    case LINEEDIT_EV_TRANSPOSE_WORD:
      {
        int start1;
        int end1;
        int start2;
        int end2;
        if (!lineedit_word_boundary
               (state,
                0,
                state->current.edit_point,
                &end2) ||
            !lineedit_word_boundary (state, 2, end2, &start2) ||
            !lineedit_word_boundary (state, 2, start2, &start1) ||
            !lineedit_word_boundary (state, 0, start1, &end1))
          return ___FIX(___INVALID_OP_ERR);
        return lineedit_transpose_chars (state, start1, end1, start2, end2);
      }

    case LINEEDIT_EV_UP:
      return lineedit_move_history_relative (state, 1);

    case LINEEDIT_EV_DOWN:
      return lineedit_move_history_relative (state, 0);

    case LINEEDIT_EV_RIGHT:
      return lineedit_move_edit_point (state, state->current.edit_point+1);

    case LINEEDIT_EV_RIGHT_WORD:
      {
        int i;
        lineedit_word_boundary (state, 0, state->current.edit_point, &i);
        return lineedit_move_edit_point (state, i);
      }

    case LINEEDIT_EV_LEFT:
      return lineedit_move_edit_point (state, state->current.edit_point-1);

    case LINEEDIT_EV_LEFT_WORD:
      {
        int i;
        lineedit_word_boundary (state, 2, state->current.edit_point, &i);
        return lineedit_move_edit_point (state, i);
      }

    case LINEEDIT_EV_HOME:
      return lineedit_move_edit_point (state, 0);

    case LINEEDIT_EV_HOME_DOC:
      return lineedit_move_history (state, state->hist_last->next);

    case LINEEDIT_EV_INSERT:
      return ___FIX(___INVALID_OP_ERR);

    case LINEEDIT_EV_DELETE:
      if (edited->length == 0)
        return lineedit_line_done (state, 1);
      else
        return lineedit_delete_chars (state, state->current.edit_point+1, 0);

    case LINEEDIT_EV_DELETE_WORD:
      {
        int i;
        if (!lineedit_word_boundary (state, 0, state->current.edit_point, &i))
          return ___FIX(___INVALID_OP_ERR);
        return lineedit_delete_chars (state, i, 0);
      }

    case LINEEDIT_EV_END:
      return lineedit_move_edit_point (state, edited->length);

    case LINEEDIT_EV_END_DOC:
      return lineedit_move_history (state, state->hist_last);

    case LINEEDIT_EV_F1:
    case LINEEDIT_EV_META_F1:
    case LINEEDIT_EV_F2:
    case LINEEDIT_EV_META_F2:
    case LINEEDIT_EV_F3:
    case LINEEDIT_EV_META_F3:
    case LINEEDIT_EV_F4:
    case LINEEDIT_EV_META_F4:
#ifdef LINEEDIT_SUPPORT_F5_TO_F12
    case LINEEDIT_EV_F5:
    case LINEEDIT_EV_META_F5:
    case LINEEDIT_EV_F6:
    case LINEEDIT_EV_META_F6:
    case LINEEDIT_EV_F7:
    case LINEEDIT_EV_META_F7:
    case LINEEDIT_EV_F8:
    case LINEEDIT_EV_META_F8:
    case LINEEDIT_EV_F9:
    case LINEEDIT_EV_META_F9:
    case LINEEDIT_EV_F10:
    case LINEEDIT_EV_META_F10:
    case LINEEDIT_EV_F11:
    case LINEEDIT_EV_META_F11:
    case LINEEDIT_EV_F12:
    case LINEEDIT_EV_META_F12:
#endif
      {
        tty_char command[7];
        command[0] = UNICODE_SHARP;
        command[1] = UNICODE_VBAR;
        command[2] = UNICODE_VBAR;
        command[3] = UNICODE_SHARP;
        command[4] = UNICODE_COMMA;
        switch (ev->event_kind)
          {
          case LINEEDIT_EV_F9:
            command[5] = 's';
            break;
          case LINEEDIT_EV_F10:
            command[5] = 'l';
            break;
          case LINEEDIT_EV_F11:
            command[5] = 'e';
            break;
          case LINEEDIT_EV_F12:
            command[5] = 'b';
            break;
          default:
            command[5] = 't';
            break;
          }
        command[6] = UNICODE_SEMICOLON;
        if (lineedit_move_edit_point (state, 0)
            == ___FIX(___NO_ERR) &&
            lineedit_delete_chars (state, edited->length, 0)
            == ___FIX(___NO_ERR) &&
            lineedit_insert_chars (state, command, 7)
            == ___FIX(___NO_ERR))
          return lineedit_line_done (state, 0);
        break;
      }
    }

  return ___FIX(___INVALID_OP_ERR);
}


___HIDDEN ___SCMOBJ lineedit_process_events
   ___P((lineedit_state *state),
        (state)
lineedit_state *state;)
{
#ifdef LINEEDIT_WITH_NONBLOCKING_IO
#define AGGREGATION_BUFFER_SIZE (80*50)
#else
#define AGGREGATION_BUFFER_SIZE 1
#endif

  ___SCMOBJ e1;
  ___SCMOBJ e2;
  tty_char aggregation_buffer[AGGREGATION_BUFFER_SIZE];
  lineedit_event ev;
  int i;
  struct lineedit_state_undo_struct previous;

  /*
   * Process events only if no input line is currently available.
   */

  if (state->input.buffer != NULL)
    return ___FIX(___NO_ERR);

  /*
   * Process events only if no line editing output is pending.
   */

  if ((e1 = lineedit_output_drain (state)) != ___FIX(___NO_ERR))
    return e1;

  /*
   * Save state so that we can undo if there is an error.
   */

  previous = state->current;

  i = 0;

  ev.event_kind = LINEEDIT_EV_NONE;

  state->current.paren_balance_trigger = 0;

  for (;;)
    {
      if ((e1 = lineedit_get_event (state, &ev)) != ___FIX(___NO_ERR) ||
          ev.event_kind == LINEEDIT_EV_NONE)
        break;

      if (ev.event_kind != LINEEDIT_EV_KEY ||
          ev.event_char < UNICODE_SPACE ||
          ev.event_char == UNICODE_RUBOUT)
        break;

      aggregation_buffer[i++] = ev.event_char;

      ev.event_kind = LINEEDIT_EV_NONE;

      if (i == AGGREGATION_BUFFER_SIZE)
        break;
    }

  e2 = lineedit_insert_chars (state, aggregation_buffer, i);
  if (e1 == ___FIX(___NO_ERR))
    e1 = e2;

  e2 = lineedit_process_single_event (state, &ev);
  if (e1 == ___FIX(___NO_ERR))
    e1 = e2;

  if (e1 == ___FIX(___INVALID_OP_ERR))
    {
      /*
       * Undo line editing operation.
       */

      tty_char c = UNICODE_BELL;

      extensible_string_set_length
        (&state->output,
         0,
         1); /* ignore error */

      state->output_wlo = state->output.length;

      state->current = previous;

      e1 = extensible_string_insert
             (&state->output,
              state->output.length,
              1,
              &c);
    }

  if (state->current.paren_balance_trigger)
    {
      /*
       * Parenthesis balancing started in this round of event
       * processing.
       */

      ___time duration;

      ___time_get_current_time (&state->current.paren_balance_end);
      ___time_from_nsecs (&duration, 0, state->paren_balance_duration_nsecs);
      ___time_add (&state->current.paren_balance_end, duration);

      state->current.paren_balance_in_progress = 1;

#ifndef LINEEDIT_WITH_NONBLOCKING_IO
      absolute_time_sleep (state->current.paren_balance_end);
      lineedit_end_paren_balance (state); /* ignore error */
#endif
    }
  else if (state->current.paren_balance_in_progress)
    {
      /*
       * Parenthesis balancing did not start in this round of event
       * processing and parenthesis balancing is in progress, check if
       * parenthesis balancing should end.
       */

      if (i > 0 || ev.event_kind != LINEEDIT_EV_NONE)
        lineedit_end_paren_balance (state); /* ignore error */
#ifdef LINEEDIT_WITH_NONBLOCKING_IO
      else
        {
          ___time now;
          ___time_get_current_time (&now);
          if (!___time_less (now, state->current.paren_balance_end))
            lineedit_end_paren_balance (state); /* ignore error */
        }
#endif
    }

  e2 = lineedit_output_drain (state);
  if (e1 == ___FIX(___NO_ERR))
    e1 = e2;

  return e1;
}


___HIDDEN ___SCMOBJ lineedit_read
   ___P((lineedit_state *state,
         tty_char *buf,
         ___stream_index len,
         ___stream_index *len_done),
        (state,
         buf,
         len,
         len_done)
lineedit_state *state;
tty_char *buf;
___stream_index len;
___stream_index *len_done;)
{
  ___SCMOBJ e;
  int rlo;
  int line_length;
  int avail;

  if ((e = lineedit_process_events (state)) != ___FIX(___NO_ERR))
    return e;

  if (state->input.buffer == NULL)
    return ___ERR_CODE_EAGAIN;

  rlo = state->input_rlo;
  line_length = state->input.length;
  avail = line_length - rlo;

  if (avail > 0)
    {
      tty_char *p = &state->input.buffer[rlo];
      tty_char *end;

      if (len > avail)
        len = avail;

      end = p + len;

      while (p < end)
        *buf++ = *p++;

      rlo += len;
    }
  else
    len = 0;

  state->input_rlo = rlo;

  if (rlo >= line_length)
    {
      extensible_string_cleanup (&state->input);
      state->input.buffer = NULL;
    }

  *len_done = len;

  return ___FIX(___NO_ERR);
}


___HIDDEN ___SCMOBJ lineedit_module_setup ___PVOID
{
  int i;

  for (i=0; i<LINEEDIT_MAX_TAB_STOP; i++)
    lineedit_mod.tab_spaces[i] = UNICODE_SPACE;

  return ___FIX(___NO_ERR);
}


___HIDDEN void lineedit_module_cleanup ___PVOID
{
}

#endif


/*---------------------------------------------------------------------------*/

/* Tty management routines */

#define TTY_BIG_SCREEN_SIZE (80*50)

___HIDDEN void tty_broadcast_signal
   ___P((tty_signal s),
        (s)
tty_signal s;)
{
  tty_state *list = tty_mod.list;

  if (list != NULL)
    {
      tty_state *state = list;
      do
        {
          state->received_signal[s] = 1;
          state = state->next;
        } while (state != list);
    }
}

___HIDDEN ___SCMOBJ tty_setup_common
   ___P((tty *t,
         tty_state *state),
        (t,
         state)
tty *t;
tty_state *state;)
{
  ___SCMOBJ e;
  tty_state *dest;

  state->refcount = 1;

  if (___getenv (___EMACS_ENV_VAR) == NULL)
    state->input_mode = LINEEDIT_INPUT_MODE;
  else
    state->input_mode = 0;

  state->term_mode = INITIAL_TERM_MODE;

  state->rlo = 0;
  state->rhi = 0;
  state->wlo = 0;
  state->whi = 0;
  state->werr = ___FIX(___NO_ERR);

  state->received_signal[USER_INTERRUPT_SIGNAL] = 0;

#ifdef USE_LINEEDIT

  /*
   * Setup line editing state.
   */

  if ((e = lineedit_setup (&state->les, state)) != ___FIX(___NO_ERR))
    {
      ___free_mem (state);
      return e;
    }

  /*
   * Switch to CBREAK mode so that terminal driver does not perform
   * its own line editing.
   */

  if (state->input_mode & LINEEDIT_INPUT_MODE)
    if ((e = tty_set_term_mode (state, CBREAK_TERM_MODE))
        != ___FIX(___NO_ERR))
      return e;

#endif

  /* add tty_state to tail of circular list of tty_state's */

  dest = tty_mod.list;

  if (dest == NULL)
    {
      state->prev = state;
      state->next = state;
      tty_mod.list = state;
    }
  else
    {
      tty_state *before_dest = dest->prev;
      state->next = dest;
      state->prev = before_dest;
      dest->prev = state;
      before_dest->next = state;
    }

  t->state = state;

  return ___FIX(___NO_ERR);
}

#ifndef USE_POSIX
#ifndef USE_WIN32
#ifndef USE_MACOS

/**************************/

#endif
#endif
#endif

#ifdef USE_POSIX

___HIDDEN ___BOOL tty_equal
   ___P((int fd1,
         int fd2),
        (fd1,
         fd2)
int fd1;
int fd2;)
{
  struct stat fd1_stat;
  struct stat fd2_stat;

#if 0
  printf ("tty_equal (%d, %d)\n", fd1, fd2);

  printf ("%d -> %s\n", fd1, ttyname(fd1));
  printf ("%d -> %s\n", fd2, ttyname(fd2));
#endif

  if (fd1 == fd2)
    return 1;

  /* check if st_rdev fields are the same */

  if (fstat (fd1, &fd1_stat) == 0 &&
      fstat (fd2, &fd2_stat) == 0 &&
#if 0
(
 printf ("  fd1_stat.st_rdev = %lld   fd2_stat.st_rdev = %lld\n",fd1_stat.st_rdev, fd2_stat.st_rdev),
 printf ("  fd1_stat.st_dev = %lld    fd2_stat.st_dev = %lld\n",fd1_stat.st_dev, fd2_stat.st_dev),
 printf ("  fd1_stat.st_ino = %ld    fd2_stat.st_ino = %ld\n",fd1_stat.st_ino, fd2_stat.st_ino),
 1) &&
#endif
      fd1_stat.st_rdev == fd2_stat.st_rdev)
    return 1;

  /* check if the descriptors both refer to the controlling terminal */

  if (tcgetpgrp (fd1) >= 0 &&
      tcgetpgrp (fd2) >= 0)
    return 1;

  return 0;
}

___HIDDEN ___SCMOBJ tty_setup_os_specific
   ___P((tty *t,
         int fd),
        (t,
         fd)
tty *t;
int fd;)
{
  tty_state *state;

  t->state = NULL;

  /* check if the same terminal was previously opened */

  state = tty_mod.list;

  if (state != NULL)
    {
      do
        {
          if (tty_equal (fd, state->fd))
            {
              //  printf ("closing tty fd=%d (same as %d)\n", fd, state->fd);//////////////////////////
              close (fd); /* ignore error */
              state->refcount++;
              t->state = state;
              return ___FIX(___NO_ERR);
            }
          state = state->next;
        } while (state != tty_mod.list);
    }

  /* need to allocate a new tty_state */

  state = ___CAST(tty_state*,
                  ___alloc_mem (sizeof (tty_state)));

  if (state == NULL)
    return ___FIX(___HEAP_OVERFLOW_ERR);

  state->received_signal[WINDOW_SIZE_SIGNAL] = 1; /* query window size */
  state->received_signal[CONTINUE_SIGNAL] = 0;

  state->fd = fd;

  if (tcgetattr (state->fd, &state->initial_termios) < 0)
    {
      ___SCMOBJ e = err_code_from_errno ();
      ___free_mem (state);
      return e;
    }

  state->encoding = LATIN1_ENCODING;

  return tty_setup_common (t, state);
}

___HIDDEN ___SCMOBJ tty_cleanup_os_specific
   ___P((tty_state *state),
        (state)
tty_state *state;)
{
  //  printf ("closing tty fd=%d\n", state->fd);//////////////////////////

  if (close (state->fd) < 0)
    return err_code_from_errno ();

  return ___FIX(___NO_ERR);
}


___HIDDEN ___SCMOBJ tty_force_open
   ___P((tty *t),
        (t)
tty *t;)
{
  ___SCMOBJ e;
  int fd;
  char term_name[L_ctermid];
  char *term_type;
  tty_state *state = t->state;

  if (state != NULL)
    return ___FIX(___NO_ERR);

  /* tty refers to the console */

  ctermid (term_name); /* get controlling terminal's name */

#if 0
  printf ("ctermid -> %s\n", term_name);fflush(stdout);//////////////
#endif

#ifdef LINEEDIT_WITH_NONBLOCKING_IO
  fd = open (term_name, O_RDWR|O_NONBLOCK);
#else
  fd = open (term_name, O_RDWR);
#endif

  if (fd < 0)
    return err_code_from_errno ();

  if ((e = tty_setup_os_specific (t, fd)) != ___FIX(___NO_ERR))//////////////////////beware... this might close the fd!!!!
    {
      close (fd); /* ignore error */
      return e;
    }

  if ((e = tty_set_terminal_type
             (t,
              ___getenv (___CONSOLE_TERMINAL_TYPE_ENV_VAR),
              DEFAULT_EMACS_BINDINGS))
      != ___FIX(___NO_ERR))
    tty_cleanup (t); /* ignore error */

#if 0
  if ((e = tty_set_encoding
             (&d->t,
              UTF8_ENCODING))
      != ___FIX(___NO_ERR))
    ;/*******************/
  if ((e = tty_set_input_mode (&d->t, CBREAK_TERM_MODE))
      != ___FIX(___NO_ERR))
    ;/*******************/
#endif

  return e;
}


___HIDDEN ___SCMOBJ tty_set_term_mode
   ___P((tty_state *state,
         tty_term_mode term_mode),
        (state,
         term_mode)
tty_state *state;
tty_term_mode term_mode;)
{
  struct termios new_termios;

  new_termios = state->initial_termios;///////////uninitialized????

  switch (term_mode)
    {
    case CBREAK_TERM_MODE:
      new_termios.c_lflag &= ~(ECHO | ICANON);
      new_termios.c_cc[VMIN] = 0;
      new_termios.c_cc[VTIME] = 0;
      break;

    case RAW_TERM_MODE:
      new_termios.c_lflag &= ~(ECHO | ICANON | IEXTEN | ISIG);
      new_termios.c_iflag &= ~(BRKINT | ICRNL | INPCK | ISTRIP | IXON | IXOFF);
      new_termios.c_cflag &= ~(CSIZE | PARENB);
      new_termios.c_cflag |= CS8;
      new_termios.c_oflag &= ~(OPOST);
      new_termios.c_cc[VMIN] = 0;
      new_termios.c_cc[VTIME] = 0;
#if 0
      glibc uses this for raw mode:
                termios-p->c_iflag &= ~(IGNBRK|BRKINT|PARMRK|ISTRIP
                                              |INLCR|IGNCR|ICRNL|IXON);
                termios-p->c_oflag &= ~OPOST;
                termios-p->c_lflag &= ~(ECHO|ECHONL|ICANON|ISIG|IEXTEN);
                termios-p->c_cflag &= ~(CSIZE|PARENB);
                termios-p->c_cflag |= CS8;
#endif
      break;
    }

  if (tcsetattr (state->fd, TCSAFLUSH, &new_termios) < 0)
    return err_code_from_errno ();

  state->term_mode = term_mode;

#ifdef USE_LINEEDIT
  state->les->linefeed_moves_to_left_margin =
    ((new_termios.c_oflag & (OPOST | ONLCR)) == (OPOST | ONLCR));
#endif

  return ___FIX(___NO_ERR);
}


___HIDDEN void tty_signal_handler
   ___P((int sig),
        (sig)
int sig;)
{
  signal (sig, tty_signal_handler);

  //  printf ("tty_signal_handler sig=%d\n",sig);fflush(stdout);/////////////

  switch (sig)
    {
    case SIGINT:
      ___raise_interrupt (___INTR_USER);////temporary hack
      tty_broadcast_signal (USER_INTERRUPT_SIGNAL);
      if (tty_mod.user_interrupt_handler != NULL)
        tty_mod.user_interrupt_handler ();
      break;

    case SIGWINCH:
      tty_broadcast_signal (WINDOW_SIZE_SIGNAL);
      {/**************/
        struct winsize size;
        if (ioctl (0, TIOCGWINSZ, &size) < 0)
          printf ("error\n");/************/
        printf ("%d rows, %d columns\n", size.ws_row, size.ws_col);
        fflush (stdout);
      }
      break;

    case SIGCONT:
      tty_broadcast_signal (CONTINUE_SIGNAL);
      if (tty_mod.console.state != NULL)
        tty_set_term_mode (tty_mod.console.state,
                           tty_mod.console.state->term_mode);
      printf ("continued\n");
      fflush (stdout);
      break;

    case SIGPIPE:
      {
        printf ("!!!!!!!!!!!SIGPIPE!!!!!!!!!!!!\n");
        fflush (stdout);
      }
      break;
    }
}


___HIDDEN ___SCMOBJ tty_read_os_specific
   ___P((tty_state *state,
         ___U8 *buf,
         ___stream_index len,
         ___stream_index *len_done),
        (state,
         buf,
         len,
         len_done)
tty_state *state;
___U8 *buf;
___stream_index len;
___stream_index *len_done;)
{
  int n;

  if ((n = read (state->fd, buf, len)) < 0)
    return err_code_from_errno ();

  *len_done = n;

  return ___FIX(___NO_ERR);
}


___HIDDEN ___SCMOBJ tty_write_os_specific
   ___P((tty_state *state,
         ___U8 *buf,
         ___stream_index len,
         ___stream_index *len_done),
        (state,
         buf,
         len,
         len_done)
tty_state *state;
___U8 *buf;
___stream_index len;
___stream_index *len_done;)
{
  int n;

#if 1
  n = write (state->fd, buf, len);
#else
  /*****************************/
  {
    static int last_time = 0;
    int now;
    struct timeval tv;
    gettimeofday (&tv, NULL);
    now = tv.tv_usec / 33000; /* simulate 300 baud */
    if (now == last_time)
      n = 0;
    else
      {
        int x = len;
        if (x >= 1)
          x = 1;
#if 0
        n = write (state->fd, buf, x);
#else
        n = 1;
#endif
        if (n == 1)
          {
#if 1
            printf ("%c", *buf); fflush(stdout);
#endif
          }
        last_time = now;
      }
  }
#endif

  if (n < 0)
    return err_code_from_errno ();

  *len_done = n;

  return ___FIX(___NO_ERR);
}

___HIDDEN ___SCMOBJ tty_module_setup_os_specific ___PVOID
{
  signal (SIGINT, tty_signal_handler);
  signal (SIGWINCH, tty_signal_handler);
  signal (SIGCONT, tty_signal_handler);
  signal (SIGPIPE, tty_signal_handler); /*************** put this somewhere else*/

  return ___FIX(___NO_ERR);
}

___HIDDEN void tty_module_cleanup_os_specific ___PVOID
{
  signal (SIGINT, SIG_DFL);
  signal (SIGWINCH, SIG_DFL);
  signal (SIGCONT, SIG_DFL);
  signal (SIGPIPE, SIG_DFL); /*************** put this somewhere else*/
}

#endif

#ifdef USE_WIN32

___HIDDEN ___SCMOBJ tty_setup_os_specific
   ___P((tty *t),
        (t)
tty *t;)
{
  tty_state *state;

  t->state = NULL;

  /* check if the same terminal was previously opened */

  state = tty_mod.list;

  /////printf ("tty_setup_os_specific\n");fflush(stdout);/////////////////////

  if (state != NULL)
    {
      /* Windows only allows one console per process. */

      state->refcount++;
      t->state = state;
      return ___FIX(___NO_ERR);
    }

  /* need to allocate a new tty_state */

  state = ___CAST(tty_state*,
                  ___alloc_mem (sizeof (tty_state)));

  if (state == NULL)
    return ___FIX(___HEAP_OVERFLOW_ERR);

  ///////////////////////////////////////////

    {
      DWORD m;
      HANDLE in;
      HANDLE out;
      HANDLE sin = GetStdHandle (STD_INPUT_HANDLE); /* ignore error */
      HANDLE sout = GetStdHandle (STD_OUTPUT_HANDLE); /* ignore error */
      HANDLE serr = GetStdHandle (STD_ERROR_HANDLE); /* ignore error */

      if (AllocConsole ())
        {
          /* restore initial standard handles */

          SetStdHandle (STD_INPUT_HANDLE, sin); /* ignore error */
          SetStdHandle (STD_OUTPUT_HANDLE, sout); /* ignore error */
          SetStdHandle (STD_ERROR_HANDLE, serr); /* ignore error */
        }

      in = CreateFile
             (_T("CONIN$"),
              GENERIC_READ | GENERIC_WRITE,
              FILE_SHARE_READ | FILE_SHARE_WRITE,
              NULL,
              OPEN_EXISTING,
              0,
              NULL);

      if (in == INVALID_HANDLE_VALUE)
        return ___FIX(___UNKNOWN_ERR);

      out = CreateFile
              (_T("CONOUT$"),
               GENERIC_READ | GENERIC_WRITE,
               FILE_SHARE_READ | FILE_SHARE_WRITE,
               NULL,
               OPEN_EXISTING,
               0,
               NULL);

      if (out == INVALID_HANDLE_VALUE)
        {
          CloseHandle (in); /* ignore error */
          return ___FIX(___UNKNOWN_ERR);
        }

  state->input = in;
  state->output = out;
  state->input_rlo = 0;
  state->input_rhi = 0;

    }

  {
#if 0
    DWORD n;
    if (WriteConsoleW (output, NULL, 0, &n, NULL))
      state->encoding = UCS2LE_ENCODING;
    else
#endif
      state->encoding = LATIN1_ENCODING;
  }

    ////printf ("tty_setup_common\n");fflush(stdout);/////////////////////

  return tty_setup_common (t, state);
}

___HIDDEN ___SCMOBJ tty_cleanup_os_specific
   ___P((tty_state *state),
        (state)
tty_state *state;)
{
  ___SCMOBJ e1 = ___FIX(___NO_ERR);
  ___SCMOBJ e2;

  if (!CloseHandle (state->input))
    e1 = err_code_from_GetLastError ();

  if (!CloseHandle (state->output))
    {
      e2 = err_code_from_GetLastError ();
      if (e1 == ___FIX(___NO_ERR))
        e1 = e2;
    }

  return e1;
}

___HIDDEN ___SCMOBJ tty_force_open
   ___P((tty *t),
        (t)
tty *t;)
{
  tty_state *state = t->state;

  if (state != NULL)
    return ___FIX(___NO_ERR);

  /////printf("force_open\n");fflush(stdout);//////////////////

  /* tty refers to the console */

  return tty_setup_os_specific (t);  /********************/

  return ___FIX(___NO_ERR);
}

___HIDDEN ___SCMOBJ tty_set_term_mode
   ___P((tty_state *state,
         tty_term_mode term_mode),
        (state,
         term_mode)
tty_state *state;
tty_term_mode term_mode;)
{
  state->term_mode = term_mode;

  return ___FIX(___NO_ERR);
}

___HIDDEN ___SCMOBJ tty_read_os_specific
   ___P((tty_state *state,
         ___U8 *buf,
         ___stream_index len,
         ___stream_index *len_done),
        (state,
         buf,
         len,
         len_done)
tty_state *state;
___U8 *buf;
___stream_index len;
___stream_index *len_done;)
{
  int i = 0;

  //printf ("tty_read_os_specific %d %d\n",state->input_rlo, state->input_rhi);fflush(stdout);////////////////////

  while (state->input_rlo >= state->input_rhi)
    {
      DWORD n;

#ifdef LINEEDIT_WITH_NONBLOCKING_IO

      if (!GetNumberOfConsoleInputEvents (state->input, &n))
        return err_code_from_GetLastError ();

      if (n <= 0)
        goto done;

      if (n > INPUT_RBUF_SIZE)
        n = INPUT_RBUF_SIZE;

#else

      n = INPUT_RBUF_SIZE;

#endif

      if (!((state->encoding == UCS2LE_ENCODING)
            ? ReadConsoleInputW
                (state->input,
                 state->input_rbuf,
                 n,
                 &n)
            : ReadConsoleInputA
                (state->input,
                 state->input_rbuf,
                 n,
                 &n)))
        return err_code_from_GetLastError ();

      ///printf ("ReadConsoleInput -> %d\n",n);fflush(stdout);////////////////////

      state->input_rlo = 0;
      state->input_rhi = n;
    }

  while (state->input_rlo < state->input_rhi)
    {
      INPUT_RECORD *ir = &state->input_rbuf[state->input_rlo];

      switch (ir->EventType)
        {
        case KEY_EVENT:
          {
            DWORD ctrl_keys;
            ___BOOL ctrl;
            ___BOOL alt;
            ___BOOL shift;
            char *seq;
            ___UCS4 c;

            if (!(ir->Event.KeyEvent.bKeyDown &&
                  ir->Event.KeyEvent.wRepeatCount > 0))
              {
                state->input_rlo++;
                break;
              }

            ctrl_keys = ir->Event.KeyEvent.dwControlKeyState;
            ctrl = ctrl_keys & (LEFT_CTRL_PRESSED|RIGHT_CTRL_PRESSED);
            alt = ctrl_keys & (LEFT_ALT_PRESSED|RIGHT_ALT_PRESSED);
            shift = ctrl_keys & SHIFT_PRESSED;

            if (state->encoding == UCS2LE_ENCODING)
              c = ir->Event.KeyEvent.uChar.UnicodeChar;
            else
              c = ir->Event.KeyEvent.uChar.AsciiChar;

            seq = NULL;

            if (!ctrl && !alt && !shift && c == 0)
              switch (ir->Event.KeyEvent.wVirtualKeyCode)
                {
                case VK_BACK:   seq = "\010";     break;
                case VK_UP:     seq = "\033[A";   break;
                case VK_DOWN:   seq = "\033[B";   break;
                case VK_RIGHT:  seq = "\033[C";   break;
                case VK_LEFT:   seq = "\033[D";   break;
                case VK_HOME:   seq = "\033[H";   break;
                case VK_INSERT: seq = "\033[2~";  break;
                case VK_DELETE: seq = "\177";     break;
                case VK_END:    seq = "\033[F";   break;
                case VK_PRIOR:  seq = "\033[5~";  break;
                case VK_NEXT:   seq = "\033[6~";  break;
                case VK_F1:     seq = "\033OP";   break;
                case VK_F2:     seq = "\033OQ";   break;
                case VK_F3:     seq = "\033OR";   break;
                case VK_F4:     seq = "\033OS";   break;
#ifdef LINEEDIT_SUPPORT_F5_TO_F12
                case VK_F5:     seq = "\033[15~"; break;
                case VK_F6:     seq = "\033[17~"; break;
                case VK_F7:     seq = "\033[18~"; break;
                case VK_F8:     seq = "\033[19~"; break;
                case VK_F9:     seq = "\033[20~"; break;
                case VK_F10:    seq = "\033[21~"; break;
                case VK_F11:    seq = "\033[23~"; break;
                case VK_F12:    seq = "\033[24~"; break;
#endif
                }

            if (seq == NULL)
              {
                if (c == 0)
                  {
                    ir->Event.KeyEvent.wRepeatCount = 0;
                    break;
                  }

                if (ctrl && c == UNICODE_SPACE)
                  c = UNICODE_NUL;

                if (state->encoding == UCS2LE_ENCODING)
                  {
                    if (len < i + 2)
                      goto done;
                    buf[i++] = c & 0xff;
                    buf[i++] = c >> 8;
                  }
                else
                  {
#if 1
                    if (len < i + 1)
                      goto done;
                    buf[i++] = c & 0xff;
#else
                    if (len < i + 2)
                      goto done;
                    buf[i++] = c & 0xff;
                    buf[i++] = ' ';
#endif
                  }
              }
            else
              {
                while (*seq != '\0')
                  {
                    if (state->encoding == UCS2LE_ENCODING)
                      {
                        if (len < i + 2)
                          goto done;
                        buf[i++] = *seq++;
                        buf[i++] = 0;
                      }
                    else
                      {
#if 1
                        if (len < i + 1)
                          goto done;
                        buf[i++] = *seq++;
#else
                        if (len < i + 2)
                          goto done;
                        buf[i++] = *seq++;
                        buf[i++] = ' ';
#endif
                      }
                  }
              }

            ir->Event.KeyEvent.wRepeatCount--;

            break;
          }

        case MOUSE_EVENT:
          {
            ___SCMOBJ e;

            if (i > 0)
              goto done; /* in case paste returns an error */

            state->input_rlo++;

#ifdef USE_LINEEDIT
            if (ir->Event.MouseEvent.dwEventFlags == 0 &&
                (ir->Event.MouseEvent.dwButtonState ==
                 (FROM_LEFT_1ST_BUTTON_PRESSED|RIGHTMOST_BUTTON_PRESSED) ||
                 ir->Event.MouseEvent.dwButtonState ==
                 FROM_LEFT_2ND_BUTTON_PRESSED))
              if ((e = lineedit_paste_from_clipboard (state->les))
                  != ___FIX(___NO_ERR))
                return e;
#endif

            goto done; /* pasted text is available to be read */
          }

        default:
          state->input_rlo++;
          break;
        }
    }

  done:

  if (i == 0)
    {
      ////printf ("returning ___ERR_CODE_EAGAIN\n");fflush(stdout);////////////////////
      return ___ERR_CODE_EAGAIN;
    }

  *len_done = i;

  return ___FIX(___NO_ERR);
}

___HIDDEN ___SCMOBJ tty_write_os_specific
   ___P((tty_state *state,
         ___U8 *buf,
         ___stream_index len,
         ___stream_index *len_done),
        (state,
         buf,
         len,
         len_done)
tty_state *state;
___U8 *buf;
___stream_index len;
___stream_index *len_done;)
{
  DWORD n;

  if (state->encoding == UCS2LE_ENCODING)
    {
      len = len >> 1;
      if (!WriteConsoleW (state->output, (wchar_t*)buf, len, &n, NULL))
        return err_code_from_GetLastError ();
      n = n << 1;
    }
  else
    {
      if (!WriteConsoleA (state->output, (char*)buf, len, &n, NULL))
        return err_code_from_GetLastError ();
    }

#if 0
  {
    CONSOLE_CURSOR_INFO ci;
    if (GetConsoleCursorInfo (state->output, &ci))
      {
        ci.bVisible = 0;
        if (!SetConsoleCursorInfo (state->output, &ci))
          ;
        ci.bVisible = 1;
        if (!SetConsoleCursorInfo (state->output, &ci))
          ;
      }
  }
#else
      {
        // make cursor visible
        CONSOLE_SCREEN_BUFFER_INFO csbi;
        if (GetConsoleScreenBufferInfo (state->output, &csbi))
          {
            SetConsoleCursorPosition (state->output, csbi.dwCursorPosition);
          }
      }
#endif

  *len_done = n;

  return ___FIX(___NO_ERR);
}

___HIDDEN BOOL WINAPI tty_console_event_handler
   ___P((DWORD dwCtrlType),
        (dwCtrlType)
DWORD dwCtrlType;)
{
  tty_state *state = tty_mod.console.state;

  switch (dwCtrlType)
    {
    case CTRL_C_EVENT:
      {
        if (state != NULL &&
            state->term_mode != CBREAK_TERM_MODE)
          return TRUE;

        tty_broadcast_signal (USER_INTERRUPT_SIGNAL);

        if (tty_mod.user_interrupt_handler != NULL)
          tty_mod.user_interrupt_handler ();

        return TRUE;
      }

    case CTRL_BREAK_EVENT:
      {
        if (state != NULL &&
            state->term_mode == RAW_TERM_MODE)
          return TRUE;

        break;
      }

    case CTRL_CLOSE_EVENT:
    case CTRL_LOGOFF_EVENT:
    case CTRL_SHUTDOWN_EVENT:
      break;
    }

  return FALSE; /* let system handle the console event */
}

___HIDDEN ___SCMOBJ tty_module_setup_os_specific ___PVOID
{
  if (!SetConsoleCtrlHandler (tty_console_event_handler, TRUE))
    return err_code_from_GetLastError ();

  return ___FIX(___NO_ERR);
}

___HIDDEN void tty_module_cleanup_os_specific ___PVOID
{
  SetConsoleCtrlHandler (tty_console_event_handler, FALSE); /* ignore error */
}

#endif

___HIDDEN void tty_setup_console
   ___P((tty *t),
        (t)
tty *t;)
{
  t->state = NULL;
}

___HIDDEN ___SCMOBJ tty_cleanup
   ___P((tty *t),
        (t)
tty *t;)
{
  ___SCMOBJ e1;
  ___SCMOBJ e2;
  tty_state *state = t->state;

  t->state = NULL; /* avoid dangling pointer */

#if 0
  if (state == NULL)
    printf ("tty_cleanup state=NULL\n");/////////////////////////
  else
    printf ("tty_cleanup refcount=%d\n",state->refcount);/////////////////////////
#endif

  if (state == NULL || --state->refcount != 0)
    return ___FIX(___NO_ERR);

#ifdef USE_LINEEDIT
  lineedit_cleanup (state->les);
#endif

  e1 = tty_set_term_mode (state, INITIAL_TERM_MODE);

  e2 = tty_cleanup_os_specific (state);
  if (e1 == ___FIX(___NO_ERR))
    e1 = e2;

  /* remove tty_state from circular list of tty_state's */

  if (state->next == state) /* only tty_state in list? */
    tty_mod.list = NULL;
  else
    {
      if (tty_mod.list == state)
        tty_mod.list = state->next;
      state->next->prev = state->prev;
      state->prev->next = state->next;
    }

  ___free_mem (state);

  return e1;
}

___HIDDEN ___SCMOBJ tty_set_terminal_type
   ___P((tty *t,
         char *terminal_type,
         ___BOOL emacs_bindings),
        (t,
         terminal_type,
         emacs_bindings)
tty *t;
char *terminal_type;
___BOOL emacs_bindings;)
{
  ___SCMOBJ e;
  tty_state *state;

  if ((e = tty_force_open (t)) != ___FIX(___NO_ERR))
    return e;

  state = t->state;

#ifdef USE_LINEEDIT
  e = lineedit_set_terminal_type (state->les, terminal_type, emacs_bindings);
#endif

  return e;
}

___HIDDEN ___SCMOBJ tty_set_encoding
   ___P((tty *t,
         tty_encoding encoding),
        (t,
         encoding)
tty *t;
tty_encoding encoding;)
{
  ___SCMOBJ e;
  tty_state *state;

  if ((e = tty_force_open (t)) != ___FIX(___NO_ERR))
    return e;

  state = t->state;

  state->encoding = encoding;

  return e;
}

___HIDDEN ___SCMOBJ tty_set_input_mode
   ___P((tty *t,
         tty_input_mode input_mode),
        (t,
         input_mode)
tty *t;
tty_input_mode input_mode;)
{
  ___SCMOBJ e;
  tty_state *state;

  if ((e = tty_force_open (t)) != ___FIX(___NO_ERR))
    return e;

  state = t->state;

  if (input_mode & RAW_INPUT_MODE)
    e = tty_set_term_mode (state, RAW_TERM_MODE);
  else
    e = tty_set_term_mode (state, CBREAK_TERM_MODE);

  if (e == ___FIX(___NO_ERR))
    state->input_mode = input_mode;

  return e;
}

___HIDDEN ___SCMOBJ tty_set_paren_balance_duration
   ___P((tty *t,
         int nsecs),
        (t,
         nsecs)
tty *t;
int nsecs;)
{
  ___SCMOBJ e;

  tty_state *state;

  if ((e = tty_force_open (t)) != ___FIX(___NO_ERR))
    return e;

  state = t->state;

#ifdef USE_LINEEDIT
  e = lineedit_set_paren_balance_duration (state->les, nsecs);
#endif

  return e;
}

___HIDDEN ___SCMOBJ tty_set_max_history_length
   ___P((tty *t,
         int max_history_length),
        (t,
         max_history_length)
tty *t;
int max_history_length;)
{
  ___SCMOBJ e;
  tty_state *state;

  if ((e = tty_force_open (t)) != ___FIX(___NO_ERR))
    return e;

  state = t->state;

#ifdef USE_LINEEDIT
  lineedit_set_max_history_length (state->les, max_history_length);
#endif

  return ___FIX(___NO_ERR);
}

___HIDDEN ___SCMOBJ tty_set_text_attrs
   ___P((tty *t,
         tty_text_attrs input_attrs,
         tty_text_attrs output_attrs),
        (t,
         input_attrs,
         output_attrs)
tty *t;
tty_text_attrs input_attrs;
tty_text_attrs output_attrs;)
{
  ___SCMOBJ e;
  tty_state *state;

  if ((e = tty_force_open (t)) != ___FIX(___NO_ERR))
    return e;

  state = t->state;

#ifdef USE_LINEEDIT
  state->les->input_attrs = input_attrs;
  state->les->output_attrs = output_attrs;
#endif

  return ___FIX(___NO_ERR);
}

___HIDDEN ___SCMOBJ tty_read_from_rbuf
   ___P((tty_state *state,
         tty_char *buf,
         ___stream_index len,
         ___stream_index *len_done),
        (state,
         buf,
         len,
         len_done)
tty_state *state;
tty_char *buf;
___stream_index len;
___stream_index *len_done;)
{
  int rlo = state->rlo;
  int rhi = state->rhi;
  ___U8 *rbuf = state->rbuf;
  int n = len;

  ////printf ("tty_read_from_rbuf n=%d\n", n);fflush(stdout);////////////////////

  switch (state->encoding)
    {
    case ASCII_ENCODING:
      {
        ___U8 byte;

        while (rlo < rhi && n > 0)
          {
            byte = rbuf[rlo++];
            if (byte > 0x7f)
              {
                rbuf[--rlo] = byte & 0x7f;
                *buf++ = UNICODE_ESCAPE;
              }
            else
              *buf++ = byte;
            n--;
          }
        break;
      }

    default: /* in case state->encoding is set incorrectly */
    case LATIN1_ENCODING:
      while (rlo < rhi && n > 0)
        {
          *buf++ = rbuf[rlo++];
          n--;
        }
      break;

    case UTF8_ENCODING:
      {
        ___U8 byte;
        ___UCS4 c;

        while (rlo < rhi && n > 0)
          {
            int start = rlo; /* in case we need to backtrack */

            byte = rbuf[rlo++];

            if (byte <= 0x7f)
              c = byte; /* ASCII character */
            else if (byte <= 0xbf || byte > 0xfd)
              c = TTY_CHAR_MAX+1; /* illegal first byte of a UTF-8 */
            else
              {
                ___U8 next;
                ___UCS4 range;

                c = byte; /* upper bits are removed later */
                range = 1<<6;

                loop:
                if (rlo >= rhi)
                  {
                    /*
                     * Not enough bytes to form a UTF-8 encoded
                     * character, leave them there for later.
                     */
                    rlo = start;
                    break; /* stop short */
                  }
                next = rbuf[rlo++];
                if (next <= 0x7f || next > 0xbf)
                  {
                    /*
                     * The byte is not a legal "after-first" byte.
                     * Leave it there so that we will try to process it
                     * as a first byte later.
                     */
                    rlo--;
                    c = TTY_CHAR_MAX+1;
                  }
                else
                  {
                    c = (c << 6) + (next & 0x3f);
                    byte <<= 1;
                    range <<= 5;
                    if (byte & 0x40)
                      goto loop;
                    c &= range-1;
                  }
              }

            if (c > TTY_CHAR_MAX)
              {
#ifdef LINEEDIT_REPORT_CHAR_ENCODING_ERRORS
                if (n == len)
                  {
                    state->rlo = rlo; /* step over the faulty character */
                    return ___FIX(___CTOS_UTF8STRING_ERR);
                  }
                rlo = start; /* error will be reported later */
                break; /* stop short */
#else
                c = UNICODE_REPLACEMENT; /* Unicode replacement character */
#endif
              }

            *buf++ = c;
            n--;
          }
        break;
      }

    case UCS2LE_ENCODING:
      while (rlo+1 < rhi && n > 0)
        {
          ___UCS4 c = rbuf[rlo++];
          c += rbuf[rlo++] << 8;
          *buf++ = c;
          n--;
        }
      break;

    case UCS2BE_ENCODING:
      while (rlo+1 < rhi && n > 0)
        {
          ___UCS4 c = rbuf[rlo++];
          c = (c << 8) + rbuf[rlo++];
          *buf++ = c;
          n--;
        }
      break;

    case UCS4LE_ENCODING:
      while (rlo+3 < rhi && n > 0)
        {
          ___UCS4 c = rbuf[rlo++];
          c += rbuf[rlo++] << 8;
          c += rbuf[rlo++] << 16;
          c += rbuf[rlo++] << 24;
          *buf++ = c;
          n--;
        }
      break;

    case UCS4BE_ENCODING:
      while (rlo+3 < rhi && n > 0)
        {
          ___UCS4 c = rbuf[rlo++];
          c = (c << 8) + rbuf[rlo++];
          c = (c << 8) + rbuf[rlo++];
          c = (c << 8) + rbuf[rlo++];
          *buf++ = c;
          n--;
        }
      break;
    }

  state->rlo = rlo;

  *len_done = len - n;

  return ___FIX(___NO_ERR);
}

___HIDDEN ___SCMOBJ tty_rbuf_fill
   ___P((tty_state *state,
         ___stream_index *len_done),
        (state,
         len_done)
tty_state *state;
___stream_index *len_done;)
{
  int rlo = state->rlo;
  int rhi = state->rhi;
  int end = TTY_RBUF_SIZE;
  ___U8 *rbuf = state->rbuf;
  ___SCMOBJ e;

  if (rlo != 0)
    {
      int i = 0;
      while (rlo < rhi)
        rbuf[i++] = rbuf[rlo++];
      rlo = 0;
      rhi = i;
      state->rlo = rlo;
      state->rhi = rhi;
    }

  if ((e = tty_read_os_specific (state, &rbuf[rhi], end-rhi, len_done))
      == ___FIX(___NO_ERR))
    state->rhi = rhi + *len_done;

  ////  printf ("tty_read_os_specific returned e=%d\n",e);fflush(stdout);////////////

  return e;
}

___HIDDEN ___SCMOBJ tty_read_noncanonical
   ___P((tty_state *state,
         tty_char *buf,
         ___stream_index len,
         ___stream_index *len_done),
        (state,
         buf,
         len,
         len_done)
tty_state *state;
tty_char *buf;
___stream_index len;
___stream_index *len_done;)
{
  ___SCMOBJ e;

#ifdef USE_LINEEDIT

  lineedit_state *les = state->les;
  tty_char *p = les->paste_text;

  /*
   * Try to get characters from pasted text first.
   */

  if (les->paste_cancel)
    {
      if (p != NULL)
        {
          ___free_mem (p);
          p = NULL;
          les->paste_text = NULL;
        }
      les->paste_cancel = 0;
    }

  if (p != NULL)
    {
      int start = les->paste_index;
      int end = start + len;
      int i = start;
      tty_char c;

      while (i < end && (c = p[i]) != UNICODE_NUL)
        {
          *buf++ = c;
          i++;
        }

      if (i < end)
        {
          /* reached null character at end of pasted text */

          ___free_mem (p);
          les->paste_text = NULL;
        }

      if (i > start)
        {
          les->paste_index = i;
          *len_done = i - start;
          return ___FIX(___NO_ERR);
        }
    }

#endif

  /*
   * Get characters from rbuf.
   */

  while ((e = tty_read_from_rbuf (state, buf, len, len_done))
         == ___FIX(___NO_ERR) &&
         *len_done == 0 &&
         (e = tty_rbuf_fill (state, len_done))
         == ___FIX(___NO_ERR) &&
         *len_done > 0) ;

  return e;
}

___HIDDEN ___SCMOBJ tty_read
   ___P((tty *t,
         tty_char *buf,
         ___stream_index len,
         ___stream_index *len_done),
        (t,
         buf,
         len,
         len_done)
tty *t;
tty_char *buf;
___stream_index len;
___stream_index *len_done;)
{
  ___SCMOBJ e;
  tty_state *state;

  ////////printf ("tty_read len=%d\n", len);fflush(stdout);////////////////////

  if ((e = tty_force_open (t)) != ___FIX(___NO_ERR))
    return e;

  state = t->state;

#ifdef USE_LINEEDIT

  if (state->input_mode & LINEEDIT_INPUT_MODE)
    return lineedit_read (state->les, buf, len, len_done);

#endif

  return tty_read_noncanonical (state, buf, len, len_done);
}


___HIDDEN ___SCMOBJ tty_write_to_wbuf
   ___P((tty_state *state,
         tty_char *buf,
         ___stream_index len,
         ___stream_index *len_done),
        (state,
         buf,
         len,
         len_done)
tty_state *state;
tty_char *buf;
___stream_index len;
___stream_index *len_done;)
{
  int whi = state->whi;
  int end = TTY_WBUF_SIZE;
  ___U8 *wbuf = state->wbuf;
  int n = len;

  switch (state->encoding)
    {
    default: /* in case state->encoding is set incorrectly */
    case ASCII_ENCODING:
    case LATIN1_ENCODING:
      while (whi < end && n > 0)
        {
          ___UCS4 c = *buf++;
          if (c > 0xff)
            {
#ifdef LINEEDIT_REPORT_CHAR_ENCODING_ERRORS
              if (n == len)
                {
                  /* The caller must step over the faulty character. */
                  return ___FIX(___STOC_LATIN1STRING_ERR);
                }
              break; /* stop short */
#else
              c = UNICODE_BELL;
#endif
            }
          wbuf[whi++] = c;
          n--;
        }
      break;

    case UTF8_ENCODING:
      while (whi < end && n > 0)
        {
          ___UCS4 c = *buf++;
          ___U8 *p = &wbuf[whi];
          int bytes = 1;

          if (c <= 0x7f)
            *p = c; /* ASCII character */
          else
            {
              ___UCS4 x = c;
              if      (x <= 0x7ff)     bytes = 2;
              else if (x <= 0xffff)    bytes = 3;
              else if (x <= 0x1fffff)  bytes = 4;
              else if (x <= 0x3ffffff) bytes = 5;
              else                     bytes = 6;
              if (whi+bytes > end)
                break; /* stop short */
              p += bytes;
              switch (bytes)
                {
                case 6:  *--p = 0x80+(x&0x3f); x >>= 6;
                case 5:  *--p = 0x80+(x&0x3f); x >>= 6;
                case 4:  *--p = 0x80+(x&0x3f); x >>= 6;
                case 3:  *--p = 0x80+(x&0x3f); x >>= 6;
                default: *--p = 0x80+(x&0x3f); x >>= 6;
                }
              *--p = 0xff - (0xff>>bytes) + x;
            }
          whi += bytes;
          n--;
        }
      break;

    case UCS2LE_ENCODING:
      while (whi+1 < end && n > 0)
        {
          ___UCS4 c = *buf++;
          wbuf[whi++] = c & 0xff;
          wbuf[whi++] = (c >> 8) & 0xff;
          n--;
        }
      break;

    case UCS2BE_ENCODING:
      while (whi+1 < end && n > 0)
        {
          ___UCS4 c = *buf++;
          wbuf[whi++] = (c >> 8) & 0xff;
          wbuf[whi++] = c & 0xff;
          n--;
        }
      break;

    case UCS4LE_ENCODING:
      while (whi+3 < end && n > 0)
        {
          ___UCS4 c = *buf++;
          wbuf[whi++] = c & 0xff;
          wbuf[whi++] = (c >> 8) & 0xff;
          wbuf[whi++] = (c >> 16) & 0xff;
          wbuf[whi++] = (c >> 24) & 0xff;
          n--;
        }
      break;

    case UCS4BE_ENCODING:
      while (whi+3 < end && n > 0)
        {
          ___UCS4 c = *buf++;
          wbuf[whi++] = (c >> 24) & 0xff;
          wbuf[whi++] = (c >> 16) & 0xff;
          wbuf[whi++] = (c >> 8) & 0xff;
          wbuf[whi++] = c & 0xff;
          n--;
        }
      break;
    }

  state->whi = whi;

  *len_done = len - n;

  return ___FIX(___NO_ERR);
}

___HIDDEN void tty_wbuf_drain
   ___P((tty_state *state),
        (state)
tty_state *state;)
{
  int wlo = state->wlo;
  int whi = state->whi;
  int end = TTY_WBUF_SIZE;
  ___U8 *wbuf = state->wbuf;
  ___stream_index len_done;
  ___SCMOBJ e;

  if (wlo < whi)
    {
      if ((e = tty_write_os_specific (state, &wbuf[wlo], whi-wlo, &len_done))
          < ___FIX(___NO_ERR))
        {
          state->werr = e;
          return;
        }

      wlo += len_done;

      if (wlo >= whi || (wlo<<1) >= end)
        {
          /* shift buffer if first half is now empty */
          int i = 0;
          while (wlo < whi)
            wbuf[i++] = wbuf[wlo++];
          wlo = 0;
          state->whi = i;
        }

      state->wlo = wlo;
    }
}

___HIDDEN ___SCMOBJ tty_write
   ___P((tty *t,
         tty_char *buf,
         ___stream_index len,
         ___stream_index *len_done),
        (t,
         buf,
         len,
         len_done)
tty *t;
tty_char *buf;
___stream_index len;
___stream_index *len_done;)
{
  ___SCMOBJ e;
  tty_state *state;

  ///////printf ("tty_write len=%d\n", len);fflush(stdout);////////////////////

  //////////////////////////////////
#if 1

#ifdef USE_LINEEDIT
  lineedit_state *les;
#endif

  if ((e = tty_force_open (t)) != ___FIX(___NO_ERR))
    return e;

  state = t->state;

#ifdef USE_LINEEDIT

  les = state->les;

  if ((e = lineedit_end_paren_balance (les)) != ___FIX(___NO_ERR))
    return e;

  if ((e = lineedit_output_set_text_attrs (les, les->output_attrs))
      != ___FIX(___NO_ERR))
    return e;

  if ((e = lineedit_output_drain (les)) != ___FIX(___NO_ERR))
    return e;

#endif

  if ((e = state->werr) != ___FIX(___NO_ERR))
    {
      state->werr = ___FIX(___NO_ERR);
      return e;
    }

  e = tty_write_to_wbuf (state, buf, len, len_done);

  tty_wbuf_drain (state); /* error code returned in state->werr */

#ifdef USE_LINEEDIT

  if (e == ___FIX(___NO_ERR))
    {
      int pos = les->current.screen_cursor + les->current.cursor_to_point;
      lineedit_track_cursor_movement (les, buf, *len_done);
      les->current.cursor_to_point = pos - les->current.screen_cursor;
    }

#endif

#else

  printf ("%c", *buf >> 8); fflush(stdout);
  *len_done = 1;
  e = ___FIX(___NO_ERR);

#endif

  return e;
}

___HIDDEN ___SCMOBJ tty_module_setup
   ___P((void (*user_interrupt_handler) ___PVOID),
        (user_interrupt_handler)
void (*user_interrupt_handler) ___PVOID;)
{
  ___SCMOBJ e;

  //  printf ("tty_module_setup\n");/////////////////////////////

  tty_mod.list = NULL;
  tty_mod.user_interrupt_handler = user_interrupt_handler;

  if ((e = tty_module_setup_os_specific ()) == ___FIX(___NO_ERR))
    {
#ifdef USE_LINEEDIT
      if ((e = lineedit_module_setup ()) == ___FIX(___NO_ERR))
        {
#endif
          tty_setup_console (&tty_mod.console);
          return ___FIX(___NO_ERR);
#ifdef USE_LINEEDIT
        }
#endif
      tty_module_cleanup_os_specific ();
    }

  return e;
}

___HIDDEN void tty_module_cleanup ___PVOID
{
  tty_cleanup (&tty_mod.console); /* ignore error */

#ifdef USE_LINEEDIT
  lineedit_module_cleanup ();
#endif

  tty_module_cleanup_os_specific ();
}

#endif

/*---------------------------------------------------------------------------*/

#if 0

void terminal_setup ___PVOID
{
  tty_module_setup (NULL); /* ignore error */
}

void terminal_cleanup ___PVOID
{
  tty_module_cleanup ();
}

tty *terminal_open ___P((int fd),(fd)
int fd;)
{
  tty *t;
  ___SCMOBJ e;

 if (!isatty (fd))
   return NULL;

 t = ___CAST(tty*,
             ___alloc_mem (sizeof (tty)));

  if (t != NULL)
    {
      int fd2 = dup (fd);

#if 0
      int fl;

      fl = fcntl (fd2, F_GETFL);
      fcntl (fd2, F_SETFL, fl | O_NONBLOCK); /* set nonblocking mode */
#endif

      if ((e = tty_setup_os_specific (t, fd2)) != ___FIX(___NO_ERR))
        {
          ___free_mem (t);
          t = NULL;
        }
      else
        {
          if ((e = tty_set_terminal_type
                     (t,
                      ___getenv (___CONSOLE_TERMINAL_TYPE_ENV_VAR),
                      DEFAULT_EMACS_BINDINGS))
              != ___FIX(___NO_ERR))
            ;
          if ((e = tty_set_encoding
                     (t,
                      UTF8_ENCODING))
              != ___FIX(___NO_ERR))
            ;
        }
    }

  return t;
}

void terminal_close ___P((tty *t),(t)
tty *t;)
{
  if (t != NULL)
    tty_cleanup (t); /* ignore error */
}

___SCMOBJ ___io_set_terminal_type
   ___P((___STREAM *stream, char *term_type, ___BOOL emacs_bindings),
        (stream, term_type, emacs_bindings)
___STREAM *stream;
char *term_type;
___BOOL emacs_bindings;)
{
  FILE *stdio_stream = stream->stdio_stream;

  if (stdio_stream == 0) /* stream is closed? */
    return ___VOID;

  if (stream->dummy1 != NULL)
    {
      ___SCMOBJ e;
      if ((e = tty_set_terminal_type (stream->dummy1, term_type, emacs_bindings))
          != ___FIX(___NO_ERR))
        ;
    }

  return ___VOID;
}

___SCMOBJ ___io_set_encoding
   ___P((___STREAM *stream, int encoding),
        (stream, encoding)
___STREAM *stream;
int encoding;)
{
  FILE *stdio_stream = stream->stdio_stream;

  if (stdio_stream == 0) /* stream is closed? */
    return ___VOID;

  if (stream->dummy1 != NULL)
    {
      ___SCMOBJ e;
      if ((e = tty_set_encoding (stream->dummy1, encoding))
          != ___FIX(___NO_ERR))
        ;
    }

  return ___VOID;
}

___SCMOBJ ___io_set_text_attributes
   ___P((___STREAM *stream, int input, int output),
        (stream, input, output)
___STREAM *stream;
int input;
int output;)
{
  FILE *stdio_stream = stream->stdio_stream;

  if (stdio_stream == 0) /* stream is closed? */
    return ___VOID;

  if (stream->dummy1 != NULL)
    {
      ___SCMOBJ e;
      if ((e = tty_set_text_attrs (stream->dummy1, input, output))
          != ___FIX(___NO_ERR))
        ;
    }

  return ___VOID;
}

___SCMOBJ ___io_set_max_history_length
   ___P((___STREAM *stream, int max_history_length),
        (stream, max_history_length)
___STREAM *stream;
int max_history_length;)
{
  FILE *stdio_stream = stream->stdio_stream;

  if (stdio_stream == 0) /* stream is closed? */
    return ___VOID;

  if (stream->dummy1 != NULL)
    {
      ___SCMOBJ e;
      if ((e = tty_set_max_history_length (stream->dummy1, max_history_length))
          != ___FIX(___NO_ERR))
        ;
    }

  return ___VOID;
}

___SCMOBJ ___io_set_paren_balance_duration
   ___P((___STREAM *stream, int nsecs),
        (stream, nsecs)
___STREAM *stream;
int nsecs;)
{
  FILE *stdio_stream = stream->stdio_stream;

  if (stdio_stream == 0) /* stream is closed? */
    return ___VOID;

  if (stream->dummy1 != NULL)
    {
      ___SCMOBJ e;
      if (nsecs >= 0)
        if ((e = tty_set_paren_balance_duration (stream->dummy1, nsecs))
            != ___FIX(___NO_ERR))
          ;
    }

  return ___VOID;
}

___SCMOBJ ___io_set_input_mode
   ___P((___STREAM *stream, int input_mode),
        (stream, input_mode)
___STREAM *stream;
int input_mode;)
{
  FILE *stdio_stream = stream->stdio_stream;

  if (stdio_stream == 0) /* stream is closed? */
    return ___VOID;

  if (stream->dummy1 != NULL)
    {
      ___SCMOBJ e;
      if ((e = tty_set_input_mode (stream->dummy1, input_mode))
          != ___FIX(___NO_ERR))
        ;
    }

  return ___VOID;
}

___SCMOBJ terminal_get ___P((tty *t),(t)
tty *t;)
{
  tty_char c;
  ___stream_index len_done;
  ___SCMOBJ e;
xxxxxxxxxxxxxxxxxxxxxx
  while ((e = tty_read (t, &c, 1, &len_done))
         == ___ERR_CODE_EAGAIN) ;

  if (e == ___FIX(0))
    return ___EOF;
  else if (e == ___FIX(1))
    return ___CHR(c);
  else
    return n;
}

___SCMOBJ terminal_put ___P((tty *t, ___UCS4 c),(t, c)
tty *t;
___UCS4 c;)
{
  tty_char ch = c;
  ___SCMOBJ n;

xxxxxxxxxxxxxxxxxxxxxxxxxx
  do
    {
      n = tty_write (t, &ch, 1);
    } while (n == ___ERR_CODE_EAGAIN ||
             n == ___FIX(0));

  if (n == ___FIX(1))
    return ___FAL;

  return n;
}

#endif

/*================================================================*/

/*---------------------------------------------------------------------------*/

/* Tty stream device. */

typedef struct ___device_tty_struct
  {
    ___device_stream base;
    tty t;
  } ___device_tty;

typedef struct ___device_tty_vtbl_struct
  {
    ___device_stream_vtbl base;
  } ___device_tty_vtbl;

extern ___SCMOBJ ___device_tty_setup/*************************************/
   ___P((___device_tty **dev,
         ___device_group *dgroup,
         int fd,
         int direction),
        ());

___HIDDEN int ___device_tty_kind
   ___P((___device *self),
        (self)
___device *self;)
{
  return ___TTY_DEVICE_KIND;
}

___HIDDEN ___SCMOBJ ___device_tty_close_virt
   ___P((___device *self,
         int direction),
        (self,
         direction)
___device *self;
int direction;)
{
  ___device_tty *d = ___CAST(___device_tty*,self);
  int is_not_closed = 0;

  if (d->base.base.read_stage != ___STAGE_CLOSED)
    is_not_closed |= ___DIRECTION_RD;

  if (d->base.base.write_stage != ___STAGE_CLOSED)
    is_not_closed |= ___DIRECTION_WR;

  if (is_not_closed == 0)
    return ___FIX(___NO_ERR);

  if ((is_not_closed & ~direction) == 0)
    {
      /* Close tty when both sides are closed. */

      ___SCMOBJ e;

      d->base.base.read_stage = ___STAGE_CLOSED; /* avoid multiple closes */
      d->base.base.write_stage = ___STAGE_CLOSED;

      if ((e = tty_cleanup (&d->t)) != ___FIX(___NO_ERR))
        return e;
    }
  else if (direction & ___DIRECTION_RD)
    d->base.base.read_stage = ___STAGE_CLOSED;
  else if (direction & ___DIRECTION_WR)
    d->base.base.write_stage = ___STAGE_CLOSED;

  return ___FIX(___NO_ERR);
}

___HIDDEN ___SCMOBJ ___device_tty_select_raw_virt
   ___P((___device_stream *self,
         ___BOOL for_writing,
         int i,
         int pass,
         ___device_select_state *state),
        (self,
         for_writing,
         i,
         pass,
         state)
___device_stream *self;
___BOOL for_writing;
int i;
int pass;
___device_select_state *state;)
{
  ___device_tty *d = ___CAST(___device_tty*,self);

  if ((for_writing ? d->base.base.write_stage : d->base.base.read_stage)
      != ___STAGE_OPEN)
    return ___FIX(___CLOSED_DEVICE_ERR);

  if (pass == ___SELECT_PASS_1)
    {
      if (d->t.state != NULL)
        {
#ifdef USE_POSIX
          device_select_add_fd (state, d->t.state->fd, for_writing);
#endif

#ifdef USE_LINEEDIT

          if (!for_writing)
            {
              lineedit_state *les = d->t.state->les;

              if (lineedit_read_ready (les))
                state->timeout = ___time_neg_infinity;
              else if (les->current.paren_balance_in_progress &&
                       ___time_less (les->current.paren_balance_end,
                                     state->timeout))
                state->timeout = les->current.paren_balance_end;
            }

#endif
        }
      return ___FIX(___SELECT_SETUP_DONE);
    }

  /* pass == ___SELECT_PASS_CHECK */

  if (d->t.state != NULL)
    {
      if (for_writing)
        {
#ifdef USE_POSIX
          if (FD_ISSET(d->t.state->fd, &state->writefds))
            state->devs[i] = NULL;
#endif

#ifdef USE_WIN32
          state->devs[i] = NULL;/////////////////////
#endif
        }
      else
        {
#ifdef USE_POSIX
          if (FD_ISSET(d->t.state->fd, &state->readfds))
            state->devs[i] = NULL;
#endif

#ifdef USE_WIN32
          state->devs[i] = NULL;/////////////////////
#endif

#ifdef USE_LINEEDIT

          {
            lineedit_state *les = d->t.state->les;

            if (lineedit_read_ready (les) ||
                (les->current.paren_balance_in_progress &&
                 state->timeout_reached &&
                 ___time_equal (les->current.paren_balance_end,
                                state->timeout)))
              state->devs[i] = NULL;
          }

#endif
        }
    }

  return ___FIX(___NO_ERR);
}

___HIDDEN ___SCMOBJ ___device_tty_release_raw_virt
   ___P((___device_stream *self),
        (self)
___device_stream *self;)
{
  return ___FIX(___NO_ERR);
}

___HIDDEN ___SCMOBJ ___device_tty_flush_write_raw_virt
   ___P((___device_stream *self),
        (self)
___device_stream *self;)
{
  ___device_tty *d = ___CAST(___device_tty*,self);

  if (d->base.base.write_stage == ___STAGE_OPEN)
    {
      /*************************/
    }

  return ___FIX(___NO_ERR);
}

___HIDDEN ___SCMOBJ ___device_tty_seek_raw_virt
   ___P((___device_stream *self,
         ___stream_index pos,
         int whence),
        (self,
         pos,
         whence)
___device_stream *self;
___stream_index pos;
int whence;)
{
  return ___FIX(___UNKNOWN_ERR);
}

___HIDDEN ___SCMOBJ ___device_tty_read_raw_virt
   ___P((___device_stream *self,
         ___U8 *buf,
         ___stream_index len,
         ___stream_index *len_done),
        (self,
         buf,
         len,
         len_done)
___device_stream *self;
___U8 *buf;
___stream_index len;
___stream_index *len_done;)
{
  ___device_tty *d = ___CAST(___device_tty*,self);
  ___SCMOBJ e;

  if (d->base.base.read_stage != ___STAGE_OPEN)
    return ___FIX(___CLOSED_DEVICE_ERR);

  ////printf ("before tty_read\n");fflush(stdout);//////////////////////

  e = tty_read (&d->t, ___CAST(tty_char*,buf), len>>1, len_done);/***********/

  /////printf ("after tty_read --> e=%d len_done=%d\n", e, *len_done);fflush(stdout);//////////////////////

  if (e == ___FIX(___NO_ERR))
    *len_done <<= 1;/********************/

  return e;
}

___HIDDEN ___SCMOBJ ___device_tty_write_raw_virt
   ___P((___device_stream *self,
         ___U8 *buf,
         ___stream_index len,
         ___stream_index *len_done),
        (self,
         buf,
         len,
         len_done)
___device_stream *self;
___U8 *buf;
___stream_index len;
___stream_index *len_done;)
{
  ___device_tty *d = ___CAST(___device_tty*,self);
  ___SCMOBJ e;

  if (d->base.base.write_stage != ___STAGE_OPEN)
    return ___FIX(___CLOSED_DEVICE_ERR);

  e = tty_write (&d->t, ___CAST(tty_char*,buf), len>>1, len_done);/***********/

  if (e == ___FIX(___NO_ERR))
    *len_done <<= 1;/********************/

  return e;
}

___HIDDEN ___SCMOBJ ___device_tty_width_virt
   ___P((___device_stream *self),
        (self)
___device_stream *self;)
{
  return ___FIX(80);
}

___HIDDEN ___SCMOBJ ___device_tty_default_options_virt
   ___P((___device_stream *self),
        (self)
___device_stream *self;)
{
#define TTY_CHAR_ENCODING ___CHAR_ENCODING_LATIN1

  return ___STREAM_DEFAULT_OPTIONS(TTY_CHAR_ENCODING,
                                   ___EOL_ENCODING_LF,
                                   ___NO_BUFFERING,
                                   TTY_CHAR_ENCODING,
                                   ___EOL_ENCODING_LF,
                                   ___NO_BUFFERING);
}

___HIDDEN ___device_tty_vtbl ___device_tty_table =
{
  {
    {
      ___device_tty_kind,
      ___device_stream_select_virt,
      ___device_stream_release_virt,
      ___device_stream_flush_write_virt,
      ___device_tty_close_virt
    },
    ___device_tty_select_raw_virt,
    ___device_tty_release_raw_virt,
    ___device_tty_flush_write_raw_virt,
    ___device_tty_seek_raw_virt,
    ___device_tty_read_raw_virt,
    ___device_tty_write_raw_virt,
    ___device_tty_width_virt,
    ___device_tty_default_options_virt
  }
};


#ifdef USE_POSIX

___SCMOBJ ___device_tty_setup_from_fd
   ___P((___device_tty **dev,
         ___device_group *dgroup,
         int fd,
         int direction),
        (dev,
         dgroup,
         fd,
         direction)
___device_tty **dev;
___device_group *dgroup;
int fd;
int direction;)
{
  ___device_tty *d;
  ___SCMOBJ e;

  d = ___CAST(___device_tty*,
              ___alloc_mem (sizeof (___device_tty)));

  if (d == NULL)
    {
      close (fd); /* ignore error */
      return ___FIX(___HEAP_OVERFLOW_ERR);
    }

  d->base.base.vtbl = &___device_tty_table;

#if 0
  {////////////////////////////////////////////
    int fl;

    fl = fcntl (fd, F_GETFL);
    fcntl (fd, F_SETFL, fl | O_NONBLOCK); /* set nonblocking mode */
  }
#endif

  if ((e = tty_setup_os_specific (&d->t, fd)) != ___FIX(___NO_ERR))
    {
      close (fd); /* ignore error */
      ___free_mem (d);
      return e;
    }

  if ((e = tty_set_terminal_type
             (&d->t,
              ___getenv (___CONSOLE_TERMINAL_TYPE_ENV_VAR),
              DEFAULT_EMACS_BINDINGS))
      != ___FIX(___NO_ERR))
    printf ("error\n");/************/

#if 0
  if ((e = tty_set_encoding
             (&d->t,
              UTF8_ENCODING))
      != ___FIX(___NO_ERR))
    ;/*******************/
  if ((e = tty_set_input_mode (&d->t, CBREAK_TERM_MODE))
      != ___FIX(___NO_ERR))
    ;/*******************/
#endif

  *dev = d;

  return ___device_stream_setup
           (&d->base,
            dgroup,
            direction,
            0);
}

#endif


#ifdef USE_WIN32

___SCMOBJ ___device_tty_setup_from_console
   ___P((___device_tty **dev,
         ___device_group *dgroup,
         int direction),
        (dev,
         dgroup,
         direction)
___device_tty **dev;
___device_group *dgroup;
int direction;)
{
  ___device_tty *d;
  ___SCMOBJ e;

  d = ___CAST(___device_tty*,
              ___alloc_mem (sizeof (___device_tty)));

  /////printf ("device_tty_setup_from_console\n");fflush(stdout);/////////////////////

  if (d == NULL)
    return ___FIX(___HEAP_OVERFLOW_ERR);

  d->base.base.vtbl = &___device_tty_table;

#if 0
  {////////////////////////////////////////////
    int fl;

    fl = fcntl (fd, F_GETFL);
    fcntl (fd, F_SETFL, fl | O_NONBLOCK); /* set nonblocking mode */
  }
#endif

  if ((e = tty_setup_os_specific (&d->t)) != ___FIX(___NO_ERR))
    {
      ___free_mem (d);
      return e;
    }

#if 0
  ////////////////////////////////

  if ((e = tty_set_terminal_type
             (&d->t,
              ___getenv (___CONSOLE_TERMINAL_TYPE_ENV_VAR),
              DEFAULT_EMACS_BINDINGS))
      != ___FIX(___NO_ERR))
    printf ("error\n");/************/

#if 0
  if ((e = tty_set_encoding
             (&d->t,
              UTF8_ENCODING))
      != ___FIX(___NO_ERR))
    ;/*******************/
  if ((e = tty_set_input_mode (&d->t, CBREAK_TERM_MODE))
      != ___FIX(___NO_ERR))
    ;/*******************/
#endif
#endif

  *dev = d;

  return ___device_stream_setup
           (&d->base,
            dgroup,
            direction,
            0);
}

#endif


___SCMOBJ ___device_tty_setup_console
   ___P((___device_tty **dev,
         ___device_group *dgroup,
         int direction),
        (dev,
         dgroup,
         direction)
___device_tty **dev;
___device_group *dgroup;
int direction;)
{
  ___device_tty *d;
  ___SCMOBJ e;

  d = ___CAST(___device_tty*,
              ___alloc_mem (sizeof (___device_tty)));

  if (d == NULL)
    return ___FIX(___HEAP_OVERFLOW_ERR);

  d->base.base.vtbl = &___device_tty_table;

  tty_setup_console (&d->t);
  ///////////////////////////////////

  *dev = d;

  return ___device_stream_setup
           (&d->base,
            dgroup,
            direction,
            0);
}


/*---------------------------------------------------------------------------*/

#ifdef USE_POSIX

/* Process stream device */

typedef struct ___device_process_struct
  {
    ___device_stream base;
    pid_t pid;     /* pid of the process */
    int fd_stdin;  /* file descriptor corresponding to process' stdin */
    int fd_stdout; /* file descriptor corresponding to process' stdout */
  } ___device_process;

typedef struct ___device_process_vtbl_struct
  {
    ___device_stream_vtbl base;
  } ___device_process_vtbl;

___HIDDEN int ___device_process_kind
   ___P((___device *self),
        (self)
___device *self;)
{
  return ___PROCESS_DEVICE_KIND;
}

___HIDDEN ___SCMOBJ ___device_process_close_virt
   ___P((___device *self,
         int direction),
        (self,
         direction)
___device *self;
int direction;)
{
  ___device_process *d = ___CAST(___device_process*,self);
  int is_not_closed = 0;

  if (d->base.base.read_stage != ___STAGE_CLOSED)
    is_not_closed |= ___DIRECTION_RD;

  if (d->base.base.write_stage != ___STAGE_CLOSED)
    is_not_closed |= ___DIRECTION_WR;

  if (is_not_closed == 0)
    return ___FIX(___NO_ERR);

  if (direction & ___DIRECTION_RD)
    {
      /* Close process' stdout */

      d->base.base.read_stage = ___STAGE_CLOSED;

      if (close (d->fd_stdout) < 0)
        return err_code_from_errno ();
    }

  if (direction & ___DIRECTION_WR)
    {
      /* Close process' stdin */

      d->base.base.write_stage = ___STAGE_CLOSED;

      if (close (d->fd_stdin) < 0)
        return err_code_from_errno ();
    }

  return ___FIX(___NO_ERR);
}

___HIDDEN ___SCMOBJ ___device_process_select_raw_virt
   ___P((___device_stream *self,
         ___BOOL for_writing,
         int i,
         int pass,
         ___device_select_state *state),
        (self,
         for_writing,
         i,
         pass,
         state)
___device_stream *self;
___BOOL for_writing;
int i;
int pass;
___device_select_state *state;)
{
  ___device_process *d = ___CAST(___device_process*,self);

  if ((for_writing ? d->base.base.write_stage : d->base.base.read_stage)
      != ___STAGE_OPEN)
    return ___FIX(___CLOSED_DEVICE_ERR);

  if (pass == ___SELECT_PASS_1)
    {
      device_select_add_fd
        (state,
         for_writing ? d->fd_stdin : d->fd_stdout,
         for_writing);
      return ___FIX(___SELECT_SETUP_DONE);
    }

  /* pass == ___SELECT_PASS_CHECK */

  if (for_writing)
    {
      if (FD_ISSET(d->fd_stdin, &state->writefds))
        state->devs[i] = NULL;
    }
  else
    {
      if (FD_ISSET(d->fd_stdout, &state->readfds))
        state->devs[i] = NULL;
    }

  return ___FIX(___NO_ERR);
}

___HIDDEN ___SCMOBJ ___device_process_release_raw_virt
   ___P((___device_stream *self),
        (self)
___device_stream *self;)
{
  return ___FIX(___NO_ERR);
}

___HIDDEN ___SCMOBJ ___device_process_flush_write_raw_virt
   ___P((___device_stream *self),
        (self)
___device_stream *self;)
{
  ___device_process *d = ___CAST(___device_process*,self);

  if (d->base.base.write_stage == ___STAGE_OPEN)
    {
#if 0
      if (fsync (d->fd_stdin) < 0)/************only works on disk files!!!!!!!*/
        return err_code_from_errno ();
#endif
    }

  return ___FIX(___NO_ERR);
}

___HIDDEN ___SCMOBJ ___device_process_seek_raw_virt
   ___P((___device_stream *self,
         ___stream_index pos,
         int whence),
        (self,
         pos,
         whence)
___device_stream *self;
___stream_index pos;
int whence;)
{
  return ___FIX(___UNKNOWN_ERR);
}

___HIDDEN ___SCMOBJ ___device_process_read_raw_virt
   ___P((___device_stream *self,
         ___U8 *buf,
         ___stream_index len,
         ___stream_index *len_done),
        (self,
         buf,
         len,
         len_done)
___device_stream *self;
___U8 *buf;
___stream_index len;
___stream_index *len_done;)
{
  ___device_process *d = ___CAST(___device_process*,self);
  int n;

  if (d->base.base.read_stage != ___STAGE_OPEN)
    return ___FIX(___CLOSED_DEVICE_ERR);

  if ((n = read (d->fd_stdout, buf, len)) < 0)
    {//printf ("read process returned %d\n",errno);fflush (stdout);/***********/
    //http://www.doc.ic.ac.uk/~mac/manuals/hpux-manual-pages/hpux/usr/man/man7/pty.7.html
    if (errno == EIO) errno = EAGAIN;
    return err_code_from_errno ();
    }

  *len_done = n;

  return ___FIX(___NO_ERR);
}

___HIDDEN ___SCMOBJ ___device_process_write_raw_virt
   ___P((___device_stream *self,
         ___U8 *buf,
         ___stream_index len,
         ___stream_index *len_done),
        (self,
         buf,
         len,
         len_done)
___device_stream *self;
___U8 *buf;
___stream_index len;
___stream_index *len_done;)
{
  ___device_process *d = ___CAST(___device_process*,self);
  int n;

  if (d->base.base.write_stage != ___STAGE_OPEN)
    return ___FIX(___CLOSED_DEVICE_ERR);

  n = write (d->fd_stdin, buf, len);

  if (n < 0)
    return err_code_from_errno ();

  *len_done = n;

  return ___FIX(___NO_ERR);
}

___HIDDEN ___SCMOBJ ___device_process_width_virt
   ___P((___device_stream *self),
        (self)
___device_stream *self;)
{
  return ___FIX(80);
}

___HIDDEN ___SCMOBJ ___device_process_default_options_virt
   ___P((___device_stream *self),
        (self)
___device_stream *self;)
{
  return ___STREAM_DEFAULT_OPTIONS(___CHAR_ENCODING_LATIN1,
                                   ___EOL_ENCODING_LF,
                                   ___BUFFERING,
                                   ___CHAR_ENCODING_LATIN1,
                                   ___EOL_ENCODING_LF,
                                   ___BUFFERING);
}

___HIDDEN ___device_process_vtbl ___device_process_table =
{
  {
    {
      ___device_process_kind,
      ___device_stream_select_virt,
      ___device_stream_release_virt,
      ___device_stream_flush_write_virt,
      ___device_process_close_virt
    },
    ___device_process_select_raw_virt,
    ___device_process_release_raw_virt,
    ___device_process_flush_write_raw_virt,
    ___device_process_seek_raw_virt,
    ___device_process_read_raw_virt,
    ___device_process_write_raw_virt,
    ___device_process_width_virt,
    ___device_process_default_options_virt
  }
};

___SCMOBJ ___device_process_setup_from_pid
   ___P((___device_process **dev,
         ___device_group *dgroup,
         pid_t pid,
         int fd_stdin,
         int fd_stdout,
         int direction),
        (dev,
         dgroup,
         pid,
         fd_stdin,
         fd_stdout,
         direction)
___device_process **dev;
___device_group *dgroup;
pid_t pid;
int fd_stdin;
int fd_stdout;
int direction;)
{
  ___device_process *d;

  d = ___CAST(___device_process*,
              ___alloc_mem (sizeof (___device_process)));

  if (d == NULL)
    return ___FIX(___HEAP_OVERFLOW_ERR);

  /*
   * Setup file descriptors to perform nonblocking I/O.
   */

  if (((direction & ___DIRECTION_RD) &&
       (fcntl (fd_stdout, F_SETFL, O_NONBLOCK) < 0)) ||
      ((direction & ___DIRECTION_WR) &&
       (fcntl (fd_stdin, F_SETFL, O_NONBLOCK) < 0))) /* set nonblocking mode */
    {
      ___SCMOBJ e = err_code_from_errno ();
      ___free_mem (&d->base.base);
      return e;
    }

  d->base.base.vtbl = &___device_process_table;
  d->pid = pid;
  d->fd_stdin = fd_stdin;
  d->fd_stdout = fd_stdout;

  *dev = d;

  return ___device_stream_setup
           (&d->base,
            dgroup,
            direction,
            0);
}

#endif


/*---------------------------------------------------------------------------*/

/* TCP client stream device */

#ifdef USE_POSIX
#define SOCKET_TYPE int
#define SOCKET_CALL_ERROR(s) ((s) < 0)
#define CONNECT_IN_PROGRESS (errno == EINPROGRESS)
#define CONNECT_WOULD_BLOCK (errno == EAGAIN)
#define NOT_CONNECTED(e) ((e) == ___FIX(___ERRNO_ERR(ENOTCONN)))
#define CLOSE_SOCKET(s) close (s)
#define ERR_CODE_FROM_SOCKET_CALL err_code_from_errno ()
#define IOCTL_SOCKET(s,cmd,argp) ioctl (s,cmd,argp)
#define SOCKET_LEN_TYPE socklen_t
#endif

#ifdef USE_WIN32
#define SOCKET_TYPE SOCKET
#define SOCKET_CALL_ERROR(s) ((s) == INVALID_SOCKET)
#define CONNECT_IN_PROGRESS ((WSAGetLastError () == WSAEALREADY) || \
(WSAGetLastError () == WSAEINVAL))
#define CONNECT_WOULD_BLOCK (WSAGetLastError () == WSAEWOULDBLOCK)
#define NOT_CONNECTED(e) ((e) == ___FIX(___WIN32_ERR(WSAENOTCONN)))
#define CLOSE_SOCKET(s) closesocket (s)
#define ERR_CODE_FROM_SOCKET_CALL err_code_from_WSAGetLastError ()
#define IOCTL_SOCKET(s,cmd,argp) ioctlsocket (s,cmd,argp)
#define SOCKET_LEN_TYPE int
#endif

#ifdef SHUT_RD
#define SHUTDOWN_RD SHUT_RD
#else
#ifdef SD_RECEIVE
#define SHUTDOWN_RD SD_RECEIVE
#else
#define SHUTDOWN_RD 0
#endif
#endif

#ifdef SHUT_WR
#define SHUTDOWN_WR SHUT_WR
#else
#ifdef SD_SEND
#define SHUTDOWN_WR SD_SEND
#else
#define SHUTDOWN_WR 1
#endif
#endif


typedef struct ___device_tcp_client_struct
  {
    ___device_stream base;
    SOCKET_TYPE s;
    struct sockaddr server_addr;
    SOCKET_LEN_TYPE server_addrlen;
    ___BOOL try_connect_again;
  } ___device_tcp_client;

typedef struct ___device_tcp_client_vtbl_struct
  {
    ___device_stream_vtbl base;
  } ___device_tcp_client_vtbl;

___HIDDEN int try_connect
   ___P((___device_tcp_client *dev),
        (dev)
___device_tcp_client *dev;)
{
  if (connect (dev->s, &dev->server_addr, dev->server_addrlen) == 0 ||
      CONNECT_IN_PROGRESS) /* establishing connection in background */
    {
      dev->try_connect_again = 0;
      return 0;
    }

  if (CONNECT_WOULD_BLOCK) /* connect must be tried again later */
    return 0;

  return -1;
}

___HIDDEN int ___device_tcp_client_kind
   ___P((___device *self),
        (self)
___device *self;)
{
  return ___TCP_CLIENT_DEVICE_KIND;
}


___HIDDEN ___SCMOBJ ___device_tcp_client_close_virt
   ___P((___device *self,
         int direction),
        (self,
         direction)
___device *self;
int direction;)
{
  ___device_tcp_client *d = ___CAST(___device_tcp_client*,self);
  int is_not_closed = 0;

  if (d->base.base.read_stage != ___STAGE_CLOSED)
    is_not_closed |= ___DIRECTION_RD;

  if (d->base.base.write_stage != ___STAGE_CLOSED)
    is_not_closed |= ___DIRECTION_WR;

  if (is_not_closed == 0)
    return ___FIX(___NO_ERR);

  if ((is_not_closed & ~direction) == 0)
    {
      /* Close socket when both sides are closed. */

      d->base.base.read_stage = ___STAGE_CLOSED; /* avoid multiple closes */
      d->base.base.write_stage = ___STAGE_CLOSED;

      if (CLOSE_SOCKET(d->s) != 0)
        return ERR_CODE_FROM_SOCKET_CALL;
    }
  else if (direction & ___DIRECTION_RD)
    {
      /* Shutdown receiving side. */

      if (shutdown (d->s, SHUTDOWN_RD) != 0)
        {
          ___SCMOBJ e = ERR_CODE_FROM_SOCKET_CALL;
          if (!NOT_CONNECTED(e))
            return e;
        }

      d->base.base.read_stage = ___STAGE_CLOSED;
    }
  else if (direction & ___DIRECTION_WR)
    {
      /* Shutdown sending side. */

      if (shutdown (d->s, SHUTDOWN_WR) != 0)
        {
          ___SCMOBJ e = ERR_CODE_FROM_SOCKET_CALL;
          if (!NOT_CONNECTED(e))
            return e;
        }

      d->base.base.write_stage = ___STAGE_CLOSED;
    }

  return ___FIX(___NO_ERR);
}

___HIDDEN ___SCMOBJ ___device_tcp_client_select_raw_virt
   ___P((___device_stream *self,
         ___BOOL for_writing,
         int i,
         int pass,
         ___device_select_state *state),
        (self,
         for_writing,
         i,
         pass,
         state)
___device_stream *self;
___BOOL for_writing;
int i;
int pass;
___device_select_state *state;)
{
  ___device_tcp_client *d = ___CAST(___device_tcp_client*,self);

  if ((for_writing ? d->base.base.write_stage : d->base.base.read_stage)
      != ___STAGE_OPEN)
    return ___FIX(___CLOSED_DEVICE_ERR);

  if (pass == ___SELECT_PASS_1)
    {
#ifdef USE_POSIX
      if (d->try_connect_again)
        device_select_add_relative_timeout (state, 0.2);
      else
        device_select_add_fd (state, d->s, for_writing);
#endif

#ifdef USE_WIN32
      //////////////////////
#endif

      return ___FIX(___SELECT_SETUP_DONE);
    }

  /* pass == ___SELECT_PASS_CHECK */

  if (d->try_connect_again)
    state->devs[i] = NULL;
  else
    {
#ifdef USE_POSIX
      if (for_writing)
        {
          if (FD_ISSET(d->s, &state->writefds))
            state->devs[i] = NULL;
        }
      else
        {
          if (FD_ISSET(d->s, &state->readfds))
            state->devs[i] = NULL;
        }
#endif

#ifdef USE_WIN32
      //////////////////////
#endif
    }

  return ___FIX(___NO_ERR);
}

___HIDDEN ___SCMOBJ ___device_tcp_client_release_raw_virt
   ___P((___device_stream *self),
        (self)
___device_stream *self;)
{
  return ___FIX(___NO_ERR);
}

___HIDDEN ___SCMOBJ ___device_tcp_client_flush_write_raw_virt
   ___P((___device_stream *self),
        (self)
___device_stream *self;)
{
  return ___FIX(___NO_ERR);
}

___HIDDEN ___SCMOBJ ___device_tcp_client_seek_raw_virt
   ___P((___device_stream *self,
         ___stream_index pos,
         int whence),
        (self,
         pos,
         whence)
___device_stream *self;
___stream_index pos;
int whence;)
{
  return ___FIX(___UNKNOWN_ERR);
}

___HIDDEN ___SCMOBJ ___device_tcp_client_read_raw_virt
   ___P((___device_stream *self,
         ___U8 *buf,
         ___stream_index len,
         ___stream_index *len_done),
        (self,
         buf,
         len,
         len_done)
___device_stream *self;
___U8 *buf;
___stream_index len;
___stream_index *len_done;)
{
  ___device_tcp_client *d = ___CAST(___device_tcp_client*,self);
  int n;

  if (d->base.base.read_stage != ___STAGE_OPEN)
    return ___FIX(___CLOSED_DEVICE_ERR);

  if (d->try_connect_again)
    {
      if (try_connect (d) == 0)
        {
          if (d->try_connect_again)
            return ___ERR_CODE_EAGAIN;
        }
      else
        return ERR_CODE_FROM_SOCKET_CALL;
    }

  if (SOCKET_CALL_ERROR(n = recv (d->s, buf, len, 0)))
    return ERR_CODE_FROM_SOCKET_CALL;

  //  printf("recv --> len=%d n=%d\n",len,n);fflush(stdout);///////////////////

  *len_done = n;

  return ___FIX(___NO_ERR);
}

___HIDDEN ___SCMOBJ ___device_tcp_client_write_raw_virt
   ___P((___device_stream *self,
         ___U8 *buf,
         ___stream_index len,
         ___stream_index *len_done),
        (self,
         buf,
         len,
         len_done)
___device_stream *self;
___U8 *buf;
___stream_index len;
___stream_index *len_done;)
{
  ___device_tcp_client *d = ___CAST(___device_tcp_client*,self);
  int n;

  if (d->base.base.write_stage != ___STAGE_OPEN)
    return ___FIX(___CLOSED_DEVICE_ERR);

  if (d->try_connect_again)
    {
      if (try_connect (d) == 0)
        {
          if (d->try_connect_again)
            return ___ERR_CODE_EAGAIN;
        }
      else
        return ERR_CODE_FROM_SOCKET_CALL;
    }

  if (SOCKET_CALL_ERROR(n = send (d->s, buf, len, 0)))
    return ERR_CODE_FROM_SOCKET_CALL;

  *len_done = n;

  return ___FIX(___NO_ERR);
}

___HIDDEN ___SCMOBJ ___device_tcp_client_width_virt
   ___P((___device_stream *self),
        (self)
___device_stream *self;)
{
  return ___FIX(80);
}

___HIDDEN ___SCMOBJ ___device_tcp_client_default_options_virt
   ___P((___device_stream *self),
        (self)
___device_stream *self;)
{
  return ___STREAM_DEFAULT_OPTIONS(___CHAR_ENCODING_LATIN1,
                                   ___EOL_ENCODING_LF,
                                   ___BUFFERING,
                                   ___CHAR_ENCODING_LATIN1,
                                   ___EOL_ENCODING_LF,
                                   ___BUFFERING);
}

___HIDDEN ___device_tcp_client_vtbl ___device_tcp_client_table =
{
  {
    {
      ___device_tcp_client_kind,
      ___device_stream_select_virt,
      ___device_stream_release_virt,
      ___device_stream_flush_write_virt,
      ___device_tcp_client_close_virt
    },
    ___device_tcp_client_select_raw_virt,
    ___device_tcp_client_release_raw_virt,
    ___device_tcp_client_flush_write_raw_virt,
    ___device_tcp_client_seek_raw_virt,
    ___device_tcp_client_read_raw_virt,
    ___device_tcp_client_write_raw_virt,
    ___device_tcp_client_width_virt,
    ___device_tcp_client_default_options_virt
  }
};


___HIDDEN int set_non_blocking
   ___P((SOCKET_TYPE s),
        (s)
SOCKET_TYPE s;)
{
#ifdef USE_ioctl

  unsigned long param = 1;

  return IOCTL_SOCKET(s, FIONBIO, &param);

#else

  int fl;

  if ((fl = fcntl (s, F_GETFL, 0)) >= 0)
    fl = fcntl (s, F_SETFL, fl | O_NONBLOCK);

  return fl;

#endif
}


___SCMOBJ ___device_tcp_client_setup_from_socket
   ___P((___device_tcp_client **dev,
         ___device_group *dgroup,
         SOCKET_TYPE s,
         struct sockaddr *server_addr,
         SOCKET_LEN_TYPE server_addrlen,
         ___BOOL try_connect_again,
         int direction),
        (dev,
         dgroup,
         s,
         server_addr,
         server_addrlen,
         try_connect_again,
         direction)
___device_tcp_client **dev;
___device_group *dgroup;
SOCKET_TYPE s;
struct sockaddr *server_addr;
SOCKET_LEN_TYPE server_addrlen;
___BOOL try_connect_again;
int direction;)
{
  ___SCMOBJ e;
  ___device_tcp_client *d;

  d = ___CAST(___device_tcp_client*,
              ___alloc_mem (sizeof (___device_tcp_client)));

  if (d == NULL)
    {
      CLOSE_SOCKET(s); /* ignore error */
      return ___FIX(___HEAP_OVERFLOW_ERR);
    }

  /*
   * Setup socket to perform nonblocking I/O.
   */

  if (set_non_blocking (s) != 0) /* set nonblocking mode */
    {
      e = ERR_CODE_FROM_SOCKET_CALL;
      CLOSE_SOCKET(s); /* ignore error */
      ___free_mem (&d->base.base);
      return e;
    }

  d->base.base.vtbl = &___device_tcp_client_table;
  d->s = s;
  d->server_addr = *server_addr;
  d->server_addrlen = server_addrlen;
  d->try_connect_again = try_connect_again;

  *dev = d;

  return ___device_stream_setup
           (&d->base,
            dgroup,
            direction,
            0);
}


___SCMOBJ ___device_tcp_client_setup_from_sockaddr
   ___P((___device_tcp_client **dev,
         ___device_group *dgroup,
         struct sockaddr *server_addr,
         SOCKET_LEN_TYPE server_addrlen,
         int options,
         int direction),
        (dev,
         dgroup,
         server_addr,
         server_addrlen,
         options,
         direction)
___device_tcp_client **dev;
___device_group *dgroup;
struct sockaddr *server_addr;
SOCKET_LEN_TYPE server_addrlen;
int options;
int direction;)
{
  ___SCMOBJ e;
  SOCKET_TYPE s;
  int keepalive_flag = (options & 1) != 0;
  int no_coalesce_flag = (options & 2) == 0;
  ___device_tcp_client *d;

#ifdef USE_POSIX
#define      TCP_NODELAY      1 //////////////////////////
#endif

  //  printf("___device_tcp_client_setup_from_sockaddr\n");fflush(stdout);////////////////////

  if (SOCKET_CALL_ERROR(s = socket (AF_INET, SOCK_STREAM, 0)))
    return ERR_CODE_FROM_SOCKET_CALL;

  if (setsockopt (s, /* keep connection alive or not */
                  SOL_SOCKET,
                  SO_KEEPALIVE,
                  ___CAST(void*,&keepalive_flag),
                  sizeof (keepalive_flag)) != 0 ||
      setsockopt (s, /* enable or disable packet coalescing algorithm */
                  IPPROTO_TCP,
                  TCP_NODELAY,
                  ___CAST(void*,&no_coalesce_flag),
                  sizeof (no_coalesce_flag)) != 0)
    {
      e = ERR_CODE_FROM_SOCKET_CALL;
      CLOSE_SOCKET(s); /* ignore error */
      return e;
    }

  if ((e = ___device_tcp_client_setup_from_socket
             (&d,
              dgroup,
              s,
              server_addr,
              server_addrlen,
              1,
              direction))
      != ___FIX(___NO_ERR))
    return e;

  *dev = d;

  if (try_connect (d) != 0)
    {
      e = ERR_CODE_FROM_SOCKET_CALL;
      ___device_cleanup (&d->base.base); /* ignore error */
      return e;
    }

  return ___FIX(___NO_ERR);
}


/*---------------------------------------------------------------------------*/

/* TCP server device. */

typedef struct ___device_tcp_server_struct
  {
    ___device base;
    SOCKET_TYPE s;
  } ___device_tcp_server;

typedef struct ___device_tcp_server_vtbl_struct
  {
    ___device_vtbl base;
  } ___device_tcp_server_vtbl;

___HIDDEN int ___device_tcp_server_kind
   ___P((___device *self),
        (self)
___device *self;)
{
  return ___TCP_SERVER_KIND;
}

___HIDDEN ___SCMOBJ ___device_tcp_server_close_virt
   ___P((___device *self,
         int direction),
        (self,
         direction)
___device *self;
int direction;)
{
  ___device_tcp_server *d = ___CAST(___device_tcp_server*,self);

  if (d->base.read_stage == ___STAGE_CLOSED)
    return ___FIX(___NO_ERR);

  if (direction & ___DIRECTION_RD)
    {
      d->base.read_stage = ___STAGE_CLOSED; /* avoid multiple closes */

      if (CLOSE_SOCKET(d->s) != 0)
        return ERR_CODE_FROM_SOCKET_CALL;
    }

  return ___FIX(___NO_ERR);
}

___HIDDEN ___SCMOBJ ___device_tcp_server_select_virt
   ___P((___device *self,
         ___BOOL for_writing,
         int i,
         int pass,
         ___device_select_state *state),
        (self,
         for_writing,
         i,
         pass,
         state)
___device *self;
___BOOL for_writing;
int i;
int pass;
___device_select_state *state;)
{
  ___device_tcp_server *d = ___CAST(___device_tcp_server*,self);

  if (pass == ___SELECT_PASS_1)
    {
#ifdef USE_POSIX
      device_select_add_fd (state, d->s, for_writing);
#endif
      return ___FIX(___SELECT_SETUP_DONE);
    }

  /* pass == ___SELECT_PASS_CHECK */

#ifdef USE_POSIX
  if (for_writing)
    {
      if (FD_ISSET(d->s, &state->writefds))
        state->devs[i] = NULL;
    }
  else
    {
      if (FD_ISSET(d->s, &state->readfds))
        state->devs[i] = NULL;
    }
#endif

#ifdef USE_WIN32
  //////////////////
#endif

  return ___FIX(___NO_ERR);
}

___HIDDEN ___SCMOBJ ___device_tcp_server_release_virt
   ___P((___device *self),
        (self)
___device *self;)
{
  return ___FIX(___NO_ERR);
}

___HIDDEN ___SCMOBJ ___device_tcp_server_flush_write_virt
   ___P((___device *self),
        (self)
___device *self;)
{
  return ___FIX(___NO_ERR);
}

___HIDDEN ___device_tcp_server_vtbl ___device_tcp_server_table =
{
  {
    ___device_tcp_server_kind,
    ___device_tcp_server_select_virt,
    ___device_tcp_server_release_virt,
    ___device_tcp_server_flush_write_virt,
    ___device_tcp_server_close_virt
  }
};

___SCMOBJ ___device_tcp_server_setup
   ___P((___device_tcp_server **dev,
         ___device_group *dgroup,
         int port_num,
         int backlog,
         int options),
        (dev,
         dgroup,
         port_num,
         backlog,
         options)
___device_tcp_server **dev;
___device_group *dgroup;
int port_num;
int backlog;
int options;)
{
  ___device_tcp_server *d;
  struct sockaddr_in addr;
  SOCKET_TYPE s;
  int reuse_address_flag;

  if (SOCKET_CALL_ERROR(s = socket (AF_INET, SOCK_STREAM, 0)))
    return ERR_CODE_FROM_SOCKET_CALL;

  addr.sin_family = AF_INET;
  addr.sin_port = htons (port_num);
  addr.sin_addr.s_addr = htonl (INADDR_ANY);

  reuse_address_flag = 1;

  if (((options & 1) &&
       setsockopt (s, /* allow reusing the same address */
                   SOL_SOCKET,
                   SO_REUSEADDR,
                   ___CAST(void*,&reuse_address_flag),
                   sizeof (reuse_address_flag)) != 0) ||
      set_non_blocking (s) != 0 || /* set nonblocking mode */
      bind (s, ___CAST(struct sockaddr*,&addr), sizeof (addr)) != 0 ||
      listen (s, backlog) != 0)
    {
      ___SCMOBJ e = ERR_CODE_FROM_SOCKET_CALL;
      CLOSE_SOCKET(s); /* ignore error */
      return e;
    }

  d = ___CAST(___device_tcp_server*,
              ___alloc_mem (sizeof (___device_tcp_server)));

  if (d == NULL)
    {
      CLOSE_SOCKET(s); /* ignore error */
      return ___FIX(___HEAP_OVERFLOW_ERR);
    }

  d->base.vtbl = &___device_tcp_server_table;
  d->base.refcount = 1;
  d->base.direction = ___DIRECTION_RD;
  d->base.read_stage = ___STAGE_OPEN;
  d->base.write_stage = ___STAGE_CLOSED;

  d->s = s;

  *dev = d;

  device_add_to_group (dgroup, &d->base);

  return ___FIX(___NO_ERR);
}


___SCMOBJ ___device_tcp_server_read
   ___P((___device_tcp_server *dev,
         ___device_group *dgroup,
         ___device_tcp_client **client),
        (dev,
         dgroup,
         client)
___device_tcp_server *dev;
___device_group *dgroup;
___device_tcp_client **client;)
{
  struct sockaddr_in addr;
  SOCKET_LEN_TYPE addrlen;
  SOCKET_TYPE s;

  if (dev->base.read_stage != ___STAGE_OPEN)
    return ___FIX(___CLOSED_DEVICE_ERR);

  addrlen = sizeof (addr);

  if ((s = accept (dev->s, ___CAST(struct sockaddr*,&addr), &addrlen)) < 0)
    return ERR_CODE_FROM_SOCKET_CALL;

  return ___device_tcp_client_setup_from_socket
           (client,
            dgroup,
            s,
            ___CAST(struct sockaddr*,&addr),
            addrlen,
            0,
            ___DIRECTION_RD|___DIRECTION_WR);
}


/*---------------------------------------------------------------------------*/

/* Directory device. */

typedef struct ___device_directory_struct
  {
    ___device base;

    int ignore_hidden;

#ifdef USE_POSIX
    DIR *dir;
#endif

#ifdef USE_WIN32
    HANDLE h;
    WIN32_FIND_DATA fdata;
#endif
  } ___device_directory;

typedef struct ___device_directory_vtbl_struct
  {
    ___device_vtbl base;
  } ___device_directory_vtbl;

___HIDDEN int ___device_directory_kind
   ___P((___device *self),
        (self)
___device *self;)
{
  return ___DIRECTORY_KIND;
}

___HIDDEN ___SCMOBJ ___device_directory_close_virt
   ___P((___device *self,
         int direction),
        (self,
         direction)
___device *self;
int direction;)
{
  ___device_directory *d = ___CAST(___device_directory*,self);

  if (d->base.read_stage == ___STAGE_CLOSED)
    return ___FIX(___NO_ERR);

  if (direction & ___DIRECTION_RD)
    {
      d->base.read_stage = ___STAGE_CLOSED; /* avoid multiple closes */

#ifdef USE_POSIX
      if (closedir (d->dir) < 0)
        return err_code_from_errno ();
#endif

#ifdef USE_WIN32
      if (!FindClose (d->h))
        return err_code_from_GetLastError ();
#endif
    }

  return ___FIX(___NO_ERR);
}

___HIDDEN ___SCMOBJ ___device_directory_select_virt
   ___P((___device *self,
         ___BOOL for_writing,
         int i,
         int pass,
         ___device_select_state *state),
        (self,
         for_writing,
         i,
         pass,
         state)
___device *self;
___BOOL for_writing;
int i;
int pass;
___device_select_state *state;)
{
  if (pass == ___SELECT_PASS_1)
    {
      state->timeout = ___time_neg_infinity;
      return ___FIX(___SELECT_SETUP_DONE);
    }

  /* pass == ___SELECT_PASS_CHECK */

  state->devs[i] = NULL;

  return ___FIX(___NO_ERR);
}

___HIDDEN ___SCMOBJ ___device_directory_release_virt
   ___P((___device *self),
        (self)
___device *self;)
{
  return ___FIX(___NO_ERR);
}

___HIDDEN ___SCMOBJ ___device_directory_flush_write_virt
   ___P((___device *self),
        (self)
___device *self;)
{
  return ___FIX(___NO_ERR);
}

___HIDDEN ___device_directory_vtbl ___device_directory_table =
{
  {
    ___device_directory_kind,
    ___device_directory_select_virt,
    ___device_directory_release_virt,
    ___device_directory_flush_write_virt,
    ___device_directory_close_virt
  }
};


#ifdef USE_POSIX
#define ___DIR_OPEN_PATH_CE_SELECT(latin1,utf8,ucs2,ucs4,native) native
#endif

#ifdef USE_WIN32
#ifdef UNICODE
#define ___DIR_OPEN_PATH_CE_SELECT(latin1,utf8,ucs2,ucs4,native) ucs2
#else
#define ___DIR_OPEN_PATH_CE_SELECT(latin1,utf8,ucs2,ucs4,native) latin1
#endif
#endif


___SCMOBJ ___device_directory_setup
   ___P((___device_directory **dev,
         ___device_group *dgroup,
         ___STRING_TYPE(___DIR_OPEN_PATH_CE_SELECT) path,
         int ignore_hidden),
        (dev,
         dgroup,
         path,
         ignore_hidden)
___device_directory **dev;
___device_group *dgroup;
___STRING_TYPE(___DIR_OPEN_PATH_CE_SELECT) path;
int ignore_hidden;)
{
  ___device_directory *d;

  d = ___CAST(___device_directory*,
              ___alloc_mem (sizeof (___device_directory)));

  if (d == NULL)
    return ___FIX(___HEAP_OVERFLOW_ERR);

  d->base.vtbl = &___device_directory_table;
  d->base.refcount = 1;
  d->base.direction = ___DIRECTION_RD;
  d->base.read_stage = ___STAGE_OPEN;
  d->base.write_stage = ___STAGE_CLOSED;

#ifdef USE_POSIX

  d->ignore_hidden = ignore_hidden;

  d->dir = opendir (path);

  if (d->dir == NULL)
    {
      ___SCMOBJ e = fnf_or_err_code_from_errno ();
      ___free_mem (d);
      return e;
    }

#endif

#ifdef USE_WIN32

  d->ignore_hidden = ignore_hidden;

  d->h = FindFirstFile (path, &d->fdata);

  if (d->h == INVALID_HANDLE_VALUE)
    {
      ___SCMOBJ e = fnf_or_err_code_from_GetLastError ();
      ___free_mem (d);
      return e;
    }

#endif

  *dev = d;

  device_add_to_group (dgroup, &d->base);

  return ___FIX(___NO_ERR);
}

___SCMOBJ ___device_directory_read
   ___P((___device_directory *dev,
         char **name),
        (dev,
         name)
___device_directory *dev;
char **name;)
{
  ___SCMOBJ e;

  if (dev->base.read_stage != ___STAGE_OPEN)
    return ___FIX(___CLOSED_DEVICE_ERR);

  for (;;)
    {
#ifdef USE_POSIX

      struct dirent *de = readdir (dev->dir);

      if (de == NULL)
        {
#if 0
          /* this seems to be broken, at least under Linux */
          if (errno != 0)
            return err_code_from_errno ();
#endif
          *name = NULL;
          e = ___FIX(___NO_ERR);
          break;
        }

      switch (dev->ignore_hidden)
        {
        default:
        case 2:
          if (de->d_name[0] == '.')
            break;

        case 1:
          if (de->d_name[0] == '.' &&
              (de->d_name[1] == '\0' ||
               (de->d_name[1] == '.' && (de->d_name[2] == '\0'))))
            break;

        case 0:
          *name = de->d_name;
          return ___FIX(___NO_ERR);
        }

#endif

#ifdef USE_WIN32

      if (1)
        {
          if (!FindNextFile (dev->h, &dev->fdata))
            return err_code_from_GetLastError ();
        }

      *name = dev->fdata.cFileName;
      if ((*name)[0] == '\0')
        {
          printf ("using 8.3 name\n");fflush(stdout);/////////////////
        *name = dev->fdata.cAlternateFileName; /* use 8.3 name */
        }

      e = ___FIX(___NO_ERR);
      break;

#endif
    }

  return e;
}


/*---------------------------------------------------------------------------*/

/* File stream device */

typedef struct ___device_file_struct
  {
    ___device_stream base;

#ifdef USE_POSIX
    int fd;
#endif

#ifdef USE_WIN32
    HANDLE h;
#endif
  } ___device_file;

typedef struct ___device_file_vtbl_struct
  {
    ___device_stream_vtbl base;
  } ___device_file_vtbl;

___HIDDEN int ___device_file_kind
   ___P((___device *self),
        (self)
___device *self;)
{
  return ___FILE_DEVICE_KIND;
}

___HIDDEN ___SCMOBJ ___device_file_close_virt
   ___P((___device *self,
         int direction),
        (self,
         direction)
___device *self;
int direction;)
{
  ___device_file *d = ___CAST(___device_file*,self);
  int is_not_closed = 0;

  if (d->base.base.read_stage != ___STAGE_CLOSED)
    is_not_closed |= ___DIRECTION_RD;

  if (d->base.base.write_stage != ___STAGE_CLOSED)
    is_not_closed |= ___DIRECTION_WR;

  if (is_not_closed == 0)
    return ___FIX(___NO_ERR);

  if ((is_not_closed & ~direction) == 0)
    {
      /* Close file when both sides are closed. */

      d->base.base.read_stage = ___STAGE_CLOSED; /* avoid multiple closes */
      d->base.base.write_stage = ___STAGE_CLOSED;

#ifdef USE_POSIX
      if (close (d->fd) < 0)
        return err_code_from_errno ();
#endif

#ifdef USE_WIN32
      if (!CloseHandle (d->h))
        return err_code_from_GetLastError ();
#endif
    }
  else if (direction & ___DIRECTION_RD)
    d->base.base.read_stage = ___STAGE_CLOSED;
  else if (direction & ___DIRECTION_WR)
    d->base.base.write_stage = ___STAGE_CLOSED;

  return ___FIX(___NO_ERR);
}

___HIDDEN ___SCMOBJ ___device_file_select_raw_virt
   ___P((___device_stream *self,
         ___BOOL for_writing,
         int i,
         int pass,
         ___device_select_state *state),
        (self,
         for_writing,
         i,
         pass,
         state)
___device_stream *self;
___BOOL for_writing;
int i;
int pass;
___device_select_state *state;)
{
  ___device_file *d = ___CAST(___device_file*,self);

  if ((for_writing ? d->base.base.write_stage : d->base.base.read_stage)
      != ___STAGE_OPEN)
    return ___FIX(___CLOSED_DEVICE_ERR);

  if (pass == ___SELECT_PASS_1)
    {
#ifdef USE_POSIX
      device_select_add_fd (state, d->fd, for_writing);
#endif
      return ___FIX(___SELECT_SETUP_DONE);
    }

  /* pass == ___SELECT_PASS_CHECK */

  if (for_writing)
    {
#ifdef USE_POSIX
      if (FD_ISSET(d->fd, &state->writefds))
        state->devs[i] = NULL;
#endif
    }
  else
    {
#ifdef USE_POSIX
      if (FD_ISSET(d->fd, &state->readfds))
        state->devs[i] = NULL;
#endif
    }

  return ___FIX(___NO_ERR);
}

___HIDDEN ___SCMOBJ ___device_file_release_raw_virt
   ___P((___device_stream *self),
        (self)
___device_stream *self;)
{
  return ___FIX(___NO_ERR);
}

___HIDDEN ___SCMOBJ ___device_file_flush_write_raw_virt
   ___P((___device_stream *self),
        (self)
___device_stream *self;)
{
  ___device_file *d = ___CAST(___device_file*,self);

  if (d->base.base.write_stage == ___STAGE_OPEN)
    {
#if 0
      if (fsync (d->fd) < 0)/************only works on disk files!!!!!!!*/
        return err_code_from_errno ();
#endif
    }

  return ___FIX(___NO_ERR);
}

___HIDDEN ___SCMOBJ ___device_file_seek_raw_virt
   ___P((___device_stream *self,
         ___stream_index pos,
         int whence),
        (self,
         pos,
         whence)
___device_stream *self;
___stream_index pos;
int whence;)
{
  ___device_file *d = ___CAST(___device_file*,self);

  if (d->base.base.write_stage == ___STAGE_OPEN)
    {
#ifdef USE_POSIX

      int p;

      if ((p = lseek (d->fd, pos, whence)) < 0)
        return err_code_from_errno ();

#endif

#ifdef USE_WIN32

      LARGE_INTEGER p;

      p.QuadPart = pos;

      p.LowPart = SetFilePointer (d->h, p.LowPart, &p.HighPart, whence);

      if (p.LowPart == INVALID_SET_FILE_POINTER &&
          GetLastError () != NO_ERROR)
        return err_code_from_GetLastError ();

#endif
    }

  return ___FIX(___NO_ERR);
}

___HIDDEN ___SCMOBJ ___device_file_read_raw_virt
   ___P((___device_stream *self,
         ___U8 *buf,
         ___stream_index len,
         ___stream_index *len_done),
        (self,
         buf,
         len,
         len_done)
___device_stream *self;
___U8 *buf;
___stream_index len;
___stream_index *len_done;)
{
  ___device_file *d = ___CAST(___device_file*,self);

  if (d->base.base.read_stage != ___STAGE_OPEN)
    return ___FIX(___CLOSED_DEVICE_ERR);

#ifdef USE_POSIX

  {
    int n;
    if ((n = read (d->fd, buf, len)) < 0)
      return err_code_from_errno ();
    *len_done = n;
  }

#endif

#ifdef USE_WIN32

  {
    DWORD n;
    if (!ReadFile (d->h, buf, len, &n, NULL))
      return err_code_from_GetLastError ();
    *len_done = n;
  }

#endif

  return ___FIX(___NO_ERR);
}

___HIDDEN ___SCMOBJ ___device_file_write_raw_virt
   ___P((___device_stream *self,
         ___U8 *buf,
         ___stream_index len,
         ___stream_index *len_done),
        (self,
         buf,
         len,
         len_done)
___device_stream *self;
___U8 *buf;
___stream_index len;
___stream_index *len_done;)
{
  ___device_file *d = ___CAST(___device_file*,self);

  if (d->base.base.write_stage != ___STAGE_OPEN)
    return ___FIX(___CLOSED_DEVICE_ERR);

#ifdef USE_POSIX

  {
    int n;
    if ((n = write (d->fd, buf, len)) < 0)
      return err_code_from_errno ();
    *len_done = n;
  }

#endif

#ifdef USE_WIN32

  {
    DWORD n;
    if (!WriteFile (d->h, buf, len, &n, NULL))
      return err_code_from_GetLastError ();
    *len_done = n;
  }

#endif

  return ___FIX(___NO_ERR);
}

___HIDDEN ___SCMOBJ ___device_file_width_virt
   ___P((___device_stream *self),
        (self)
___device_stream *self;)
{
  return ___FIX(80);
}

___HIDDEN ___SCMOBJ ___device_file_default_options_virt
   ___P((___device_stream *self),
        (self)
___device_stream *self;)
{
  return ___STREAM_DEFAULT_OPTIONS(___CHAR_ENCODING_LATIN1,
                                   ___EOL_ENCODING_LF,
                                   ___BUFFERING,
                                   ___CHAR_ENCODING_LATIN1,
                                   ___EOL_ENCODING_LF,
                                   ___BUFFERING);
}

___HIDDEN ___device_file_vtbl ___device_file_table =
{
  {
    {
      ___device_file_kind,
      ___device_stream_select_virt,
      ___device_stream_release_virt,
      ___device_stream_flush_write_virt,
      ___device_file_close_virt
    },
    ___device_file_select_raw_virt,
    ___device_file_release_raw_virt,
    ___device_file_flush_write_raw_virt,
    ___device_file_seek_raw_virt,
    ___device_file_read_raw_virt,
    ___device_file_write_raw_virt,
    ___device_file_width_virt,
    ___device_file_default_options_virt
  }
};


#ifdef USE_POSIX

___HIDDEN ___SCMOBJ ___device_file_setup_from_fd
   ___P((___device_file **dev,
         ___device_group *dgroup,
         int fd,
         int direction),
        (dev,
         dgroup,
         fd,
         direction)
___device_file **dev;
___device_group *dgroup;
int fd;
int direction;)
{
  ___device_file *d;

  d = ___CAST(___device_file*,
              ___alloc_mem (sizeof (___device_file)));

  if (d == NULL)
    {
      close (fd); /* ignore error */
      return ___FIX(___HEAP_OVERFLOW_ERR);
    }

  d->base.base.vtbl = &___device_file_table;
  d->fd = fd;

  *dev = d;

  return ___device_stream_setup
           (&d->base,
            dgroup,
            direction,
            0);
}

#endif


#ifdef USE_WIN32

___HIDDEN ___SCMOBJ ___device_file_setup_from_handle
   ___P((___device_file **dev,
         ___device_group *dgroup,
         HANDLE h,
         int direction),
        (dev,
         dgroup,
         h,
         direction)
___device_file **dev;
___device_group *dgroup;
HANDLE h;
int direction;)
{
  ___device_file *d;

  d = ___CAST(___device_file*,
              ___alloc_mem (sizeof (___device_file)));

  if (d == NULL)
    {
      CloseHandle (h); /* ignore error */
      return ___FIX(___HEAP_OVERFLOW_ERR);
    }

  d->base.base.vtbl = &___device_file_table;
  d->h = h;

  *dev = d;

  return ___device_stream_setup
           (&d->base,
            dgroup,
            direction,
            0);
}

#endif


/*---------------------------------------------------------------------------*/

#ifdef USE_POSIX

___HIDDEN int ___device_stream_kind_from_fd
   ___P((int fd),
        (fd)
int fd;)
{
  struct stat s;

  /*
   * Determine what kind of device is attached to the file descriptor
   * (tty, socket, or regular file).
   */

  if (isatty (fd))
    return ___TTY_DEVICE_KIND;

  if (fstat (fd, &s) < 0)
    return ___NONE_KIND;

  if (S_ISREG(s.st_mode))
    return ___FILE_DEVICE_KIND;

#if 0
  if (S_ISDIR(s.st_mode))
    return ???;
#endif

  if (S_ISCHR(s.st_mode))
    return ___TTY_DEVICE_KIND;////////////////
  //    return ___FILE_DEVICE_KIND;//////////////////

#if 0
  if (S_ISBLK(s.st_mode))
    return ???;
#endif

  if (S_ISFIFO(s.st_mode))
    return ___FILE_DEVICE_KIND;//////////////

#if 0
  if (S_ISLNK(s.st_mode))
    return ???;
#endif

  if (S_ISSOCK(s.st_mode))
    return ___TCP_CLIENT_DEVICE_KIND;

  return ___NONE_KIND;
}


___HIDDEN int ___device_stream_direction_from_fd
   ___P((int fd),
        (fd)
int fd;)
{
  int direction = ___DIRECTION_RD|___DIRECTION_WR;
  char buf[1];

  /*
   * A "read" and "write" of 0 bytes is attempted to tell which
   * directions the file descriptor supports.
   */

  if (read (fd, buf, 0) < 0)
    direction &= ~___DIRECTION_RD;

  if (write (fd, buf, 0) < 0)
    direction &= ~___DIRECTION_WR;

  /*
   * It is likely that on some systems a zero length "read" and
   * "write" will return an error, regardless of the possible
   * operations.  In this case assume that both operations are
   * possible.
   */

  if (direction == 0)
    direction = ___DIRECTION_RD|___DIRECTION_WR;

  return direction;
}


___SCMOBJ ___device_stream_setup_from_fd
   ___P((___device_stream **dev,
         ___device_group *dgroup,
         int fd,
         int kind,
         int direction),
        (dev,
         dgroup,
         fd,
         kind,
         direction)
___device_stream **dev;
___device_group *dgroup;
int fd;
int kind;
int direction;)
{
  ___SCMOBJ e = ___FIX(___UNKNOWN_ERR);

  if (kind == ___NONE_KIND)
    kind = ___device_stream_kind_from_fd (fd);

#if 0
  printf ("fd=%d kind=%d\n", fd, kind);fflush(stdout);///////////
#endif

  if (direction == 0)
    direction = ___device_stream_direction_from_fd (fd);

  switch (kind)
    {
    case ___TTY_DEVICE_KIND:
      {
        ___device_tty *d;
        if ((e = ___device_tty_setup_from_fd
                   (&d,
                    dgroup,
                    fd,
                    direction))
            == ___FIX(___NO_ERR))
          *dev = ___CAST(___device_stream*,d);
        break;
      }

    case ___TCP_CLIENT_DEVICE_KIND:
      {
        ___device_tcp_client *d;
        struct sockaddr server_addr;
        if ((e = ___device_tcp_client_setup_from_socket
                   (&d,
                    dgroup,
                    fd,
                    &server_addr,
                    0,
                    0,
                    direction))
            == ___FIX(___NO_ERR))
          *dev = ___CAST(___device_stream*,d);
        break;
      }

    case ___FILE_DEVICE_KIND:
      {
        ___device_file *d;
        if ((e = ___device_file_setup_from_fd
                   (&d,
                    dgroup,
                    fd,
                    direction))
            == ___FIX(___NO_ERR))
          *dev = ___CAST(___device_stream*,d);
        break;
      }

    default:
      {
        close (fd); /* ignore error */
        e = ___FIX(___UNKNOWN_ERR);
        break;
      }
    }

  return e;
}


___HIDDEN void device_translate_flags
   ___P((int flags,
         int *fl,
         int *direction),
        (flags,
         fl,
         direction)
int flags;
int *fl;
int *direction;)
{
  int f;

  switch ((flags >> 4) & 3)
    {
    default:
    case 1:
      f = O_RDONLY;
      *direction = ___DIRECTION_RD;
      break;
    case 2:
      f = O_WRONLY;
      *direction = ___DIRECTION_WR;
      break;
    case 3:
      f = O_RDWR;
      *direction = ___DIRECTION_RD|___DIRECTION_WR;
      break;
    }

  if (flags & 16)
    f |= O_APPEND;

  switch ((flags >> 1) & 3)
    {
    default:
    case 0: break;
    case 1: f |= O_CREAT; break;
    case 2: f |= O_CREAT|O_EXCL; break;
    }

  if (flags & 1)
    f |= O_TRUNC;

  f |= O_NONBLOCK;

  *fl = f;
}

#endif


#ifdef USE_WIN32

___HIDDEN int ___device_stream_kind_from_handle
   ___P((HANDLE h),
        (h)
HANDLE h;)
{
  DWORD n;
  CONSOLE_CURSOR_INFO cinfo;
  DCB dcb;
  BY_HANDLE_FILE_INFORMATION finfo;

  if (GetNumberOfConsoleInputEvents (h, &n))
    return ___TTY_DEVICE_KIND;

  if (GetConsoleCursorInfo (h, &cinfo))
    return ___TTY_DEVICE_KIND;

  if (GetCommState (h, &dcb))
    return ___SERIAL_DEVICE_KIND;

  if (GetFileType (h) == FILE_TYPE_PIPE)
#if 0
    return ___PIPE_DEVICE_KIND;
#else
    return ___FILE_DEVICE_KIND;////////////////////
#endif

  if (GetFileInformationByHandle (h, &finfo))
    return ___FILE_DEVICE_KIND;

  return ___NONE_KIND;
}


___HIDDEN int ___device_stream_direction_from_handle
   ___P((HANDLE h),
        (h)
HANDLE h;)
{
  DWORD n;
  int direction = ___DIRECTION_RD|___DIRECTION_WR;

  /*
   * A "ReadFile" and "WriteFile" of 0 bytes is attempted to tell which
   * directions the file handle supports.
   */

  if (!ReadFile (h, NULL, 0, &n, NULL))
    direction &= ~___DIRECTION_RD;

  if (!WriteFile (h, NULL, 0, &n, NULL))
    direction &= ~___DIRECTION_WR;

  /*
   * It is likely that on some systems a zero length "ReadFile" and
   * "write" will return an error, regardless of the possible
   * operations.  In this case assume that both operations are
   * possible.
   */

  if (direction == 0)
    direction = ___DIRECTION_RD|___DIRECTION_WR;

  return direction;
}


___SCMOBJ ___device_stream_setup_from_handle
   ___P((___device_stream **dev,
         ___device_group *dgroup,
         HANDLE h,
         int kind,
         int direction),
        (dev,
         dgroup,
         h,
         kind,
         direction)
___device_stream **dev;
___device_group *dgroup;
HANDLE h;
int kind;
int direction;)
{
  ___SCMOBJ e = ___FIX(___UNKNOWN_ERR);

  if (kind == ___NONE_KIND)
    kind = ___device_stream_kind_from_handle (h);

  ////printf ("kind=%d\n",kind);fflush(stdout);////////////////////

  if (direction == 0)
    direction = ___device_stream_direction_from_handle (h);

  switch (kind)
    {
    case ___TTY_DEVICE_KIND:
      {
        ___device_tty *d;
        if ((e = ___device_tty_setup_from_console
                   (&d,
                    dgroup,
                    direction))
            == ___FIX(___NO_ERR))
          *dev = ___CAST(___device_stream*,d);
        break;
      }

    case ___FILE_DEVICE_KIND:
      {
        ___device_file *d;
        if ((e = ___device_file_setup_from_handle
                   (&d,
                    dgroup,
                    h,
                    direction))
            == ___FIX(___NO_ERR))
          *dev = ___CAST(___device_stream*,d);
        break;
      }

    default:
      {
        CloseHandle (h); /* ignore error */
        e = ___FIX(___UNKNOWN_ERR);
        break;
      }
    }

  return e;
}


___HIDDEN void device_translate_flags
   ___P((int flags,
         DWORD *access_mode,
         DWORD *share_mode,
         DWORD *creation_mode,
         DWORD *attributes,
         int *direction),
        (flags,
         access_mode,
         share_mode,
         creation_mode,
         attributes,
         direction)
int flags;
DWORD *access_mode;
DWORD *share_mode;
DWORD *creation_mode;
DWORD *attributes;
int *direction;)
{
  DWORD am;
  DWORD cm;

  switch ((flags >> 4) & 3)
    {
    default:
    case 1:
      am = GENERIC_READ;
      *direction = ___DIRECTION_RD;
      break;
    case 2:
      am = GENERIC_WRITE;
      *direction = ___DIRECTION_WR;
      break;
    case 3:
      am = GENERIC_READ|GENERIC_WRITE;
      *direction = ___DIRECTION_RD|___DIRECTION_WR;
      break;
    }

#if 0
  //////////////////////// not supported by Windows
  if (flags & 16)
    f |= O_APPEND;
#endif

  switch ((flags >> 1) & 3)
    {
    default:
    case 0:
      if (flags & 1)
        cm = TRUNCATE_EXISTING;
      else
        cm = OPEN_EXISTING;
      break;

    case 1:
      if (flags & 1)
        cm = CREATE_ALWAYS; /////////////////////// I'm not sure this truncates an existing file.
      else
        cm = OPEN_ALWAYS;
      break;

    case 2:
      cm = CREATE_NEW;
      break;
    }

  *access_mode = am;
  *share_mode = FILE_SHARE_READ;
  *creation_mode = cm;
  *attributes = FILE_ATTRIBUTE_NORMAL;
}

#endif


/*---------------------------------------------------------------------------*/

#ifdef USE_POSIX

////////////////////////////////
#define USE_pipe

typedef struct half_duplex_pipe
  {
    int reading_fd;
    int writing_fd;
  } half_duplex_pipe;

typedef struct full_duplex_pipe
  {
    half_duplex_pipe input;
    half_duplex_pipe output;
  } full_duplex_pipe;


___HIDDEN int open_half_duplex_pipe
   ___P((half_duplex_pipe *hdp),
        (hdp)
half_duplex_pipe *hdp;)
{
  int fds[2];

#ifdef USE_pipe
  if (pipe (fds) < 0)
    return -1;
#endif

#ifdef USE_socketpair
  if (socketpair (AF_UNIX, SOCK_STREAM, 0, fds) < 0)
    return -1;
#endif

  hdp->reading_fd = fds[0];
  hdp->writing_fd = fds[1];

  return 0;
}

___HIDDEN void close_half_duplex_pipe
   ___P((half_duplex_pipe *hdp,
         int end),
        (hdp,
         end)
half_duplex_pipe *hdp;
int end;)
{
  if (end != 1 && hdp->reading_fd >= 0)
    {
      close (hdp->reading_fd); /* ignore error */
      hdp->reading_fd = -1;
    }

  if (end != 0 && hdp->writing_fd >= 0)
    {
      close (hdp->writing_fd); /* ignore error */
      hdp->writing_fd = -1;
    }
}

___HIDDEN int open_pseudo_terminal_master
   ___P((int *master_fd_ptr,
         int *slave_fd_ptr),
        (master_fd_ptr,
         slave_fd_ptr)
int *master_fd_ptr;
int *slave_fd_ptr;)
{
#ifdef USE_openpty

  return openpty (master_fd_ptr, slave_fd_ptr, NULL, NULL, NULL);

#else
#ifdef USE_getpt

  *slave_fd_ptr = -1;
  return *master_fd_ptr = getpt ();

#else

  *slave_fd_ptr = -1;
  return *master_fd_ptr = open ("/dev/ptmx", O_RDWR);

#endif
#endif
}


#ifndef USE_openpty
/***************************/
#define __USE_XOPEN
#define __USE_GNU
#include <stdlib.h>
#include <stropts.h>
extern char *ptsname (int __fd);
#endif

___HIDDEN int open_pseudo_terminal_slave
   ___P((int master_fd,
         int slave_fd,
         int *slave_fd_ptr),
        (master_fd,
         slave_fd,
         slave_fd_ptr)
int master_fd;
int slave_fd;
int *slave_fd_ptr;)
{
#ifndef USE_openpty
#ifndef USE_ptsname

  errno = xxx;/********************/
  return -1;

#endif
#endif

#ifdef USE_openpty

  int fd = slave_fd;

    {
      int tmp;

      if (1
          /*
          !isastream (fd) ||
          (ioctl (fd, I_PUSH, "ptem") >= 0 &&
           ioctl (fd, I_PUSH, "ldterm") >= 0)
          */
)
        {
          struct termios tios;

          if (tcgetattr (fd, &tios) >= 0)
            {
              tios.c_lflag &= ~(ECHO | ICANON | IEXTEN | ISIG);
              tios.c_iflag &= ~(BRKINT | INLCR | ICRNL | INPCK | ISTRIP | IXON | IXOFF);
              tios.c_cflag &= ~(CSIZE | PARENB);
              tios.c_cflag |= CS8;
              tios.c_oflag &= ~(OPOST | ONLCR 
                                       /*********************** | OCRNL */
);
              if (tcsetattr (fd, TCSANOW, &tios) >= 0)
                {
                  *slave_fd_ptr = fd;
                  return 0;
                }
            }
        }

      tmp = errno;
      close (fd); /* ignore error */
      errno = tmp;
    }

  return -1;

#endif

#ifdef USE_ptsname

  int fd;
  char *name;

  if (grantpt (master_fd) >= 0 &&
      unlockpt (master_fd) >= 0 &&
      (name = ptsname (master_fd)) != NULL &&
      (fd = open (name, O_RDWR)) >= 0)
    {
      int tmp;

      if (!isastream (fd) ||
          (ioctl (fd, I_PUSH, "ptem") >= 0 &&
           ioctl (fd, I_PUSH, "ldterm") >= 0))
        {
          struct termios tios;

          if (tcgetattr (fd, &tios) >= 0)
            {
              tios.c_lflag &= ~(ECHO | ICANON | IEXTEN | ISIG);
              tios.c_iflag &= ~(BRKINT | INLCR | ICRNL | INPCK | ISTRIP | IXON | IXOFF);
              tios.c_cflag &= ~(CSIZE | PARENB);
              tios.c_cflag |= CS8;
              tios.c_oflag &= ~(OPOST | ONLCR | OCRNL);
              if (tcsetattr (fd, TCSANOW, &tios) >= 0)
                {
                  *slave_fd_ptr = fd;
                  return 0;
                }
            }
        }

      tmp = errno;
      close (fd); /* ignore error */
      errno = tmp;
    }

  return -1;

#endif
}


___HIDDEN int open_full_duplex_pipe1
   ___P((full_duplex_pipe *fdp,
         ___BOOL use_pty),
        (fdp,
         use_pty)
full_duplex_pipe *fdp;
___BOOL use_pty;)
{
  fdp->input.reading_fd = -1;
  fdp->input.writing_fd = -1;
  fdp->output.reading_fd = -1;
  fdp->output.writing_fd = -1;

  if (use_pty)
    {
      int master_fd;
      int slave_fd;
      if (open_pseudo_terminal_master (&master_fd, &slave_fd) >= 0)
        {
          int master_fd_dup;
          int tmp;
          if ((master_fd_dup = dup (master_fd)) >= 0)
            {
              fdp->input.writing_fd = master_fd;
              fdp->output.reading_fd = master_fd_dup;
              fdp->input.reading_fd = slave_fd;
              return 0;
            }
          tmp = errno;
          close (master_fd); /* ignore error */
          if (slave_fd >= 0)
            close (slave_fd); /* ignore error */
          errno = tmp;
        }
    }
  else
    {
      if (open_half_duplex_pipe (&fdp->input) >= 0)
        {
          if (open_half_duplex_pipe (&fdp->output) >= 0)
            return 0;
          close_half_duplex_pipe (&fdp->input, 2);
        }
    }

  return -1;
}


___HIDDEN int open_full_duplex_pipe2
   ___P((full_duplex_pipe *fdp,
         ___BOOL use_pty),
        (fdp,
         use_pty)
full_duplex_pipe *fdp;
___BOOL use_pty;)
{
  if (use_pty)
    {
      int slave_fd;
      if (setsid () >= 0 &&
          open_pseudo_terminal_slave (fdp->input.writing_fd,
                                      fdp->input.reading_fd,
                                      &slave_fd) >= 0)
        {
          int slave_fd_dup;
          int tmp;
          if ((slave_fd_dup = dup (slave_fd)) >= 0)
            {
              fdp->input.reading_fd = slave_fd;
              fdp->output.writing_fd = slave_fd_dup;
              return 0;
            }
          tmp = errno;
          close (slave_fd); /* ignore error */
          errno = tmp;
        }
    }
  else
    return 0;

  return -1;
}


___SCMOBJ ___device_stream_setup_from_process
   ___P((___device_stream **dev,
         ___device_group *dgroup,
         char **argv,
         char **env,
         int options),
        (dev,
         dgroup,
         argv,
         env,
         options)
___device_stream **dev;
___device_group *dgroup;
char **argv;
char **env;
int options;)
{
  ___SCMOBJ e;
  int direction;
  pid_t pid = 0;
  ___device_process *d;
  half_duplex_pipe hdp_errno;
  full_duplex_pipe fdp;
  int execvp_errno;
  int n;

  e = ___FIX(___NO_ERR);

  if (open_half_duplex_pipe (&hdp_errno) < 0)
    e = err_code_from_errno ();
  else
    {
      if (open_full_duplex_pipe1 (&fdp, options & 2) < 0)
        e = err_code_from_errno ();
      else
        {
          if ((pid = fork ()) < 0)
            {
              e = err_code_from_errno ();
              close_half_duplex_pipe (&fdp.input, 2);
              close_half_duplex_pipe (&fdp.output, 2);
            }
        }

      if (e != ___FIX(___NO_ERR))
        close_half_duplex_pipe (&hdp_errno, 2);
    }

  if (e != ___FIX(___NO_ERR))
    return e;

  if (pid == 0)
    {
      /* child process */

      if (open_full_duplex_pipe2 (&fdp, options & 2) >= 0 &&
(
#if 0
 printf ("getpid -> %d\n", getpid()),
 printf ("getsid (0) -> %d\n", getsid(0)),
 printf ("tcgetpgrp (%d) -> %d\n", fdp.input.reading_fd, tcgetpgrp (fdp.input.reading_fd)),
 printf ("tcgetpgrp (%d) -> %d\n", fdp.output.writing_fd, tcgetpgrp (fdp.output.writing_fd)),
 /*
 printf ("tcgetsid (%d) -> %d\n", fdp.input.reading_fd, tcgetsid (fdp.input.reading_fd)),
 printf ("tcgetsid (%d) -> %d\n", fdp.output.reading_fd, tcgetsid (fdp.output.reading_fd)),
 */
#endif
 1) &&
          dup2 (fdp.input.reading_fd, STDIN_FILENO) >= 0 &&
          dup2 (fdp.output.writing_fd, STDOUT_FILENO) >= 0 &&
          ((options & 1) == 0 ||
           dup2 (fdp.output.writing_fd, STDERR_FILENO) >= 0) &&
          fcntl (hdp_errno.writing_fd, F_SETFD, FD_CLOEXEC) >= 0)
        {
          close_half_duplex_pipe (&fdp.input, 1);
          close_half_duplex_pipe (&fdp.output, 0);
          close_half_duplex_pipe (&hdp_errno, 0);

          {
            extern char **environ;
            if (env != NULL)
              environ = env;
          }

          execvp (argv[0], argv);

          /* the execvp failed, errno will be returned to parent */
        }

      /* return the errno to the parent process */

      execvp_errno = errno;

      write (hdp_errno.writing_fd, &execvp_errno, sizeof (execvp_errno));

      close_half_duplex_pipe (&fdp.input, 0);
      close_half_duplex_pipe (&fdp.output, 1);
      close_half_duplex_pipe (&hdp_errno, 1);

      _exit (1);
    }

  /* parent process */

  close_half_duplex_pipe (&fdp.input, 0);
  close_half_duplex_pipe (&fdp.output, 1);
  close_half_duplex_pipe (&hdp_errno, 1);

  n = read (hdp_errno.reading_fd, &execvp_errno, sizeof (execvp_errno));

  if (n < 0)
    e = err_code_from_errno ();
  else if (n == sizeof (execvp_errno))
    {
      errno = execvp_errno;
      e = err_code_from_errno ();
    }
  else if (n != 0)
    e = ___FIX(___UNKNOWN_ERR);
  else
    {
      direction = ___DIRECTION_RD|___DIRECTION_WR;

      e = ___device_process_setup_from_pid
            (&d,
             dgroup,
             pid,
             fdp.input.writing_fd,
             fdp.output.reading_fd,
             direction);

      *dev = ___CAST(___device_stream*,d);
    }

  if (e != ___FIX(___NO_ERR))
    {
      close_half_duplex_pipe (&fdp.input, 1);
      close_half_duplex_pipe (&fdp.output, 0);
    }

  close_half_duplex_pipe (&hdp_errno, 0);

  return e;
}

#endif


#ifdef USE_POSIX
#define ___STREAM_OPEN_PATH_CE_SELECT(latin1,utf8,ucs2,ucs4,native) native
#endif

#ifdef USE_WIN32
#ifdef UNICODE
#define ___STREAM_OPEN_PATH_CE_SELECT(latin1,utf8,ucs2,ucs4,native) ucs2
#else
#define ___STREAM_OPEN_PATH_CE_SELECT(latin1,utf8,ucs2,ucs4,native) latin1
#endif
#endif

___SCMOBJ ___device_stream_setup_from_path
   ___P((___device_stream **dev,
         ___device_group *dgroup,
         ___STRING_TYPE(___STREAM_OPEN_PATH_CE_SELECT) path,
         int flags,
         int mode),
        (dev,
         dgroup,
         path,
         flags,
         mode)
___device_stream **dev;
___device_group *dgroup;
___STRING_TYPE(___STREAM_OPEN_PATH_CE_SELECT) path;
int flags;
int mode;)
{
#ifdef USE_POSIX

  int fl;
  int direction;
  int fd;

  device_translate_flags (flags,
                          &fl,
                          &direction);

  if ((fd = open (path, fl, mode)) < 0)
    return fnf_or_err_code_from_errno ();

  return ___device_stream_setup_from_fd
           (dev,
            dgroup,
            fd,
            ___NONE_KIND,
            direction);

#endif

#ifdef USE_WIN32

  DWORD access_mode;
  DWORD share_mode;
  DWORD creation_mode;
  DWORD attributes;
  int direction;
  HANDLE h;

  device_translate_flags (flags,
                          &access_mode,
                          &share_mode,
                          &creation_mode,
                          &attributes,
                          &direction);

  h = CreateFile (path,
                  access_mode,
                  share_mode,
                  NULL,
                  creation_mode,
                  attributes,
                  NULL);

  if (h == INVALID_HANDLE_VALUE)
    return fnf_or_err_code_from_GetLastError ();

  return ___device_stream_setup_from_handle
           (dev,
            dgroup,
            h,
            ___NONE_KIND,
            direction);

#endif
}


/*---------------------------------------------------------------------------*/

/* I/O module. */

typedef struct io_module_struct
  {
    ___device_group *dgroup;
  } io_module;

___HIDDEN io_module io_mod;


/*   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   */

/* Device operations. */

___SCMOBJ ___os_device_kind
   ___P((___SCMOBJ dev),
        (dev)
___SCMOBJ dev;)
{
  ___device *d = ___CAST(___device*,___FIELD(dev,___FOREIGN_PTR));

  return ___FIX(___device_kind (d));
}


___SCMOBJ ___os_device_force_output
   ___P((___SCMOBJ dev),
        (dev)
___SCMOBJ dev;)
{
  ___device *d = ___CAST(___device*,___FIELD(dev,___FOREIGN_PTR));

  return ___device_flush_write (d);
}


___SCMOBJ ___os_device_close
   ___P((___SCMOBJ dev,
         ___SCMOBJ direction),
        (dev,
         direction)
___SCMOBJ dev;
___SCMOBJ direction;)
{
  ___device *d = ___CAST(___device*,___FIELD(dev,___FOREIGN_PTR));

  return ___device_close (d, ___INT(direction));
}


/*   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   */

/* Stream device operations. */

___SCMOBJ ___os_device_stream_seek
   ___P((___SCMOBJ dev,
         ___SCMOBJ pos,
         ___SCMOBJ whence),
        (dev,
         pos,
         whence)
___SCMOBJ dev;
___SCMOBJ pos;
___SCMOBJ whence;)
{
  ___device_stream *d =
    ___CAST(___device_stream*,___FIELD(dev,___FOREIGN_PTR));
  ___S32 p;
  ___SCMOBJ e;

  if ((e = ___SCMOBJ_to_S32 (pos, &p, 1)) == ___FIX(___NO_ERR))
    e = ___device_stream_seek (d, p, ___INT(whence));

  return e;
}


___SCMOBJ ___os_device_stream_read
   ___P((___SCMOBJ dev,
         ___SCMOBJ buffer,
         ___SCMOBJ lo,
         ___SCMOBJ hi),
        (dev,
         buffer,
         lo,
         hi)
___SCMOBJ dev;
___SCMOBJ buffer;
___SCMOBJ lo;
___SCMOBJ hi;)
{
  ___device_stream *d =
    ___CAST(___device_stream*,___FIELD(dev,___FOREIGN_PTR));
  ___stream_index len_done;
  ___SCMOBJ e;

  if ((e = ___device_stream_read
             (d,
              ___CAST(___U8*,___BODY_AS(buffer,___tSUBTYPED)) + ___INT(lo),
              ___INT(hi) - ___INT(lo),
              &len_done))
      == ___FIX(___NO_ERR))
    return ___FIX(len_done);

  return e;
}


___SCMOBJ ___os_device_stream_write
   ___P((___SCMOBJ dev,
         ___SCMOBJ buffer,
         ___SCMOBJ lo,
         ___SCMOBJ hi),
        (dev,
         buffer,
         lo,
         hi)
___SCMOBJ dev;
___SCMOBJ buffer;
___SCMOBJ lo;
___SCMOBJ hi;)
{
  ___device_stream *d =
    ___CAST(___device_stream*,___FIELD(dev,___FOREIGN_PTR));
  ___stream_index len_done;
  ___SCMOBJ e;

  if ((e = ___device_stream_write
             (d,
              ___CAST(___U8*,___BODY_AS(buffer,___tSUBTYPED)) + ___INT(lo),
              ___INT(hi) - ___INT(lo),
              &len_done))
      == ___FIX(___NO_ERR))
    return ___FIX(len_done);

  return e;
}


___SCMOBJ ___os_device_stream_width
   ___P((___SCMOBJ dev),
        (dev)
___SCMOBJ dev;)
{
  ___device_stream *d =
    ___CAST(___device_stream*,___FIELD(dev,___FOREIGN_PTR));

  return ___device_stream_width (d);
}


___SCMOBJ ___os_device_stream_default_options
   ___P((___SCMOBJ dev),
        (dev)
___SCMOBJ dev;)
{
  ___device_stream *d =
    ___CAST(___device_stream*,___FIELD(dev,___FOREIGN_PTR));

  return ___device_stream_default_options (d);
}


/*   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   */

/*
 * Procedure called by the Scheme runtime when a device is no longer
 * reachable.
 */

___HIDDEN ___SCMOBJ device_cleanup_from_ptr
   ___P((void *ptr),
        (ptr)
void *ptr;)
{
  return ___device_cleanup (___CAST(___device*,ptr));
}


/*   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   */

/* Opening a predefined device (stdin, stdout, stderr, console, etc). */

___SCMOBJ ___os_device_stream_open_predefined
   ___P((___SCMOBJ index,
         ___SCMOBJ flags),
        (index,
         flags)
___SCMOBJ index;
___SCMOBJ flags;)
{
  ___SCMOBJ e;
  ___device_stream *dev;
  ___SCMOBJ result;

#ifdef USE_POSIX

  int fl;
  int direction;
  int fd;

  device_translate_flags (___INT(flags),
                          &fl,
                          &direction);

  switch (___INT(index))
    {
    case -4:
      {
        ___device_tty *d;

        if ((e = ___device_tty_setup_console
                   (&d,
                    io_mod.dgroup,
                    direction))
            != ___FIX(___NO_ERR))
          return e;

        dev = ___CAST(___device_stream*,d);

        break;
      }

    default:
      {
        switch (___INT(index))
          {
          case -1:
            fd = STDIN_FILENO;
            break;
          case -2:
            fd = STDOUT_FILENO;
            break;
          case -3:
            fd = STDERR_FILENO;
            break;
          default:
            fd = ___INT(index);
            break;
          }

        /*
         * The file descriptor must be dup'ed so that the standard
         * stdin/stdout/stderr are not closed.
         */

        //        printf ("tty fd=%d\n", fd);/////////////////
        if ((fd = dup (fd)) < 0)
          return err_code_from_errno ();
        //        printf ("  new tty fd=%d\n", fd);/////////////////////

        if ((e = ___device_stream_setup_from_fd
                   (&dev,
                    io_mod.dgroup,
                    fd,
                    ___NONE_KIND,
                    direction))
            != ___FIX(___NO_ERR))
          return e;

        break;
      }
    }

#endif

#ifdef USE_WIN32

  DWORD access_mode;
  DWORD share_mode;
  DWORD creation_mode;
  DWORD attributes;
  int direction;
  HANDLE h;

  device_translate_flags (___INT(flags),
                          &access_mode,
                          &share_mode,
                          &creation_mode,
                          &attributes,
                          &direction);

  /////printf ("predef %d\n",___INT(index));fflush(stdout);/////////////////////

  switch (___INT(index))
    {
    case -4:
      {
        ___device_tty *d;

        if ((e = ___device_tty_setup_console
                   (&d,
                    io_mod.dgroup,
                    direction))
            != ___FIX(___NO_ERR))
          return e;

        dev = ___CAST(___device_stream*,d);

        break;
      }

    default:
      {
        switch (___INT(index))
          {
          default:
          case -1:
            h = GetStdHandle (STD_INPUT_HANDLE);
            break;
          case -2:
            h = GetStdHandle (STD_OUTPUT_HANDLE);
            break;
          case -3:
            h = GetStdHandle (STD_ERROR_HANDLE);
            break;
          }

        /////////////// duplicate the handle?

        if ((e = ___device_stream_setup_from_handle
                   (&dev,
                    io_mod.dgroup,
                    h,
#if 0
                    ___FILE_DEVICE_KIND,
#else
                    ___NONE_KIND,
#endif
                    direction))
            != ___FIX(___NO_ERR))
          return e;

        break;
      }
    }

#endif

  e = ___NONNULLPOINTER_to_SCMOBJ
        (dev,
         ___FAL,
         device_cleanup_from_ptr,
         &result,
         ___RETURN_POS);

  if (e != ___FIX(___NO_ERR))
    {
      ___device_cleanup (___CAST(___device*,dev)); /* ignore error */
      return e;
    }

  ___release_scmobj (result);

  return result;
}


/*   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   */

/* Opening a path. */

___SCMOBJ ___os_device_stream_open_path
   ___P((___SCMOBJ path,
         ___SCMOBJ flags,
         ___SCMOBJ mode),
        (path,
         flags,
         mode)
___SCMOBJ path;
___SCMOBJ flags;
___SCMOBJ mode;)
{
  ___SCMOBJ e;
  ___SCMOBJ result;
  ___device_stream *dev;
  void *cpath;

  if ((e = ___SCMOBJ_to_NONNULLSTRING
             (path,
              &cpath,
              1,
              ___CE(___STREAM_OPEN_PATH_CE_SELECT),
              0))
      != ___FIX(___NO_ERR))
    result = e;
  else
    {
      ___STRING_TYPE(___STREAM_OPEN_PATH_CE_SELECT) p =
        ___CAST(___STRING_TYPE(___STREAM_OPEN_PATH_CE_SELECT),cpath);

      if ((e = ___device_stream_setup_from_path
                 (&dev,
                  io_mod.dgroup,
                  p,
                  ___INT(flags),
                  ___INT(mode)))
          != ___FIX(___NO_ERR))
        result = e;
      else
        {
          if ((e = ___NONNULLPOINTER_to_SCMOBJ
                     (dev,
                      ___FAL,
                      device_cleanup_from_ptr,
                      &result,
                      ___RETURN_POS))
              != ___FIX(___NO_ERR))
            {
              ___device_cleanup (___CAST(___device*,dev)); /* ignore error */
              result = e;
            }
        }

      ___release_string (cpath);
    }

  ___release_scmobj (result);

  return result;
}


/*   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   */

/* Opening a process. */

___SCMOBJ ___os_device_stream_open_process
   ___P((___SCMOBJ path_and_args,
         ___SCMOBJ environment,
         ___SCMOBJ options),
        (path_and_args,
         environment,
         options)
___SCMOBJ path_and_args;
___SCMOBJ environment;
___SCMOBJ options;)
{
  ___SCMOBJ e;
  ___device_stream *dev;
  ___SCMOBJ result;
  char **argv;
  char **env = NULL;

  if ((e = ___SCMOBJ_to_NONNULLCHARSTRINGLIST (path_and_args, &argv, 1))
      == ___FIX(___NO_ERR))
    {
      if (environment == ___FAL ||
          (e = ___SCMOBJ_to_NONNULLCHARSTRINGLIST (environment, &env, 2))
          == ___FIX(___NO_ERR))
        {
#ifdef USE_POSIX
          //////////////////////////////
          e = ___device_stream_setup_from_process
                (&dev,
                 io_mod.dgroup,
                 argv,
                 env,
                 ___INT(options));
#endif

          if (env != NULL)
            ___release_string_list (env);

          if (e == ___FIX(___NO_ERR))
            {
              if ((e = ___NONNULLPOINTER_to_SCMOBJ
                         (dev,
                          ___FAL,
                          device_cleanup_from_ptr,
                          &result,
                          ___RETURN_POS))
                  == ___FIX(___NO_ERR))
                {
                  ___release_string_list (argv);
                  ___release_scmobj (result);
                  return result;
                }
            }
        }

      ___release_string_list (argv);
    }

  return e;
}


/*   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   */

/* Access to host information. */

___SCMOBJ ___os_host_info
   ___P((___SCMOBJ host),
        (host)
___SCMOBJ host;)
{
#ifndef USE_gethostbyname

  return ___FIX(___UNKNOWN_ERR);

#endif

#ifdef USE_gethostbyname

  ___SCMOBJ e;
  ___SCMOBJ vect;
  ___SCMOBJ lst;
  ___SCMOBJ x;
  ___SCMOBJ p;
  int i;
  struct hostent *he;
  char *chost;

  /*
   * Convert the Scheme string to a C "char*" string.  If an invalid
   * character is seen then return an error.
   */

  if ((e = ___SCMOBJ_to_NONNULLCHARSTRING (host, &chost, 1))
      != ___FIX(___NO_ERR))
    return e;

#ifdef USE_POSIX

  errno = 0; /* in case the h_errno ends up being NETDB_SUCCESS
              * incorrectly which will be treated as NETDB_INTERNAL
              * (see err_code_from_h_errno)
              */

#endif

  if ((he = gethostbyname (chost)) == 0)
    {
#ifdef USE_POSIX
      e = err_code_from_h_errno ();
#endif

#ifdef USE_WIN32
      e = err_code_from_WSAGetLastError ();
#endif
    }

  ___release_string (chost);

  if (e != ___FIX(___NO_ERR))
    return e;

  vect = ___make_vector (4, ___FAL, ___STILL);

  if (___FIXNUMP(vect))
    return ___FIX(___CTOS_HEAP_OVERFLOW_ERR+___RETURN_POS);/************/

  /* convert h_name to string */

  if ((e = ___CHARSTRING_to_SCMOBJ (he->h_name, &x, ___RETURN_POS))
      != ___FIX(___NO_ERR))
    {
      ___release_scmobj (vect);
      return e;
    }

  ___FIELD(vect,1) = x;
  ___release_scmobj (x);

  /* convert h_aliases to strings */

  i = 0;
  while (he->h_aliases[i] != 0)
    i++;

  lst = ___NUL;
  while (i-- > 0)
    {
      if ((e = ___CHARSTRING_to_SCMOBJ (he->h_aliases[i], &x, ___RETURN_POS))
          != ___FIX(___NO_ERR))
        {
          ___release_scmobj (lst);
          ___release_scmobj (vect);
          return e;
        }

      p = ___make_pair (x, lst, ___STILL);

      ___release_scmobj (x);
      ___release_scmobj (lst);

      if (___FIXNUMP(p))
        {
          ___release_scmobj (vect);
          return ___FIX(___CTOS_HEAP_OVERFLOW_ERR+___RETURN_POS);/*******************/
        }

      lst = p;
    }

  ___FIELD(vect,2) = lst;
  ___release_scmobj (lst);

  /* convert h_addr_list to u8vectors */

  i = 0;
  while (he->h_addr_list[i] != 0)
    i++;

  lst = ___NUL;
  while (i-- > 0)
    {
      switch (he->h_addrtype)
        {
        case AF_INET:
          {
            unsigned long host_addr;

            host_addr =
              ntohl (___CAST(struct in_addr*,he->h_addr_list[i])->s_addr);

            x = ___alloc_scmobj (___sU8VECTOR, 4, ___STILL);

            if (___FIXNUMP(x))
              {
                ___release_scmobj (lst);
                ___release_scmobj (vect);
                return ___FIX(___HEAP_OVERFLOW_ERR);/***************/
              }

            ___U8VECTORSET(x,___FIX(0),___FIX((host_addr>>24)&0xff))
            ___U8VECTORSET(x,___FIX(1),___FIX((host_addr>>16)&0xff))
            ___U8VECTORSET(x,___FIX(2),___FIX((host_addr>>8)&0xff))
            ___U8VECTORSET(x,___FIX(3),___FIX(host_addr&0xff))

            break;
          }

        default:
#ifdef USE_IPV6
        case AF_INET6:
#endif
          continue; /* skip this address */
        }

      p = ___make_pair (x, lst, ___STILL);

      ___release_scmobj (x);
      ___release_scmobj (lst);

      if (___FIXNUMP(p))
        {
          ___release_scmobj (vect);
          return ___FIX(___CTOS_HEAP_OVERFLOW_ERR+___RETURN_POS);/*******************/
        }

      lst = p;
    }

  ___FIELD(vect,3) = lst;
  ___release_scmobj (lst);
  ___release_scmobj (vect);

  /* guarantee that at least one address is returned */

  if (lst == ___NUL)
    return ___FIX(___H_ERRNO_ERR(NO_ADDRESS));

  return vect;

#endif
}


/*   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   */

/* Access to service information. */

___SCMOBJ ___os_service_info
   ___P((___SCMOBJ service,
         ___SCMOBJ protocol),
        (service,
         protocol)
___SCMOBJ service;
___SCMOBJ protocol;)
{
#ifndef USE_getservbyname

  return ___FIX(___UNKNOWN_ERR);

#endif

#ifdef USE_getservbyname

  ___SCMOBJ e;
  ___SCMOBJ vect;
  ___SCMOBJ lst;
  ___SCMOBJ x;
  ___SCMOBJ p;
  int i;
  struct servent *se;
  char *cservice;
  char *cprotocol;

  /*
   * Convert the Scheme string to a C "char*" string.  If an invalid
   * character is seen then return an error.
   */

  if (!___FIXNUMP(service))
    if ((e = ___SCMOBJ_to_NONNULLCHARSTRING (service, &cservice, 1))
        != ___FIX(___NO_ERR))
      return e;

  if ((e = ___SCMOBJ_to_CHARSTRING (protocol, &cprotocol, 2))
      != ___FIX(___NO_ERR))
    {
      if (!___FIXNUMP(service))
        ___release_string (cservice);
      return e;
    }

#ifdef USE_POSIX

  errno = 0; /* in case the h_errno ends up being NETDB_SUCCESS
              * incorrectly which will be treated as NETDB_INTERNAL
              * (see err_code_from_h_errno)
              */

#endif

  if (___FIXNUMP(service))
    se = getservbyport (___INT(service), cprotocol);
  else
    se = getservbyname (cservice, cprotocol);

  if (se == 0)
    {
#ifdef USE_POSIX
      e = err_code_from_h_errno ();
#endif

#ifdef USE_WIN32
      e = err_code_from_WSAGetLastError ();
#endif
    }

  if (cprotocol != 0)
    ___release_string (cprotocol);

  if (!___FIXNUMP(service))
    ___release_string (cservice);

  if (e != ___FIX(___NO_ERR))
    return e;

  vect = ___make_vector (5, ___FAL, ___STILL);

  if (___FIXNUMP(vect))
    return ___FIX(___CTOS_HEAP_OVERFLOW_ERR+___RETURN_POS);/************/

  /* convert s_name to string */

  if ((e = ___CHARSTRING_to_SCMOBJ (se->s_name, &x, ___RETURN_POS))
      != ___FIX(___NO_ERR))
    {
      ___release_scmobj (vect);
      return e;
    }

  ___FIELD(vect,1) = x;
  ___release_scmobj (x);

  /* convert s_aliases to strings */

  i = 0;
  while (se->s_aliases[i] != 0)
    i++;

  lst = ___NUL;
  while (i-- > 0)
    {
      if ((e = ___CHARSTRING_to_SCMOBJ (se->s_aliases[i], &x, ___RETURN_POS))
          != ___FIX(___NO_ERR))
        {
          ___release_scmobj (lst);
          ___release_scmobj (vect);
          return e;
        }

      p = ___make_pair (x, lst, ___STILL);

      ___release_scmobj (x);
      ___release_scmobj (lst);

      if (___FIXNUMP(p))
        {
          ___release_scmobj (vect);
          return ___FIX(___CTOS_HEAP_OVERFLOW_ERR+___RETURN_POS);/*******************/
        }

      lst = p;
    }

  ___FIELD(vect,2) = lst;
  ___release_scmobj (lst);

  /* convert s_port to integer */

  ___FIELD(vect,3) = ___FIX(ntohs (se->s_port));

  /* convert s_name to string */

  if ((e = ___CHARSTRING_to_SCMOBJ (se->s_proto, &x, ___RETURN_POS))
      != ___FIX(___NO_ERR))
    {
      ___release_scmobj (vect);
      return e;
    }

  ___FIELD(vect,4) = x;
  ___release_scmobj (x);

  ___release_scmobj (vect);

  return vect;

#endif
}


/*   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   */

/* Access to protocol information. */

___SCMOBJ ___os_protocol_info
   ___P((___SCMOBJ protocol),
        (protocol)
___SCMOBJ protocol;)
{
#ifndef USE_getprotobyname

  return ___FIX(___UNKNOWN_ERR);

#endif

#ifdef USE_getprotobyname

  ___SCMOBJ e;
  ___SCMOBJ vect;
  ___SCMOBJ lst;
  ___SCMOBJ x;
  ___SCMOBJ p;
  int i;
  struct protoent *pe;
  char *cprotocol;

  /*
   * Convert the Scheme string to a C "char*" string.  If an invalid
   * character is seen then return an error.
   */

  if (!___FIXNUMP(protocol))
    if ((e = ___SCMOBJ_to_NONNULLCHARSTRING (protocol, &cprotocol, 1))
        != ___FIX(___NO_ERR))
      return e;

#ifdef USE_POSIX

  errno = 0; /* in case the h_errno ends up being NETDB_SUCCESS
              * incorrectly which will be treated as NETDB_INTERNAL
              * (see err_code_from_h_errno)
              */

#endif

  if (___FIXNUMP(protocol))
    pe = getprotobynumber (___INT(protocol));
  else
    pe = getprotobyname (cprotocol);

  if (pe == 0)
    {
#ifdef USE_POSIX
      e = err_code_from_h_errno ();
#endif

#ifdef USE_WIN32
      e = err_code_from_WSAGetLastError ();
#endif
    }

  if (!___FIXNUMP(protocol))
    ___release_string (cprotocol);

  if (e != ___FIX(___NO_ERR))
    return e;

  vect = ___make_vector (4, ___FAL, ___STILL);

  if (___FIXNUMP(vect))
    return ___FIX(___CTOS_HEAP_OVERFLOW_ERR+___RETURN_POS);/************/

  /* convert p_name to string */

  if ((e = ___CHARSTRING_to_SCMOBJ (pe->p_name, &x, ___RETURN_POS))
      != ___FIX(___NO_ERR))
    {
      ___release_scmobj (vect);
      return e;
    }

  ___FIELD(vect,1) = x;
  ___release_scmobj (x);

  /* convert p_aliases to strings */

  i = 0;
  while (pe->p_aliases[i] != 0)
    i++;

  lst = ___NUL;
  while (i-- > 0)
    {
      if ((e = ___CHARSTRING_to_SCMOBJ (pe->p_aliases[i], &x, ___RETURN_POS))
          != ___FIX(___NO_ERR))
        {
          ___release_scmobj (lst);
          ___release_scmobj (vect);
          return e;
        }

      p = ___make_pair (x, lst, ___STILL);

      ___release_scmobj (x);
      ___release_scmobj (lst);

      if (___FIXNUMP(p))
        {
          ___release_scmobj (vect);
          return ___FIX(___CTOS_HEAP_OVERFLOW_ERR+___RETURN_POS);/*******************/
        }

      lst = p;
    }

  ___FIELD(vect,2) = lst;
  ___release_scmobj (lst);

  /* convert p_proto to integer */

  ___FIELD(vect,3) = ___FIX(pe->p_proto);

  ___release_scmobj (vect);

  return vect;

#endif
}


/*   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   */

/* Access to network information. */

___SCMOBJ ___os_network_info
   ___P((___SCMOBJ network),
        (network)
___SCMOBJ network;)
{
#ifndef USE_getnetbyname

  return ___FIX(___UNKNOWN_ERR);

#endif
}


/*   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   */

/* Opening a TCP/IP client. */

___SCMOBJ ___os_device_tcp_client_open
   ___P((___SCMOBJ server_addr,
         ___SCMOBJ port_num,
         ___SCMOBJ options),
        (server_addr,
         port_num,
         options)
___SCMOBJ server_addr;
___SCMOBJ port_num;
___SCMOBJ options;)
{
  ___SCMOBJ e;
  ___device_tcp_client *dev;
  ___SCMOBJ result;
  struct sockaddr *addr;
  SOCKET_LEN_TYPE addrlen;
  struct sockaddr_in addr_in;

#ifdef USE_IPV6
  struct sockaddr_in6 addr_in6;
#endif

  if (___U8VECTORLENGTH(server_addr) == ___FIX(4))
    {
      addr_in.sin_family = AF_INET;
      addr_in.sin_port = htons (___INT(port_num));
      addr_in.sin_addr.s_addr =
        htonl ((___INT(___U8VECTORREF(server_addr,___FIX(0)))<<24) +
               (___INT(___U8VECTORREF(server_addr,___FIX(1)))<<16) +
               (___INT(___U8VECTORREF(server_addr,___FIX(2)))<<8) +
               ___INT(___U8VECTORREF(server_addr,___FIX(3))));
      addr = ___CAST(struct sockaddr*,&addr_in);
      addrlen = sizeof (addr_in);
    }
  else
    {
#ifdef USE_IPV6
      int i;
      addr_in6.sin6_family = AF_INET6;
      addr_in6.sin6_port = htons (___INT(port_num));
      for (i=0; i<16; i++)
        addr_in6.sin6_addr.s6_addr[i] =
          ___INT(___U8VECTORREF(server_addr,___FIX(i)));
      addr = ___CAST(struct sockaddr*,&addr_in6);
      addrlen = sizeof (addr_in6);
#else
      return ___FIX(___UNKNOWN_ERR);
#endif
    }

  e = ___device_tcp_client_setup_from_sockaddr
        (&dev,
         io_mod.dgroup,
         addr,
         addrlen,
         ___INT(options),
         ___DIRECTION_RD|___DIRECTION_WR);

  if (e != ___FIX(___NO_ERR))
    return e;

  e = ___NONNULLPOINTER_to_SCMOBJ
        (dev,
         ___FAL,
         device_cleanup_from_ptr,
         &result,
         ___RETURN_POS);

  if (e != ___FIX(___NO_ERR))
    {
      ___device_cleanup (___CAST(___device*,dev)); /* ignore error */
      return e;
    }

  ___release_scmobj (result);

  return result;
}


/*   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   */

/* Opening and reading a TCP/IP server. */

___SCMOBJ ___os_device_tcp_server_open
   ___P((___SCMOBJ port_num,
         ___SCMOBJ backlog,
         ___SCMOBJ options),
        (port_num,
         backlog,
         options)
___SCMOBJ port_num;
___SCMOBJ backlog;
___SCMOBJ options;)
{
  ___SCMOBJ e;
  ___device_tcp_server *dev;
  ___SCMOBJ result;

  e = ___device_tcp_server_setup
        (&dev,
         io_mod.dgroup,
         ___INT(port_num),
         ___INT(backlog),
         ___INT(options));

  if (e != ___FIX(___NO_ERR))
    return e;

  e = ___NONNULLPOINTER_to_SCMOBJ
        (dev,
         ___FAL,
         device_cleanup_from_ptr,
         &result,
         ___RETURN_POS);

  if (e != ___FIX(___NO_ERR))
    {
      ___device_cleanup (___CAST(___device*,dev)); /* ignore error */
      return e;
    }

  ___release_scmobj (result);

  return result;
}


___SCMOBJ ___os_device_tcp_server_read
   ___P((___SCMOBJ dev),
        (dev)
___SCMOBJ dev;)
{
  ___device_tcp_server *d =
    ___CAST(___device_tcp_server*,___FIELD(dev,___FOREIGN_PTR));
  ___SCMOBJ e;
  ___device_tcp_client *client;
  ___SCMOBJ result;

  if ((e = ___device_tcp_server_read (d, io_mod.dgroup, &client))
      != ___FIX(___NO_ERR))
    return e;

  e = ___NONNULLPOINTER_to_SCMOBJ
        (client,
         ___FAL,
         device_cleanup_from_ptr,
         &result,
         ___RETURN_POS);

  if (e != ___FIX(___NO_ERR))
    {
      ___device_cleanup (___CAST(___device*,d)); /* ignore error */
      return e;
    }

  ___release_scmobj (result);

  return result;
}


/*   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   */

/* Opening and reading a directory. */

___SCMOBJ ___os_device_directory_open_path
   ___P((___SCMOBJ path,
         ___SCMOBJ ignore_hidden),
        (path,
         ignore_hidden)
___SCMOBJ path;
___SCMOBJ ignore_hidden;)
{
  ___SCMOBJ e;
  ___SCMOBJ result;
  ___device_directory *dev;
  void *cpath;

  if ((e = ___SCMOBJ_to_NONNULLSTRING
             (path,
              &cpath,
              1,
              ___CE(___DIR_OPEN_PATH_CE_SELECT),
              0))
      != ___FIX(___NO_ERR))
    result = e;
  else
    {
      ___STRING_TYPE(___DIR_OPEN_PATH_CE_SELECT) p =
        ___CAST(___STRING_TYPE(___DIR_OPEN_PATH_CE_SELECT),cpath);

      if ((e = ___device_directory_setup
                 (&dev,
                  io_mod.dgroup,
                  p,
                  ___INT(ignore_hidden)))
          != ___FIX(___NO_ERR))
        result = e;
      else
        {
          if ((e = ___NONNULLPOINTER_to_SCMOBJ
                     (dev,
                      ___FAL,
                      device_cleanup_from_ptr,
                      &result,
                      ___RETURN_POS))
              != ___FIX(___NO_ERR))
            {
              ___device_cleanup (___CAST(___device*,dev)); /* ignore error */
              result = e;
            }
        }

      ___release_string (cpath);
    }

  ___release_scmobj (result);

  return result;
}


___SCMOBJ ___os_device_directory_read
   ___P((___SCMOBJ dev),
        (dev)
___SCMOBJ dev;)
{
  ___device_directory *d =
    ___CAST(___device_directory*,___FIELD(dev,___FOREIGN_PTR));
  ___SCMOBJ e;
  char *name;///////////////////////////////////
  ___SCMOBJ result;

  if ((e = ___device_directory_read (d, &name)) != ___FIX(___NO_ERR))
    return e;

  if (name == NULL)
    return ___EOF;

  if ((e = ___CHARSTRING_to_SCMOBJ (name, &result, ___RETURN_POS))
      != ___FIX(___NO_ERR))
    return e;

  ___release_scmobj (result);

  return result;
}


/*   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   */

/* Tty device operations. */

___SCMOBJ ___os_device_tty_type_set
   ___P((___SCMOBJ dev,
         ___SCMOBJ term_type,
         ___SCMOBJ emacs_bindings),
        (dev,
         term_type,
         emacs_bindings)
___SCMOBJ dev;
___SCMOBJ term_type;
___SCMOBJ emacs_bindings;)
{
  return ___FIX(___UNKNOWN_ERR);/////////////////////////////////////////
}


___SCMOBJ ___os_device_tty_text_attributes_set
   ___P((___SCMOBJ dev,
         ___SCMOBJ input,
         ___SCMOBJ output),
        (dev,
         input,
         output)
___SCMOBJ dev;
___SCMOBJ input;
___SCMOBJ output;)
{
  return ___FIX(___UNKNOWN_ERR);/////////////////////////////////////////
}


___SCMOBJ ___os_device_tty_history
   ___P((___SCMOBJ dev),
        (dev)
___SCMOBJ dev;)
{
  return ___FIX(___UNKNOWN_ERR);/////////////////////////////////////////
}


___SCMOBJ ___os_device_tty_history_set
   ___P((___SCMOBJ dev,
         ___SCMOBJ history),
        (dev,
         history)
___SCMOBJ dev;
___SCMOBJ history;)
{
  return ___FIX(___UNKNOWN_ERR);/////////////////////////////////////////
}


___SCMOBJ ___os_device_tty_max_history_length_set
   ___P((___SCMOBJ dev,
         ___SCMOBJ max_length),
        (dev,
         max_length)
___SCMOBJ dev;
___SCMOBJ max_length;)
{
  return ___FIX(___UNKNOWN_ERR);/////////////////////////////////////////
}


___SCMOBJ ___os_device_tty_paren_balance_duration_set
   ___P((___SCMOBJ dev,
         ___SCMOBJ duration),
        (dev,
         duration)
___SCMOBJ dev;
___SCMOBJ duration;)
{
  return ___FIX(___UNKNOWN_ERR);/////////////////////////////////////////
}


___SCMOBJ ___os_device_tty_mode_set
   ___P((___SCMOBJ dev,
         ___SCMOBJ mode),
        (dev,
         mode)
___SCMOBJ dev;
___SCMOBJ mode;)
{
  return ___FIX(___UNKNOWN_ERR);/////////////////////////////////////////
}


/*   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   */

/* Waiting for I/O to become possible on a set of devices. */

___SCMOBJ ___os_condvar_select
   ___P((___SCMOBJ run_queue,
         ___SCMOBJ timeout),
        (run_queue,
         timeout)
___SCMOBJ run_queue;
___SCMOBJ timeout;)
{
#define MAX_CONDVARS SELECT_DEVS_MAX

  /////////////////////////////////////
#define ___BTQ_DEQ_NEXT 1
#define ___BTQ_DEQ_PREV 2
#define ___BTQ_COLOR    3
#define ___BTQ_PARENT   4
#define ___BTQ_LEFT     5
#define ___BTQ_RIGHT    6
#define ___BTQ_LEFTMOST 6
#define ___BTQ_OWNER    7
#define ___CONDVAR_NAME 8

  ___SCMOBJ e;
  ___time to;
  ___device *devs[MAX_CONDVARS];
  ___SCMOBJ condvars[MAX_CONDVARS];
  int read_pos;
  int write_pos;
  ___SCMOBJ condvar;
  int i;
  int j;

  if (timeout == ___FAL)
    to = ___time_neg_infinity;
  else if (timeout == ___TRU)
    to = ___time_pos_infinity;
  else
    ___time_from_seconds (&to, ___F64VECTORREF(timeout,___FIX(0)));

  read_pos = 0;
  write_pos = MAX_CONDVARS;
  condvar = ___FIELD(run_queue,___BTQ_DEQ_NEXT);

  while (condvar != run_queue)
    {
      ___SCMOBJ owner = ___FIELD(condvar,___BTQ_OWNER);
      if (read_pos < write_pos)
        {
          if (owner & ___FIX(2))
            condvars[--write_pos] = condvar;
          else
            condvars[read_pos++] = condvar;
          ___FIELD(condvar,___BTQ_OWNER) = owner & ~___FIX(1);
        }
      else
        {
          to = ___time_neg_infinity;
          ___FIELD(condvar,___BTQ_OWNER) = owner | ___FIX(1);
        }
      condvar = ___FIELD(condvar,___BTQ_DEQ_NEXT);
    }

  i = 0;

  while (i < read_pos)
    {
      devs[i] = ___CAST(___device*,
                        ___FIELD(___FIELD(condvars[i],___CONDVAR_NAME),
                                 ___FOREIGN_PTR));
      i++;
    }

  j = MAX_CONDVARS;

  while (j > write_pos)
    {
      j--;
      devs[i] = ___CAST(___device*,
                        ___FIELD(___FIELD(condvars[j],___CONDVAR_NAME),
                                 ___FOREIGN_PTR));
      i++;
    }

  e = ___device_select (devs, read_pos, MAX_CONDVARS-write_pos, to);

  i = 0;

  while (i < read_pos)
    {
      if (devs[i] == NULL)
        {
          condvar = condvars[i];
          ___FIELD(condvar,___BTQ_OWNER) |= ___FIX(1);
        }
      i++;
    }

  j = MAX_CONDVARS;

  while (j > write_pos)
    {
      j--;
      if (devs[i] == NULL)
        {
          condvar = condvars[j];
          ___FIELD(condvar,___BTQ_OWNER) |= ___FIX(1);
        }
      i++;
    }

  return e;
}


/*   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   */

/*
 * Decoding and encoding of a buffer of Scheme characters to a buffer
 * of bytes.
 */

#define unicode_LF 10
#define unicode_CR 13
#define char_EOL unicode_LF
#define unicode_BOM 0xfeff

#define bytes_per_LATIN1 1
#define max_LATIN1       0xff

#define bytes_per_UTF8 1 /* optimization for single byte case */
#define max_UTF8       0x7f

#define bytes_per_UCS2 2
#define max_UCS2       0xffff

#define bytes_per_UCS4 4
#define max_UCS4       0x7fffffff

#define DECODE_EOL(loop_label) \
if (c != unicode_LF) \
  { \
    if (c != unicode_CR) \
      { \
        options += ___DECODE_STATE_NONE-___DECODE_STATE(options); \
        *chi++ = c; \
        if (chi < end) \
          goto loop_label; \
      } \
    else \
      { \
        int eol = ___EOL_ENCODING(options); \
        if (eol != ___EOL_ENCODING_LF) \
          { \
            if (eol != ___EOL_ENCODING_CR) \
              { \
                int rs = ___DECODE_STATE(options); \
                if (rs == ___DECODE_STATE_LF) \
                  { \
                    options += \
                      ___DECODE_STATE_NONE-___DECODE_STATE_LF; \
                    goto loop_label; \
                  } \
                options += ___DECODE_STATE_CR-rs; \
              } \
            c = char_EOL; \
          } \
        *chi++ = c; \
        if (chi < end) \
          goto loop_label; \
      } \
  } \
else \
  { \
    int eol = ___EOL_ENCODING(options); \
    if (eol != ___EOL_ENCODING_CR) \
      { \
        if (eol != ___EOL_ENCODING_LF) \
          { \
            int rs = ___DECODE_STATE(options); \
            if (rs == ___DECODE_STATE_CR) \
              { \
                options += \
                  ___DECODE_STATE_NONE-___DECODE_STATE_CR; \
                goto loop_label; \
              } \
            options += ___DECODE_STATE_LF-rs; \
          } \
        c = char_EOL; \
      } \
    *chi++ = c; \
    if (chi < end) \
      goto loop_label; \
  }

#define DECODE_CHARS_LOOP(loop_label,bytes_per_char,max_char,err_code,get_char) \
\
loop_label: \
blo += bytes_per_char; \
if (blo <= bhi) \
  { \
    c = get_char(-1); \
    if (max_char <= ___MAX_CHR || \
        c <= ___MAX_CHR) \
      { \
        DECODE_EOL(loop_label) \
      } \
    else \
      { \
        blo -= bytes_per_char; \
        e = err_code; \
      } \
  } \
else \
  { \
    blo -= bytes_per_char; \
    if (bytes_per_char > 1 && \
        blo < bhi &&   /* at least one byte left in buffer? */ \
        eof != ___FAL) /* end-of-file was reached? */ \
      e = err_code; /* trailing partial! */ \
  } \
break;

#define ENCODE_EOL(loop_label,bytes_per_char,put_char) \
switch (___EOL_ENCODING(options)) \
  { \
  case ___EOL_ENCODING_LFCR: \
    bhi += bytes_per_char; \
    if (bhi > end) \
      { \
        bhi -= 2*bytes_per_char; \
        clo--; \
        goto encode_chars_end; \
      } \
    put_char(-2,unicode_LF); \
  case ___EOL_ENCODING_CR: \
    put_char(-1,unicode_CR); \
    break; \
  case ___EOL_ENCODING_CRLF: \
    bhi += bytes_per_char; \
    if (bhi > end) \
      { \
        bhi -= 2*bytes_per_char; \
        clo--; \
        goto encode_chars_end; \
      } \
    put_char(-2,unicode_CR); \
  default: \
    put_char(-1,unicode_LF); \
    break; \
  } \
if (___LINE_BUFFERED(options)) \
  goto encode_chars_end; /* must empty byte buffer */ \
else if (clo < chi) \
  goto loop_label;

#define ENCODE_CHARS_LOOP(loop_label,bytes_per_char,max_char,err_code,put_char) \
\
loop_label: \
c = *clo++; \
if (___MAX_CHR <= max_char || \
    c <= max_char) \
  { \
    bhi += bytes_per_char; \
    if (bhi <= end) \
      { \
        if (c != char_EOL) \
          { \
            put_char(-1,c); \
            if (clo < chi) \
              goto loop_label; \
          } \
        else \
          { \
            ENCODE_EOL(loop_label,bytes_per_char,put_char); \
          } \
      } \
    else \
      { \
        bhi -= bytes_per_char; \
        clo--; \
        goto encode_chars_end; \
      } \
  } \
else \
  { \
    e = err_code; \
    goto encode_chars_end; \
  } \
break;

#define get_LATIN1(i) \
blo[(i)*bytes_per_LATIN1]

#define get_UTF8(i) \
blo[(i)*bytes_per_UTF8]

#define get_UCS2BE(i) \
(___CAST(___UCS2,blo[(i)*bytes_per_UCS2+0]) << 8) + \
___CAST(___UCS2,blo[(i)*bytes_per_UCS2+1])

#define get_UCS2LE(i) \
(___CAST(___UCS2,blo[(i)*bytes_per_UCS2+1]) << 8) + \
___CAST(___UCS2,blo[(i)*bytes_per_UCS2+0])

#define get_UCS4BE(i) \
(((((___CAST(___UCS4,blo[(i)*bytes_per_UCS4+0]) << 8) + \
    ___CAST(___UCS4,blo[(i)*bytes_per_UCS4+1])) << 8) + \
  ___CAST(___UCS4,blo[(i)*bytes_per_UCS4+2])) << 8) + \
___CAST(___UCS4,blo[(i)*bytes_per_UCS4+3])

#define get_UCS4LE(i) \
(((((___CAST(___UCS4,blo[(i)*bytes_per_UCS4+3]) << 8) + \
    ___CAST(___UCS4,blo[(i)*bytes_per_UCS4+2])) << 8) + \
  ___CAST(___UCS4,blo[(i)*bytes_per_UCS4+1])) << 8) + \
___CAST(___UCS4,blo[(i)*bytes_per_UCS4+0])

#define put_LATIN1(i,c) \
bhi[(i)*bytes_per_LATIN1] = (c);

#define put_UTF8(i,c) \
bhi[(i)*bytes_per_UTF8] = (c);

#define put_UCS2BE(i,c) \
bhi[(i)*bytes_per_UCS2+1] = (c) & 0xff; \
bhi[(i)*bytes_per_UCS2+0] = ((c)>>8) & 0xff;

#define put_UCS2LE(i,c) \
bhi[(i)*bytes_per_UCS2+0] = (c) & 0xff; \
bhi[(i)*bytes_per_UCS2+1] = ((c)>>8) & 0xff;

#define put_UCS4BE(i,c) \
bhi[(i)*bytes_per_UCS4+3] = (c) & 0xff; \
bhi[(i)*bytes_per_UCS4+2] = ((c)>>8) & 0xff; \
bhi[(i)*bytes_per_UCS4+1] = ((c)>>16) & 0xff; \
bhi[(i)*bytes_per_UCS4+0] = ((c)>>24) & 0xff;

#define put_UCS4LE(i,c) \
bhi[(i)*bytes_per_UCS4+0] = (c) & 0xff; \
bhi[(i)*bytes_per_UCS4+1] = ((c)>>8) & 0xff; \
bhi[(i)*bytes_per_UCS4+2] = ((c)>>16) & 0xff; \
bhi[(i)*bytes_per_UCS4+3] = ((c)>>24) & 0xff;

//////////////////////////////////////////////////
#define ___PORT_MUTEX                1
#define ___PORT_RKIND                2
#define ___PORT_WKIND                3
#define ___PORT_NAME                 4
#define ___PORT_READ_DATUM           5
#define ___PORT_WRITE_DATUM          6
#define ___PORT_NEWLINE              7
#define ___PORT_FORCE_OUTPUT         8
#define ___PORT_CLOSE                9
#define ___PORT_ROPTIONS             10
#define ___PORT_RTIMEOUT             11
#define ___PORT_RTIMEOUT_THUNK       12
#define ___PORT_SET_RTIMEOUT         13
#define ___PORT_WOPTIONS             14
#define ___PORT_WTIMEOUT             15
#define ___PORT_WTIMEOUT_THUNK       16
#define ___PORT_SET_WTIMEOUT         17

#define ___PORT_OBJECT_OTHER1        18
#define ___PORT_OBJECT_OTHER2        19
#define ___PORT_OBJECT_OTHER3        20

#define ___PORT_CHAR_RBUF            18
#define ___PORT_CHAR_RLO             19
#define ___PORT_CHAR_RHI             20
#define ___PORT_CHAR_RCHARS          21
#define ___PORT_CHAR_RLINES          22
#define ___PORT_CHAR_RCURLINE        23
#define ___PORT_CHAR_RBUF_FILL       24
#define ___PORT_CHAR_PEEK_EOFP       25

#define ___PORT_CHAR_WBUF            26
#define ___PORT_CHAR_WLO             27
#define ___PORT_CHAR_WHI             28
#define ___PORT_CHAR_WCHARS          29
#define ___PORT_CHAR_WLINES          30
#define ___PORT_CHAR_WCURLINE        31
#define ___PORT_CHAR_WBUF_DRAIN      32
#define ___PORT_OUTPUT_WIDTH         33

#define ___PORT_CHAR_OTHER1          34
#define ___PORT_CHAR_OTHER2          35
#define ___PORT_CHAR_OTHER3          36
#define ___PORT_CHAR_OTHER4          37
#define ___PORT_CHAR_OTHER5          38

#define ___PORT_BYTE_RBUF            34
#define ___PORT_BYTE_RLO             35
#define ___PORT_BYTE_RHI             36
#define ___PORT_BYTE_RBUF_FILL       37

#define ___PORT_BYTE_WBUF            38
#define ___PORT_BYTE_WLO             39
#define ___PORT_BYTE_WHI             40
#define ___PORT_BYTE_WBUF_DRAIN      41

#define ___PORT_BYTE_OTHER1          42
#define ___PORT_BYTE_OTHER2          43

#define ___PORT_RDEVICE_CONDVAR      42
#define ___PORT_WDEVICE_CONDVAR      43

#define ___PORT_DEVICE_OTHER1        44
#define ___PORT_DEVICE_OTHER2        45

#define ___C ___CS_SELECT(___U8,___U16,___U32)

___SCMOBJ ___os_port_decode_chars
   ___P((___SCMOBJ port,
         ___SCMOBJ eof),
        (port,
         eof)
___SCMOBJ port;
___SCMOBJ eof;)
{
  ___SCMOBJ e;
  ___SCMOBJ result;
  ___C *char_rbuf =
     ___CAST(___C*,___BODY_AS(___FIELD(port,___PORT_CHAR_RBUF),___tSUBTYPED));
  ___C *chi =
     char_rbuf + ___INT(___FIELD(port,___PORT_CHAR_RHI));
  ___C *end =
     char_rbuf + ___INT(___STRINGLENGTH(___FIELD(port,___PORT_CHAR_RBUF)));
  ___U8 *byte_rbuf =
     ___CAST(___U8*,___BODY_AS(___FIELD(port,___PORT_BYTE_RBUF),___tSUBTYPED));
  ___U8 *blo =
     byte_rbuf + ___INT(___FIELD(port,___PORT_BYTE_RLO));
  ___U8 *bhi =
     byte_rbuf + ___INT(___FIELD(port,___PORT_BYTE_RHI));
  ___UCS4 c;
  int options = ___INT(___FIELD(port,___PORT_ROPTIONS));

  if ((___INT(___FIELD(port,___PORT_RKIND)) & ___TTY_DEVICE_KIND) ==
      ___TTY_DEVICE_KIND)
    {
      ////////////////////////// this special handling should be removed
      /*
       * Characters read from tty stream devices are represented as 8,
       * 16 or 32 bit unsigned integers of the machine's endianness.
       * It is the tty stream device driver that decodes the bytes
       * received from the device using the encoding specified in the
       * ___PORT_ROPTIONS field.
       */

      options = (options - ___CHAR_ENCODING(options)) +
                ___CS_SELECT(___CHAR_ENCODING_U8,
                             ___CHAR_ENCODING_U16,
                             ___CHAR_ENCODING_U32);
    }

  e = ___FIX(___NO_ERR);

  ___FIELD(port,___PORT_CHAR_RCHARS) =
    ___FIELD(port,___PORT_CHAR_RCHARS) + ___FIX(chi-char_rbuf);

  chi = char_rbuf;

  /* fill character buffer as much as possible */

  if (chi < end)
    {
      switch (___CHAR_ENCODING(options))
        {
        default:
        case ___CHAR_ENCODING_ASCII:
        case ___CHAR_ENCODING_LATIN1:
          DECODE_CHARS_LOOP(decode_next_LATIN1,
                            bytes_per_LATIN1,
                            0xff,
                            ___FIX(___CTOS_LATIN1STRING_ERR),
                            get_LATIN1);

        case ___CHAR_ENCODING_UTF8:
          {
            decode_next_UTF8:
            blo += bytes_per_UTF8;
            if (blo <= bhi)
              {
                c = get_UTF8(-1);
                if (c <= 0x7f)
                  {
                    DECODE_EOL(decode_next_UTF8);
                  }
                else if (c <= 0xbf || c > 0xfd)
                  {
                    blo -= bytes_per_UTF8;
                    e = ___FIX(___CTOS_UTF8STRING_ERR);
                  }
                else
                  {
                    ___U8* orig_blo = blo;
                    ___U8 b0 = c;
                    int bits = 6;
                    while (b0 & 0x40)
                      {
                        ___U8 next = *blo++;
                        if (blo > bhi)
                          {
                            blo = orig_blo-bytes_per_UTF8;
                            if (eof != ___FAL) /* end-of-file reached? */
                              e = ___FIX(___CTOS_UTF8STRING_ERR);
                            goto end_UTF8;
                          }
                        if (next <= 0x7f || next > 0xbf)
                          {
                            blo = orig_blo-bytes_per_UTF8;
                            e = ___FIX(___CTOS_UTF8STRING_ERR);
                            goto end_UTF8;
                          }
                        c = (c << 6) + (next & 0x3f);
                        b0 <<= 1;
                        bits += 5;
                      }
                    c &= (___CAST(___UCS4,1)<<bits)-1;
                    if (c >= 0x80 &&
                        c >= (___CAST(___UCS4,1)<<(bits-5)) &&
                        c <= ___MAX_CHR)
                      {
                        options +=
                          ___DECODE_STATE_NONE-___DECODE_STATE(options);
                        *chi++ = c;
                        if (chi < end)
                          goto decode_next_UTF8;
                      }
                    else
                      {
                        blo = orig_blo-bytes_per_UTF8;
                        e = ___FIX(___CTOS_UTF8STRING_ERR);
                      }
                    end_UTF8:;
                  }
              }
            else
              {
                blo -= bytes_per_UTF8;
                if (blo < bhi &&   /* at least one byte left in buffer? */
                    eof != ___FAL) /* end-of-file was reached? */
                  e = ___FIX(___CTOS_UTF8STRING_ERR);
              }
          }

        case ___CHAR_ENCODING_UCS2:
          {
            blo += bytes_per_UCS2;
            if (blo <= bhi)
              {
                ___UCS4 cle;
                c = get_UCS2BE(-1);
                if (c == unicode_BOM)
                  {
                    options += ___CHAR_ENCODING_UCS2BE-___CHAR_ENCODING_UCS2;
                    goto decode_next_UCS2BE;
                  }
                cle = ((c&0xff) << 8) +
                      ((c>>8)&0xff);
                if (cle == unicode_BOM)
                  {
                    options += ___CHAR_ENCODING_UCS2LE-___CHAR_ENCODING_UCS2;
                    goto decode_next_UCS2LE;
                  }
                blo -= bytes_per_UCS2;
#ifdef DEFAULT_TO_BIG_ENDIAN
                options += ___CHAR_ENCODING_UCS2BE-___CHAR_ENCODING_UCS2;
                goto decode_next_UCS2BE;
#else
                options += ___CHAR_ENCODING_UCS2LE-___CHAR_ENCODING_UCS2;
                goto decode_next_UCS2LE;
#endif
              }
            else
              {
                blo -= bytes_per_UCS2;
                if (bytes_per_UCS2 > 1 &&
                    blo < bhi &&   /* at least one byte left in buffer? */
                    eof != ___FAL) /* end-of-file was reached? */
                  e = ___FIX(___CTOS_UCS2STRING_ERR); /* trailing partial! */
              }
            break;
          }

        case ___CHAR_ENCODING_UCS2BE:
          DECODE_CHARS_LOOP(decode_next_UCS2BE,
                            bytes_per_UCS2,
                            0xffff,
                            ___FIX(___CTOS_UCS2STRING_ERR),
                            get_UCS2BE);

        case ___CHAR_ENCODING_UCS2LE:
          DECODE_CHARS_LOOP(decode_next_UCS2LE,
                            bytes_per_UCS2,
                            0xffff,
                            ___FIX(___CTOS_UCS2STRING_ERR),
                            get_UCS2LE);

        case ___CHAR_ENCODING_UCS4:
          {
            blo += bytes_per_UCS4;
            if (blo <= bhi)
              {
                ___UCS4 cle;
                c = get_UCS4BE(-1);
                if (c == unicode_BOM)
                  {
                    options += ___CHAR_ENCODING_UCS4BE-___CHAR_ENCODING_UCS4;
                    goto decode_next_UCS4BE;
                  }
                cle = ((((((c&0xff) << 8) +
                          ((c>>8)&0xff)) << 8) +
                        ((c>>16)&0xff)) << 8) +
                      ((c>>24)&0xff);
                if (cle == unicode_BOM)
                  {
                    options += ___CHAR_ENCODING_UCS4LE-___CHAR_ENCODING_UCS4;
                    goto decode_next_UCS4LE;
                  }
                blo -= bytes_per_UCS4;
#ifdef DEFAULT_TO_BIG_ENDIAN
                options += ___CHAR_ENCODING_UCS4BE-___CHAR_ENCODING_UCS4;
                goto decode_next_UCS4BE;
#else
                options += ___CHAR_ENCODING_UCS4LE-___CHAR_ENCODING_UCS4;
                goto decode_next_UCS4LE;
#endif
              }
            else
              {
                blo -= bytes_per_UCS4;
                if (bytes_per_UCS4 > 1 &&
                    blo < bhi &&   /* at least one byte left in buffer? */
                    eof != ___FAL) /* end-of-file was reached? */
                  e = ___FIX(___CTOS_UCS4STRING_ERR); /* trailing partial! */
              }
            break;
          }

        case ___CHAR_ENCODING_UCS4BE:
          DECODE_CHARS_LOOP(decode_next_UCS4BE,
                            bytes_per_UCS4,
                            0xffffffff,
                            ___FIX(___CTOS_UCS4STRING_ERR),
                            get_UCS4BE);

        case ___CHAR_ENCODING_UCS4LE:
          DECODE_CHARS_LOOP(decode_next_UCS4LE,
                            bytes_per_UCS4,
                            0xffffffff,
                            ___FIX(___CTOS_UCS4STRING_ERR),
                            get_UCS4LE);
        }
    }

  /* 
   * either the character buffer is full (chi == end) or no more
   * characters can be extracted from the byte buffer either because
   * the remaining bytes are not long enough to form a character (e ==
   * ___FIX(___NO_ERR)) or don't form a valid character (e !=
   * ___FIX(___NO_ERR)).
   */

  if (chi == char_rbuf)
    {
      /* could not extract a single character from the byte buffer */

      if (e == ___FIX(___NO_ERR))
        {
          /*
           * the remaining bytes in the buffer (possibly zero bytes)
           * are not long enough to form a character so make some room
           * at the end of the byte buffer to be filled with new bytes.
           */

          ___U8 *p = byte_rbuf;

          if (p != blo)
            {
              while (blo < bhi)
                *p++ = *blo++;
              blo = byte_rbuf;
              bhi = p;
            }
        }

      result = ___FAL; /* need to fill byte buffer */
    }
  else
    {
      e = ___FIX(___NO_ERR); /* only report errors at head of char buffer */
      result = ___TRU; /* no need to fill byte buffer */
    }

  ___FIELD(port,___PORT_CHAR_RLO) = ___FIX(0);
  ___FIELD(port,___PORT_CHAR_RHI) = ___FIX(chi-char_rbuf);
  ___FIELD(port,___PORT_BYTE_RLO) = ___FIX(blo-byte_rbuf);
  ___FIELD(port,___PORT_BYTE_RHI) = ___FIX(bhi-byte_rbuf);
  ___FIELD(port,___PORT_ROPTIONS) = ___FIX(options);

  if (e != ___FIX(___NO_ERR))
    return e;

  return result;
}


___SCMOBJ ___os_port_encode_chars
   ___P((___SCMOBJ port),
        (port)
___SCMOBJ port;)
{
  ___SCMOBJ e;
  ___SCMOBJ result;
  ___C *char_wbuf =
     ___CAST(___C*,___BODY_AS(___FIELD(port,___PORT_CHAR_WBUF),___tSUBTYPED));
  ___C *clo =
     char_wbuf + ___INT(___FIELD(port,___PORT_CHAR_WLO));
  ___C *chi =
     char_wbuf + ___INT(___FIELD(port,___PORT_CHAR_WHI));
  ___U8 *byte_wbuf =
     ___CAST(___U8*,___BODY_AS(___FIELD(port,___PORT_BYTE_WBUF),___tSUBTYPED));
  ___U8 *bhi =
     byte_wbuf + ___INT(___FIELD(port,___PORT_BYTE_WHI));
  ___U8 *end =
     byte_wbuf + ___HD_BYTES(___HEADER(___FIELD(port,___PORT_BYTE_WBUF)));
  ___UCS4 c;
  int options = ___INT(___FIELD(port,___PORT_WOPTIONS));

  if ((___INT(___FIELD(port,___PORT_WKIND)) & ___TTY_DEVICE_KIND) ==
      ___TTY_DEVICE_KIND)
    {
      ////////////////////////// this special handling should be removed
      /*
       * Characters written to tty stream devices are represented as
       * 8, 16 or 32 bit unsigned integers of the machine's
       * endianness.  When the tty stream device driver transfers the
       * characters to the device, the characters will be encoded
       * using the encoding specified in the ___PORT_WOPTIONS field.
       */

      options = (options - ___CHAR_ENCODING(options)) +
                ___CS_SELECT(___CHAR_ENCODING_U8,
                             ___CHAR_ENCODING_U16,
                             ___CHAR_ENCODING_U32);
    }

  e = ___FIX(___NO_ERR);

  result = ___TRU; /* default is to empty byte buffer */

  /* empty character buffer as much as possible */

  if (clo < chi)
    {
      switch (___CHAR_ENCODING(options))
        {
        default:
        case ___CHAR_ENCODING_ASCII:
        case ___CHAR_ENCODING_LATIN1:
          ENCODE_CHARS_LOOP(encode_next_LATIN1,
                            bytes_per_LATIN1,
                            max_LATIN1,
                            ___FIX(___CTOS_LATIN1STRING_ERR),
                            put_LATIN1);

        case ___CHAR_ENCODING_UTF8:
          {
            encode_next_UTF8:
            c = *clo++;
            if (___MAX_CHR <= max_UTF8 ||
                c <= max_UTF8)
              {
                bhi += bytes_per_UTF8;
                if (bhi <= end)
                  {
                    if (c != char_EOL)
                      {
                        put_UTF8(-1,c);
                        if (clo < chi)
                          goto encode_next_UTF8;
                      }
                    else
                      {
                        ENCODE_EOL(encode_next_UTF8,bytes_per_UTF8,put_UTF8);
                      }
                  }
                else
                  {
                    bhi -= bytes_per_UTF8;
                    clo--;
                    goto encode_chars_end;
                  }
              }
            else
              {
                ___U8 *p;
                int bytes;
                if      (c <= 0x7ff)      bytes = 2;
                else if (c <= 0xffff)     bytes = 3;
                else if (c <= 0x1fffff)   bytes = 4;
                else if (c <= 0x3ffffff)  bytes = 5;
                else if (c <= 0x7fffffff) bytes = 6;
                else
                  {
                    e = ___FIX(___STOC_UTF8STRING_ERR);
                    goto encode_chars_end;
                  }
                p = bhi + bytes;
                if (p <= end)
                  {
                    bhi = p;
                    switch (bytes)
                      {
                      case 6:  *--p = 0x80+(c&0x3f); c >>= 6;
                      case 5:  *--p = 0x80+(c&0x3f); c >>= 6;
                      case 4:  *--p = 0x80+(c&0x3f); c >>= 6;
                      case 3:  *--p = 0x80+(c&0x3f); c >>= 6;
                      default: *--p = 0x80+(c&0x3f); c >>= 6;
                      }
                    *--p = 0xff - (0xff>>bytes) + c;
                    if (clo < chi)
                      goto encode_next_UTF8;
                  }
                else
                  {
                    clo--;
                    goto encode_chars_end;
                  }
              }
            break;
          }

        case ___CHAR_ENCODING_UCS2:
          bhi += bytes_per_UCS2;
          if (bhi > end)
            {
              bhi -= bytes_per_UCS2;
              goto encode_chars_end;
            }
#ifdef DEFAULT_TO_BIG_ENDIAN
          put_UCS2BE(-1,unicode_BOM);
          options += ___CHAR_ENCODING_UCS2BE-___CHAR_ENCODING_UCS2;
          ___FIELD(port,___PORT_WOPTIONS) = ___FIX(options);
          goto encode_next_UCS2BE;
#else
          put_UCS2LE(-1,unicode_BOM);
          options += ___CHAR_ENCODING_UCS2LE-___CHAR_ENCODING_UCS2;
          ___FIELD(port,___PORT_WOPTIONS) = ___FIX(options);
          goto encode_next_UCS2LE;
#endif

        case ___CHAR_ENCODING_UCS2BE:
          ENCODE_CHARS_LOOP(encode_next_UCS2BE,
                            bytes_per_UCS2,
                            max_UCS2,
                            ___FIX(___CTOS_UCS2STRING_ERR),
                            put_UCS2BE);

        case ___CHAR_ENCODING_UCS2LE:
          ENCODE_CHARS_LOOP(encode_next_UCS2LE,
                            bytes_per_UCS2,
                            max_UCS2,
                            ___FIX(___CTOS_UCS2STRING_ERR),
                            put_UCS2LE);

        case ___CHAR_ENCODING_UCS4:
          bhi += bytes_per_UCS4;
          if (bhi > end)
            {
              bhi -= bytes_per_UCS4;
              goto encode_chars_end;
            }
#ifdef DEFAULT_TO_BIG_ENDIAN
          put_UCS4BE(-1,unicode_BOM);
          options += ___CHAR_ENCODING_UCS4BE-___CHAR_ENCODING_UCS4;
          ___FIELD(port,___PORT_WOPTIONS) = ___FIX(options);
          goto encode_next_UCS4BE;
#else
          put_UCS4LE(-1,unicode_BOM);
          options += ___CHAR_ENCODING_UCS4LE-___CHAR_ENCODING_UCS4;
          ___FIELD(port,___PORT_WOPTIONS) = ___FIX(options);
          goto encode_next_UCS4LE;
#endif

        case ___CHAR_ENCODING_UCS4BE:
          ENCODE_CHARS_LOOP(encode_next_UCS4BE,
                            bytes_per_UCS4,
                            max_UCS4,
                            ___FIX(___CTOS_UCS4STRING_ERR),
                            put_UCS4BE);

        case ___CHAR_ENCODING_UCS4LE:
          ENCODE_CHARS_LOOP(encode_next_UCS4LE,
                            bytes_per_UCS4,
                            max_UCS4,
                            ___FIX(___CTOS_UCS4STRING_ERR),
                            put_UCS4LE);
        }
    }

  /* character buffer has been fully emptied */

  ___FIELD(port,___PORT_CHAR_WCHARS) =
    ___FIELD(port,___PORT_CHAR_WCHARS) + ___FIX(chi-char_wbuf);

  clo = char_wbuf;
  chi = char_wbuf;

  result = ___FAL; /* do not empty byte buffer */

  encode_chars_end:

  ___FIELD(port,___PORT_CHAR_WLO) = ___FIX(clo-char_wbuf);
  ___FIELD(port,___PORT_CHAR_WHI) = ___FIX(chi-char_wbuf);
  ___FIELD(port,___PORT_BYTE_WHI) = ___FIX(bhi-byte_wbuf);

  if (e != ___FIX(___NO_ERR))
    return e;

  return result;
}


/*   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   */

/* Access to file information. */

___SCMOBJ ___os_file_info
   ___P((___SCMOBJ path,
         ___SCMOBJ chase),
        (path,
         chase)
___SCMOBJ path;
___SCMOBJ chase;)
{
  ___SCMOBJ e;
  ___SCMOBJ result;
  ___SCMOBJ x;
  void *cpath;

#ifndef USE_POSIX
#ifndef USE_WIN32

  return ___FIX(___UNKNOWN_ERR);

#endif
#endif

#ifdef USE_POSIX

  struct stat s;

#define ___INFO_PATH_CE_SELECT(latin1,utf8,ucs2,ucs4,native) native

  if ((e = ___SCMOBJ_to_NONNULLSTRING
             (path,
              &cpath,
              1,
              ___CE(___INFO_PATH_CE_SELECT),
              0))
      == ___FIX(___NO_ERR))
    {
      if (((chase == ___FAL)
           ? lstat (___CAST(___STRING_TYPE(___INFO_PATH_CE_SELECT),cpath), &s)
           : stat (___CAST(___STRING_TYPE(___INFO_PATH_CE_SELECT),cpath), &s))
          < 0)
        {
          e = fnf_or_err_code_from_errno ();
          ___release_string (cpath);
          return e;
        }

      ___release_string (cpath);

      result = ___make_vector (14, ___FAL, ___STILL);

      if (___FIXNUMP(result))
        return ___FIX(___CTOS_HEAP_OVERFLOW_ERR+___RETURN_POS);/**********/

      if (S_ISREG(s.st_mode))
        ___FIELD(result,1) = ___FIX(1);
      else if (S_ISDIR(s.st_mode))
        ___FIELD(result,1) = ___FIX(2);
      else if (S_ISCHR(s.st_mode))
        ___FIELD(result,1) = ___FIX(3);
      else if (S_ISBLK(s.st_mode))
        ___FIELD(result,1) = ___FIX(4);
      else if (S_ISFIFO(s.st_mode))
        ___FIELD(result,1) = ___FIX(5);
      else if (S_ISLNK(s.st_mode))
        ___FIELD(result,1) = ___FIX(6);
      else if (S_ISSOCK(s.st_mode))
        ___FIELD(result,1) = ___FIX(7);
      else
        ___FIELD(result,1) = ___FIX(0);

      if ((e = ___ULONGLONG_to_SCMOBJ (s.st_dev, &x, ___RETURN_POS))
          != ___FIX(___NO_ERR))
        {
          ___release_scmobj (result);
          return e;
        }

      ___FIELD(result,2) = x;
      ___release_scmobj (x);

      if ((e = ___LONGLONG_to_SCMOBJ (s.st_ino, &x, ___RETURN_POS))
          != ___FIX(___NO_ERR))
        {
          ___release_scmobj (result);
          return e;
        }

      ___FIELD(result,3) = x;
      ___release_scmobj (x);

      ___FIELD(result,4) =
        ___FIX(s.st_mode & (S_ISUID|S_ISGID|S_ISVTX|S_IRWXU|S_IRWXG|S_IRWXO));

      if ((e = ___ULONGLONG_to_SCMOBJ (s.st_nlink, &x, ___RETURN_POS))
          != ___FIX(___NO_ERR))
        {
          ___release_scmobj (result);
          return e;
        }

      ___FIELD(result,5) = x;
      ___release_scmobj (x);

      ___FIELD(result,6) = ___FIX(s.st_uid);

      ___FIELD(result,7) = ___FIX(s.st_gid);

      if ((e = ___LONGLONG_to_SCMOBJ (s.st_size, &x, ___RETURN_POS))
          != ___FIX(___NO_ERR))
        {
          ___release_scmobj (result);
          return e;
        }

      ___FIELD(result,8) = x;
      ___release_scmobj (x);

      if ((e = ___F64_to_SCMOBJ (s.st_atime, &x, ___RETURN_POS))
          != ___FIX(___NO_ERR))
        {
          ___release_scmobj (result);
          return e;
        }

      ___FIELD(result,9) = x;
      ___release_scmobj (x);

      if ((e = ___F64_to_SCMOBJ (s.st_mtime, &x, ___RETURN_POS))
          != ___FIX(___NO_ERR))
        {
          ___release_scmobj (result);
          return e;
        }

      ___FIELD(result,10) = x;
      ___release_scmobj (x);

      if ((e = ___F64_to_SCMOBJ (s.st_ctime, &x, ___RETURN_POS))
          != ___FIX(___NO_ERR))
        {
          ___release_scmobj (result);
          return e;
        }

      ___FIELD(result,11) = x;
      ___release_scmobj (x);

#define FILE_ATTRIBUTE_READ_ONLY 1
#define FILE_ATTRIBUTE_DIRECTORY 16
#define FILE_ATTRIBUTE_NORMAL    128

      ___FIELD(result,12) =
        ___FIX(S_ISDIR(s.st_mode)
               ? FILE_ATTRIBUTE_DIRECTORY
               : FILE_ATTRIBUTE_NORMAL);

      if ((e = ___F64_to_SCMOBJ (NEG_INFINITY, &x, ___RETURN_POS))
          != ___FIX(___NO_ERR))
        {
          ___release_scmobj (result);
          return e;
        }

      ___FIELD(result,13) = x;
      ___release_scmobj (x);

      ___release_scmobj (result);

      return result;
    }

#endif

#ifdef USE_WIN32

  e = ___FIX(___UNKNOWN_ERR);

#endif

  return e;
}


/*   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   */

/* Access to user information. */

___SCMOBJ ___os_user_info
   ___P((___SCMOBJ user),
        (user)
___SCMOBJ user;)
{
  ___SCMOBJ e = ___FIX(___NO_ERR);
  ___SCMOBJ result;
  ___SCMOBJ x;
  void *cuser = 0;

#ifndef USE_POSIX
#ifndef USE_WIN32

  return ___FIX(___UNKNOWN_ERR);

#endif
#endif

#ifdef USE_POSIX

  struct passwd *p;

#define ___USER_CE_SELECT(latin1,utf8,ucs2,ucs4,native) native

  if (___FIXNUMP(user) ||
      (e = ___SCMOBJ_to_NONNULLSTRING
             (user,
              &cuser,
              1,
              ___CE(___USER_CE_SELECT),
              0))
      == ___FIX(___NO_ERR))
    {
      if (___FIXNUMP(user))
        {
          if ((p = getpwuid (___INT(user)))
              == 0)
            {
              e = err_code_from_errno ();
              return e;
            }
        }
      else
        {
          if ((p = getpwnam (___CAST(___STRING_TYPE(___USER_CE_SELECT),cuser)))
              == 0)
            {
              e = err_code_from_errno ();
              ___release_string (cuser);
              return e;
            }

          ___release_string (cuser);
        }

      result = ___make_vector (6, ___FAL, ___STILL);

      if (___FIXNUMP(result))
        return ___FIX(___CTOS_HEAP_OVERFLOW_ERR+___RETURN_POS);/**********/

      if ((e = ___NONNULLCHARSTRING_to_SCMOBJ
                 (p->pw_name,
                  &x,
                  ___RETURN_POS))
          != ___FIX(___NO_ERR))
        {
          ___release_scmobj (result);
          return e;
        }

      ___FIELD(result,1) = x;
      ___release_scmobj (x);

      ___FIELD(result,2) = ___FIX(p->pw_uid);

      ___FIELD(result,3) = ___FIX(p->pw_gid);

      if ((e = ___NONNULLCHARSTRING_to_SCMOBJ
                 (p->pw_dir,
                  &x,
                  ___RETURN_POS))
          != ___FIX(___NO_ERR))
        {
          ___release_scmobj (result);
          return e;
        }

      ___FIELD(result,4) = x;
      ___release_scmobj (x);

      if ((e = ___NONNULLCHARSTRING_to_SCMOBJ
                 (p->pw_shell,
                  &x,
                  ___RETURN_POS))
          != ___FIX(___NO_ERR))
        {
          ___release_scmobj (result);
          return e;
        }

      ___FIELD(result,5) = x;
      ___release_scmobj (x);

      ___release_scmobj (result);

      return result;
    }

#endif

#ifdef USE_WIN32

  e = ___FIX(___UNKNOWN_ERR);

#endif

  return e;
}


/*   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   */

/* Access to group information. */

___SCMOBJ ___os_group_info
   ___P((___SCMOBJ group),
        (group)
___SCMOBJ group;)
{
  ___SCMOBJ e = ___FIX(___NO_ERR);
  ___SCMOBJ result;
  ___SCMOBJ x;
  void *cgroup = 0;

#ifndef USE_POSIX
#ifndef USE_WIN32

  return ___FIX(___UNKNOWN_ERR);

#endif
#endif

#ifdef USE_POSIX

  struct group *g;

#define ___GROUP_CE_SELECT(latin1,utf8,ucs2,ucs4,native) native

  if (___FIXNUMP(group) ||
      (e = ___SCMOBJ_to_NONNULLSTRING
             (group,
              &cgroup,
              1,
              ___CE(___GROUP_CE_SELECT),
              0))
      == ___FIX(___NO_ERR))
    {
      if (___FIXNUMP(group))
        {
          if ((g = getgrgid (___INT(group)))
              == 0)
            {
              e = err_code_from_errno ();
              return e;
            }
        }
      else
        {
          if ((g = getgrnam (___CAST(___STRING_TYPE(___GROUP_CE_SELECT),cgroup)))
              == 0)
            {
              e = err_code_from_errno ();
              ___release_string (cgroup);
              return e;
            }

          ___release_string (cgroup);
        }

      result = ___make_vector (3, ___FAL, ___STILL);

      if (___FIXNUMP(result))
        return ___FIX(___CTOS_HEAP_OVERFLOW_ERR+___RETURN_POS);/**********/

      if ((e = ___NONNULLCHARSTRING_to_SCMOBJ
                 (g->gr_name,
                  &x,
                  ___RETURN_POS))
          != ___FIX(___NO_ERR))
        {
          ___release_scmobj (result);
          return e;
        }

      ___FIELD(result,1) = x;
      ___release_scmobj (x);

      ___FIELD(result,2) = ___FIX(g->gr_gid);

      if ((e = ___NONNULLCHARSTRINGLIST_to_SCMOBJ
                 (g->gr_mem,
                  &x,
                  ___RETURN_POS))
          != ___FIX(___NO_ERR))
        {
          ___release_scmobj (result);
          return e;
        }

      ___FIELD(result,3) = x;
      ___release_scmobj (x);

      ___release_scmobj (result);

      return result;
    }

#endif

#ifdef USE_WIN32

  e = ___FIX(___UNKNOWN_ERR);

#endif

  return e;
}


/*   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   */

/* Filesystem operations. */

___SCMOBJ ___os_create_directory
   ___P((___SCMOBJ path,
         ___SCMOBJ mode),
        (path,
         mode)
___SCMOBJ path;
___SCMOBJ mode;)
{
  ___SCMOBJ e;
  void *cpath;

#ifndef USE_POSIX
#ifndef USE_WIN32

  e = ___FIX(___UNKNOWN_ERR);

#endif
#endif

#ifdef USE_POSIX

#define ___CREATE_DIRECTORY_PATH_CE_SELECT(latin1,utf8,ucs2,ucs4,native) native

  if ((e = ___SCMOBJ_to_NONNULLSTRING
             (path,
              &cpath,
              1,
              ___CE(___CREATE_DIRECTORY_PATH_CE_SELECT),
              0))
      == ___FIX(___NO_ERR))
    {
      if (mkdir (___CAST(___STRING_TYPE(___CREATE_DIRECTORY_PATH_CE_SELECT),cpath), ___INT(mode)) < 0)
        e = fnf_or_err_code_from_errno ();
      ___release_string (cpath);
    }

#endif

#ifdef USE_WIN32

#ifdef UNICODE
#define ___CREATE_DIRECTORY_PATH_CE_SELECT(latin1,utf8,ucs2,ucs4,native) ucs2
#else
#define ___CREATE_DIRECTORY_PATH_CE_SELECT(latin1,utf8,ucs2,ucs4,native) latin1
#endif

  if ((e = ___SCMOBJ_to_NONNULLSTRING
             (path,
              &cpath,
              1,
              ___CE(___CREATE_DIRECTORY_PATH_CE_SELECT),
              0))
      == ___FIX(___NO_ERR))
    {
      if (!CreateDirectory (___CAST(___STRING_TYPE(___CREATE_DIRECTORY_PATH_CE_SELECT),cpath), NULL))
        e = fnf_or_err_code_from_GetLastError ();
      ___release_string (cpath);
    }

#endif

  return e;
}


___SCMOBJ ___os_create_fifo
   ___P((___SCMOBJ path,
         ___SCMOBJ mode),
        (path,
         mode)
___SCMOBJ path;
___SCMOBJ mode;)
{
  ___SCMOBJ e;
  void *cpath;

#ifndef USE_POSIX
#ifndef USE_WIN32

  e = ___FIX(___UNKNOWN_ERR);

#endif
#endif

#ifdef USE_POSIX

#define ___CREATE_FIFO_PATH_CE_SELECT(latin1,utf8,ucs2,ucs4,native) native

  if ((e = ___SCMOBJ_to_NONNULLSTRING
             (path,
              &cpath,
              1,
              ___CE(___CREATE_FIFO_PATH_CE_SELECT),
              0))
      == ___FIX(___NO_ERR))
    {
      if (mkfifo (___CAST(___STRING_TYPE(___CREATE_FIFO_PATH_CE_SELECT),cpath), ___INT(mode)) < 0)
        e = fnf_or_err_code_from_errno ();
      ___release_string (cpath);
    }

#endif

#ifdef USE_WIN32

#ifdef UNICODE
#define ___CREATE_FIFO_PATH_CE_SELECT(latin1,utf8,ucs2,ucs4,native) ucs2
#else
#define ___CREATE_FIFO_PATH_CE_SELECT(latin1,utf8,ucs2,ucs4,native) latin1
#endif

  if ((e = ___SCMOBJ_to_NONNULLSTRING
             (path,
              &cpath,
              1,
              ___CE(___CREATE_FIFO_PATH_CE_SELECT),
              0))
      == ___FIX(___NO_ERR))
    {
      e = ___FIX(___UNKNOWN_ERR);
      ___release_string (cpath);
    }

#endif

  return e;
}


___SCMOBJ ___os_create_link
   ___P((___SCMOBJ path1,
         ___SCMOBJ path2),
        (path1,
         path2)
___SCMOBJ path1;
___SCMOBJ path2;)
{
  ___SCMOBJ e;
  void *cpath1;
  void *cpath2;

#ifndef USE_POSIX
#ifndef USE_WIN32

  e = ___FIX(___UNKNOWN_ERR);

#endif
#endif

#ifdef USE_POSIX

#define ___CREATE_LINK_PATH_CE_SELECT(latin1,utf8,ucs2,ucs4,native) native

  if ((e = ___SCMOBJ_to_NONNULLSTRING
             (path1,
              &cpath1,
              1,
              ___CE(___CREATE_LINK_PATH_CE_SELECT),
              0))
      == ___FIX(___NO_ERR))
    {
      if ((e = ___SCMOBJ_to_NONNULLSTRING
                 (path2,
                  &cpath2,
                  2,
                  ___CE(___CREATE_LINK_PATH_CE_SELECT),
                  0))
          == ___FIX(___NO_ERR))
        {
          if (link (___CAST(___STRING_TYPE(___CREATE_LINK_PATH_CE_SELECT),cpath1),
                    ___CAST(___STRING_TYPE(___CREATE_LINK_PATH_CE_SELECT),cpath2))
              < 0)
            e = fnf_or_err_code_from_errno ();
          ___release_string (cpath2);
        }
      ___release_string (cpath1);
    }
 
#endif

#ifdef USE_WIN32

#ifdef UNICODE
#define ___CREATE_LINK_PATH_CE_SELECT(latin1,utf8,ucs2,ucs4,native) ucs2
#else
#define ___CREATE_LINK_PATH_CE_SELECT(latin1,utf8,ucs2,ucs4,native) latin1
#endif

  if ((e = ___SCMOBJ_to_NONNULLSTRING
             (path1,
              &cpath1,
              1,
              ___CE(___CREATE_LINK_PATH_CE_SELECT),
              0))
      == ___FIX(___NO_ERR))
    {
      if ((e = ___SCMOBJ_to_NONNULLSTRING
                 (path2,
                  &cpath2,
                  2,
                  ___CE(___CREATE_LINK_PATH_CE_SELECT),
                  0))
          == ___FIX(___NO_ERR))
        {
          e = ___FIX(___UNKNOWN_ERR);
          ___release_string (cpath2);
        }
      ___release_string (cpath1);
    }

#endif

  return e;
}


___SCMOBJ ___os_create_symbolic_link
   ___P((___SCMOBJ path1,
         ___SCMOBJ path2),
        (path1,
         path2)
___SCMOBJ path1;
___SCMOBJ path2;)
{
  ___SCMOBJ e;
  void *cpath1;
  void *cpath2;

#ifndef USE_POSIX
#ifndef USE_WIN32

  e = ___FIX(___UNKNOWN_ERR);

#endif
#endif

#ifdef USE_POSIX

#define ___CREATE_SYMLINK_PATH_CE_SELECT(latin1,utf8,ucs2,ucs4,native) native

  if ((e = ___SCMOBJ_to_NONNULLSTRING
             (path1,
              &cpath1,
              1,
              ___CE(___CREATE_SYMLINK_PATH_CE_SELECT),
              0))
      == ___FIX(___NO_ERR))
    {
      if ((e = ___SCMOBJ_to_NONNULLSTRING
                 (path2,
                  &cpath2,
                  2,
                  ___CE(___CREATE_SYMLINK_PATH_CE_SELECT),
                  0))
          == ___FIX(___NO_ERR))
        {
          if (symlink (___CAST(___STRING_TYPE(___CREATE_SYMLINK_PATH_CE_SELECT),cpath1),
                       ___CAST(___STRING_TYPE(___CREATE_SYMLINK_PATH_CE_SELECT),cpath2))
              < 0)
            e = fnf_or_err_code_from_errno ();
          ___release_string (cpath2);
        }
      ___release_string (cpath1);
    }
 
#endif

#ifdef USE_WIN32

#ifdef UNICODE
#define ___CREATE_SYMLINK_PATH_CE_SELECT(latin1,utf8,ucs2,ucs4,native) ucs2
#else
#define ___CREATE_SYMLINK_PATH_CE_SELECT(latin1,utf8,ucs2,ucs4,native) latin1
#endif

  if ((e = ___SCMOBJ_to_NONNULLSTRING
             (path1,
              &cpath1,
              1,
              ___CE(___CREATE_SYMLINK_PATH_CE_SELECT),
              0))
      == ___FIX(___NO_ERR))
    {
      if ((e = ___SCMOBJ_to_NONNULLSTRING
                 (path2,
                  &cpath2,
                  2,
                  ___CE(___CREATE_SYMLINK_PATH_CE_SELECT),
                  0))
          == ___FIX(___NO_ERR))
        {
          e = ___FIX(___UNKNOWN_ERR);
          ___release_string (cpath2);
        }
      ___release_string (cpath1);
    }

#endif

  return e;
}


___SCMOBJ ___os_delete_directory
   ___P((___SCMOBJ path),
        (path)
___SCMOBJ path;)
{
  ___SCMOBJ e;
  void *cpath;

#ifndef USE_POSIX
#ifndef USE_WIN32

  e = ___FIX(___UNKNOWN_ERR);

#endif
#endif

#ifdef USE_POSIX

#define ___DELETE_DIRECTORY_PATH_CE_SELECT(latin1,utf8,ucs2,ucs4,native) native

  if ((e = ___SCMOBJ_to_NONNULLSTRING
             (path,
              &cpath,
              1,
              ___CE(___DELETE_DIRECTORY_PATH_CE_SELECT),
              0))
      == ___FIX(___NO_ERR))
    {
      if (rmdir (___CAST(___STRING_TYPE(___DELETE_DIRECTORY_PATH_CE_SELECT),cpath)) < 0)
        e = fnf_or_err_code_from_errno ();
      ___release_string (cpath);
    }

#endif

#ifdef USE_WIN32

#ifdef UNICODE
#define ___DELETE_DIRECTORY_PATH_CE_SELECT(latin1,utf8,ucs2,ucs4,native) ucs2
#else
#define ___DELETE_DIRECTORY_PATH_CE_SELECT(latin1,utf8,ucs2,ucs4,native) latin1
#endif

  if ((e = ___SCMOBJ_to_NONNULLSTRING
             (path,
              &cpath,
              1,
              ___CE(___DELETE_DIRECTORY_PATH_CE_SELECT),
              0))
      == ___FIX(___NO_ERR))
    {
      if (!RemoveDirectory (___CAST(___STRING_TYPE(___DELETE_DIRECTORY_PATH_CE_SELECT),cpath)))
        e = fnf_or_err_code_from_GetLastError ();
      ___release_string (cpath);
    }

#endif

  return e;
}


___SCMOBJ ___os_set_current_directory
   ___P((___SCMOBJ path),
        (path)
___SCMOBJ path;)
{
  ___SCMOBJ e;
  void *cpath;

#ifndef USE_POSIX
#ifndef USE_WIN32

  e = ___FIX(___UNKNOWN_ERR);

#endif
#endif

#ifdef USE_POSIX

#define ___SET_CURRENT_DIRECTORY_PATH_CE_SELECT(latin1,utf8,ucs2,ucs4,native) native

  if ((e = ___SCMOBJ_to_NONNULLSTRING
             (path,
              &cpath,
              1,
              ___CE(___SET_CURRENT_DIRECTORY_PATH_CE_SELECT),
              0))
      == ___FIX(___NO_ERR))
    {
      if (chdir (___CAST(___STRING_TYPE(___SET_CURRENT_DIRECTORY_PATH_CE_SELECT),cpath)) < 0)
        e = fnf_or_err_code_from_errno ();
      ___release_string (cpath);
    }

#endif

#ifdef USE_WIN32

#ifdef UNICODE
#define ___SET_CURRENT_DIRECTORY_PATH_CE_SELECT(latin1,utf8,ucs2,ucs4,native) ucs2
#else
#define ___SET_CURRENT_DIRECTORY_PATH_CE_SELECT(latin1,utf8,ucs2,ucs4,native) latin1
#endif

  if ((e = ___SCMOBJ_to_NONNULLSTRING
             (path,
              &cpath,
              1,
              ___CE(___SET_CURRENT_DIRECTORY_PATH_CE_SELECT),
              0))
      == ___FIX(___NO_ERR))
    {
      if (!SetCurrentDirectory (___CAST(___STRING_TYPE(___SET_CURRENT_DIRECTORY_PATH_CE_SELECT),cpath)))
        e = fnf_or_err_code_from_GetLastError ();
      ___release_string (cpath);
    }

#endif

  return e;
}


___SCMOBJ ___os_rename_file
   ___P((___SCMOBJ path1,
         ___SCMOBJ path2),
        (path1,
         path2)
___SCMOBJ path1;
___SCMOBJ path2;)
{
  ___SCMOBJ e;
  void *cpath1;
  void *cpath2;

#ifndef USE_POSIX
#ifndef USE_WIN32

  e = ___FIX(___UNKNOWN_ERR);

#endif
#endif

#ifdef USE_POSIX

#define ___RENAME_FILE_PATH_CE_SELECT(latin1,utf8,ucs2,ucs4,native) native

  if ((e = ___SCMOBJ_to_NONNULLSTRING
             (path1,
              &cpath1,
              1,
              ___CE(___RENAME_FILE_PATH_CE_SELECT),
              0))
      == ___FIX(___NO_ERR))
    {
      if ((e = ___SCMOBJ_to_NONNULLSTRING
                 (path2,
                  &cpath2,
                  2,
                  ___CE(___RENAME_FILE_PATH_CE_SELECT),
                  0))
          == ___FIX(___NO_ERR))
        {
          if (rename (___CAST(___STRING_TYPE(___RENAME_FILE_PATH_CE_SELECT),cpath1),
                      ___CAST(___STRING_TYPE(___RENAME_FILE_PATH_CE_SELECT),cpath2))
              < 0)
            e = fnf_or_err_code_from_errno ();
          ___release_string (cpath2);
        }
      ___release_string (cpath1);
    }
 
#endif

#ifdef USE_WIN32

#ifdef UNICODE
#define ___RENAME_FILE_PATH_CE_SELECT(latin1,utf8,ucs2,ucs4,native) ucs2
#else
#define ___RENAME_FILE_PATH_CE_SELECT(latin1,utf8,ucs2,ucs4,native) latin1
#endif

  if ((e = ___SCMOBJ_to_NONNULLSTRING
             (path1,
              &cpath1,
              1,
              ___CE(___RENAME_FILE_PATH_CE_SELECT),
              0))
      == ___FIX(___NO_ERR))
    {
      if ((e = ___SCMOBJ_to_NONNULLSTRING
                 (path2,
                  &cpath2,
                  2,
                  ___CE(___RENAME_FILE_PATH_CE_SELECT),
                  0))
          == ___FIX(___NO_ERR))
        {
          if (!MoveFile (___CAST(___STRING_TYPE(___RENAME_FILE_PATH_CE_SELECT),cpath1),
                         ___CAST(___STRING_TYPE(___RENAME_FILE_PATH_CE_SELECT),cpath2)))
            e = fnf_or_err_code_from_GetLastError ();
          ___release_string (cpath2);
        }
      ___release_string (cpath1);
    }

#endif

  return e;
}


___SCMOBJ ___os_copy_file
   ___P((___SCMOBJ path1,
         ___SCMOBJ path2),
        (path1,
         path2)
___SCMOBJ path1;
___SCMOBJ path2;)
{
  ___SCMOBJ e;
  void *cpath1;
  void *cpath2;

#ifndef USE_POSIX
#ifndef USE_WIN32

  e = ___FIX(___UNKNOWN_ERR);

#endif
#endif

#ifdef USE_POSIX

#define ___COPY_FILE_PATH_CE_SELECT(latin1,utf8,ucs2,ucs4,native) native

  if ((e = ___SCMOBJ_to_NONNULLSTRING
             (path1,
              &cpath1,
              1,
              ___CE(___COPY_FILE_PATH_CE_SELECT),
              0))
      == ___FIX(___NO_ERR))
    {
      if ((e = ___SCMOBJ_to_NONNULLSTRING
                 (path2,
                  &cpath2,
                  2,
                  ___CE(___COPY_FILE_PATH_CE_SELECT),
                  0))
          == ___FIX(___NO_ERR))
        {
          ;
          ___release_string (cpath2);
        }
      ___release_string (cpath1);
    }
 
#endif

#ifdef USE_WIN32

#ifdef UNICODE
#define ___COPY_FILE_PATH_CE_SELECT(latin1,utf8,ucs2,ucs4,native) ucs2
#else
#define ___COPY_FILE_PATH_CE_SELECT(latin1,utf8,ucs2,ucs4,native) latin1
#endif

  if ((e = ___SCMOBJ_to_NONNULLSTRING
             (path1,
              &cpath1,
              1,
              ___CE(___COPY_FILE_PATH_CE_SELECT),
              0))
      == ___FIX(___NO_ERR))
    {
      if ((e = ___SCMOBJ_to_NONNULLSTRING
                 (path2,
                  &cpath2,
                  2,
                  ___CE(___COPY_FILE_PATH_CE_SELECT),
                  0))
          == ___FIX(___NO_ERR))
        {
          if (!CopyFile (___CAST(___STRING_TYPE(___COPY_FILE_PATH_CE_SELECT),cpath1),
                         ___CAST(___STRING_TYPE(___COPY_FILE_PATH_CE_SELECT),cpath2),
                         1))
            e = fnf_or_err_code_from_GetLastError ();
          ___release_string (cpath2);
        }
      ___release_string (cpath1);
    }

#endif

  return e;
}


___SCMOBJ ___os_delete_file
   ___P((___SCMOBJ path),
        (path)
___SCMOBJ path;)
{
  ___SCMOBJ e;
  void *cpath;

#ifndef USE_POSIX
#ifndef USE_WIN32

  e = ___FIX(___UNKNOWN_ERR);

#endif
#endif

#ifdef USE_POSIX

#define ___DELETE_FILE_PATH_CE_SELECT(latin1,utf8,ucs2,ucs4,native) native

  if ((e = ___SCMOBJ_to_NONNULLSTRING
             (path,
              &cpath,
              1,
              ___CE(___DELETE_FILE_PATH_CE_SELECT),
              0))
      == ___FIX(___NO_ERR))
    {
      if (unlink (___CAST(___STRING_TYPE(___DELETE_FILE_PATH_CE_SELECT),cpath))
          < 0)
        e = fnf_or_err_code_from_errno ();
      ___release_string (cpath);
    }

#endif

#ifdef USE_WIN32

#ifdef UNICODE
#define ___DELETE_FILE_PATH_CE_SELECT(latin1,utf8,ucs2,ucs4,native) ucs2
#else
#define ___DELETE_FILE_PATH_CE_SELECT(latin1,utf8,ucs2,ucs4,native) latin1
#endif

  if ((e = ___SCMOBJ_to_NONNULLSTRING
             (path,
              &cpath,
              1,
              ___CE(___DELETE_FILE_PATH_CE_SELECT),
              0))
      == ___FIX(___NO_ERR))
    {
      if (!DeleteFile (___CAST(___STRING_TYPE(___DELETE_FILE_PATH_CE_SELECT),cpath)))
        e = fnf_or_err_code_from_GetLastError ();
      ___release_string (cpath);
    }

#endif

  return e;
}


#if 0
//////////////////////////////////////////////////

{
  WIN32_FILE_ATTRIBUTE_DATA info;
  HANDLE hFile;
  hFile = CreateFile
            (TEXT("c:\\soft"),
             0,
             0,
             NULL,
             OPEN_EXISTING,
             FILE_ATTRIBUTE_NORMAL,
             NULL);
  if (hFile == INVALID_HANDLE_VALUE)
    error ("CreateFile");
  else
    {
      BY_HANDLE_FILE_INFORMATION info2;
      if (!GetFileInformationByHandle (hFile, &info2))
        error ("GetFileInformationByHandle");
      else
        {
          printf ("attrs = %d\n", info2.dwFileAttributes);
          //  FILETIME ftCreationTime; 
          //  FILETIME ftLastAccessTime; 
          //  FILETIME ftLastWriteTime; 
          printf ("dwVolumeSerialNumber = %d\n", info2.dwVolumeSerialNumber);
          printf ("nFileSizeHigh = %d\n", info2.nFileSizeHigh);
          printf ("nFileSizeLow = %d\n", info2.nFileSizeLow);
          printf ("nNumberOfLinks = %d\n", info2.nNumberOfLinks);
          printf ("nFileIndexHigh = %d\n", info2.nFileIndexHigh);
          printf ("nFileIndexLow = %d\n", info2.nFileIndexLow);
        }
      CloseHandle (hFile);
    }

  {
    int len = 10;
    int n;
    char buf[10];
    if ((n = GetCurrentDirectory (len, buf)) < len)
      printf ("cd = %s\n", buf);
  }

  if (!CreateDirectory (TEXT("foo"), NULL))
    error ("CreateDirectory");

  if (!SetCurrentDirectory (TEXT("foo2")))
    error ("SetCurrentDirectory");

  if (!MoveFile (TEXT("foo"), TEXT("foo2")))
    error ("MoveFile");

  if (!RemoveDirectory (TEXT("foo2")))
    error ("RemoveDirectory");

  if (!CopyFile (TEXT("test.c"), TEXT("test2.c"), 1))
    error ("CopyFile");

  if (!DeleteFile (TEXT("test3.c")))
    error ("DeleteFile");

  if (!GetFileAttributesEx
       (TEXT("test.c"),
        GetFileExInfoStandard,
        &info))
    error ("GetFileAttributesEx");
  else
    {
      printf ("attrs = %d\n", info.dwFileAttributes);
      //  FILETIME   ftCreationTime; 
      //  FILETIME   ftLastAccessTime; 
      //  FILETIME   ftLastWriteTime; 
      printf ("size hi = %d\n", info.nFileSizeHigh);
      printf ("size lo = %d\n", info.nFileSizeLow);
    }
#endif


/*   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   */

/* I/O module setup and cleanup. */

___HIDDEN ___SCMOBJ io_module_setup ___PVOID
{
  ___SCMOBJ e;

  ////printf ("io_module_setup\n");fflush(stdout);/////////////////////////////

  if ((e = ___device_group_setup (&io_mod.dgroup)) == ___FIX(___NO_ERR))
    {
      if ((e = tty_module_setup (NULL)) == ___FIX(___NO_ERR))
        {
#ifdef USE_WIN32

#define WINSOCK_MAJOR 1
#define WINSOCK_MINOR 1

    WSADATA winsock_data;

    if (WSAStartup (MAKEWORD(WINSOCK_MAJOR, WINSOCK_MINOR), &winsock_data))
      {
        printf ("WSAStartup failed\n"); fflush (stdout);///////////////////
      }

    if (LOBYTE(winsock_data.wVersion) != WINSOCK_MINOR ||
        HIBYTE(winsock_data.wVersion) != WINSOCK_MAJOR)
      {
        WSACleanup (); /* ignore error */
        printf ("WSAStartup did not get correct version\n"); fflush (stdout);///////////////////
      }

#endif
          return ___FIX(___NO_ERR);
        }
      ___device_group_cleanup (io_mod.dgroup);
    }

  return e;
}


___HIDDEN void io_module_cleanup ___PVOID
{
  //  printf ("io_module_cleanup\n");/////////////////////////////

#ifdef USE_WIN32

  WSACleanup (); /* ignore error */

#endif

  tty_module_cleanup ();
  ___device_group_cleanup (io_mod.dgroup);
}


/*---------------------------------------------------------------------------*/

___SCMOBJ ___setup_os ___PVOID
{
  ___SCMOBJ e;

  /*
   * To perform correct cleanup when the program terminates an
   * "atexit (___cleanup)" is performed in "setup_io" in certain
   * environments.  There must not be any possibility of program
   * termination through "exit (...)" between the "atexit (...)"
   * and the entry of "___setup_mem".  This guarantees that
   * "___cleanup" does not access dangling pointers.
   */

#if 0
  MessageBeep (MB_OK);/////////////////////
  MessageBeep (MB_OK);
  MessageBeep (MB_OK);
  MessageBeep (MB_OK);
  MessageBeep (MB_OK);
#endif

  setup_os_specific ();
  setup_fp ();
  setup_process_times ();
  setup_alloc_mem ();
  setup_interrupt_handling ();/***************/

  if ((e = io_module_setup ()) == ___FIX(___NO_ERR))
    {
      setup_dynamic_load ();
#ifdef USE_POSIX
      signal (SIGPIPE, SIG_IGN); /*************** put this somewhere else*/
#endif
      return ___FIX(___NO_ERR);
    }

  cleanup_interrupt_handling ();
  cleanup_alloc_mem ();
  cleanup_process_times ();
  cleanup_fp ();
  cleanup_os_specific ();

  return e;
}


void ___cleanup_os ___PVOID
{
  cleanup_dynamic_load ();
  io_module_cleanup ();
  cleanup_interrupt_handling ();
  cleanup_alloc_mem ();
  cleanup_process_times ();
  cleanup_fp ();
  cleanup_os_specific ();
}


/*---------------------------------------------------------------------------*/
